Package: xplainfi
Title: Feature Importance Methods for Global Explanations
Version: 1.1.0
Authors@R: 
    person("Lukas", "Burk", , "cran@lukasburk.de", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-7528-3795"))
Description: Provides a consistent interface for common feature importance
    methods as described in Ewald et al. (2024)
    <doi:10.1007/978-3-031-63797-1_22>, including permutation feature
    importance (PFI), conditional and relative feature importance (CFI,
    RFI), leave one covariate out (LOCO), and Shapley additive global
    importance (SAGE), as well as feature sampling mechanisms to support
    conditional importance methods.
License: LGPL (>= 3)
URL: https://mlr-org.github.io/xplainfi/,
        https://github.com/mlr-org/xplainfi
BugReports: https://github.com/mlr-org/xplainfi/issues
Depends: R (>= 4.1.0)
Imports: checkmate, cli, data.table (>= 1.15.0), mirai, mlr3 (>=
        1.1.0), mlr3fselect, mvtnorm, paradox (>= 1.0.0), R6, stats,
        utils
Suggests: arf, DiagrammeR, foreach, future, future.apply, ggplot2,
        glue, gower, knitr, knockoff, lgr, mlr3data, mlr3learners,
        partykit, patchwork, ranger, rmarkdown, rpart, testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/Needs/website: rmarkdown, here, jsonlite, cpi
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-02-26 19:18:17 UTC; Lukas
Author: Lukas Burk [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7528-3795>)
Maintainer: Lukas Burk <cran@lukasburk.de>
Repository: CRAN
Date/Publication: 2026-02-26 19:40:02 UTC
Built: R 4.6.0; ; 2026-04-28 13:18:12 UTC; unix

Package: xplainfi
Title: Feature Importance Methods for Global Explanations
Version: 1.2.0
Authors@R: 
    person("Lukas", "Burk", , "cran@lukasburk.de", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0001-7528-3795"))
Description: Provides a consistent interface for common feature importance
    methods as described in Ewald et al. (2024)
    <doi:10.1007/978-3-031-63797-1_22>, including permutation feature
    importance (PFI), conditional and relative feature importance (CFI,
    RFI), leave one covariate out (LOCO), and Shapley additive global
    importance (SAGE), as well as feature sampling mechanisms to support
    conditional importance methods.
License: LGPL (>= 3)
URL: https://mlr-org.github.io/xplainfi/,
        https://github.com/mlr-org/xplainfi
BugReports: https://github.com/mlr-org/xplainfi/issues
Depends: R (>= 4.1.0)
Imports: checkmate, cli, data.table (>= 1.15.0), mirai, mlr3 (>=
        1.1.0), mlr3fselect, mvtnorm, paradox (>= 1.0.0), R6, stats,
        utils
Suggests: arf, DiagrammeR, foreach, future, future.apply, ggplot2,
        glue, gower, knitr, knockoff, lgr, mlr3data, mlr3learners,
        partykit, patchwork, ranger, rmarkdown, rpart, testthat (>=
        3.0.0), withr
VignetteBuilder: knitr
Config/Needs/website: rmarkdown, here, jsonlite, cpi
Config/roxygen2/markdown: TRUE
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-07-24 16:43:13 UTC; Lukas
Author: Lukas Burk [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-7528-3795>)
Maintainer: Lukas Burk <cran@lukasburk.de>
Repository: CRAN
Date/Publication: 2026-07-24 17:30:02 UTC
Built: R 4.6.0; ; 2026-07-24 19:32:17 UTC; unix

# Generated by roxygen2: do not edit by hand

export(CFI)
export(ConditionalARFSampler)
export(ConditionalCtreeSampler)
export(ConditionalGaussianSampler)
export(ConditionalKNNSampler)
export(ConditionalSAGE)
export(ConditionalSampler)
export(FeatureImportanceMethod)
export(FeatureSampler)
export(KnockoffGaussianSampler)
export(KnockoffSampler)
export(LOCO)
export(MarginalPermutationSampler)
export(MarginalReferenceSampler)
export(MarginalSAGE)
export(MarginalSampler)
export(PFI)
export(PerturbationImportance)
export(RFI)
export(SAGE)
export(WVIM)
export(check_groups)
export(rsmp_all_test)
export(sim_dgp_confounded)
export(sim_dgp_correlated)
export(sim_dgp_ewald)
export(sim_dgp_independent)
export(sim_dgp_interactions)
export(sim_dgp_mediated)
export(wvim_design_matrix)
export(xplain_opt)
if (getRversion() < "4.4.0") export(`%||%`)
import(data.table)
importFrom(R6,R6Class)
importFrom(checkmate,"%??%")
importFrom(cli,cli)
importFrom(mlr3,PredictionClassif)
importFrom(mlr3,PredictionRegr)
importFrom(mlr3,as_resample_result)
importFrom(mlr3,assert_task)
importFrom(mlr3,resample)
importFrom(mlr3,rsmp)
importFrom(mlr3fselect,fs)
importFrom(mlr3fselect,fselect)
importFrom(mvtnorm,rmvnorm)
importFrom(paradox,ps)
importFrom(stats,binom.test)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(tools,toRd)
importFrom(utils,bibentry)
importFrom(utils,packageVersion)
importFrom(utils,person)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# `xplainfi`

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN
status](https://www.r-pkg.org/badges/version/xplainfi)](https://CRAN.R-project.org/package=xplainfi)
[![R-CMD-check](https://github.com/mlr-org/xplainfi/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mlr-org/xplainfi/actions/workflows/R-CMD-check.yaml)
[![codecov](https://codecov.io/gh/mlr-org/xplainfi/graph/badge.svg?token=QIQDMP3AM7)](https://app.codecov.io/gh/mlr-org/xplainfi)
<!-- badges: end -->

The goal of `xplainfi` is to collect common feature importance methods
under a unified and extensible interface.

It is built around [mlr3](https://mlr-org.com/) as available
abstractions for learners, tasks, measures, etc. greatly simplify the
implementation of importance measures.

## Installation

Install `xplainfi` from CRAN:

``` r
install.packages("xplainfi")
```

Or install from [R-universe](https://mlr-org.r-universe.dev):

``` r
install.packages("xplainfi", repos = c("https://mlr-org.r-universe.dev", "https://cloud.r-project.org"))
```

The latest development version of `xplainfi` can be installed with
`pak`:

``` r
# install.packages(pak)
pak::pak("mlr-org/xplainfi")
```

## Example: PFI

Here is a basic example on how to calculate PFI for an untrained learner
and task, using cross-validation for resampling and computing PFI within
each resampling iteration 10 times on the `friedman1` task (see
`?mlbench::mlbench.friedman1`).

The `friedman1` task has the following structure:

$$y = 10 \sin(\pi x_1 x_2) + 20(x_3 - 0.5)^2 + 10x_4 + 5x_5 + \varepsilon$$

Where $x_{\{1,2,3,4,5\}}$ are named `important1` through `important5` in
the `Task`, with additional numbered `unimportant` features without
effect on $y$.

``` r
library(xplainfi)
library(mlr3learners)
#> Loading required package: mlr3

task = tgen("friedman1")$generate(1000)
learner = lrn("regr.ranger", num.trees = 100)
measure = msr("regr.mse")

pfi = PFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = rsmp("cv", folds = 3),
    n_repeats = 30
)
```

Compute and print PFI scores:

``` r
pfi$compute()
pfi$importance()
#> Key: <feature>
#>          feature   importance
#>           <char>        <num>
#>  1:   important1  8.183995584
#>  2:   important2  7.481268675
#>  3:   important3  1.571760349
#>  4:   important4 12.585739572
#>  5:   important5  2.810875567
#>  6: unimportant1  0.030667439
#>  7: unimportant2 -0.002837696
#>  8: unimportant3 -0.044922079
#>  9: unimportant4 -0.060054450
#> 10: unimportant5  0.060148388
```

If it aids interpretation, importances can also be calculated as the
*ratio* rather than the *difference* between the baseline and
post-permutation losses:

``` r
pfi$importance(relation = "ratio")
#> Key: <feature>
#>          feature importance
#>           <char>      <num>
#>  1:   important1  2.6987668
#>  2:   important2  2.5598945
#>  3:   important3  1.3294180
#>  4:   important4  3.6278508
#>  5:   important5  1.5874860
#>  6: unimportant1  1.0067957
#>  7: unimportant2  0.9994507
#>  8: unimportant3  0.9905990
#>  9: unimportant4  0.9874657
#> 10: unimportant5  1.0126572
```

When PFI is computed based on resampling with multiple iterations, and /
or multiple permutation iterations, the individual scores can be
retrieved as a `data.table`:

``` r
str(pfi$scores())
#> Classes 'data.table' and 'data.frame':   900 obs. of  6 variables:
#>  $ feature          : chr  "important1" "important1" "important1" "important1" ...
#>  $ iter_rsmp        : int  1 1 1 1 1 1 1 1 1 1 ...
#>  $ iter_repeat      : int  1 2 3 4 5 6 7 8 9 10 ...
#>  $ regr.mse_baseline: num  4.56 4.56 4.56 4.56 4.56 ...
#>  $ regr.mse_post    : num  12.3 11.9 11.3 12.1 13.6 ...
#>  $ importance       : num  7.77 7.33 6.74 7.56 9.06 ...
#>  - attr(*, ".internal.selfref")=<externalptr>
```

Where `iter_rsmp` corresponds to the resampling iteration, i.e., 3 for
3-fold cross-validation, and `iter_repeat` corresponds to the
permutation iteration within each resampling iteration, 5 in this case.
While `pfi$importance()` contains the means across all iterations,
`pfi$scores()` allows you to manually visualize or aggregate them in any
way you see fit.

For example:

``` r
library(ggplot2)

ggplot(
    pfi$scores(),
    aes(x = importance, y = reorder(feature, importance))
) +
    geom_boxplot(color = "#f44560", fill = alpha("#f44560", 0.4)) +
    labs(
        title = "Permutation Feature Importance on Friedman1",
        subtitle = "Computed over 3-fold CV with 5 permutations per iteration using Random Forest",
        x = "Importance",
        y = "Feature"
    ) +
    theme_minimal(base_size = 16) +
    theme(
        plot.title.position = "plot",
        panel.grid.major.y = element_blank()
    )
```

<img src="man/figures/README-pfi-plot-1.png" alt="" width="100%" />

If the measure in question needs to be maximized rather than minimized
(like $R^2$), the internal importance calculation takes that into
account via the `$minimize` property of the measure and calculates
importances such that the intuition “performance improvement” -\>
“higher importance score” still holds:

``` r
pfi = PFI$new(
    task = task,
    learner = learner,
    measure = msr("regr.rsq")
)
#> ℹ No <Resampling> provided, using `resampling = rsmp("holdout", ratio = 2/3)`
#>   (test set size: 333)

pfi$compute()
pfi$importance()
#> Key: <feature>
#>          feature   importance
#>           <char>        <num>
#>  1:   important1  0.329915393
#>  2:   important2  0.297695022
#>  3:   important3  0.063613087
#>  4:   important4  0.493673768
#>  5:   important5  0.121794662
#>  6: unimportant1  0.003972813
#>  7: unimportant2  0.002157623
#>  8: unimportant3 -0.002780577
#>  9: unimportant4  0.001914150
#> 10: unimportant5  0.001366645
```

See `vignette("xplainfi")` for more examples.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 7,
	fig.height = 5
)

## ----setup--------------------------------------------------------------------
library(xplainfi)
library(DiagrammeR)
library(mlr3learners)
set.seed(123)

## ----overview-table, echo=FALSE-----------------------------------------------
dgp_overview <- data.frame(
	DGP = c(
		"sim_dgp_correlated",
		"sim_dgp_mediated",
		"sim_dgp_confounded",
		"sim_dgp_interactions",
		"sim_dgp_independent",
		"sim_dgp_ewald"
	),
	Challenge = c(
		"Spurious correlation",
		"Mediation effects",
		"Confounding",
		"Interaction effects",
		"Baseline (no challenges)",
		"Mixed effects"
	),
	`PFI Behavior` = c(
		"High for spurious x2",
		"Shows total effects",
		"Biased upward",
		"Low (no main effects)",
		"Accurate",
		"Mixed"
	),
	`CFI Behavior` = c(
		"Low for spurious x2",
		"Shows direct effects",
		"Less biased",
		"High (captures interactions)",
		"Accurate",
		"Mixed"
	),
	check.names = FALSE
)
knitr::kable(dgp_overview, caption = "Overview of simulation settings and expected method behavior")

## ----dag-correlated, echo=FALSE, fig.cap="DAG for correlated features DGP", fig.width=10, fig.height=4----
grViz(
	"
  digraph Correlated {
    rankdir=LR;
    graph [ranksep=1.5];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightcoral', label='X₁\n(β=2.0)'];
    X2 [fillcolor='pink', label='X₂\n(β=0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=1.0)'];
    X4 [fillcolor='lightgray', label='X₄\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    X1 -> X2 [color=red, style=bold, label='r≈0.9'];
    X1 -> Y [label='2.0'];
    X2 -> Y [style=dashed, color=gray, label='0'];
    X3 -> Y [label='1.0'];
    X4 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X3; X4}
    {rank=same; X2}
    {rank=sink; Y}
  }"
)

## ----correlated-example-------------------------------------------------------
set.seed(123)
task <- sim_dgp_correlated(n = 500)

# Check correlation between X1 and X2
cor(task$data()[, c("x1", "x2")])

# True coefficients: x1=2.0, x2=0, x3=1.0, x4=0
# Note: x2 is highly correlated with x1 but has NO causal effect!

## ----dag-mediated, echo=FALSE, fig.cap="DAG for mediated effects DGP", fig.width=10, fig.height=4----
grViz(
	"
  digraph Mediated {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    E [fillcolor='orange', label='Exposure\n(β=0)'];
    D [fillcolor='lightblue', label='Direct\n(β=0.5)'];
    M [fillcolor='yellow', label='Mediator\n(β=1.5)'];
    N [fillcolor='lightgray', label='Noise\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    E -> M [label='0.8', color=purple, penwidth=2];
    D -> M [label='0.6', color=blue];
    D -> Y [label='0.5', color=blue];
    M -> Y [label='1.5', color=purple, penwidth=2];
    N -> Y [style=dashed, color=gray];
    
    {rank=source; E; D; N}
    {rank=same; M}
    {rank=sink; Y}
  }"
)

## ----mediated-example---------------------------------------------------------
set.seed(123)
task <- sim_dgp_mediated(n = 500)

# Calculate total effect of exposure
# Total effect = 0.8 * 1.5 = 1.2 (through mediator)
# Direct effect = 0 (no direct path to Y)

## ----dag-confounded, echo=FALSE, fig.cap="DAG for confounding DGP", fig.width=10, fig.height=5----
grViz(
	"
  digraph Confounded {
    rankdir=LR;
    graph [ranksep=1.2, nodesep=0.8];
    node [shape=circle, style=filled, fontsize=14, width=1.2];

    H [fillcolor='red', label='H\n(Confounder)', style='filled,dashed'];
    X1 [fillcolor='lightcoral', label='X₁\n(β=1.0)'];
    P [fillcolor='pink', label='Proxy\n(β=0)'];
    I [fillcolor='lightblue', label='Independent\n(β=1.0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];

    H -> X1 [color=red, label='1.0'];
    H -> P [color=red, style=dashed, label='1.0'];
    H -> Y [color=red, label='1.0', penwidth=2];
    X1 -> Y [label='1.0'];
    I -> Y [label='1.0'];

    {rank=source; H}
    {rank=same; X1; P; I}
    {rank=sink; Y}
  }"
)

## ----confounded-example-------------------------------------------------------
set.seed(123)
# Hidden confounder scenario (default)
task_hidden <- sim_dgp_confounded(n = 500, hidden = TRUE)
task_hidden$feature_names # proxy available but not confounder

# Observable confounder scenario
task_observed <- sim_dgp_confounded(n = 500, hidden = FALSE)
task_observed$feature_names # both confounder and proxy available

## ----dag-interactions, echo=FALSE, fig.cap="DAG for interaction effects DGP", fig.width=10, fig.height=4----
grViz(
	"
  digraph Interaction {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='orange', label='X₁\n(β=0)'];
    X2 [fillcolor='orange', label='X₂\n(β=0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=1.0)'];
    N1 [fillcolor='lightgray', label='N₁\n(β=0)'];
    N2 [fillcolor='lightgray', label='N₂\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    INT [fillcolor='red', shape=diamond, label='X₁×X₂\n(β=2.0)', width=1.5];
    
    X1 -> INT [color=red, penwidth=2];
    X2 -> INT [color=red, penwidth=2];
    INT -> Y [color=red, label='2.0', penwidth=2];
    X3 -> Y [label='1.0'];
    N1 -> Y [style=dashed, color=gray];
    N2 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X2; X3; N1; N2}
    {rank=same; INT}
    {rank=sink; Y}
  }"
)

## ----interactions-example-----------------------------------------------------
set.seed(123)
task <- sim_dgp_interactions(n = 500)

# Note: X1 and X2 have NO main effects
# Their importance comes ONLY through their interaction

## ----dag-independent, echo=FALSE, fig.cap="DAG for independent features DGP", fig.width=10, fig.height=4----
grViz(
	"
  digraph Independent {
    rankdir=LR;
    graph [ranksep=1.5];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightblue', label='X₁\n(β=2.0)'];
    X2 [fillcolor='lightblue', label='X₂\n(β=1.0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=0.5)'];
    N1 [fillcolor='lightgray', label='N₁\n(β=0)'];
    N2 [fillcolor='lightgray', label='N₂\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    X1 -> Y [label='2.0', penwidth=3];
    X2 -> Y [label='1.0', penwidth=2];
    X3 -> Y [label='0.5'];
    N1 -> Y [style=dashed, color=gray];
    N2 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X2; X3; N1; N2}
    {rank=sink; Y}
  }"
)

## ----independent-example------------------------------------------------------
set.seed(123)
task <- sim_dgp_independent(n = 500)

# All methods should rank features consistently:
# important1 > important2 > important3 > unimportant1,2 (approx. 0)

## ----dag-ewald, echo=FALSE, fig.cap="DAG for Ewald et al. (2024) DGP", fig.width=10, fig.height=4----
grViz(
	"
  digraph Ewald {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightgray', label='X₁\n(β=0)'];
    X2 [fillcolor='lightgray', label='X₂\n(β=0)'];
    X3 [fillcolor='lightgray', label='X₃\n(β=0)'];
    X4 [fillcolor='lightblue', label='X₄\n(β=1.0)'];
    X5 [fillcolor='lightblue', label='X₅\n(β=1.0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    INT [fillcolor='red', shape=diamond, label='X₄×X₅\n(β=1.0)', width=1.5];
    
    X1 -> X2 [color=gray, label='≈1.0'];
    X3 -> X4 [color=gray, label='≈1.0'];
    X4 -> Y [label='1.0'];
    X5 -> Y [label='1.0'];
    X4 -> INT [color=red];
    X5 -> INT [color=red];
    INT -> Y [color=red, label='1.0'];
    
    {rank=source; X1; X3; X5}
    {rank=same; X2; X4}
    {rank=same; INT}
    {rank=sink; Y}
  }"
)

## ----ewald-example------------------------------------------------------------
sim_dgp_ewald(n = 500)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 7,
	fig.height = 5
)

## ----load-libraries-----------------------------------------------------------
library(xplainfi)
library(mlr3)
library(mlr3learners)
library(data.table)

# Create a task for demonstration
task_mixed = tsk("penguins")
task_numeric = sim_dgp_correlated(n = 200)

## ----feature-types------------------------------------------------------------
# Check supported feature types for different samplers
task_mixed$feature_types
permutation = MarginalPermutationSampler$new(task_mixed)
permutation$feature_types

## ----sample-methods-----------------------------------------------------------
# Sample from stored task (using row_ids)
sampled_task = permutation$sample(
	feature = "bill_length",
	row_ids = 40:45
)
sampled_task

# Sample from "external" data
test_data = task_mixed$data(rows = 40:45)
sampled_external = permutation$sample_newdata(
	feature = "bill_length",
	newdata = test_data
)
sampled_external

## ----permutation-example------------------------------------------------------
# Create permutation sampler
permutation = MarginalPermutationSampler$new(task_mixed)

# Sample a continuous feature
original = task_mixed$data(rows = 1:10)
sampled = permutation$sample("bill_length", row_ids = 1:10)

# Compare original and sampled values
data.table(
	original_bill = original$bill_length,
	sampled_bill = sampled$bill_length,
	sex = original$sex # Unchanged
)

## ----marginal-ref-example-----------------------------------------------------
# Create marginal reference sampler with n_samples reference pool
marginal_ref = MarginalReferenceSampler$new(task_mixed, n_samples = 30L)

# Sample a feature - each row gets values from a randomly sampled reference row
original = task_mixed$data(rows = 1:5)
sampled = marginal_ref$sample("bill_length", row_ids = 1:5)

# Compare
data.table(
	original_bill = original$bill_length,
	sampled_bill = sampled$bill_length,
	sex = original$sex # Unchanged
)

## ----correlation-preservation-------------------------------------------------
# Sample with MarginalPermutationSampler (breaks correlations)
perm = MarginalPermutationSampler$new(task_numeric)
sampled_perm = perm$sample(c("x1", "x2"), row_ids = 1:10)

# Sample with MarginalReferenceSampler (preserves within-row correlations)
ref = MarginalReferenceSampler$new(task_numeric, n_samples = 50L)
sampled_ref = ref$sample(c("x1", "x2"), row_ids = 1:10)

# Check correlations
cor_original = cor(task_numeric$data()$x1, task_numeric$data()$x2)
cor_perm = cor(sampled_perm$x1, sampled_perm$x2)
cor_ref = cor(sampled_ref$x1, sampled_ref$x2)

data.table(
	method = c("Original", "Permutation", "Reference"),
	correlation = c(cor_original, cor_perm, cor_ref)
)

## ----gaussian-sampler---------------------------------------------------------
# Create Gaussian conditional sampler
gaussian = ConditionalGaussianSampler$new(task_numeric)

# Sample x1 conditioned on other features
sampled = gaussian$sample(
	feature = "x1",
	row_ids = 1:10,
	conditioning_set = c("x2", "x3", "x4")
)

# Compare original and conditionally sampled values
original = task_numeric$data(rows = 1:10)
data.table(
	original = original$x1,
	sampled = sampled$x1,
	x2 = original$x2 # Conditioning feature (unchanged)
)

## ----arf-sampler--------------------------------------------------------------
# Create ARF sampler (works with full task including categorical features)
arf = ConditionalARFSampler$new(task_mixed, num_trees = 20, verbose = FALSE)

# Sample island conditioned on body measurements
sampled = arf$sample(
	feature = "island",
	row_ids = 1:10,
	conditioning_set = c("bill_length", "body_mass")
)

# Compare original and sampled island
original = task_mixed$data(rows = 1:10)
data.table(
	original_island = original$island,
	sampled_island = sampled$island,
	bill_length = original$bill_length, # Conditioning feature
	body_mass = original$body_mass # Conditioning feature
)

## ----ctree-sampler------------------------------------------------------------
# Create ctree sampler
ctree = ConditionalCtreeSampler$new(task_mixed)

# Sample with default parameters
sampled = ctree$sample(
	feature = "bill_length",
	row_ids = 1:10,
	conditioning_set = "island"
)

original = task_mixed$data(rows = 1:10)
data.table(
	island = original$island, # Conditioning feature
	original = original$bill_length,
	sampled = sampled$bill_length
)

## ----knn-sampler-numeric------------------------------------------------------
# Create kNN sampler with k=5 neighbors
knn_numeric = ConditionalKNNSampler$new(task_numeric, k = 5)

# Sample x1 based on nearest neighbors in (x2, x3) space
sampled_numeric = knn_numeric$sample(
	feature = "x1",
	row_ids = 1:5,
	conditioning_set = c("x2", "x3")
)

original_numeric = task_numeric$data(rows = 1:5)
data.table(
	x2 = original_numeric$x2,
	x3 = original_numeric$x3,
	original_x1 = original_numeric$x1,
	sampled_x1 = sampled_numeric$x1
)

## ----knn-sampler-mixed--------------------------------------------------------
# Use task with categorical features
knn_mixed = ConditionalKNNSampler$new(task_mixed, k = 5)

# Sample bill_length conditioning on island (categorical) and body_mass (numeric)
sampled_mixed = knn_mixed$sample(
	feature = "bill_length",
	row_ids = 1:5,
	conditioning_set = c("island", "body_mass")
)

original_mixed = task_mixed$data(rows = 1:5)
data.table(
	island = original_mixed$island,
	body_mass = original_mixed$body_mass,
	original_bill = original_mixed$bill_length,
	sampled_bill = sampled_mixed$bill_length
)

## ----knockoff-sampler---------------------------------------------------------
# Create Gaussian knockoff sampler (using task_numeric from earlier)
knockoff = KnockoffGaussianSampler$new(task_numeric)

# Generate knockoffs
original = task_numeric$data(rows = 1:5)
knockoffs = knockoff$sample(
	feature = task_numeric$feature_names,
	row_ids = 1:5
)

# Original vs knockoff values
data.table(
	x1_original = original$x1,
	x1_knockoff = knockoffs$x1,
	x2_original = original$x2,
	x2_knockoff = knockoffs$x2
)

## ----cfi-knockoff, eval = FALSE-----------------------------------------------
# # CFI with knockoff sampler for conditional independence testing
# cfi_knockoff = CFI$new(
# 	task = task_numeric,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse"),
# 	sampler = knockoff
# )
# 
# # Compute importance with CPI-based inference
# cfi_knockoff$compute()
# cfi_knockoff$importance(ci_method = "cpi")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 8,
	fig.height = 6
)
set.seed(123)
# Quiet down
lgr::get_logger("mlr3")$set_threshold("warn")
options("xplain.progress" = interactive())

## ----setup--------------------------------------------------------------------
library(xplainfi)
library(mlr3)
library(mlr3learners)
library(data.table)
library(ggplot2)

## ----setup-problem------------------------------------------------------------
task <- tgen("friedman1")$generate(n = 300)
learner <- lrn("regr.ranger", num.trees = 100)
measure <- msr("regr.mse")
resampling <- rsmp("cv", folds = 3)

## ----pfi-basic----------------------------------------------------------------
pfi <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 10
)

pfi$compute()
pfi$importance()

## ----pfi-parameters-----------------------------------------------------------
pfi_stable <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 50
)

pfi_stable$compute()
pfi_stable$importance()

## ----pif-nrepeats-------------------------------------------------------------
pfi_stable$scores()[feature == "important2", ] |>
	ggplot(aes(y = importance, x = factor(iter_rsmp))) +
	geom_boxplot() +
	labs(
		title = "PFI variability within resampling iterations",
		subtitle = "Setting n_repeats higher improves PFI estimates",
		y = "PFI score (important2)",
		x = "Resampling iteration (3-fold CV)"
	) +
	theme_minimal()

## ----pfi-scores-tmp, echo=FALSE-----------------------------------------------
pfi_important2_range = round(range(pfi_stable$scores()[feature == "important2", importance]), 2)

## ----pfi-ratio----------------------------------------------------------------
pfi_stable$importance(relation = "ratio")

## ----loco-basic---------------------------------------------------------------
loco = LOCO$new(
  task = task,
  learner = learner,
  measure = measure,
  resampling = resampling
)

loco$compute()
loco$importance()

## ----samplers-demo------------------------------------------------------------
arf_sampler <- ConditionalARFSampler$new(task)

sample_data <- task$data(rows = 1:5)
sample_data[, .(important1, important2)]

## ----conditional-sampling-----------------------------------------------------
sampled_conditional <- arf_sampler$sample_newdata(
	feature = "important1",
	newdata = sample_data,
	conditioning_set = c("important2", "important3")
)

sample_data[, .(important1, important2, important3)]
sampled_conditional[, .(important1, important2, important3)]

## ----detailed-scores----------------------------------------------------------
pfi$scores() |>
	head(10) |>
	knitr::kable(digits = 4, caption = "Detailed PFI scores (first 10 rows)")

## ----scoring-summary----------------------------------------------------------
pfi$scores()[, .(
	features = uniqueN(feature),
	resampling_folds = uniqueN(iter_rsmp),
	permutation_iters = uniqueN(iter_repeat),
	total_scores = .N
)]

## ----detailed-scores-ratio----------------------------------------------------
pfi$scores(relation = "ratio") |>
	head(10) |>
	knitr::kable(digits = 4, caption = "PFI scores using the ratio (first 10 rows)")

## ----pfi-obs-scores-----------------------------------------------------------
pfi$obs_loss()

## ----pretrained-pfi-----------------------------------------------------------
resampling_holdout <- rsmp("holdout")$instantiate(task)
learner_trained <- lrn("regr.ranger", num.trees = 100)
learner_trained$train(task, row_ids = resampling_holdout$train_set(1))

pfi_pretrained <- PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_holdout,
	n_repeats = 10
)

pfi_pretrained$compute()
pfi_pretrained$importance()

## ----pretrained-custom--------------------------------------------------------
# Simulate: learner was trained elsewhere, we have new data to use
new_data <- tgen("friedman1")$generate(n = 100)

# Same as rsmp_all_test(task)
resampling_custom <- rsmp("custom")$instantiate(
	new_data,
	train_sets = list(integer(0)),
	test_sets = list(new_data$row_ids)
)

pfi_newdata <- PFI$new(
	task = new_data,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_custom,
	n_repeats = 10
)

pfi_newdata$compute()
pfi_newdata$importance()

## ----pretrained-error, error = TRUE-------------------------------------------
try({
PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = rsmp("cv", folds = 3)
)
})

## ----parallel-future, eval = FALSE--------------------------------------------
# library(future)
# plan("multisession", workers = 2)
# 
# # PFI with parallelization across features
# pfi_parallel = PFI$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse"),
# 	n_repeats = 10
# )
# pfi_parallel$compute()
# pfi_parallel$importance()
# 
# # LOCO with parallelization (uses mlr3fselect internally)
# loco_parallel = LOCO$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse")
# )
# loco_parallel$compute()
# loco_parallel$importance()

## ----parallel-mirai, eval = FALSE---------------------------------------------
# library(mirai)
# daemons(n = 2)
# 
# # Same PFI/LOCO code works with mirai backend
# pfi_parallel = PFI$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse"),
# 	n_repeats = 10
# )
# pfi_parallel$compute()
# pfi_parallel$importance()
# 
# # Clean up daemons when done
# daemons(0)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 8,
	fig.height = 6
)
set.seed(123)
# Quiet down
lgr::get_logger("mlr3")$set_threshold("warn")
options("xplain.progress" = interactive())

## ----setup--------------------------------------------------------------------
library(xplainfi)
library(mlr3)
library(mlr3learners)
library(data.table)
library(ggplot2)

## ----setup-problem------------------------------------------------------------
task <- tgen("friedman1")$generate(n = 300)
learner <- lrn("regr.ranger", num.trees = 100)
measure <- msr("regr.mse")
resampling <- rsmp("cv", folds = 3)

## ----pfi-basic----------------------------------------------------------------
pfi <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 10
)

pfi$compute()
pfi$importance()

## ----pfi-parameters-----------------------------------------------------------
pfi_stable <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 50
)

pfi_stable$compute()
pfi_stable$importance()

## ----pif-nrepeats-------------------------------------------------------------
pfi_stable$scores()[feature == "important2", ] |>
	ggplot(aes(y = importance, x = factor(iter_rsmp))) +
	geom_boxplot() +
	labs(
		title = "PFI variability within resampling iterations",
		subtitle = "Setting n_repeats higher improves PFI estimates",
		y = "PFI score (important2)",
		x = "Resampling iteration (3-fold CV)"
	) +
	theme_minimal()

## ----pfi-scores-tmp, echo=FALSE-----------------------------------------------
pfi_important2_range = round(range(pfi_stable$scores()[feature == "important2", importance]), 2)

## ----pfi-ratio----------------------------------------------------------------
pfi_stable$importance(relation = "ratio")

## ----loco-basic---------------------------------------------------------------
loco <- LOCO$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 10
)

loco$compute()
loco$importance()

## ----samplers-demo------------------------------------------------------------
arf_sampler <- ConditionalARFSampler$new(task)

sample_data <- task$data(rows = 1:5)
sample_data[, .(important1, important2)]

## ----conditional-sampling-----------------------------------------------------
sampled_conditional <- arf_sampler$sample_newdata(
	feature = "important1",
	newdata = sample_data,
	conditioning_set = c("important2", "important3")
)

sample_data[, .(important1, important2, important3)]
sampled_conditional[, .(important1, important2, important3)]

## ----detailed-scores----------------------------------------------------------
pfi$scores() |>
	head(10) |>
	knitr::kable(digits = 4, caption = "Detailed PFI scores (first 10 rows)")

## ----scoring-summary----------------------------------------------------------
pfi$scores()[, .(
	features = uniqueN(feature),
	resampling_folds = uniqueN(iter_rsmp),
	permutation_iters = uniqueN(iter_repeat),
	total_scores = .N
)]

## ----detailed-scores-ratio----------------------------------------------------
pfi$scores(relation = "ratio") |>
	head(10) |>
	knitr::kable(digits = 4, caption = "PFI scores using the ratio (first 10 rows)")

## ----pfi-obs-scores-----------------------------------------------------------
pfi$obs_loss()

## ----pretrained-pfi-----------------------------------------------------------
resampling_holdout <- rsmp("holdout")$instantiate(task)
learner_trained <- lrn("regr.ranger", num.trees = 100)
learner_trained$train(task, row_ids = resampling_holdout$train_set(1))

pfi_pretrained <- PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_holdout,
	n_repeats = 10
)

pfi_pretrained$compute()
pfi_pretrained$importance()

## ----pretrained-custom--------------------------------------------------------
# Simulate: learner was trained elsewhere, we have new data to use
new_data <- tgen("friedman1")$generate(n = 100)

# Same as rsmp_all_test(task)
resampling_custom <- rsmp("custom")$instantiate(
	new_data,
	train_sets = list(integer(0)),
	test_sets = list(new_data$row_ids)
)

pfi_newdata <- PFI$new(
	task = new_data,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_custom,
	n_repeats = 10
)

pfi_newdata$compute()
pfi_newdata$importance()

## ----pretrained-error, error = TRUE-------------------------------------------
try({
PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = rsmp("cv", folds = 3)
)
})

## ----parallel-future, eval = FALSE--------------------------------------------
# library(future)
# plan("multisession", workers = 2)
# 
# # PFI with parallelization across features
# pfi_parallel = PFI$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse"),
# 	n_repeats = 10
# )
# pfi_parallel$compute()
# pfi_parallel$importance()
# 
# # LOCO with parallelization (uses mlr3fselect internally)
# loco_parallel = LOCO$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse")
# )
# loco_parallel$compute()
# loco_parallel$importance()

## ----parallel-mirai, eval = FALSE---------------------------------------------
# library(mirai)
# daemons(n = 2)
# 
# # Same PFI/LOCO code works with mirai backend
# pfi_parallel = PFI$new(
# 	task,
# 	learner = lrn("regr.ranger"),
# 	measure = msr("regr.mse"),
# 	n_repeats = 10
# )
# pfi_parallel$compute()
# pfi_parallel$importance()
# 
# # Clean up daemons when done
# daemons(0)

#' Check group specification
#' @param groups (`list`) A (named) list of groups
#' @param all_features (character()) All available feature names from the task.
#'
#' @family utilities
#' @export
#' @return The input list `group`, with each element now named.
#' @examples
#' task <- sim_dgp_interactions(n = 100)
#' task$feature_names
#'
#' # Intended use
#' groups1 = list(effects = c("x1", "x2", "x3"), noise = c("noise1", "noise2"))
#' check_groups(groups1, task$feature_names)
#'
#' # Names are auto-generated where needed
#' check_groups(list(a = "x1",  c("x2", "x1")), task$feature_names)
#'
#' # Examples for cases that throw errors:
#'
#' # Unexpected features
#' groups2 = list(effects = c("x1", "foo", "bar", "x1"))
#' try(check_groups(groups2, task$feature_names))
#' # Too deeply nested
#' groups3 = list(effects = c("x1", "x2", "x3"), noise = c("noise1", list(c("noise2"))))
#' try(check_groups(groups2, task$feature_names))
#'
check_groups = function(groups, all_features) {
  # Unlist non-recursively so we can fail if the result is not a vector
  group_features = unlist(groups, use.names = FALSE, recursive = FALSE)

  if (is.list(group_features)) {
    cli::cli_abort(c(
      x = "Group{?s} is nested too deeply: {.val {names(which(sapply(groups, is.list)))}}",
      i = "Ensure each group consists of a vector of features, not a list."
    ))
  }

  # TODO: Figure out whether it's an error or just weird if groups have overlapping features (interpreation changes at least)
  if (anyDuplicated(group_features) > 0) {
    cli::cli_alert_warning(
      "Feature{?s} is specified in multiple groups: {.val {group_features[anyDuplicated(group_features)]}}"
    )
  }

  extra_feature = setdiff(group_features, all_features)
  if (length(extra_feature) > 0) {
    cli::cli_abort(
      "Feature{?s} specified in {.code groups} not in provided {.cls Task}: {.val {extra_feature}}"
    )
  }

  # Check all are named
  nm = names(groups)
  unnamed = which(nm == "")
  if (length(unnamed) > 0) {
    cli::cli_inform(c(
      "Not all groups are named",
      i = "Group{?s} {.val {as.character(unnamed)}} {?is/are} named automatically"
    ))
    # Override only names where there are non and use k-th letter of alphabet corresponding
    # to position of groups with missing name
    names(groups)[unnamed] = paste0(c("Group", LETTERS[unnamed]), collapse = "")
  }

  groups
}

#' @title Conditional Feature Sampler
#'
#' @description Base class for conditional sampling methods where features
#' are sampled conditionally on other features. This is an abstract class
#' that should be extended by concrete implementations.
#'
#' @export
ConditionalSampler = R6Class(
  "ConditionalSampler",
  inherit = FeatureSampler,
  public = list(
    #' @description
    #' Creates a new instance of the ConditionalSampler class
    #' @param task ([mlr3::Task]) Task to sample from
    #' @param conditioning_set (`character` | `NULL`) Default conditioning set to use in `$sample()`.
    initialize = function(task, conditioning_set = NULL) {
      super$initialize(task)
      self$label = "Conditional sampler"

      # Define param_set with conditioning_set parameter for all conditional samplers
      self$param_set = paradox::ps(
        conditioning_set = paradox::p_uty(default = NULL)
      )

      # Store conditioning_set if provided
      if (!is.null(conditioning_set)) {
        self$param_set$set_values(conditioning_set = conditioning_set)
      }
    },

    #' @description
    #' Sample from stored task conditionally on other features.
    #' @param feature (`character`) Feature(s) to sample.
    #' @param row_ids (`integer()` | `NULL`) Row IDs to use. If `NULL`, uses all rows.
    #' @param conditioning_set (`character` | `NULL`) Features to condition on.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param ... Additional arguments passed to the sampler implementation.
    #' @return Modified copy with sampled feature(s).
    sample = function(feature, row_ids = NULL, conditioning_set = NULL, samples_per_row = 1L, ...) {
      checkmate::assert_count(samples_per_row, positive = TRUE)
      data_copy = private$.get_task_data_by_row_id(row_ids)

      # Determine conditioning set (note: NULL is different than character(0))
      # Priority:
      # 1) function argument,
      # 2) stored param_set value,
      # 3) default (all other features) (! important behavior expected by CFI implementation!)
      conditioning_set = resolve_param(
        conditioning_set,
        self$param_set$values$conditioning_set,
        setdiff(self$task$feature_names, feature)
      )

      if (xplain_opt("debug")) {
        cli::cli_alert_info(
          "Resolved conditioning_set: {.val {conditioning_set}}"
        )
      }

      private$.sample_conditional(
        data_copy,
        feature,
        conditioning_set,
        samples_per_row = samples_per_row,
        ...
      )
    },

    #' @description
    #' Sample from external data conditionally.
    #' @param feature (`character`) Feature(s) to sample.
    #' @param newdata ([`data.table`][data.table::data.table]) External data to use.
    #' @param conditioning_set (`character` | `NULL`) Features to condition on.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param ... Additional arguments passed to the sampler implementation.
    #' @return Modified copy with sampled feature(s).
    sample_newdata = function(
      feature,
      newdata,
      conditioning_set = NULL,
      samples_per_row = 1L,
      ...
    ) {
      checkmate::assert_count(samples_per_row, positive = TRUE)
      # Create copy to avoid modifying original
      if (inherits(newdata, "data.table")) {
        data_copy = data.table::copy(newdata)
      } else {
        data_copy = as.data.table(newdata)
      }

      # Determine conditioning set (note: NULL is different than character(0))
      # Priority:
      # 1) function argument,
      # 2) stored param_set value,
      # 3) default (all other features) (! important behavior expected by CFI implementation!)
      conditioning_set = resolve_param(
        conditioning_set,
        self$param_set$values$conditioning_set,
        setdiff(self$task$feature_names, feature)
      )

      if (xplain_opt("debug")) {
        cli::cli_alert_info(
          "Resolved conditioning_set: {.val {conditioning_set}}"
        )
      }

      private$.sample_conditional(
        data_copy,
        feature,
        conditioning_set,
        samples_per_row = samples_per_row,
        ...
      )
    }
  ),

  private = list(
    # @description
    # Internal method for conditional sampling. Must be implemented by subclasses.
    # @param data (`data.table`) Data to sample from (copy that can be modified).
    # @param feature (`character`) Feature(s) to sample.
    # @param conditioning_set (`character` | `NULL`) Features to condition on.
    # @param samples_per_row (`integer(1)`) Number of draws per input row.
    # @param ... Additional sampler-specific arguments.
    .sample_conditional = function(data, feature, conditioning_set, samples_per_row = 1L, ...) {
      cli::cli_abort(c(
        "Abstract method",
        i = "Subclasses must implement the {.fn .sample_conditional} method"
      ))
    }
  )
)

# Tests for KnockoffSampler

test_that("KnockoffSampler initialization works", {
  skip_if_not_installed("knockoff")

  n = 100
  task = tgen("friedman1")$generate(n = n)
  sampler = KnockoffSampler$new(task)

  expect_s3_class(sampler, "KnockoffSampler")
  expect_equal(sampler$label, "Knockoff sampler")
  expect_s3_class(sampler$param_set, "ParamSet")
  expect_true("knockoff_fun" %in% sampler$param_set$ids())

  # Check that knockoff matrix was created
  expect_true(data.table::is.data.table(sampler$x_tilde))
  expect_equal(nrow(sampler$x_tilde), n)
  # x_tilde has feature columns + internal ..row_id column
  expect_equal(ncol(sampler$x_tilde), length(task$feature_names) + 1L)
  expect_equal(setdiff(names(sampler$x_tilde), "..row_id"), task$feature_names)
})

test_that("KnockoffSampler sampling works", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 100)
  sampler = KnockoffSampler$new(task)
  data = task$data()

  # Test single feature sampling
  sampled_data = sampler$sample("important1")

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Sampled feature values are from knockoff matrix
  expect_identical(sampled_data$important1, sampler$x_tilde$important1)
})

test_that("KnockoffSampler handles multiple features", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 100)
  sampler = KnockoffSampler$new(task)
  data = task$data()

  features = c("important1", "important2")
  sampled_data = sampler$sample(features)

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Check that sampled feature values are from knockoff matrix
  for (feat in features) {
    expect_identical(sampled_data[[feat]], sampler$x_tilde[[feat]])
  }
})

test_that("KnockoffSampler works with different numeric tasks", {
  skip_if_not_installed("knockoff")

  # Friedman1 regression task (all numeric)
  task_friedman = tgen("friedman1")$generate(n = 100)
  sampler_friedman = KnockoffSampler$new(task_friedman)
  sampled_friedman = sampler_friedman$sample("important1")
  expect_sampler_output_structure(sampled_friedman, task_friedman, nrows = 100)

  # Circle task with specified dimensions (all numeric)
  task_circle = tgen("circle", d = 4)$generate(n = 80)
  sampler_circle = KnockoffSampler$new(task_circle)
  sampled_circle = sampler_circle$sample("x1")
  expect_sampler_output_structure(sampled_circle, task_circle, nrows = 80)

  # Custom numeric task
  set.seed(123)
  custom_data = data.table::data.table(
    x1 = rnorm(50),
    x2 = runif(50),
    x3 = rexp(50),
    y = rnorm(50)
  )
  task_custom = as_task_regr(custom_data, target = "y")
  sampler_custom = KnockoffSampler$new(task_custom)
  sampled_custom = sampler_custom$sample("x1")
  expect_sampler_output_structure(sampled_custom, task_custom, nrows = 50)
})

test_that("KnockoffSampler knockoff matrix properties", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 100)
  sampler = KnockoffSampler$new(task)

  # Knockoff matrix should have same dimensions as feature matrix (plus ..row_id column)
  feature_data = task$data(cols = task$feature_names)
  expect_equal(nrow(sampler$x_tilde), nrow(feature_data))
  expect_equal(ncol(sampler$x_tilde), ncol(feature_data) + 1L)
  expect_equal(setdiff(names(sampler$x_tilde), "..row_id"), names(feature_data))

  # Knockoff values should be numeric (since task is all numeric)
  expect_true(all(sapply(sampler$x_tilde, is.numeric)))

  # Knockoff values should be finite
  expect_true(all(sapply(sampler$x_tilde, function(x) all(is.finite(x)))))
})

test_that("KnockoffSampler custom knockoff function", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 50)

  # Custom knockoff function that returns scaled versions
  custom_knockoff_fun = function(x) {
    x_matrix = as.matrix(x)
    noise = matrix(rnorm(prod(dim(x_matrix)), sd = 0.1), nrow = nrow(x_matrix))
    return(x_matrix + noise)
  }

  sampler = KnockoffSampler$new(task, knockoff_fun = custom_knockoff_fun)

  expect_true(is.function(sampler$param_set$values$knockoff_fun))
  expect_true(data.table::is.data.table(sampler$x_tilde))
  expect_equal(nrow(sampler$x_tilde), 50)
  expect_equal(ncol(sampler$x_tilde), length(task$feature_names) + 1L)

  data = task$data()
  sampled_data = sampler$sample("important1")

  expect_true(data.table::is.data.table(sampled_data))
  expect_equal(nrow(sampled_data), 50)
  expect_false(identical(sampled_data$important1, data$important1))
  expect_identical(sampled_data$important1, sampler$x_tilde$important1)
})

test_that("KnockoffSampler reproducibility", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 80)

  set.seed(123)
  sampler1 = KnockoffSampler$new(task)

  set.seed(123)
  sampler2 = KnockoffSampler$new(task)

  # Dimensions should be identical
  expect_equal(dim(sampler1$x_tilde), dim(sampler2$x_tilde))
  expect_equal(names(sampler1$x_tilde), names(sampler2$x_tilde))
  expect_true("..row_id" %in% names(sampler1$x_tilde))

  # Sampling should be consistent within each sampler
  sampled1a = sampler1$sample("important1")
  sampled1b = sampler1$sample("important1")
  expect_identical(sampled1a$important1, sampled1b$important1)
})

test_that("KnockoffSampler edge cases", {
  skip_if_not_installed("knockoff")

  # Test with minimum viable task (single feature)
  # Seeded: knockoff::create.second_order warns only for some random samples,
  # so pin the RNG to a state that reliably triggers the expected warning.
  set.seed(1L)
  single_feature_data = data.table::data.table(
    x1 = rnorm(30),
    y = rnorm(30)
  )
  task_single = as_task_regr(single_feature_data, target = "y")

  expect_warning({
    sampler_single = KnockoffSampler$new(task_single)
  })

  expect_equal(ncol(sampler_single$x_tilde), 2L) # x1 + ..row_id
  expect_equal(setdiff(names(sampler_single$x_tilde), "..row_id"), "x1")

  sampled_single = sampler_single$sample("x1")
  expect_true(data.table::is.data.table(sampled_single))
  expect_equal(nrow(sampled_single), 30)
})

test_that("KnockoffSampler fails with non-numeric features", {
  skip_if_not_installed("knockoff")

  # Factor features
  data_with_factor = data.table::data.table(
    x1 = rnorm(50),
    x2 = factor(sample(c("A", "B", "C"), 50, replace = TRUE)),
    x3 = rnorm(50),
    y = rnorm(50)
  )
  task_factor = as_task_regr(data_with_factor, target = "y")
  expect_error(KnockoffSampler$new(task_factor), class = "error")

  # Character features
  data_with_char = data.table::data.table(
    x1 = rnorm(50),
    x2 = sample(letters[1:5], 50, replace = TRUE),
    x3 = rnorm(50),
    y = rnorm(50)
  )
  task_char = as_task_regr(data_with_char, target = "y")
  expect_error(KnockoffSampler$new(task_char), class = "error")
})

test_that("KnockoffGaussianSampler works with all-numeric features", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 100)
  sampler = KnockoffGaussianSampler$new(task)

  expect_s3_class(sampler, "KnockoffGaussianSampler")
  expect_s3_class(sampler, "KnockoffSampler")
  expect_equal(sampler$label, "Gaussian Knockoff sampler")

  expect_true(data.table::is.data.table(sampler$x_tilde))
  expect_equal(nrow(sampler$x_tilde), 100)
  expect_equal(ncol(sampler$x_tilde), length(task$feature_names) + 1L)

  data = task$data()
  sampled_data = sampler$sample("important1")

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_identical(sampled_data$important1, sampler$x_tilde$important1)

  # Test row_ids
  expect_equal(nrow(sampler$sample("important1", row_ids = 1:10)), 10)
  expect_equal(nrow(sampler$sample("important1", row_ids = 32)), 1)

  # Non-sampled features unchanged
  non_sampled = c(
    "important2",
    "important3",
    "important4",
    "important5",
    "unimportant1",
    "unimportant2",
    "unimportant3",
    "unimportant4",
    "unimportant5"
  )
  expect_equal(
    sampler$sample("important1", row_ids = 32)[, ..non_sampled],
    data[32, ..non_sampled]
  )

  # Multiple features
  features = c("important1", "important2", "important3")
  sampled_multi = sampler$sample(features)

  for (feat in features) {
    expect_false(identical(sampled_multi[[feat]], data[[feat]]))
    expect_identical(sampled_multi[[feat]], sampler$x_tilde[[feat]])
  }
})

test_that("KnockoffGaussianSampler fails with factor features", {
  skip_if_not_installed("knockoff")

  data_with_factor = data.table::data.table(
    x1 = rnorm(50),
    x2 = factor(sample(c("A", "B", "C"), 50, replace = TRUE)),
    x3 = rnorm(50),
    y = rnorm(50)
  )
  task_factor = as_task_regr(data_with_factor, target = "y")

  expect_error(
    KnockoffGaussianSampler$new(task_factor),
    regexp = "following unsupported feature types"
  )
})

test_that("KnockoffSampler works with multiple iterations", {
  skip_if_not_installed("knockoff")

  task = sim_dgp_ewald(n = 100)

  withr::with_seed(42, {
    sampler_default = KnockoffSampler$new(task = task)
  })
  withr::with_seed(42, {
    sampler_1 = KnockoffSampler$new(task = task, iters = 1)
  })
  sampler_5 = KnockoffSampler$new(task = task, iters = 5)

  # x_tilde generated with same seed, same length
  expect_equal(
    sampler_default$sample("x1"),
    sampler_1$sample("x1")
  )

  x_orig = task$data(rows = 2)
  x_sampled = sampler_1$sample("x1", row_ids = 2)

  # Other values are the same (excluding sampled feature x1)
  non_sampled_cols = setdiff(names(x_sampled), "x1")
  expect_equal(x_orig[, ..non_sampled_cols], x_sampled[, ..non_sampled_cols])

  # Multiple samples via samples_per_row
  x_orig = task$data(rows = c(1, 1, 1))
  x_sampled = sampler_5$sample("x1", row_ids = 1, samples_per_row = 3L)

  expect_equal(x_sampled$x2, x_orig$x2)

  # Sampling more draws than knockoff iters triggers replacement warning
  expect_warning(
    sampled_1_repeats <- sampler_1$sample("x1", row_ids = 1, samples_per_row = 3L),
    regexp = "requested more often than they are present"
  )
  expect_equal(nrow(sampled_1_repeats), 3L)
})

test_that("KnockoffGaussianSampler preserves feature types", {
  skip_if_not_installed("knockoff")
  test_sampler_feature_types(KnockoffGaussianSampler)
})

test_that("KnockoffSampler obeys draw-major order under samples_per_row > 1", {
  skip_if_not_installed("knockoff")
  set.seed(123L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = KnockoffSampler$new(task, iters = 3L)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 3L
  )
})

test_that("KnockoffSampler warns and samples with replacement when samples_per_row > iters", {
  skip_if_not_installed("knockoff")
  set.seed(123L)
  task = mlr3::tgen("friedman1")$generate(n = 20)
  sampler = KnockoffSampler$new(task, iters = 2L)

  expect_warning(
    out <- sampler$sample("important1", row_ids = task$row_ids, samples_per_row = 5L),
    regexp = "requested more often than they are present"
  )
  expect_equal(nrow(out), 20L * 5L)
})

# Tests for MarginalReferenceSampler

test_that("MarginalReferenceSampler initialization works", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalReferenceSampler$new(task)

  expect_s3_class(sampler, "MarginalReferenceSampler")
  expect_s3_class(sampler, "MarginalSampler")
  expect_equal(sampler$label, "Marginal reference sampler")
  expect_s3_class(sampler$param_set, "ParamSet")

  # Reference data should be stored
  expect_true(data.table::is.data.table(sampler$reference_data))
  expect_equal(nrow(sampler$reference_data), 100)
  expect_named(sampler$reference_data, task$feature_names)
})

test_that("MarginalReferenceSampler sampling works", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalReferenceSampler$new(task)
  data = task$data()

  # Test single feature sampling
  sampled_data = sampler$sample("x1")

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Sampled values should come from reference data
  expect_true(all(sampled_data$x1 %in% sampler$reference_data$x1))
})

test_that("MarginalReferenceSampler handles multiple features", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalReferenceSampler$new(task)
  data = task$data()

  features = c("x1", "x2", "x3")
  sampled_data = sampler$sample(features)

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Sampled values should come from reference data
  for (feat in features) {
    expect_true(all(sampled_data[[feat]] %in% sampler$reference_data[[feat]]))
  }
})

test_that("MarginalReferenceSampler with n_samples", {
  task = tgen("circle", d = 5)$generate(n = 100)

  # Create sampler with subsampled reference data
  sampler = MarginalReferenceSampler$new(task, n_samples = 50L)

  expect_equal(nrow(sampler$reference_data), 50)

  sampled_data = sampler$sample("x1", row_ids = 1:10)

  expect_sampler_output_structure(sampled_data, task, nrows = 10)

  # All sampled values should be from the task (reference is subset of task)
  expect_true(all(sampled_data$x1 %in% task$data()$x1))
})

test_that("MarginalReferenceSampler sample_newdata works", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalReferenceSampler$new(task)

  newdata = tgen("circle", d = 5)$generate(n = 20)$data()
  sampled_data = sampler$sample_newdata("x1", newdata = newdata)

  expect_sampler_output_structure(sampled_data, task, nrows = 20)

  # Sampled values should come from reference data
  expect_true(all(sampled_data$x1 %in% sampler$reference_data$x1))

  # Original newdata should be unchanged
  expect_equal(nrow(newdata), 20)
})

test_that("MarginalReferenceSampler preserves within-row dependencies", {
  set.seed(123)
  n = 100
  x1 = rnorm(n)
  data = data.table::data.table(
    x1 = x1,
    x2 = x1 * 2, # Perfect correlation
    x3 = x1 * 3,
    y = x1 + rnorm(n, sd = 0.1)
  )
  task = as_task_regr(data, target = "y")

  sampler = MarginalReferenceSampler$new(task)

  # Sample multiple correlated features
  sampled_data = sampler$sample(c("x1", "x2", "x3"), row_ids = 1:50)

  # Within each sampled row, the relationships should be preserved
  expect_true(all(abs(sampled_data$x2 - sampled_data$x1 * 2) < 1e-10))
  expect_true(all(abs(sampled_data$x3 - sampled_data$x1 * 3) < 1e-10))
})

test_that("MarginalReferenceSampler vs MarginalPermutationSampler difference", {
  set.seed(123)
  n = 100
  x1 = rnorm(n)
  data = data.table::data.table(
    x1 = x1,
    x2 = x1 * 2 + rnorm(n, sd = 0.1),
    y = x1 + rnorm(n, sd = 0.1)
  )
  task = as_task_regr(data, target = "y")

  # MarginalReferenceSampler preserves within-row correlation
  marginal_ref = MarginalReferenceSampler$new(task)
  sampled_ref = marginal_ref$sample(c("x1", "x2"))
  cor_ref = cor(sampled_ref$x1, sampled_ref$x2)

  # MarginalPermutationSampler breaks all correlations
  permutation = MarginalPermutationSampler$new(task)
  set.seed(456)
  sampled_perm = permutation$sample(c("x1", "x2"))
  cor_perm = cor(sampled_perm$x1, sampled_perm$x2)

  # MarginalReferenceSampler should preserve correlation better
  expect_gt(abs(cor_ref), abs(cor_perm) * 0.5)
})

test_that("MarginalReferenceSampler works with different task types", {
  # Regression task
  task_regr = tgen("circle", d = 4)$generate(n = 100)
  sampler_regr = MarginalReferenceSampler$new(task_regr)
  sampled_regr = sampler_regr$sample("x1")
  expect_sampler_output_structure(sampled_regr, task_regr, nrows = 100)

  # Binary classification task
  task_classif = tsk("sonar")
  sampler_classif = MarginalReferenceSampler$new(task_classif)
  sampled_classif = sampler_classif$sample("V1")
  expect_sampler_output_structure(sampled_classif, task_classif, nrows = task_classif$nrow)

  # Multiclass classification task
  task_multi = tsk("iris")
  sampler_multi = MarginalReferenceSampler$new(task_multi)
  sampled_multi = sampler_multi$sample("Sepal.Length")
  expect_sampler_output_structure(sampled_multi, task_multi, nrows = 150)
})

test_that("MarginalReferenceSampler handles n_samples edge cases", {
  task = tgen("circle", d = 5)$generate(n = 100)

  # n_samples larger than task size
  sampler = MarginalReferenceSampler$new(task, n_samples = 200L)
  expect_equal(nrow(sampler$reference_data), 100) # Capped at task size

  # n_samples = 1
  sampler_small = MarginalReferenceSampler$new(task, n_samples = 1L)
  expect_equal(nrow(sampler_small$reference_data), 1)
})

test_that("MarginalReferenceSampler preserves feature types", {
  test_sampler_feature_types(MarginalReferenceSampler)
})

test_that("MarginalReferenceSampler obeys draw-major order under samples_per_row > 1", {
  set.seed(123L)
  n = 10L
  dt = data.table::data.table(
    y = rnorm(n),
    x = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = MarginalReferenceSampler$new(task)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

# Tests for ConditionalGaussianSampler

test_that("ConditionalGaussianSampler initialization works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalGaussianSampler$new(task)

  expect_s3_class(sampler, "ConditionalGaussianSampler")
  expect_s3_class(sampler, "ConditionalSampler")
  expect_equal(sampler$label, "Gaussian Conditional Sampler")
  expect_s3_class(sampler$param_set, "ParamSet")

  # Check stored statistics
  checkmate::expect_numeric(sampler$mu, len = task$n_features, any.missing = FALSE)
  checkmate::expect_matrix(sampler$sigma, nrows = task$n_features, ncols = task$n_features)
  expect_true(isSymmetric(sampler$sigma))
})

test_that("ConditionalGaussianSampler rejects non-numeric tasks", {
  task = tsk("penguins")
  expect_error(ConditionalGaussianSampler$new(task), "unsupported feature types")
})

test_that("ConditionalGaussianSampler sampling works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalGaussianSampler$new(task)

  # Marginal sampling
  expect_marginal_sampling(sampler, feature = "important1", row_ids = 1:20)

  # Conditional sampling with single conditioning feature
  expect_conditional_sampling(
    sampler,
    feature = "important2",
    conditioning_set = "important1",
    row_ids = 1:20
  )

  # Conditional sampling with multiple features
  expect_conditional_sampling(
    sampler,
    feature = c("important2", "important3"),
    conditioning_set = "important1",
    row_ids = 1:20
  )
})

test_that("ConditionalGaussianSampler sample_newdata works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalGaussianSampler$new(task)
  test_data = task$data(rows = 1:10)

  sampled = sampler$sample_newdata(
    feature = c("important2", "important3"),
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 10)
  expect_feature_type_consistency(sampled, task)
  expect_non_sampled_unchanged(sampled, test_data, "important1")
})

test_that("ConditionalGaussianSampler handles single observation", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalGaussianSampler$new(task)
  test_data = task$data(rows = 1)

  sampled = sampler$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 1)
  expect_non_sampled_unchanged(sampled, test_data, "important1")
})

test_that("ConditionalGaussianSampler is reproducible with seed", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalGaussianSampler$new(task)
  test_data = task$data(rows = 1:10)

  sampled1 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  sampled2 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  expect_identical(sampled1$important2, sampled2$important2)
})

test_that("ConditionalGaussianSampler conditioning_set parameter behavior", {
  task = tgen("friedman1")$generate(n = 100)
  test_conditioning_set_behavior(ConditionalGaussianSampler, task)
})

test_that("ConditionalGaussianSampler preserves feature types", {
  test_sampler_feature_types(ConditionalGaussianSampler)
})

test_that("ConditionalGaussianSampler obeys draw-major order under samples_per_row > 1", {
  set.seed(123L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = ConditionalGaussianSampler$new(task, conditioning_set = c("x2", "tag"))

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

# Tests for ConditionalARFSampler

test_that("ConditionalARFSampler initialization works", {
  skip_if_not_installed("arf")

  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = ConditionalARFSampler$new(task)

  expect_s3_class(sampler, "ConditionalARFSampler")
  expect_s3_class(sampler, "ConditionalSampler")
  expect_equal(sampler$label, "Adversarial Random Forest sampler")
  expect_s3_class(sampler$param_set, "ParamSet")

  # Check expected parameters
  expected_params = c(
    "conditioning_set",
    "finite_bounds",
    "round",
    "stepsize",
    "verbose",
    "parallel"
  )
  expect_true(all(expected_params %in% sampler$param_set$ids()))

  # Check that ARF model was fitted
  expect_s3_class(sampler$arf_model, "ranger")
  expect_type(sampler$psi, "list")
})

test_that("ConditionalARFSampler sampling works", {
  skip_if_not_installed("arf")

  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = ConditionalARFSampler$new(task)

  # Marginal sampling
  expect_marginal_sampling(sampler, feature = "x1", row_ids = 1:20)

  # Conditional sampling
  expect_conditional_sampling(
    sampler,
    feature = "x1",
    conditioning_set = c("x2", "x3"),
    row_ids = 1:20
  )

  # Multiple features
  expect_conditional_sampling(
    sampler,
    feature = c("x1", "x2"),
    conditioning_set = "x3",
    row_ids = 1:20
  )
})

test_that("ConditionalARFSampler handles empty conditioning set", {
  skip_if_not_installed("arf")

  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = ConditionalARFSampler$new(task, conditioning_set = character(0))
  data = task$data()

  sampled = sampler$sample("x1")

  expect_sampler_output_structure(sampled, task, nrows = 100)

  # Sampled feature should differ from original
  expect_sampled_features_changed(sampled, data, "x1")

  # Non-sampled features should remain unchanged
  expect_non_sampled_unchanged(sampled, data, c("x2", "x3", "x4", "x5"))

  # Verify empty conditioning is stored
  expect_equal(sampler$param_set$values$conditioning_set, character(0))
})

test_that("ConditionalARFSampler works with different task types", {
  skip_if_not_installed("arf")

  # Regression task
  task_regr = tgen("circle", d = 4)$generate(n = 100)
  sampler_regr = ConditionalARFSampler$new(task_regr)
  expect_conditional_sampling(sampler_regr, "x1", "x2", row_ids = 1:20)

  # Binary classification task
  task_classif = tsk("sonar")
  sampler_classif = ConditionalARFSampler$new(task_classif)
  expect_conditional_sampling(sampler_classif, "V1", "V2", row_ids = 1:20)

  # Multiclass classification task
  task_multi = tsk("iris")
  sampler_multi = ConditionalARFSampler$new(task_multi)
  expect_conditional_sampling(sampler_multi, "Sepal.Length", "Sepal.Width", row_ids = 1:20)
})

test_that("ConditionalARFSampler param_set structure", {
  skip_if_not_installed("arf")

  task = tgen("circle", d = 5)$generate(n = 50)
  sampler = ConditionalARFSampler$new(task, finite_bounds = "no")

  # Check that removed/hardcoded parameters are not exposed
  hardcoded = c("evidence_row_mode", "sample_NAs", "nomatch", "n_synth")
  expect_false(any(hardcoded %in% sampler$param_set$ids()))

  # Check parameter types and defaults
  expect_equal(sampler$param_set$params[id == "finite_bounds"]$cls, "ParamFct")
  expect_equal(sampler$param_set$params[id == "finite_bounds"]$default[[1]], "no")
  expect_true(sampler$param_set$params[id == "round"]$default[[1]])
  expect_equal(sampler$param_set$params[id == "stepsize"]$default[[1]], 0)
  expect_false(sampler$param_set$params[id == "verbose"]$default[[1]])
  expect_false(sampler$param_set$params[id == "parallel"]$default[[1]])
})

test_that("ConditionalARFSampler parameter priority and storage", {
  skip_if_not_installed("arf")

  task = tgen("circle", d = 5)$generate(n = 50)
  sampler = ConditionalARFSampler$new(
    task,
    conditioning_set = "x2",
    verbose = FALSE,
    parallel = FALSE
  )

  # Stored parameters should be used
  sampled1 = sampler$sample("x1")
  expect_sampler_output_structure(sampled1, task, nrows = 50)

  # Override at call time
  sampled2 = sampler$sample("x1", verbose = TRUE)
  expect_sampler_output_structure(sampled2, task, nrows = 50)

  # Stored parameters remain unchanged
  expect_false(sampler$param_set$values$verbose)
  expect_false(sampler$param_set$values$parallel)
  expect_equal(sampler$param_set$values$conditioning_set, "x2")
})

test_that("ConditionalARFSampler errors when parallel = TRUE but no backend registered", {
  skip_if_not_installed("arf")

  task = tsk("iris")
  sampler = ConditionalARFSampler$new(task, parallel = FALSE, verbose = FALSE)
  # Simulate deserialization scenario: parallel was TRUE when saved
  sampler$param_set$set_values(parallel = TRUE)

  # Ensure no parallel backend (registerDoSEQ counts as unregistered)
  foreach::registerDoSEQ()

  expect_error(
    sampler$sample("Sepal.Length", conditioning_set = "Sepal.Width"),
    "no parallel backend is registered"
  )

  # Calling $sample() with parallel = FALSE override should work
  expect_no_error(
    sampler$sample("Sepal.Length", conditioning_set = "Sepal.Width", parallel = FALSE)
  )
})

test_that("ConditionalARFSampler conditioning_set parameter behavior", {
  skip_if_not_installed("arf")

  task = tgen("friedman1")$generate(n = 100)
  test_conditioning_set_behavior(ConditionalARFSampler, task, verbose = FALSE)
})

test_that("ConditionalARFSampler preserves feature types", {
  skip_if_not_installed("arf")
  test_sampler_feature_types(ConditionalARFSampler, verbose = FALSE)
})

test_that("ConditionalARFSampler obeys draw-major order under samples_per_row > 1", {
  skip_if_not_installed("arf")
  set.seed(123L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = ConditionalARFSampler$new(task, conditioning_set = c("x2", "tag"), verbose = FALSE)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

test_that("ConditionalARFSampler calls arf::forge with n_synth > 1, not duplicated evidence", {
  skip_if_not_installed("arf")
  set.seed(123L)
  task = mlr3::tgen("2dnormals")$generate(n = 50L)
  sampler = ConditionalARFSampler$new(task, conditioning_set = "x2", verbose = FALSE)

  forge_calls = list()
  real_forge = arf::forge # capture before mocking to avoid recursion
  recorder = function(...) {
    args = list(...)
    forge_calls[[length(forge_calls) + 1L]] <<- list(
      n_synth = args$n_synth,
      nrow_evidence = if (is.null(args$evidence)) 0L else nrow(args$evidence)
    )
    real_forge(...) # delegate to real implementation
  }

  testthat::with_mocked_bindings(
    forge = recorder,
    .package = "arf",
    {
      sampler$sample("x1", row_ids = task$row_ids[1:10], samples_per_row = 5L)
    }
  )

  expect_length(forge_calls, 1L)
  expect_equal(forge_calls[[1L]]$n_synth, 5L)
  expect_equal(forge_calls[[1L]]$nrow_evidence, 10L)
})

# Tests for ConditionalCtreeSampler

test_that("ConditionalCtreeSampler initialization works", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalCtreeSampler$new(task)

  expect_s3_class(sampler, "ConditionalCtreeSampler")
  expect_s3_class(sampler, "ConditionalSampler")
  expect_equal(sampler$label, "Conditional Inference Tree Conditional Sampler")
  expect_s3_class(sampler$param_set, "ParamSet")

  # Check default parameters
  expect_equal(sampler$param_set$values$mincriterion, 0.95)
  expect_equal(sampler$param_set$values$minsplit, 20L)
  expect_equal(sampler$param_set$values$minbucket, 7L)
  expect_true(sampler$param_set$values$use_cache)

  # Check tree cache initialized
  expect_type(sampler$tree_cache, "environment")
})

test_that("ConditionalCtreeSampler works with custom parameters", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)

  sampler = ConditionalCtreeSampler$new(
    task,
    mincriterion = 0.90,
    minsplit = 10L,
    minbucket = 5L,
    use_cache = FALSE
  )

  expect_equal(sampler$param_set$values$mincriterion, 0.90)
  expect_equal(sampler$param_set$values$minsplit, 10L)
  expect_equal(sampler$param_set$values$minbucket, 5L)
  expect_false(sampler$param_set$values$use_cache)
})

test_that("ConditionalCtreeSampler sampling works", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalCtreeSampler$new(task)
  data = task$data()

  # Marginal sampling
  expect_marginal_sampling(sampler, feature = "important1", row_ids = 1:20)

  # Conditional sampling
  expect_conditional_sampling(
    sampler,
    feature = "important2",
    conditioning_set = "important1",
    row_ids = 1:20
  )

  # Multiple features
  expect_conditional_sampling(
    sampler,
    feature = c("important2", "important3"),
    conditioning_set = "important1",
    row_ids = 1:20
  )

  # Multiple conditioning features
  expect_conditional_sampling(
    sampler,
    feature = "important3",
    conditioning_set = c("important1", "important2"),
    row_ids = 1:20
  )
})

test_that("ConditionalCtreeSampler sample_newdata works", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalCtreeSampler$new(task)
  test_data = task$data(rows = 1:10)

  sampled = sampler$sample_newdata(
    feature = c("important2", "important3"),
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 10)
  expect_feature_type_consistency(sampled, task)
  expect_non_sampled_unchanged(sampled, test_data, "important1")
})

test_that("ConditionalCtreeSampler caching works", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  test_data = task$data(rows = 1:10)

  # With caching enabled
  sampler_cached = ConditionalCtreeSampler$new(task, use_cache = TRUE)

  # First call should build tree
  expect_length(ls(sampler_cached$tree_cache), 0)
  sampled1 = sampler_cached$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )
  expect_length(ls(sampler_cached$tree_cache), 1)

  # Second call should use cached tree
  sampled2 = sampler_cached$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )
  expect_length(ls(sampler_cached$tree_cache), 1)

  # Without caching
  sampler_uncached = ConditionalCtreeSampler$new(task, use_cache = FALSE)
  sampled3 = sampler_uncached$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )
  expect_length(ls(sampler_uncached$tree_cache), 0)
})

test_that("ConditionalCtreeSampler handles single observation", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalCtreeSampler$new(task)
  test_data = task$data(rows = 1)

  sampled = sampler$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 1)
  expect_non_sampled_unchanged(sampled, test_data, "important1")
})

test_that("ConditionalCtreeSampler is reproducible with seed", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalCtreeSampler$new(task)
  test_data = task$data(rows = 1:10)

  sampled1 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  sampled2 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  expect_identical(sampled1$important2, sampled2$important2)
})

test_that("ConditionalCtreeSampler conditioning_set parameter behavior", {
  skip_if_not_installed("partykit")
  task = tgen("friedman1")$generate(n = 100)
  test_conditioning_set_behavior(ConditionalCtreeSampler, task)
})

test_that("ConditionalCtreeSampler preserves feature types", {
  skip_if_not_installed("partykit")
  test_sampler_feature_types(ConditionalCtreeSampler)
})

test_that("ConditionalCtreeSampler obeys draw-major order under samples_per_row > 1", {
  skip_if_not_installed("partykit")
  set.seed(123L)
  n = 30L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = ConditionalCtreeSampler$new(task, conditioning_set = c("x2", "tag"))

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 3L
  )
})

# Custom testthat expectations for FeatureSampler testing
#
# These follow testthat 3e conventions.
# See: https://testthat.r-lib.org/articles/custom-expectation.html

# -----------------------------------------------------------------------------
# Helper: Generate task from sampler's supported feature types
# -----------------------------------------------------------------------------

#' Generate a test task based on sampler's supported feature types
#'
#' @param supported_types Character vector of supported feature types
#' @param n Number of observations
#' @return An mlr3 regression task
generate_test_task = function(supported_types, n = 100) {
  xdat = data.table::data.table(
    x_num1 = rnorm(n),
    x_num2 = runif(n)
  )

  if ("integer" %in% supported_types) {
    xdat[, x_int := sample(1L:10L, n, replace = TRUE)]
  }
  if ("factor" %in% supported_types) {
    xdat[, x_fct := factor(sample(c("a", "b", "c"), n, replace = TRUE))]
  }
  if ("ordered" %in% supported_types) {
    xdat[,
      x_ord := ordered(
        sample(c("low", "mid", "high"), n, replace = TRUE),
        levels = c("low", "mid", "high")
      )
    ]
  }
  if ("logical" %in% supported_types) {
    xdat[, x_lgl := sample(c(TRUE, FALSE), n, replace = TRUE)]
  }

  xdat[, y := x_num1 + 0.5 * x_num2 + rnorm(n, sd = 0.1)]
  mlr3::as_task_regr(xdat, target = "y")
}

# -----------------------------------------------------------------------------
# expect_feature_type_consistency
# -----------------------------------------------------------------------------

#' Expect sampled output has consistent feature types with task
#'
#' Compares feature classes in sampled data against task$feature_types.
#'
#' @param sampled A data.table returned from sampler$sample()
#' @param task The mlr3 task used for sampling
#' @return Invisibly returns sampled data for piping
expect_feature_type_consistency = function(sampled, task) {
  expected_classes = stats::setNames(
    task$feature_types$type,
    task$feature_types$id
  )
  actual_classes = sapply(
    sampled[, task$feature_names, with = FALSE],
    function(x) class(x)[1]
  )

  mismatches = expected_classes != actual_classes
  if (any(mismatches)) {
    bad_feats = names(expected_classes)[mismatches]
    msg = glue::glue(
      "Feature type mismatch in sampled data.\n",
      "Features with wrong type: {paste(bad_feats, collapse = ', ')}\n",
      "Expected: {paste(expected_classes[mismatches], collapse = ', ')}\n",
      "Actual: {paste(actual_classes[mismatches], collapse = ', ')}"
    )
  } else {
    msg = ""
  }

  expect(!any(mismatches), msg)
  invisible(sampled)
}

# -----------------------------------------------------------------------------
# expect_non_sampled_unchanged
# -----------------------------------------------------------------------------

#' Expect non-sampled features are unchanged after sampling
#'
#' Verifies that features not passed to $sample() remain identical
#' between the original data and the sampled output.
#'
#' @param sampled A data.table returned from sampler$sample()
#' @param original The original data before sampling
#' @param features Character vector of feature names that should be unchanged
#' @return Invisibly returns sampled data for piping
expect_non_sampled_unchanged = function(sampled, original, features) {
  if (length(features) == 0) {
    expect(TRUE, "No features to check")
    return(invisible(sampled))
  }

  for (feat in features) {
    ok = identical(sampled[[feat]], original[[feat]])
    expect(
      ok,
      glue::glue(
        "Feature '{feat}' was modified in sampled data.\n",
        "Non-sampled features must remain unchanged during sampling."
      )
    )
  }

  invisible(sampled)
}

# -----------------------------------------------------------------------------
# expect_sampled_features_changed
# -----------------------------------------------------------------------------

#' Expect sampled features differ from original (stochastic check)
#'
#' Verifies that at least some values in sampled features differ from original.
#' This is a probabilistic check - with sufficient data variability, randomly
#' sampling identical values is extremely unlikely.
#'
#' @param sampled A data.table returned from sampler$sample()
#' @param original The original data before sampling
#' @param sampled_features Character vector of feature names that were sampled
#' @return Invisibly returns sampled data for piping
expect_sampled_features_changed = function(sampled, original, sampled_features) {
  for (feat in sampled_features) {
    ok = !identical(sampled[[feat]], original[[feat]])
    expect(
      ok,
      glue::glue(
        "Sampled feature '{feat}' is identical to original.\n",
        "This suggests the sampler did not modify the feature.\n",
        "(Note: This could theoretically be a false positive with very low probability)"
      )
    )
  }

  invisible(sampled)
}

# -----------------------------------------------------------------------------
# expect_sampler_output_structure
# -----------------------------------------------------------------------------

#' Expect sampler output has correct structure
#'
#' Verifies the sampled data is a data.table with correct columns and dimensions.
#'
#' @param sampled A data.table returned from sampler$sample()
#' @param task The mlr3 task used for sampling
#' @param nrows Expected number of rows (NULL to skip check)
#' @return Invisibly returns sampled data for piping
expect_sampler_output_structure = function(sampled, task, nrows = NULL) {
  expect(
    data.table::is.data.table(sampled),
    glue::glue(
      "Sampled data is not a data.table.\n",
      "Actual class: {paste(class(sampled), collapse = ', ')}"
    )
  )

  expected_cols = c(task$target_names, task$feature_names)
  actual_cols = names(sampled)

  expect(
    identical(actual_cols, expected_cols),
    glue::glue(
      "Sampled data has incorrect columns.\n",
      "Expected: {paste(expected_cols, collapse = ', ')}\n",
      "Actual: {paste(actual_cols, collapse = ', ')}"
    )
  )

  if (!is.null(nrows)) {
    expect(
      nrow(sampled) == nrows,
      glue::glue(
        "Sampled data has incorrect number of rows.\n",
        "Expected: {nrows}\n",
        "Actual: {nrow(sampled)}"
      )
    )
  }

  invisible(sampled)
}

# -----------------------------------------------------------------------------
# expect_marginal_sampling
# -----------------------------------------------------------------------------

#' Expect conditional sampler handles marginal sampling correctly
#'
#' Tests that a conditional sampler works correctly with an empty conditioning
#' set (character(0)), which should trigger marginal sampling behavior.
#'
#' @param sampler A ConditionalSampler instance
#' @param feature Feature to sample
#' @param row_ids Row IDs to sample
#' @return Invisibly returns sampled data for piping
expect_marginal_sampling = function(sampler, feature, row_ids = 1:10) {
  expect(
    inherits(sampler, "ConditionalSampler"),
    "Sampler is not a ConditionalSampler. Marginal sampling test only applies to conditional samplers."
  )

  original = sampler$task$data(rows = row_ids)

  # Sample with empty conditioning set (marginal sampling)
  sampled = sampler$sample(
    feature = feature,
    row_ids = row_ids,
    conditioning_set = character(0)
  )

  # Check structure
  expect_sampler_output_structure(sampled, sampler$task, nrows = length(row_ids))

  # Check types
  expect_feature_type_consistency(sampled, sampler$task)

  # Non-sampled features should remain unchanged. The key test here is that
  # conditional samplers can handle an empty conditioning set without error.
  non_sampled = setdiff(sampler$task$feature_names, feature)
  expect_non_sampled_unchanged(sampled, original, non_sampled)

  invisible(sampled)
}

# -----------------------------------------------------------------------------
# expect_conditional_sampling
# -----------------------------------------------------------------------------

#' Expect conditional sampler handles conditional sampling correctly
#'
#' Tests that a conditional sampler correctly preserves conditioning features
#' and modifies sampled features.
#'
#' @param sampler A ConditionalSampler instance
#' @param feature Feature(s) to sample
#' @param conditioning_set Features to condition on
#' @param row_ids Row IDs to sample
#' @return Invisibly returns sampled data for piping
expect_conditional_sampling = function(sampler, feature, conditioning_set, row_ids = 1:10) {
  expect(
    inherits(sampler, "ConditionalSampler"),
    "Sampler is not a ConditionalSampler. Conditional sampling test only applies to conditional samplers."
  )

  original = sampler$task$data(rows = row_ids)

  sampled = sampler$sample(
    feature = feature,
    row_ids = row_ids,
    conditioning_set = conditioning_set
  )

  # Check structure
  expect_sampler_output_structure(sampled, sampler$task, nrows = length(row_ids))

  # Check types
  expect_feature_type_consistency(sampled, sampler$task)

  # Non-sampled features must be unchanged
  expect_non_sampled_unchanged(sampled, original, conditioning_set)

  # Sampled features should change (stochastic check)
  expect_sampled_features_changed(sampled, original, feature)

  invisible(sampled)
}

# -----------------------------------------------------------------------------
# Omnibus test functions (combine multiple expectations)
# -----------------------------------------------------------------------------

#' Run comprehensive feature type tests for a sampler class
#'
#' Generates a task based on the sampler's supported feature types and tests
#' that all sampling operations preserve correct types.
#'
#' @param sampler_class R6 class for the sampler to test
#' @param ... Additional arguments passed to sampler constructor
#' @return NULL (used for side effects via testthat expectations)
test_sampler_feature_types = function(sampler_class, ...) {
  supported_types = sampler_class$public_fields$feature_types
  task = generate_test_task(supported_types)

  sampler = sampler_class$new(task, ...)
  is_conditional = inherits(sampler, "ConditionalSampler")

  # Test sampling each feature
  # Use enough rows to make false positives from identical sampling negligible,
  # especially for low-cardinality features like ordered factors
  test_row_ids = 1:50

  for (feat in task$feature_names) {
    sampled = sampler$sample(feat, row_ids = test_row_ids)
    expect_sampler_output_structure(sampled, task, nrows = length(test_row_ids))
    expect_feature_type_consistency(sampled, task)

    # For conditional samplers, also test with explicit conditioning set
    if (is_conditional) {
      other_feats = setdiff(task$feature_names, feat)
      if (length(other_feats) >= 1) {
        expect_conditional_sampling(
          sampler,
          feature = feat,
          conditioning_set = other_feats[1],
          row_ids = test_row_ids
        )
      }
      # Also test marginal case
      expect_marginal_sampling(sampler, feature = feat, row_ids = test_row_ids)
    }
  }

  invisible(NULL)
}

#' Test conditioning_set parameter behavior for conditional samplers
#'
#' Verifies that a conditional sampler correctly:
#' 1. Stores conditioning_set in param_set when provided during initialization
#' 2. Can sample without specifying conditioning_set (uses stored value)
#' 3. Can override conditioning_set in $sample() calls
#' 4. Handles NULL conditioning_set (defaults to all other features)
#' 5. Handles empty conditioning_set character(0) (marginal sampling)
#'
#' @param sampler_class R6 class for the sampler to test
#' @param task mlr3 Task to use for testing (must have at least 3 features)
#' @param ... Additional arguments passed to sampler constructor
#' @return NULL (used for side effects via testthat expectations)
test_conditioning_set_behavior = function(sampler_class, task, ...) {
  features = task$feature_names
  checkmate::assert_true(
    length(features) >= 3,
    .var.name = "task must have at least 3 features"
  )

  target_feature = features[1]
  cond_set_1 = features[2]
  cond_set_2 = features[3]
  other_features = setdiff(features, target_feature)

  # Test 1: conditioning_set stored in param_set when provided
  sampler_with_cond = sampler_class$new(task, conditioning_set = cond_set_1, ...)
  expect_identical(
    sampler_with_cond$param_set$values$conditioning_set,
    cond_set_1
  )

  # Test 2: Can sample using stored conditioning_set
  original_data = task$data(rows = 1:5)
  result_stored = sampler_with_cond$sample(feature = target_feature, row_ids = 1:5)
  expect_sampler_output_structure(result_stored, task, nrows = 5)
  expect_non_sampled_unchanged(result_stored, original_data, cond_set_1)

  # Test 3: Can override conditioning_set in $sample() call
  result_override = sampler_with_cond$sample(
    feature = target_feature,
    row_ids = 1:5,
    conditioning_set = cond_set_2
  )
  expect_sampler_output_structure(result_override, task, nrows = 5)
  expect_non_sampled_unchanged(result_override, original_data, cond_set_2)

  # Test 4: NULL conditioning_set during initialization
  sampler_no_cond = sampler_class$new(task, ...)
  expect_null(sampler_no_cond$param_set$values$conditioning_set)

  # Test 5: Can specify conditioning_set in $sample() when not set during init
  result_specified = sampler_no_cond$sample(
    feature = target_feature,
    row_ids = 1:5,
    conditioning_set = cond_set_1
  )
  expect_non_sampled_unchanged(result_specified, original_data, cond_set_1)

  # Test 6: NULL conditioning_set should default to all other features
  # Verify by checking that all other features remain unchanged (they were conditioned on)
  result_null = sampler_no_cond$sample(
    feature = target_feature,
    row_ids = 1:5,
    conditioning_set = NULL
  )
  expect_sampler_output_structure(result_null, task, nrows = 5)
  # If NULL defaults to all other features as conditioning set, they should be unchanged
  expect_non_sampled_unchanged(result_null, original_data, other_features)

  # Test 7: character(0) should result in empty conditioning set (marginal sampling)
  # This tests that the sampler handles empty conditioning set without error
  result_empty = sampler_no_cond$sample(
    feature = target_feature,
    row_ids = 1:5,
    conditioning_set = character(0)
  )
  expect_sampler_output_structure(result_empty, task, nrows = 5)
  # The sampled feature should still be modified
  expect_sampled_features_changed(result_empty, original_data, target_feature)

  invisible(NULL)
}

# -----------------------------------------------------------------------------
# expect_draw_major_row_order
# -----------------------------------------------------------------------------

# Assert that a sampler obeys the draw-major contract:
# `sampler$sample(feature, row_ids, samples_per_row = k)` returns
# `samples_per_row * length(row_ids)` rows, organized so that for every
# d in 1..samples_per_row, rows `(d-1)*n + 1 .. d*n` correspond positionally
# to `row_ids` for any column that the sampler did NOT perturb.
#
# `tag_column` must be a feature on `task` that the sampler will not touch
# (because it's not in `feature` and, where applicable, not in the
# conditioning set used to construct the sampler).
#
# @param sampler `FeatureSampler` instance constructed against `task`.
# @param task `mlr3::Task` containing `feature` plus `tag_column`.
# @param feature character. Feature(s) being sampled.
# @param tag_column character(1). Identifier feature whose values must pass
#   through unchanged in draw-major order.
# @param samples_per_row integer(1).
# @param row_ids integer(). Defaults to all task row IDs.
expect_draw_major_row_order = function(
  sampler,
  task,
  feature,
  tag_column,
  samples_per_row,
  row_ids = task$row_ids
) {
  out = sampler$sample(feature, row_ids = row_ids, samples_per_row = samples_per_row)

  n = length(row_ids)
  testthat::expect_equal(
    nrow(out),
    n * samples_per_row,
    info = sprintf("sampler: %s", class(sampler)[[1L]])
  )

  expected_tag = rep(task$data(rows = row_ids)[[tag_column]], times = samples_per_row)
  testthat::expect_identical(
    out[[tag_column]],
    expected_tag,
    info = sprintf("draw-major tag order, sampler: %s", class(sampler)[[1L]])
  )

  invisible(out)
}

# Tests for MarginalPermutationSampler

test_that("MarginalPermutationSampler initialization works", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalPermutationSampler$new(task)

  expect_s3_class(sampler, "MarginalPermutationSampler")
  expect_s3_class(sampler, "MarginalSampler")
  expect_equal(sampler$label, "Permutation sampler")
  expect_s3_class(sampler$param_set, "ParamSet")
})

test_that("MarginalPermutationSampler sampling works", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalPermutationSampler$new(task)
  data = task$data()

  # Test single feature sampling
  sampled_data = sampler$sample("x1")

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Permuted values come from original distribution
  expect_setequal(sampled_data$x1, data$x1)

  # Other features unchanged
  expect_non_sampled_unchanged(sampled_data, data, c("x2", "x3", "x4", "x5"))
})

test_that("MarginalPermutationSampler handles multiple features", {
  task = tgen("circle", d = 5)$generate(n = 100)
  sampler = MarginalPermutationSampler$new(task)
  data = task$data()

  features = c("x1", "x2", "x3")
  sampled_data = sampler$sample(features)

  expect_sampler_output_structure(sampled_data, task, nrows = 100)
  expect_feature_type_consistency(sampled_data, task)

  # Permuted values come from original distribution
  for (feat in features) {
    expect_setequal(sampled_data[[feat]], data[[feat]])
  }

  # Non-sampled features unchanged
  expect_non_sampled_unchanged(sampled_data, data, c("x4", "x5"))
})

test_that("MarginalPermutationSampler works with different task types", {
  # Regression task
  task_regr = tgen("circle", d = 4)$generate(n = 100)
  sampler_regr = MarginalPermutationSampler$new(task_regr)
  sampled_regr = sampler_regr$sample("x1")
  expect_sampler_output_structure(sampled_regr, task_regr, nrows = 100)

  # Binary classification task
  task_classif = tsk("sonar")
  sampler_classif = MarginalPermutationSampler$new(task_classif)
  sampled_classif = sampler_classif$sample("V1")
  expect_sampler_output_structure(sampled_classif, task_classif, nrows = task_classif$nrow)

  # Multiclass classification task
  task_multi = tsk("iris")
  sampler_multi = MarginalPermutationSampler$new(task_multi)
  sampled_multi = sampler_multi$sample("Sepal.Length")
  expect_sampler_output_structure(sampled_multi, task_multi, nrows = 150)
})

test_that("MarginalPermutationSampler preserves feature types", {
  test_sampler_feature_types(MarginalPermutationSampler)
})

test_that("MarginalPermutationSampler obeys draw-major order under samples_per_row > 1", {
  set.seed(123L)
  n = 10L
  dt = data.table::data.table(
    y = rnorm(n),
    x = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = MarginalPermutationSampler$new(task)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

test_that("$sample() does not mutate the task data", {
  task = sim_dgp_independent(n = 50)
  before = task$data(rows = task$row_ids)[["important1"]]

  sampler = MarginalPermutationSampler$new(task)
  out = sampler$sample("important1", row_ids = task$row_ids)

  # task is unchanged after sampling (sampler must work on a disposable copy)
  expect_identical(task$data(rows = task$row_ids)[["important1"]], before)
  checkmate::expect_data_table(out, nrows = task$nrow)
})

test_that("$sample_newdata() does not mutate the caller's newdata", {
  task = sim_dgp_independent(n = 50)
  sampler = MarginalPermutationSampler$new(task)

  newdata = task$data()
  snapshot = data.table::copy(newdata)

  sampler$sample_newdata("important1", newdata = newdata)

  expect_identical(newdata, snapshot)
})

# Tests for ConditionalKNNSampler

test_that("ConditionalKNNSampler initialization works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalKNNSampler$new(task, k = 5L)

  expect_s3_class(sampler, "ConditionalKNNSampler")
  expect_s3_class(sampler, "ConditionalSampler")
  expect_equal(sampler$label, "k-Nearest Neighbors Conditional Sampler")
  expect_s3_class(sampler$param_set, "ParamSet")

  # Check k parameter
  expect_equal(sampler$param_set$values$k, 5L)
})

test_that("ConditionalKNNSampler works with default k", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalKNNSampler$new(task) # Default k = 5L

  expect_equal(sampler$param_set$values$k, 5L)
})

test_that("ConditionalKNNSampler sampling works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalKNNSampler$new(task, k = 5L)

  # Marginal sampling
  expect_marginal_sampling(sampler, feature = "important1", row_ids = 1:20)

  # Conditional sampling
  expect_conditional_sampling(
    sampler,
    feature = "important2",
    conditioning_set = "important1",
    row_ids = 1:20
  )

  # Multiple features
  expect_conditional_sampling(
    sampler,
    feature = c("important2", "important3"),
    conditioning_set = "important1",
    row_ids = 1:20
  )
})

test_that("ConditionalKNNSampler sample_newdata works", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalKNNSampler$new(task, k = 5L)
  test_data = task$data(rows = 1:10)

  sampled = sampler$sample_newdata(
    feature = c("important2", "important3"),
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 10)
  expect_feature_type_consistency(sampled, task)
  expect_non_sampled_unchanged(sampled, test_data, "important1")
})

test_that("ConditionalKNNSampler handles different k values", {
  task = tgen("friedman1")$generate(n = 100)
  test_data = task$data(rows = 1:10)

  # Test with k=1 (nearest neighbor)
  sampler_k1 = ConditionalKNNSampler$new(task, k = 1L)
  expect_equal(sampler_k1$param_set$values$k, 1L)

  # Test with k=20
  sampler_k20 = ConditionalKNNSampler$new(task, k = 20L)
  expect_equal(sampler_k20$param_set$values$k, 20L)

  # Both should work
  sampled_k1 = sampler_k1$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )
  sampled_k20 = sampler_k20$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled_k1, task, nrows = 10)
  expect_sampler_output_structure(sampled_k20, task, nrows = 10)
})

test_that("ConditionalKNNSampler handles k > n_train", {
  task = tgen("friedman1")$generate(n = 50)
  sampler = ConditionalKNNSampler$new(task, k = 100L)
  test_data = task$data(rows = 1:5)

  # Should still work (uses all available training data)
  sampled = sampler$sample_newdata(
    feature = "important2",
    newdata = test_data,
    conditioning_set = "important1"
  )

  expect_sampler_output_structure(sampled, task, nrows = 5)
})

test_that("ConditionalKNNSampler is reproducible with seed", {
  task = tgen("friedman1")$generate(n = 100)
  sampler = ConditionalKNNSampler$new(task, k = 5L)
  test_data = task$data(rows = 1:10)

  sampled1 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  sampled2 = withr::with_seed(123, {
    sampler$sample_newdata(
      feature = "important2",
      newdata = test_data,
      conditioning_set = "important1"
    )
  })

  expect_identical(sampled1$important2, sampled2$important2)
})

test_that("ConditionalKNNSampler conditioning_set parameter behavior", {
  task = tgen("friedman1")$generate(n = 100)
  test_conditioning_set_behavior(ConditionalKNNSampler, task, k = 5L)
})

test_that("ConditionalKNNSampler preserves feature types", {
  skip_if_not_installed("gower")
  test_sampler_feature_types(ConditionalKNNSampler, k = 5L)
})

test_that("ConditionalKNNSampler works with categorical features using Gower distance", {
  skip_if_not_installed("gower")

  task = tsk("penguins")
  data = task$data()
  sampler = ConditionalKNNSampler$new(task, k = 5L)
  test_data = task$data(rows = 1:10)

  # Sample conditioning on categorical feature
  sampled = sampler$sample_newdata(
    feature = "bill_length",
    newdata = test_data,
    conditioning_set = "island"
  )

  expect_sampler_output_structure(sampled, task, nrows = 10)
  expect_non_sampled_unchanged(sampled, test_data, "island")

  # Sampled values should come from training data
  expect_true(all(sampled$bill_length %in% data$bill_length))
})

test_that("ConditionalKNNSampler works with mixed numeric and categorical conditioning", {
  skip_if_not_installed("gower")

  task = tsk("penguins")
  sampler = ConditionalKNNSampler$new(task, k = 5L)
  test_data = task$data(rows = 1:10)

  # Sample conditioning on both numeric and factor features
  sampled = sampler$sample_newdata(
    feature = "bill_length",
    newdata = test_data,
    conditioning_set = c("island", "body_mass")
  )

  expect_sampler_output_structure(sampled, task, nrows = 10)
  expect_non_sampled_unchanged(sampled, test_data, c("island", "body_mass"))
})

test_that("ConditionalKNNSampler obeys draw-major order under samples_per_row > 1", {
  set.seed(123L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  sampler = ConditionalKNNSampler$new(task, conditioning_set = c("x2", "tag"), k = 3L)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

test_that("ConditionalKNNSampler Gower branch obeys draw-major order (samples_per_row > 1)", {
  skip_if_not_installed("gower")
  set.seed(503L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    g = factor(sample(c("a", "b", "c"), n, replace = TRUE)),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  # factor in conditioning set -> Gower distance branch
  sampler = ConditionalKNNSampler$new(task, conditioning_set = "g", k = 3L)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

test_that("ConditionalKNNSampler marginal-fallback branch obeys draw-major order (samples_per_row > 1)", {
  set.seed(617L)
  n = 20L
  dt = data.table::data.table(
    y = rnorm(n),
    x1 = rnorm(n),
    x2 = rnorm(n),
    tag = seq_len(n) + 0.5
  )
  task = mlr3::as_task_regr(dt, target = "y")
  # conditioning_set = character(0) is non-NULL so it is stored in param_set;
  # the helper's $sample() call (which passes conditioning_set = NULL) resolves via
  # resolve_param(NULL, stored = character(0), default) -> character(0), hitting the
  # length(conditioning_set) == 0 marginal-fallback branch at ConditionalKNNSampler.R:125.
  sampler = ConditionalKNNSampler$new(task, conditioning_set = character(0), k = 3L)

  expect_draw_major_row_order(
    sampler,
    task,
    feature = "x1",
    tag_column = "tag",
    samples_per_row = 4L
  )
})

#' @title Williamson's Variable Importance Measure (WVIM)
#'
#' @description
#' Base class generalizing refit-based variable importance measures.
#' Default corresponds to leaving out each feature `n_repeats = 1` times, which
#' corresponds to LOCO (Leave One Covariate Out).
#'
#' @examplesIf requireNamespace("ranger", quietly = TRUE) && requireNamespace("mlr3learners", quietly = TRUE)
#' library(mlr3)
#' library(mlr3learners)
#'
#' task <- sim_dgp_correlated(n = 500)
#'
#' # Group correlated features together, independent features separately
#' groups <- list(
#'   correlated = c("x1", "x2"),
#'   independent = c("x3", "x4")
#' )
#'
#' wvim <- WVIM$new(
#'   task = task,
#'   learner = lrn("regr.ranger", num.trees = 10),
#'   groups = groups
#' )
#' wvim$compute()
#' wvim$importance()
#' @export
#' @importFrom mlr3fselect fselect fs
#' @keywords internal
WVIM = R6Class(
  "WVIM",
  inherit = FeatureImportanceMethod,
  public = list(
    #' @field direction (`character(1)`) Either "leave-out" or "leave-in".
    direction = NULL,
    #' @field design (`logical()`) Feature selection design matrix where `TRUE` equals "left in" and `FALSE` "left out".
    #' 	Columns correspond to `task$feature_names` and the number of rows corresponds to `length(features) * n_repeats`.
    #'  The base matrix is created by [wvim_design_matrix] and then replicated `n_repeats` times before.
    design = NULL,
    #' @field instance (`FSelectInstanceBatchSingleCrit`) The `mlr3fselect` feature selection instance containing
    #'   also the archive of all evaluations, possible useful for future use. Only stored if `store_instance` is `TRUE`.
    instance = NULL,
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' @param task,learner,measure,resampling,features,groups Passed to `FeatureImportanceMethod` for construction.
    #' @param direction (`character(1)`) Either "leave-out" or "leave-in".
    #' @param label (`character(1)`) Method label.
    #' @param n_repeats (`integer(1)`: `1L`) Number of refit iterations per resampling iteration.
    #'   This may be useful for some stochastic learners, but in general it is advisable to increase resampling iterations instead of
    #'   repeatedly refitting on the same data.
    #' @param batch_size (`integer(1)`: `NULL`) Number of design points (refits) dispatched per internal
    #'   `mlr3::benchmark()` call by the `mlr3fselect::fs("design_points")` fselector. `NULL` (default) keeps
    #'   the bbotk default of one design point per call, i.e. sequential single-refit evaluation. Set to a
    #'   positive integer to batch refits so mlr3's `future`/`mirai` parallelization can spread them across
    #'   workers; a sensible value is the number of available workers/daemons.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      direction = c("leave-out", "leave-in"),
      label = "Williamson's Variable Importance Measure (WVIM)",
      n_repeats = 1L,
      batch_size = NULL
    ) {
      # Should this go in the param_set?
      direction = match.arg(direction)
      self$direction = direction
      checkmate::assert_int(n_repeats, lower = 1L)
      checkmate::assert_int(batch_size, lower = 1L, null.ok = TRUE)
      private$.batch_size = batch_size

      # params
      ps = ps(
        relation = paradox::p_fct(c("difference", "ratio"), default = "difference"),
        n_repeats = paradox::p_int(lower = 1L, default = 1L)
      )
      ps$values = list(
        relation = "difference",
        n_repeats = n_repeats
      )

      # If learner is already trained, warn that the model will be ignored
      # Note that resample() clones the learner anyway later.
      if (!is.null(learner$model)) {
        cli::cli_warn(c(
          "{.cls Learner} is already trained, which is not compatible with refit-based importance methods",
          "!" = "Given {.cls Learner} will be cloned.",
          "i" = "Provide an untrained {.cls Learner} to avoid this warning."
        ))
      }

      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        groups = groups,
        param_set = ps,
        label = label
      )

      # Add Lei et al. (2018) inference to CI methods registry
      private$.ci_methods = c(private$.ci_methods, "lei")
    },

    #' @description
    #' Get aggregated importance scores.
    #' Extends the base `$importance()` method to support `ci_method = "lei"`.
    #'
    #' This implements distribution-free inference based on Lei et al. (2018),
    #' testing observation-wise loss differences using the Wilcoxon signed-rank test by default.
    #'
    #' Lei et al. (2018) proposed this method specifically for LOCO with L1 (absolute) loss,
    #' median aggregation, and a single train/test split (holdout). The inference is
    #' conditional on the training data, requiring i.i.d. test observations from a single
    #' split. While the aggregation function, statistical test, and resampling strategy are
    #' parameterizable, deviating from these defaults may invalidate the theoretical guarantees.
    #'
    #' For a comprehensive overview of inference methods, see
    #' `vignette("inference", package = "xplainfi")`.
    #'
    #' @param relation (`character(1)`) How to relate perturbed scores to originals ("difference" or "ratio"). If `NULL`, uses stored parameter value.
    #' @param standardize (`logical(1)`: `FALSE`) If `TRUE`, importances are standardized by the highest score so all scores fall in `[-1, 1]`.
    #' @param ci_method (`character(1)`: `"none"`) Variance estimation method. In addition to base methods (`"none"`, `"raw"`, `"nadeau_bengio"`, `"quantile"`),
    #'   WVIM methods support `"lei"` for distribution-free inference (Lei et al., 2018).
    #' @param conf_level (`numeric(1)`: `0.95`) Confidence level to use for confidence interval construction when `ci_method != "none"`.
    #' @param alternative (`character(1)`: `"two.sided"`) Type of alternative hypothesis for statistical tests.
    #'   `"greater"` tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
    #'   `"two.sided"` tests H0: importance = 0 vs H1: importance != 0.
    #' @param test (`character(1)`: `"wilcoxon"`) Test to use for Lei et al. inference. One of `"wilcoxon"`, `"t"`, `"fisher"`, or `"binomial"`.
    #'   Only used when `ci_method = "lei"`.
    #' @param B (`integer(1)`: `1999`) Number of replications for Fisher test. Only used when `ci_method = "lei"` and `test = "fisher"`.
    #' @param aggregator (`function`: `stats::median`) Aggregation function for computing the point estimate from observation-wise
    #'   importance values. Defaults to `stats::median` as proposed by Lei et al. (2018). Only used when `ci_method = "lei"`.
    #' @param p_adjust (`character(1)`: `"none"`) Method for p-value adjustment for multiple comparisons.
    #'   Accepts any method supported by [stats::p.adjust.methods], e.g. `"holm"`, `"bonferroni"`, `"BH"`, `"none"`.
    #'   When `"bonferroni"`, confidence intervals are also adjusted (alpha/k).
    #'   For other correction methods (e.g. `"holm"`, `"BH"`), only p-values are adjusted;
    #'   confidence intervals remain at the nominal `conf_level` because these sequential/adaptive
    #'   procedures do not have a clean per-comparison alpha for CI construction.
    #' @param ... Additional arguments passed to the base method.
    #' @return ([data.table][data.table::data.table]) Aggregated importance scores.
    #' @references `r print_bib("lei_2018")`
    importance = function(
      relation = NULL,
      standardize = FALSE,
      ci_method = c("none", "raw", "nadeau_bengio", "quantile", "lei"),
      conf_level = 0.95,
      alternative = c("two.sided", "greater"),
      test = c("wilcoxon", "t", "fisher", "binomial"),
      B = 1999,
      aggregator = NULL,
      p_adjust = "none",
      ...
    ) {
      if (length(ci_method) > 1) {
        ci_method = ci_method[1]
      }
      alternative = match.arg(alternative)

      if (ci_method == "lei") {
        if (is.null(private$.scores)) {
          cli::cli_inform(c(x = "No importances computed yet!"))
          return(invisible(NULL))
        }

        # Default aggregator for Lei et al. is median
        aggregator = aggregator %||% stats::median

        test = match.arg(test)
        agg_importance = importance_loco(
          conf_level = conf_level,
          alternative = alternative,
          test = test,
          aggregator = aggregator,
          p_adjust = p_adjust,
          B = B,
          method_obj = self
        )

        if (standardize) {
          agg_importance[, importance := importance / max(abs(importance), na.rm = TRUE)]
        }

        setkeyv(agg_importance, "feature")
        return(agg_importance[])
      }

      # Default to parent method for other ci_methods
      super$importance(
        relation = relation,
        standardize = standardize,
        ci_method = ci_method,
        conf_level = conf_level,
        alternative = alternative,
        p_adjust = p_adjust,
        ...
      )
    },

    #' @description
    #' Computes leave-out or leave-in feature importance.
    # @param design (`data.table`) A design matrix indicating which features are to be left in or out
    # at each step. Can be created with [wvim_design_matrix].
    #' `wvim_design_matrix(task$feature_names, "leave-out")` corresponds to LOCO.
    #' @param store_models,store_backends (`logical(1)`: `TRUE`) Whether to store fitted models / data backends, passed to [mlr3::resample] internally
    #'   backends in resample result.
    #'   Required for some measures, but may increase memory footprint.
    #' @param store_instance (`logical(1)`: `FALSE`) Whether to store the [mlr3fselect] instance in `$instance`.
    compute = function(store_models = TRUE, store_backends = TRUE, store_instance = FALSE) {
      if (is.null(self$groups)) {
        feature_or_groups = self$features
      } else {
        feature_or_groups = self$groups
      }

      design = wvim_design_matrix(
        all_features = self$task$feature_names,
        feature_names = feature_or_groups,
        direction = self$direction
      )

      design = data.table::rbindlist(replicate(
        n = self$param_set$values$n_repeats,
        design,
        simplify = FALSE
      ))
      private$.compute_wvim(
        design = design,
        store_models = store_models,
        store_backends = store_backends,
        store_instance = store_instance
      )
    }
  ),

  private = list(
    # Number of design points per internal benchmark() call; NULL = bbotk default (1).
    # Forwarded to fs("design_points", batch_size=) so future/mirai can parallelize refits.
    .batch_size = NULL,

    # Compute baseline scores for WVIM, which depend on the direction
    # "leave-out" -> baseline is the "full" model with all features in the task used
    # "leave-in" -> baseline is the "empty" model, so we swap in the featureless learner
    .compute_baseline = function(store_models = TRUE, store_backends = TRUE) {
      # Correct featureless learner depends on task type, there's no "surv.featureless" though
      learner = switch(
        self$direction,
        "leave-out" = self$learner,
        "leave-in" = lrn(paste(self$task$task_type, "featureless", sep = "."))
      )
      # Initial resampling
      self$resample_result = resample(
        self$task,
        learner,
        self$resampling,
        store_models = store_models,
        store_backends = store_backends
      )
      # Prepare baseline scores
      scores_baseline = self$resample_result$score(self$measure)[,
        .SD,
        .SDcols = c("iteration", self$measure$id)
      ]
      setnames(scores_baseline, old = self$measure$id, "score_baseline")
      setnames(scores_baseline, old = "iteration", "iter_rsmp")
      scores_baseline[]
    },

    .compute_wvim = function(
      design,
      store_models = TRUE,
      store_backends = TRUE,
      store_instance = FALSE
    ) {
      scores_baseline = private$.compute_baseline(store_backends = store_backends)

      # Fselect section.
      # batch_size is set post-hoc rather than passed to fs(): the fselector's
      # paradox ParamSet rejects an explicit NULL value, but NULL is exactly the
      # "use bbotk default (1 design point per benchmark call)" sentinel we want.
      fselector = mlr3fselect::fs("design_points", design = design)
      if (!is.null(private$.batch_size)) {
        fselector$param_set$values$batch_size = private$.batch_size
      }
      instance = mlr3fselect::fselect(
        fselector = fselector,
        task = self$task,
        learner = self$learner,
        resampling = self$resampling,
        measures = self$measure,
        store_models = store_models
      )

      archive_dt = as.data.table(instance$archive)
      # NOTE: `as.data.table(archive)$batch_nr` is the bbotk *batch* index
      # (one value per `eval_batch` chunk of `batch_size` design points), not
      # a per-design-point id. Keying on it joins N design points to the same
      # group when `batch_size > 1` (data.table cartesian blow-up / wrong
      # scores). The resample-result `uhash` is unique per evaluated design
      # point and matches the benchmark result, regardless of `batch_size`.
      archive_base = data.table::data.table(
        uhash = vapply(archive_dt$resample_result, function(rr) rr$uhash, character(1)),
        # Match features in fselect instance to features left in or out, and
        # assign group names to "feature" var for consistency
        feature = private$.foi_chr(archive_dt$features)
      )

      # Only keep instance if requested, otherwise would just increase memory footprint
      if (store_instance) {
        self$instance = instance
      }

      # Scores and predictions ---
      scores = instance$archive$benchmark_result$score(self$measure)
      setnames(scores, self$measure$id, "score_post")
      setnames(scores, "iteration", "iter_rsmp")
      # `nr` = per-design-point resample-result index in the combined
      # benchmark result; kept (not renamed) to map obs_loss below.

      # merge baseline scores and post-modification scores
      scores = scores[scores_baseline, on = "iter_rsmp"]
      # attach the feature-of-interest per design point via the unique uhash
      scores = archive_base[scores, on = "uhash"]
      # Regain n_repeats: enumerate the n_repeats replicates of each
      # (resampling iter, feature) in evaluation (= design-row) order.
      scores[, iter_repeat := seq_len(.N), by = c("iter_rsmp", "feature")]

      # Sanity check for iter refit
      #scores[, .(iter_rsmp, nr, iter_repeat, feature)][feature == "important1"]
      private$.scores = scores[, .(feature, iter_rsmp, iter_repeat, score_baseline, score_post)]

      # Extract prediction for storage
      # self$predictions needs: iter_rsmp, iter_repeat/refit, feature, prediction
      predictions = copy(scores)[, .(iter_rsmp, iter_repeat, feature, prediction_test)]
      setnames(predictions, "prediction_test", "prediction")
      setkeyv(predictions, cols = c("feature", "iter_rsmp"))
      self$predictions = predictions[, .(iter_rsmp, iter_repeat, feature, prediction)]

      # obs losses ----
      if (has_obs_loss(self$measure)) {
        obs_loss_vals = instance$archive$benchmark_result$obs_loss(self$measure)
        # `obs_loss()$resample_result` is the per-design-point index, equal
        # to the benchmark score `nr`. Map it to the design point's feature
        # and repeat via `scores`, so iter_repeat stays consistent with
        # `$scores()` (avoids the old `%% n_repeats` reconstruction, which
        # also broke under `batch_size > 1`).
        dp_map = unique(scores[, .(nr, feature, iter_repeat)])
        obs_loss_vals = dp_map[obs_loss_vals, on = c(nr = "resample_result")]

        setnames(obs_loss_vals, "iteration", "iter_rsmp")
        setnames(obs_loss_vals, self$measure$id, "loss_post")

        private$.obs_losses = obs_loss_vals[, .(
          feature,
          iter_rsmp,
          iter_repeat,
          row_ids,
          loss_post
        )]
      }
    },

    # This is hackier than I'd like but oh well, it gets the job done.
    # @param feature (list()) List-column of feature names in fselect instance
    # Goal is to create a column for the feature of interest OR group name
    # for the features of interest. Input is always list of features "left in",
    # so we need to flip (setdiff) for the other direction.
    # And for grouped features we do extra shenanigans to get the group names
    .foi_chr = function(features) {
      if (is.null(self$groups)) {
        switch(
          self$direction,
          "leave-in" = unlist(features),
          "leave-out" = vapply(
            features,
            \(x) {
              setdiff(self$task$feature_names, x)
            },
            character(1)
          )
        )
      } else {
        groups_tbl = private$.groups_tbl()
        vapply(
          features,
          \(x) {
            if (self$direction == "leave-out") {
              x = setdiff(self$task$feature_names, x)
            }
            feature_chr = paste0(x, collapse = ";")
            groups_tbl[feature_chr == groups_tbl$features, group]
          },
          character(1)
        )
      }
    }
  )
)

#' @title Leave-One-Covariate-Out (LOCO)
#'
#' @description
#' Calculates Leave-One-Covariate-Out (LOCO) scores.
#'
#' @details
#' LOCO measures feature importance by comparing model performance with and without
#' each feature. For each feature, the model is retrained without that feature and
#' the performance difference (reduced_model_loss - full_model_loss) indicates the
#' feature's importance. Higher values indicate more important features.
#'
#' @examples
#' library(mlr3)
#' library(mlr3learners)
#'
#' task <- sim_dgp_correlated(n = 500)
#'
#' loco <- LOCO$new(
#'   task = task,
#'   learner = lrn("regr.rpart"),
#'   measure = msr("regr.mse")
#' )
#' loco$compute()
#' loco$importance()
#' @export
#'
#' @references `r print_bib("lei_2018")`
LOCO = R6Class(
  "LOCO",
  inherit = WVIM,
  public = list(
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' @param task ([mlr3::Task]) Task to compute importance for.
    #' @param learner ([mlr3::Learner]) Learner to use for prediction.
    #' @param measure ([mlr3::Measure]: `NULL`) Measure to use for scoring. Defaults to
    #'   `classif.ce` for classification and `regr.mse` for regression.
    #' @param resampling ([mlr3::Resampling]) Resampling strategy. Defaults to holdout.
    #' @param features (`character()`) Features to compute importance for. Defaults to all features.
    #' @param n_repeats (`integer(1)`: `1L`) Number of refit iterations per resampling iteration.
    #'   This may be useful for some stochastic learners, but in general it is advisable to increase resampling iterations instead of
    #'   repeatedly refitting on the same data.
    #' @param batch_size (`integer(1)`: `NULL`) Number of refits dispatched per internal
    #'   `mlr3::benchmark()` call. `NULL` (default) keeps sequential single-refit evaluation; set to a
    #'   positive integer (e.g. the number of `future`/`mirai` workers) to enable parallel refits.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      n_repeats = 1L,
      batch_size = NULL
    ) {
      if (!is.null(features)) {
        # LOCO specifically does not "allow" grouped features
        checkmate::assert_character(features)
        checkmate::assert_subset(features, choices = task$feature_names)
      }

      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        direction = "leave-out",
        label = "Leave-One-Covariate-Out (LOCO)",
        n_repeats = n_repeats,
        batch_size = batch_size
      )
    },

    #' @description
    #' Compute LOCO importances.
    #'
    #' @param store_models,store_backends (`logical(1)`: `TRUE`) Whether to store fitted models / data backends, passed to [mlr3::resample] internally
    compute = function(store_models = TRUE, store_backends = TRUE) {
      design = wvim_design_matrix(
        all_features = self$task$feature_names,
        feature_names = self$features,
        direction = self$direction
      )

      design = data.table::rbindlist(replicate(
        n = self$param_set$values$n_repeats,
        design,
        simplify = FALSE
      ))

      private$.compute_wvim(design, store_models = store_models, store_backends = store_backends)
    }
  )
)

# =============================================================================
# CFI Tests using higher-level test helpers
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality
# -----------------------------------------------------------------------------

test_that("CFI default behavior with minimal parameters", {
  skip_if_not_installed("arf")

  test_default_behavior(CFI, task_type = "regr", n_repeats = 1L)
})

test_that("CFI basic workflow with classification", {
  task = tgen("2dnormals")$generate(n = 100)

  cfi = CFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )
  checkmate::expect_r6(cfi, c("FeatureImportanceMethod", "PerturbationImportance", "CFI"))

  cfi$compute()
  expect_method_output(cfi)
})

test_that("CFI uses ConditionalARFSampler by default", {
  skip_if_not_installed("arf")

  task = tgen("xor")$generate(n = 100)

  cfi = CFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce")
  )

  checkmate::expect_r6(cfi$sampler, "ConditionalARFSampler")
  expect_equal(cfi$label, "Conditional Feature Importance")
})

test_that("CFI featureless learner produces zero importance", {
  skip_if_not_installed("arf")

  test_featureless_zero_importance(CFI, task_type = "classif", n_repeats = 1L)
})

# -----------------------------------------------------------------------------
# Repeats and scores
# -----------------------------------------------------------------------------

test_that("CFI multiple repeats and scores structure", {
  task = tgen("friedman1")$generate(n = 100)

  test_n_repeats_and_scores(
    CFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    n_repeats = 2L,
    sampler = ConditionalGaussianSampler$new(task)
  )
})

test_that("CFI single feature", {
  task = tgen("friedman1")$generate(n = 100)

  test_single_feature(
    CFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    feature = "important4",
    n_repeats = 2L,
    sampler = ConditionalGaussianSampler$new(task)
  )
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("CFI friedman1 produces sensible ranking", {
  skip_if_not_installed("arf")

  test_friedman1_sensible_ranking(
    CFI,
    n_repeats = 5L
  )
})

# -----------------------------------------------------------------------------
# Resampling
# -----------------------------------------------------------------------------

test_that("CFI with resampling", {
  task = tgen("2dnormals")$generate(n = 100)

  cfi = CFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    resampling = rsmp("cv", folds = 3),
    n_repeats = 2L,
    sampler = ConditionalGaussianSampler$new(task)
  )
  cfi$compute()
  expect_method_output(cfi)
})

# -----------------------------------------------------------------------------
# Parameter validation
# -----------------------------------------------------------------------------

test_that("CFI parameter validation", {
  task = tgen("2dnormals")$generate(n = 50)

  test_parameter_validation(
    CFI,
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    sampler = ConditionalGaussianSampler$new(task)
  )
})

# -----------------------------------------------------------------------------
# Grouped importance
# -----------------------------------------------------------------------------

test_that("CFI with feature groups", {
  task = tgen("friedman1")$generate(n = 100)

  groups = list(
    important_features = c("important1", "important2", "important3"),
    unimportant_features = c("unimportant1", "unimportant2", "unimportant3")
  )

  test_grouped_importance(
    CFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    groups = groups,
    expected_classes = c("FeatureImportanceMethod", "PerturbationImportance", "CFI"),
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Custom samplers
# -----------------------------------------------------------------------------

test_that("CFI with custom ARF sampler", {
  skip_if_not_installed("arf")

  task = tgen("spirals")$generate(n = 100)

  cfi = CFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    sampler = ConditionalARFSampler$new(task),
    n_repeats = 1L
  )
  checkmate::expect_r6(cfi$sampler, "ConditionalARFSampler")
  cfi$compute()
  expect_importance_dt(cfi$importance(), features = cfi$features)
})

test_that("CFI with KnockoffSampler and KnockoffGaussianSampler", {
  skip_if_not_installed("knockoff")

  task = tgen("friedman1")$generate(n = 150)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  # Test with KnockoffSampler
  cfi_ko = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    sampler = KnockoffSampler$new(task),
    n_repeats = 1L
  )
  checkmate::expect_r6(cfi_ko$sampler, "KnockoffSampler")
  cfi_ko$compute()
  expect_importance_dt(cfi_ko$importance(), features = cfi_ko$features)

  # Test with KnockoffGaussianSampler
  cfi_gko = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    sampler = KnockoffGaussianSampler$new(task),
    n_repeats = 1L
  )
  checkmate::expect_r6(cfi_gko$sampler, "KnockoffGaussianSampler")
  cfi_gko$compute()
  expect_importance_dt(cfi_gko$importance(), features = cfi_gko$features)
})

# -----------------------------------------------------------------------------
# CFI-specific: CPI variance method
# -----------------------------------------------------------------------------

test_that("CFI importance is approximately equivalent at seed 42 across the samples_per_row refactor", {
  skip_if_not_installed("arf")
  skip_if_not_installed("ranger")
  skip_if_not_installed("mlr3learners")

  # Regression guard for the samples_per_row refactor (sampling-repeat-efficiency
  # branch vs main). The two code paths differ in RNG consumption order:
  #   - old: arf::forge(n_synth = 1) on evidence replicated n_repeats times
  #   - new: arf::forge(n_synth = n_repeats) on unique evidence rows
  # So values are not bitwise identical but are draws from the same distribution.
  #
  # Reference values captured via cross::run_branches() with n_repeats = 50.
  # Max absolute diff observed across features: 0.0804 (on x1, the largest-importance
  # signal feature, ~2.4% relative). Signal features (x1, x3) agreed within ~3%.
  # Noise features (x2, x4) agreed within ~9e-4 absolute.
  # Tolerance = max(0.01, 3 * max_abs_diff) ~= 0.241 gives headroom for run-to-run
  # variation if the test ever executes on a different RNG implementation.
  #
  # To regenerate after an R / arf / ranger toolchain bump: run this test body via
  # `cross::run_branches(branches = c("main", "<this-branch>"), { ... })` and paste
  # the current-branch importance vector below; recompute the tolerance from the
  # max observed absolute diff.
  expected_features = c("x1", "x2", "x3", "x4")
  expected_importance = c(
    3.39659680872389,
    -0.00555866600268278,
    1.39995939734368,
    0.00696214631206585
  )

  set.seed(42L)
  task = sim_dgp_correlated(n = 200L, r = 0.7)
  cfi = CFI$new(
    task = task,
    learner = lrn("regr.ranger", num.trees = 100L),
    measure = msr("regr.mse"),
    sampler = ConditionalARFSampler$new(task, verbose = FALSE),
    n_repeats = 50L
  )
  cfi$compute()
  got = cfi$importance()

  expect_equal(got$feature, expected_features)
  expect_equal(got$importance, expected_importance, tolerance = 0.241)
})

test_that("CFI with CPI variance method using KnockoffGaussianSampler", {
  skip_if_not_installed("ranger")
  skip_if_not_installed("mlr3learners")
  skip_if_not_installed("knockoff")

  # Use n=500 for stable p-value comparisons
  task = sim_dgp_correlated(n = 500, r = 0.7)
  learner = lrn("regr.ranger", num.trees = 100)
  measure = msr("regr.mse")
  resampling = rsmp("holdout")

  gaussian_sampler = KnockoffGaussianSampler$new(task)
  cfi = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    sampler = gaussian_sampler,
    n_repeats = 1L
  )

  # Check that CPI is in the variance methods registry
  expect_contains(cfi$.__enclos_env__$private$.ci_methods, "cpi")

  cfi$compute()

  # Test CPI variance method
  cpi_result = cfi$importance(ci_method = "cpi")

  # Check structure
  expect_importance_dt(cpi_result, features = cfi$features)
  expected_cols = c(
    "feature",
    "importance",
    "se",
    "statistic",
    "p.value",
    "conf_lower",
    "conf_upper"
  )
  expect_contains(names(cpi_result), expected_cols)

  # Check that all values are finite (except conf_upper which is Inf for one-sided)
  checkmate::expect_numeric(cpi_result$importance, finite = TRUE)
  checkmate::expect_numeric(cpi_result$se, finite = TRUE, lower = 0)
  checkmate::expect_numeric(cpi_result$statistic, finite = TRUE)
  checkmate::expect_numeric(cpi_result$p.value, finite = TRUE, lower = 0, upper = 1)
  checkmate::expect_numeric(cpi_result$conf_lower, finite = TRUE)
  checkmate::expect_numeric(cpi_result$conf_upper, finite = TRUE)

  # For correlated DGP, important features should have smaller p-values
  important_pvals = cpi_result[feature %in% c("x1", "x3")]$p.value
  unimportant_pvals = cpi_result[feature %in% c("x2", "x4")]$p.value
  expect_lt(mean(important_pvals), mean(unimportant_pvals))
})

test_that("CFI with CPI and p_adjust = 'BH' adjusts p-values", {
  skip_if_not_installed("knockoff")

  task = sim_dgp_correlated(n = 200, r = 0.7)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")
  resampling = rsmp("holdout")

  gaussian_sampler = KnockoffGaussianSampler$new(task)
  cfi = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    sampler = gaussian_sampler,
    n_repeats = 1L
  )
  cfi$compute()

  cpi_none = cfi$importance(ci_method = "cpi")
  cpi_bh = cfi$importance(ci_method = "cpi", p_adjust = "BH")

  # Point estimates should be identical
  expect_equal(cpi_none$importance, cpi_bh$importance)

  # BH-adjusted p-values should be >= unadjusted
  expect_true(all(cpi_bh$p.value >= cpi_none$p.value - 1e-10))

  # CIs should be unchanged (BH only adjusts p-values, not CIs)
  expect_equal(cpi_none$conf_lower, cpi_bh$conf_lower)
  expect_equal(cpi_none$conf_upper, cpi_bh$conf_upper)
})

test_that("CPI with non-t test reports se = NA", {
  skip_if_not_installed("knockoff")

  task = sim_dgp_correlated(n = 200, r = 0.7)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")
  resampling = rsmp("holdout")

  gaussian_sampler = KnockoffGaussianSampler$new(task)
  cfi = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    sampler = gaussian_sampler,
    n_repeats = 1L
  )
  cfi$compute()

  # Default test is t-test, se should be finite
  cpi_t = cfi$importance(ci_method = "cpi", test = "t")
  checkmate::expect_numeric(cpi_t$se, finite = TRUE, lower = 0)

  # Wilcoxon test: se should be NA
  cpi_wilcox = suppressWarnings(cfi$importance(ci_method = "cpi", test = "wilcoxon"))
  expect_true(all(is.na(cpi_wilcox$se)))
})

test_that("CFI with CPI warning on problematic resampling", {
  skip_if_not_installed("knockoff")

  task = sim_dgp_correlated(n = 50)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")
  resampling = rsmp("subsampling", repeats = 5)

  gaussian_sampler = KnockoffGaussianSampler$new(task)
  cfi = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    sampler = gaussian_sampler,
    n_repeats = 1L
  )
  cfi$compute()

  # Should warn about non-holdout resampling
  expect_warning(
    cpi_result <- cfi$importance(ci_method = "cpi"),
    regexp = "single test set"
  )

  expect_importance_dt(cpi_result, features = cfi$features)

  # CV also warns (overlapping training data)
  cfi_cv = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = rsmp("cv", folds = 5),
    sampler = gaussian_sampler,
    n_repeats = 1L
  )
  cfi_cv$compute()
  expect_warning(
    cfi_cv$importance(ci_method = "cpi"),
    regexp = "single test set"
  )

  # With holdout (single split), should be silent
  cfi_holdout = CFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = rsmp("holdout"),
    sampler = gaussian_sampler,
    n_repeats = 1L
  )
  cfi_holdout$compute()
  expect_silent(cfi_holdout$importance(ci_method = "cpi"))
})

# =============================================================================
# RFI Tests using higher-level test helpers
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality
# -----------------------------------------------------------------------------

test_that("RFI default behavior with minimal parameters", {
  skip_if_not_installed("arf")

  # RFI warns when conditioning_set is not specified (defaults to empty)
  expect_warning(
    test_default_behavior(RFI, task_type = "regr", n_repeats = 1L)
  )
})

test_that("RFI basic workflow with classification", {
  task = tgen("2dnormals")$generate(n = 100)

  rfi = RFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = "x2",
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )
  checkmate::expect_r6(rfi, c("FeatureImportanceMethod", "PerturbationImportance", "RFI"))

  rfi$compute()
  expect_method_output(rfi)
})

test_that("RFI uses ConditionalARFSampler by default", {
  skip_if_not_installed("arf")

  task = tgen("xor")$generate(n = 100)

  rfi = RFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = "x2"
  )

  checkmate::expect_r6(rfi$sampler, "ConditionalARFSampler")
  expect_equal(rfi$label, "Relative Feature Importance")
})

test_that("RFI featureless learner produces zero importance", {
  skip_if_not_installed("arf")

  test_featureless_zero_importance(
    RFI,
    task_type = "classif",
    conditioning_set = "x1",
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Repeats and scores
# -----------------------------------------------------------------------------

test_that("RFI multiple repeats and scores structure", {
  task = tgen("friedman1")$generate(n = 100)

  test_n_repeats_and_scores(
    RFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    n_repeats = 2L,
    conditioning_set = "important1",
    sampler = ConditionalGaussianSampler$new(task)
  )
})

test_that("RFI single feature", {
  task = tgen("friedman1")$generate(n = 100)

  test_single_feature(
    RFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    feature = "important4",
    n_repeats = 2L,
    conditioning_set = c("important1", "important2"),
    sampler = ConditionalGaussianSampler$new(task)
  )
})

# -----------------------------------------------------------------------------
# Relation parameter
# -----------------------------------------------------------------------------

test_that("RFI difference vs ratio relations", {
  # Seed needed: 33-row holdout + rpart on 2dnormals can yield a zero
  # baseline classif.ce, which makes the ratio relation 0/0 = NaN.
  # Flaky under covr where the RNG state differs from R CMD check.
  withr::local_seed(42)
  task = tgen("2dnormals")$generate(n = 100)

  test_relation_parameter(
    RFI,
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = character(0),
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("RFI friedman1 produces sensible ranking", {
  skip_if_not_installed("arf")

  test_friedman1_sensible_ranking(
    RFI,
    conditioning_set = "important1",
    n_repeats = 5L
  )
})

# -----------------------------------------------------------------------------
# Parameter validation
# -----------------------------------------------------------------------------

test_that("RFI parameter validation", {
  task = tgen("2dnormals")$generate(n = 50)

  test_parameter_validation(
    RFI,
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = "x1",
    sampler = ConditionalGaussianSampler$new(task)
  )
})

test_that("RFI rejects invalid conditioning_set", {
  task = tgen("2dnormals")$generate(n = 50)

  expect_error(RFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = c("nonexistent_feature"),
    sampler = ConditionalGaussianSampler$new(task)
  ))
})

# -----------------------------------------------------------------------------
# Grouped importance
# -----------------------------------------------------------------------------

test_that("RFI with feature groups", {
  task = tgen("friedman1")$generate(n = 100)

  groups = list(
    early_important = c("important1", "important2"),
    late_important = c("important3", "important4"),
    noise = c("unimportant1", "unimportant2")
  )

  test_grouped_importance(
    RFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    groups = groups,
    expected_classes = c("FeatureImportanceMethod", "PerturbationImportance", "RFI"),
    conditioning_set = "important5",
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Custom samplers
# -----------------------------------------------------------------------------

test_that("RFI with custom ARF sampler", {
  skip_if_not_installed("arf")

  task = tgen("spirals")$generate(n = 100)

  rfi = RFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    sampler = ConditionalARFSampler$new(task),
    conditioning_set = "x1",
    n_repeats = 1L
  )
  checkmate::expect_r6(rfi$sampler, "ConditionalARFSampler")
  rfi$compute()
  expect_importance_dt(rfi$importance(), features = rfi$features)
})

# -----------------------------------------------------------------------------
# RFI-specific: conditioning_set behavior
# -----------------------------------------------------------------------------

test_that("RFI with custom conditioning set", {
  task = tgen("friedman1")$generate(n = 100)
  conditioning_set = c("important1", "important2")

  rfi = RFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    conditioning_set = conditioning_set,
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )

  expect_identical(rfi$param_set$values$conditioning_set, conditioning_set)

  rfi$compute()
  expect_importance_dt(rfi$importance(), features = rfi$features)
})

test_that("RFI with empty conditioning set (equivalent to PFI)", {
  task = tgen("friedman1")$generate(n = 100)

  # RFI with empty conditioning set warns
  expect_warning({
    rfi = RFI$new(
      task = task,
      learner = lrn("regr.rpart"),
      measure = msr("regr.mse"),
      sampler = ConditionalGaussianSampler$new(task),
      n_repeats = 1L
    )
  })

  expect_length(rfi$param_set$values$conditioning_set, 0)

  rfi$compute()
  expect_importance_dt(rfi$importance(), features = rfi$features)
})

test_that("RFI with single conditioning feature", {
  task = tgen("2dnormals")$generate(n = 100)

  rfi = RFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    conditioning_set = "x1",
    sampler = ConditionalGaussianSampler$new(task),
    n_repeats = 1L
  )

  expect_length(rfi$param_set$values$conditioning_set, 1)
  expect_equal(rfi$param_set$values$conditioning_set, "x1")

  rfi$compute()
  expect_importance_dt(rfi$importance(), features = rfi$features)
})

test_that("RFI different conditioning sets produce different results", {
  task = tgen("friedman1")$generate(n = 100)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")
  sampler = ConditionalGaussianSampler$new(task)

  # RFI with empty conditioning set
  rfi_empty = RFI$new(
    task = task,
    learner = learner,
    measure = measure,
    conditioning_set = character(0),
    sampler = sampler,
    n_repeats = 2L
  )

  # RFI with one conditioning feature
  rfi_one = RFI$new(
    task = task,
    learner = learner,
    measure = measure,
    conditioning_set = "important1",
    sampler = sampler,
    n_repeats = 2L
  )

  # RFI with multiple conditioning features
  rfi_multi = RFI$new(
    task = task,
    learner = learner,
    measure = measure,
    conditioning_set = c("important1", "important2"),
    sampler = sampler,
    n_repeats = 2L
  )

  rfi_empty$compute()
  rfi_one$compute()
  rfi_multi$compute()

  result_empty = rfi_empty$importance()
  result_one = rfi_one$importance()
  result_multi = rfi_multi$importance()

  # All should be valid importance tables
  expect_importance_dt(result_empty, features = rfi_empty$features)
  expect_importance_dt(result_one, features = rfi_one$features)
  expect_importance_dt(result_multi, features = rfi_multi$features)

  # Results should generally be different
  expect_false(all(abs(result_empty$importance - result_one$importance) < 1e-10))
  expect_false(all(abs(result_one$importance - result_multi$importance) < 1e-10))
})

# =============================================================================
# Tests for pretrained learner support
# =============================================================================

# -----------------------------------------------------------------------------
# assert_pretrained() unit tests
# -----------------------------------------------------------------------------

test_that("assert_pretrained returns FALSE for untrained learner", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  resampling = rsmp("holdout")$instantiate(task)

  expect_false(assert_pretrained(learner, task, resampling))
})

test_that("assert_pretrained returns TRUE for valid pretrained setup", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  resampling = rsmp("holdout")$instantiate(task)

  expect_true(assert_pretrained(learner, task, resampling))
})

test_that("assert_pretrained errors when trained learner has multi-fold resampling", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  resampling = rsmp("cv", folds = 3)$instantiate(task)

  expect_error(assert_pretrained(learner, task, resampling), "not compatible")
})

# -----------------------------------------------------------------------------
# PFI with pretrained learner
# -----------------------------------------------------------------------------

test_that("PFI works with pretrained learner", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("regr.mse")

  learner$train(task)

  pfi = PFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_repeats = 3L
  )

  pfi$compute()
  expect_importance_dt(pfi$importance(), features = pfi$features)
  checkmate::expect_r6(pfi$resample_result, "ResampleResult")

  # Original learner must still have its model after compute
  expect_false(is.null(learner$model))
})

test_that("PFI pretrained and non-pretrained produce comparable results", {
  set.seed(123)
  task = tsk("mtcars")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("regr.mse")

  # Pretrained path
  learner_pre = lrn("regr.rpart")
  learner_pre$train(task)

  set.seed(1)
  pfi_pre = PFI$new(
    task = task,
    learner = learner_pre,
    measure = measure,
    resampling = resampling,
    n_repeats = 5L
  )
  pfi_pre$compute()

  # Non-pretrained path (fresh learner, same resampling)
  set.seed(1)
  pfi_fresh = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = measure,
    resampling = resampling,
    n_repeats = 5L
  )
  pfi_fresh$compute()

  # Both should produce valid importance tables with same features
  expect_importance_dt(pfi_pre$importance(), features = pfi_pre$features)
  expect_importance_dt(pfi_fresh$importance(), features = pfi_fresh$features)
  expect_setequal(pfi_pre$importance()$feature, pfi_fresh$importance()$feature)
})

test_that("PFI with pretrained learner errors on non-instantiated resampling", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  measure = msr("regr.mse")

  expect_error(
    PFI$new(
      task = task,
      learner = learner,
      measure = measure,
      resampling = rsmp("holdout")
    ),
    "instantiated"
  )
})

test_that("PFI with pretrained learner errors on multi-fold resampling at construction", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  measure = msr("regr.mse")

  expect_error(
    PFI$new(
      task = task,
      learner = learner,
      measure = measure,
      resampling = rsmp("cv", folds = 3)$instantiate(task),
      n_repeats = 2L
    ),
    "not compatible"
  )
})

test_that("PFI with pretrained learner works for classification", {
  task = tsk("penguins")
  learner = lrn("classif.rpart", predict_type = "prob")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("classif.ce")

  learner$train(task)

  pfi = PFI$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_repeats = 3L
  )

  pfi$compute()
  expect_importance_dt(pfi$importance(), features = pfi$features)
})

# -----------------------------------------------------------------------------
# MarginalSAGE with pretrained learner
# -----------------------------------------------------------------------------

test_that("MarginalSAGE works with pretrained learner", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("regr.mse")

  learner$train(task)

  sage = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_permutations = 2L,
    n_samples = 20L
  )

  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
  checkmate::expect_r6(sage$resample_result, "ResampleResult")
})

test_that("MarginalSAGE pretrained and non-pretrained produce comparable results", {
  set.seed(123)
  task = tsk("mtcars")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("regr.mse")

  # Pretrained path
  learner_pre = lrn("regr.rpart")
  learner_pre$train(task)

  set.seed(1)
  sage_pre = MarginalSAGE$new(
    task = task,
    learner = learner_pre,
    measure = measure,
    resampling = resampling,
    n_permutations = 3L,
    n_samples = 20L
  )
  sage_pre$compute()

  # Non-pretrained path
  set.seed(1)
  sage_fresh = MarginalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = measure,
    resampling = resampling,
    n_permutations = 3L,
    n_samples = 20L
  )
  sage_fresh$compute()

  expect_importance_dt(sage_pre$importance(), features = sage_pre$features)
  expect_importance_dt(sage_fresh$importance(), features = sage_fresh$features)
  expect_setequal(sage_pre$importance()$feature, sage_fresh$importance()$feature)
})

test_that("MarginalSAGE with pretrained learner errors on non-instantiated resampling", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  measure = msr("regr.mse")

  expect_error(
    MarginalSAGE$new(
      task = task,
      learner = learner,
      measure = measure,
      resampling = rsmp("holdout"),
      n_permutations = 2L,
      n_samples = 20L
    ),
    "instantiated"
  )
})

test_that("MarginalSAGE with pretrained learner errors on multi-fold resampling at construction", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)
  measure = msr("regr.mse")

  expect_error(
    MarginalSAGE$new(
      task = task,
      learner = learner,
      measure = measure,
      resampling = rsmp("cv", folds = 3)$instantiate(task),
      n_permutations = 2L,
      n_samples = 20L
    ),
    "not compatible"
  )
})

test_that("MarginalSAGE with pretrained learner works for classification", {
  task = tsk("penguins")
  learner = lrn("classif.rpart", predict_type = "prob")
  resampling = rsmp("holdout")$instantiate(task)
  measure = msr("classif.ce")

  learner$train(task)

  sage = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_permutations = 2L,
    n_samples = 20L
  )

  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
})

# -----------------------------------------------------------------------------
# LOCO warns for pretrained learner (refit-based methods)
# -----------------------------------------------------------------------------

test_that("LOCO warns when given a pretrained learner", {
  task = tsk("mtcars")
  learner = lrn("regr.rpart")
  learner$train(task)

  expect_warning(
    LOCO$new(
      task = task,
      learner = learner,
      measure = msr("regr.mse")
    ),
    "already trained"
  )
})

# =============================================================================
# ConditionalSAGE Tests
# =============================================================================
#
# Strategy for test speed:
# - Use ConditionalGaussianSampler (fast) for most tests
# - Use sim_dgp_* functions (4-5 features) instead of friedman1 (10 features)
# - Only use ARF sampler in tests specifically testing ARF behavior
# - ARF-specific tests get skip_on_cran()

# -----------------------------------------------------------------------------
# Basic functionality
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE basic workflow with Gaussian sampler", {
  # Use small task (4 features) and fast Gaussian sampler
  task = sim_dgp_correlated(n = 100)
  sampler = ConditionalGaussianSampler$new(task)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )

  checkmate::expect_r6(sage, c("FeatureImportanceMethod", "SAGE", "ConditionalSAGE"))
  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
})

test_that("ConditionalSAGE works with classification tasks", {
  # Binary classification - 2dnormals has 2 features
  task_binary = tgen("2dnormals")$generate(n = 50)
  sampler = ConditionalGaussianSampler$new(task_binary)

  sage_binary = ConditionalSAGE$new(
    task = task_binary,
    learner = lrn("classif.rpart", predict_type = "prob"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )
  checkmate::expect_r6(sage_binary, c("FeatureImportanceMethod", "SAGE", "ConditionalSAGE"))
  sage_binary$compute()
  expect_importance_dt(sage_binary$importance(), features = sage_binary$features)
})

test_that("ConditionalSAGE multiclass classification", {
  skip_on_cran() # multiclass with 3 features is slower
  task_multi = tgen("cassini")$generate(n = 50)
  sampler = ConditionalGaussianSampler$new(task_multi)

  sage_multi = ConditionalSAGE$new(
    task = task_multi,
    learner = lrn("classif.rpart", predict_type = "prob"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )
  sage_multi$compute()
  expect_importance_dt(sage_multi$importance(), features = sage_multi$features)
  expect_length(task_multi$class_names, 3L)
})

test_that("ConditionalSAGE featureless learner produces zero importance", {
  # Use small task (4 features) and fast Gaussian sampler
  task = sim_dgp_correlated(n = 100)
  sampler = ConditionalGaussianSampler$new(task)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.featureless"),
    measure = msr("regr.mse"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )

  sage$compute()
  result = sage$importance()

  expect_importance_dt(result, features = sage$features)
  # All importance values should be essentially zero
  checkmate::expect_numeric(result$importance, lower = -1e-10, upper = 1e-10)
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE produces sensible ranking", {
  # Use sim_dgp_independent (5 features) with Gaussian sampler
  task = sim_dgp_independent(n = 200)
  sampler = ConditionalGaussianSampler$new(task)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )

  sage$compute()
  result = sage$importance()
  expect_importance_dt(result, features = sage$features)

  # Important features should have higher mean importance than unimportant
  important_scores = result[grepl("^important", feature)]$importance
  unimportant_scores = result[grepl("^unimportant", feature)]$importance
  expect_gt(mean(important_scores), mean(unimportant_scores))
})

# -----------------------------------------------------------------------------
# Sampler behavior
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE uses ConditionalARFSampler by default", {
  skip_if_not_installed("arf")

  # Just check default sampler class - no compute needed
  task = sim_dgp_correlated(n = 50)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    n_permutations = 2L,
    n_samples = 20L
  )

  checkmate::expect_r6(sage$sampler, "ConditionalARFSampler")
  expect_equal(sage$label, "Conditional SAGE")
})

test_that("ConditionalSAGE with ARF sampler computes correctly", {
  skip_on_cran() # ARF sampling is slow
  skip_if_not_installed("arf")

  # Use small task for ARF test
  task = sim_dgp_correlated(n = 50)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    n_permutations = 2L,
    n_samples = 20L
  )

  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
})

test_that("ConditionalSAGE with custom ARF sampler settings", {
  skip_on_cran() # ARF sampling is slow
  skip_if_not_installed("arf")

  task = sim_dgp_correlated(n = 50)
  custom_sampler = ConditionalARFSampler$new(task, finite_bounds = "local")

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    sampler = custom_sampler,
    n_permutations = 2L,
    n_samples = 20L
  )

  checkmate::expect_r6(sage$sampler, "ConditionalSampler")
  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
})

# -----------------------------------------------------------------------------
# Parameter validation
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE requires predict_type='prob' for classification", {
  task = tgen("2dnormals")$generate(n = 50)

  expect_error(
    ConditionalSAGE$new(
      task = task,
      learner = lrn("classif.rpart", predict_type = "response"),
      sampler = ConditionalGaussianSampler$new(task)
    ),
    "Classification learners require probability predictions for SAGE."
  )
})

# -----------------------------------------------------------------------------
# Batching (skip_on_cran - tests implementation detail)
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE batching produces consistent results", {
  skip_on_cran() # tests implementation detail, not core functionality
  skip_if_not_installed("withr")

  # Use small task with Gaussian sampler for faster batching test
  task = sim_dgp_correlated(n = 50)
  sampler = ConditionalGaussianSampler$new(task)

  # Results should be identical with or without batching
  result_batch = withr::with_seed(42, {
    sage = ConditionalSAGE$new(
      task = task,
      learner = lrn("regr.rpart"),
      sampler = ConditionalGaussianSampler$new(task),
      n_permutations = 2L,
      n_samples = 20L
    )
    sage$compute(batch_size = 1)
    sage$importance()
  })

  result_normal = withr::with_seed(42, {
    sage = ConditionalSAGE$new(
      task = task,
      learner = lrn("regr.rpart"),
      sampler = ConditionalGaussianSampler$new(task),
      n_permutations = 2L,
      n_samples = 20L
    )
    sage$compute()
    sage$importance()
  })

  expect_equal(result_batch$importance, result_normal$importance, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# n_samples parameter
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE with custom n_samples", {
  # Use small task with Gaussian sampler
  task = sim_dgp_correlated(n = 50)
  sampler = ConditionalGaussianSampler$new(task)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    sampler = sampler,
    n_permutations = 2L,
    n_samples = 20L
  )
  expect_equal(sage$param_set$values$n_samples, 20L)

  sage$compute()
  expect_importance_dt(sage$importance(), features = sage$features)
})

# -----------------------------------------------------------------------------
# Convergence tracking (skip_on_cran - tests advanced feature)
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE SE tracking in convergence_history", {
  skip_on_cran() # tests convergence tracking feature, not core SAGE

  # Use small task with Gaussian sampler
  task = sim_dgp_correlated(n = 50)
  sampler = ConditionalGaussianSampler$new(task)

  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    sampler = sampler,
    n_permutations = 6L,
    n_samples = 20L
  )

  # Compute with early stopping to get convergence history
  sage$compute(early_stopping = TRUE, se_threshold = 0.05, check_interval = 2L)

  # Check that convergence_history exists and has SE column
  expect_false(is.null(sage$convergence_history))
  expect_contains(colnames(sage$convergence_history), "se")

  # Check structure of convergence_history
  expected_cols = c("n_permutations", "feature", "importance", "se")
  expect_setequal(colnames(sage$convergence_history), expected_cols)

  # SE values should be non-negative and finite
  se_values = sage$convergence_history$se
  checkmate::expect_numeric(se_values, lower = 0, finite = TRUE)

  # All features should be represented in convergence history
  expect_setequal(
    unique(sage$convergence_history$feature),
    sage$features
  )
})

# -----------------------------------------------------------------------------
# Regression guard for the samples_per_row migration (+ underlying ARF fix)
# -----------------------------------------------------------------------------

test_that("ConditionalSAGE noise feature receives ~0 importance with ARFSampler", {
  skip_if_not_installed("arf")
  skip_if_not_installed("ranger")
  skip_if_not_installed("mlr3learners")

  # `sim_dgp_correlated`: y = 2 * x1 + x3 + noise; x1 correlated with x2 (r),
  # x4 is pure noise (independent, no causal effect). The expected SAGE pattern:
  #   - x1 dominant (signal, coef 2)
  #   - x3 next (signal, coef 1)
  #   - x2 some attribution via Shapley (correlated with x1 but no causal effect)
  #   - x4 ~ 0 (pure noise)
  #
  # A regression in the multi-draw conditional sampling path tends to inflate
  # all SAGE values (notably for x4) because empty-coalition predictions become
  # garbage. This test pins both the signal/noise pattern and concrete values
  # captured on the post-fix branch.
  set.seed(42L)
  task = sim_dgp_correlated(n = 200L, r = 0.7)
  sage = ConditionalSAGE$new(
    task = task,
    learner = lrn("regr.ranger", num.trees = 100L),
    measure = msr("regr.mse"),
    sampler = ConditionalARFSampler$new(task, verbose = FALSE),
    n_permutations = 30L,
    n_samples = 30L
  )
  sage$compute()
  imp = sage$importance()

  # Correctness: noise feature x4 stays near zero, signal features dominate.
  expect_lt(abs(imp[feature == "x4", importance]), 0.3)
  expect_gt(imp[feature == "x1", importance], 0.5)
  expect_gt(imp[feature == "x1", importance], imp[feature == "x4", importance] + 1.0)

  # Regression guard. Captured on the post-fix branch at the same settings.
  # Tolerance generous enough to absorb run-to-run MC variation under different
  # R / arf / ranger versions; if it ever breaks, regenerate by running the same
  # block and pasting `imp$importance` below.
  expected_features = c("x1", "x2", "x3", "x4")
  expected_importance = c(
    1.75718693344525,
    0.359611168661264,
    0.651142439812633,
    -0.0157528350794578
  )
  expect_equal(imp$feature, expected_features)
  expect_equal(imp$importance, expected_importance, tolerance = 0.5)
})

test_that("importance() accepts all ci_method values", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )

  pfi$compute()

  # Test that all variance methods work
  imp_none = pfi$importance(ci_method = "none")
  imp_raw = pfi$importance(ci_method = "raw")
  expect_warning(
    imp_nb <- pfi$importance(ci_method = "nadeau_bengio")
  )
  imp_quantile = pfi$importance(ci_method = "quantile")

  expect_importance_dt(imp_none, features = pfi$features)
  expect_importance_dt(imp_raw, features = pfi$features)
  expect_importance_dt(imp_nb, features = pfi$features)
  expect_importance_dt(imp_quantile, features = pfi$features)
})

test_that("ci_method='none' produces no variance columns", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )

  pfi$compute()
  imp_none = pfi$importance(ci_method = "none")

  # Check that only feature and importance columns exist
  expect_named(imp_none, c("feature", "importance"))
})

test_that("raw CIs are narrower than nadeau_bengio corrected CIs", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11, ratio = 0.8),
    n_repeats = 3
  )

  pfi$compute()

  # Use two-sided to compare finite CI widths
  imp_raw = pfi$importance(ci_method = "raw", alternative = "two.sided")
  imp_nb = pfi$importance(ci_method = "nadeau_bengio", alternative = "two.sided")

  # Calculate CI widths
  width_raw = imp_raw$conf_upper - imp_raw$conf_lower
  width_nb = imp_nb$conf_upper - imp_nb$conf_lower

  # Raw CIs should be narrower than corrected ones on average
  # Compare the mean widths instead of individual features
  # The nadeau_bengio correction factor should make CIs wider on average
  expect_true(mean(width_nb) > mean(width_raw))
})

test_that("nadeau_bengio correction requires appropriate resampling", {
  task = sim_dgp_independent(n = 100)

  # Cross-validation is not supported for nadeau_bengio
  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("cv", folds = 3),
    n_repeats = 2
  )

  pfi$compute()

  # Should error for unsupported resampling
  expect_warning(
    pfi$importance(ci_method = "nadeau_bengio"),
    regexp = "recommended for resampling types"
  )

  # But raw variance should still work
  imp_raw = pfi$importance(ci_method = "raw")
  expect_importance_dt(imp_raw, features = pfi$features)
})

test_that("confidence level parameter works correctly", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )

  pfi$compute()

  # Test different confidence levels with two-sided CIs to compare widths
  imp_90 = pfi$importance(ci_method = "raw", conf_level = 0.90, alternative = "two.sided")
  imp_95 = pfi$importance(ci_method = "raw", conf_level = 0.95, alternative = "two.sided")
  imp_99 = pfi$importance(ci_method = "raw", conf_level = 0.99, alternative = "two.sided")

  # Calculate CI widths
  width_90 = imp_90$conf_upper - imp_90$conf_lower
  width_95 = imp_95$conf_upper - imp_95$conf_lower
  width_99 = imp_99$conf_upper - imp_99$conf_lower

  # Higher confidence level should produce wider CIs (on average)
  expect_true(mean(width_90) < mean(width_95))
  expect_true(mean(width_95) < mean(width_99))
})

test_that("variance estimation works with bootstrap resampling", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("bootstrap", repeats = 11),
    n_repeats = 2
  )

  pfi$compute()

  # Both raw and nadeau_bengio should work with bootstrap
  imp_raw = pfi$importance(ci_method = "raw")
  imp_nb = pfi$importance(ci_method = "nadeau_bengio")

  expect_importance_dt(imp_raw, features = pfi$features)
  expect_importance_dt(imp_nb, features = pfi$features)

  # Verify variance columns exist
  expect_true(all(c("se", "conf_lower", "conf_upper") %in% names(imp_raw)))
  expect_true(all(c("se", "conf_lower", "conf_upper") %in% names(imp_nb)))
})

test_that("quantile variance method works", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )

  pfi$compute()

  # Use two-sided for testing finite CI bounds
  imp_quantile = pfi$importance(ci_method = "quantile", alternative = "two.sided")

  # Check structure
  expect_importance_dt(imp_quantile, features = pfi$features)

  # Quantile method only returns confidence bounds, not se/statistic/p.value
  expected_cols = c("feature", "importance", "conf_lower", "conf_upper")
  expect_named(imp_quantile, expected_cols)

  # All CIs should be valid intervals (two-sided has finite bounds)
  expect_true(all(imp_quantile$conf_lower <= imp_quantile$conf_upper))

  # Point estimates should be between lower and upper bounds (or close)
  # Due to using mean vs quantiles, this is not guaranteed but usually holds
  expect_true(all(
    imp_quantile$importance >= imp_quantile$conf_lower |
      abs(imp_quantile$importance - imp_quantile$conf_lower) < 0.01
  ))
  expect_true(all(
    imp_quantile$importance <= imp_quantile$conf_upper |
      abs(imp_quantile$importance - imp_quantile$conf_upper) < 0.01
  ))
})

test_that("quantile CIs differ from parametric methods", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 15),
    n_repeats = 3
  )

  pfi$compute()

  # Use two-sided to compare finite CI bounds
  imp_raw = pfi$importance(ci_method = "raw", alternative = "two.sided")
  imp_quantile = pfi$importance(ci_method = "quantile", alternative = "two.sided")

  # Point estimates should be the same (both use mean)
  expect_equal(imp_raw$importance, imp_quantile$importance)

  # CIs should generally differ between methods
  # (quantile is non-parametric, raw assumes normality)
  expect_false(all(imp_raw$conf_lower == imp_quantile$conf_lower))
  expect_false(all(imp_raw$conf_upper == imp_quantile$conf_upper))
})

test_that("alternative='greater' produces one-sided CIs and tests", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11),
    n_repeats = 3
  )

  pfi$compute()

  # Test raw method with greater alternative
  imp_raw = pfi$importance(ci_method = "raw", alternative = "greater")

  # Should have statistic and p.value columns
  expect_true(all(c("statistic", "p.value") %in% names(imp_raw)))

  # Upper bound should be Inf for one-sided
  expect_true(all(is.infinite(imp_raw$conf_upper)))
  expect_true(all(imp_raw$conf_upper > 0)) # Inf, not -Inf

  # Lower bound should be finite
  expect_true(all(is.finite(imp_raw$conf_lower)))

  # Test nadeau_bengio with greater alternative
  imp_nb = pfi$importance(ci_method = "nadeau_bengio", alternative = "greater")
  expect_true(all(is.infinite(imp_nb$conf_upper)))
  expect_true(all(c("statistic", "p.value") %in% names(imp_nb)))

  # Test quantile with greater alternative (no statistic/p.value)
  imp_quantile = pfi$importance(ci_method = "quantile", alternative = "greater")
  expect_true(all(is.infinite(imp_quantile$conf_upper)))
  # Quantile method doesn't have statistic/p.value
  expect_false("statistic" %in% names(imp_quantile))
  expect_false("p.value" %in% names(imp_quantile))
})

test_that("alternative='two.sided' produces two-sided CIs and tests", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11),
    n_repeats = 3
  )

  pfi$compute()

  # Test raw method with two.sided alternative
  imp_raw = pfi$importance(ci_method = "raw", alternative = "two.sided")

  # Should have statistic and p.value columns
  expect_true(all(c("statistic", "p.value") %in% names(imp_raw)))

  # Both bounds should be finite
  expect_true(all(is.finite(imp_raw$conf_lower)))
  expect_true(all(is.finite(imp_raw$conf_upper)))

  # Test nadeau_bengio
  imp_nb = pfi$importance(ci_method = "nadeau_bengio", alternative = "two.sided")
  expect_true(all(is.finite(imp_nb$conf_upper)))
  expect_true(all(c("statistic", "p.value") %in% names(imp_nb)))

  # Test quantile (no statistic/p.value for quantile method)
  imp_quantile = pfi$importance(ci_method = "quantile", alternative = "two.sided")
  expect_true(all(is.finite(imp_quantile$conf_upper)))
  expect_false("statistic" %in% names(imp_quantile))
  expect_false("p.value" %in% names(imp_quantile))
})

test_that("two-sided p-values are larger than one-sided for positive importance", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11),
    n_repeats = 3
  )

  pfi$compute()

  imp_greater = pfi$importance(ci_method = "raw", alternative = "greater")
  imp_twosided = pfi$importance(ci_method = "raw", alternative = "two.sided")

  # For features with positive importance, two-sided p-values should be ~2x one-sided
  positive_mask = imp_greater$importance > 0
  if (any(positive_mask)) {
    expect_true(all(
      imp_twosided$p.value[positive_mask] >= imp_greater$p.value[positive_mask]
    ))
  }

  # Test statistics should be identical regardless of alternative
  expect_equal(imp_greater$statistic, imp_twosided$statistic)
})

test_that("p_adjust = 'bonferroni' adjusts p-values and CIs for raw", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11),
    n_repeats = 3
  )
  pfi$compute()

  imp_none = pfi$importance(ci_method = "raw", alternative = "two.sided")
  imp_bonf = pfi$importance(ci_method = "raw", alternative = "two.sided", p_adjust = "bonferroni")

  # p-values should be larger (or equal) after Bonferroni correction
  # (filter out NAs from features with zero variance)
  valid = is.finite(imp_none$p.value) & is.finite(imp_bonf$p.value)
  expect_true(all(imp_bonf$p.value[valid] >= imp_none$p.value[valid] - 1e-10))

  # CIs should be wider with Bonferroni correction
  width_none = imp_none$conf_upper - imp_none$conf_lower
  width_bonf = imp_bonf$conf_upper - imp_bonf$conf_lower
  expect_true(all(width_bonf >= width_none - 1e-10))

  # Point estimates and SEs should be unchanged
  expect_equal(imp_none$importance, imp_bonf$importance)
  expect_equal(imp_none$se, imp_bonf$se)
})

test_that("p_adjust = 'BH' adjusts only p-values for nadeau_bengio", {
  task = sim_dgp_independent(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 11, ratio = 0.8),
    n_repeats = 3
  )
  pfi$compute()

  imp_none = pfi$importance(ci_method = "nadeau_bengio", alternative = "two.sided")
  imp_bh = pfi$importance(ci_method = "nadeau_bengio", alternative = "two.sided", p_adjust = "BH")

  # CIs should be identical (BH does not adjust CIs)
  expect_equal(imp_none$conf_lower, imp_bh$conf_lower)
  expect_equal(imp_none$conf_upper, imp_bh$conf_upper)

  # Point estimates and SEs should be unchanged
  expect_equal(imp_none$importance, imp_bh$importance)
  expect_equal(imp_none$se, imp_bh$se)
})

test_that("invalid p_adjust value is rejected", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )
  pfi$compute()

  expect_error(
    pfi$importance(ci_method = "raw", p_adjust = "invalid_method"),
    regexp = "p_adjust"
  )
})

test_that("unknown arguments to $importance() are rejected", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )
  pfi$compute()

  expect_error(
    pfi$importance(ci_method = "raw", tset = "t"),
    regexp = "Unknown argument"
  )
})

test_that("default alternative is 'two.sided'", {
  task = sim_dgp_independent(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 2
  )

  pfi$compute()

  # Default should be one-sided
  imp_default = pfi$importance(ci_method = "raw")
  imp_greater = pfi$importance(ci_method = "raw", alternative = "two.sided")

  expect_equal(imp_default$conf_lower, imp_greater$conf_lower)
  expect_equal(imp_default$conf_upper, imp_greater$conf_upper)
  expect_equal(imp_default$p.value, imp_greater$p.value)
})

#' @title Feature Sampler Class
#'
#' @description Base class for implementing different sampling strategies
#' for feature importance methods like PFI and CFI
#'
#' @export
FeatureSampler = R6Class(
  "FeatureSampler",
  public = list(
    #' @field task ([mlr3::Task]) Original task.
    task = NULL,
    #' @field label (`character(1)`) Name of the sampler.
    label = NULL,
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    #'   Will be checked against the provided [mlr3::Task] to ensure compatibility.
    feature_types = c(
      "numeric",
      "factor",
      "ordered",
      "integer",
      "logical",
      "Date",
      "POSIXct",
      "character"
    ),
    #' @field param_set ([paradox::ParamSet]) Parameter set for the sampler.
    param_set = NULL,

    #' @description
    #' Creates a new instance of the FeatureSampler class
    #' @param task ([mlr3::Task]) Task to sample from
    initialize = function(task) {
      self$task = mlr3::assert_task(task, feature_types = self$feature_types)
      # Initialize empty param_set - subclasses should define their own
      self$param_set = paradox::ps()
    },

    #' @description
    #' Sample values for feature(s) from stored task
    #' @param feature (`character`) Feature name(s) to sample (can be single or multiple). Must match those in the stored [Task][mlr3::Task].
    #' @param row_ids (`integer()`: `NULL`) Row IDs of the stored [Task][mlr3::Task] to use as basis for sampling.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples to draw per input row.
    #'   When `> 1`, the returned data.table has `samples_per_row * length(row_ids)` rows in **draw-major**
    #'   order: rows `1..n` are draw 1 across all input rows in order, rows `n+1..2n` are draw 2, and so on.
    #' @return Modified copy of the input features with the feature(s) sampled. A
    #'   [data.table][data.table::data.table] with `samples_per_row * length(row_ids)` rows in draw-major order.
    sample = function(feature, row_ids = NULL, samples_per_row = 1L) {
      cli::cli_abort(c(
        "Abstract method",
        i = "Use a concrete implementation."
      ))
    },
    #' @description
    #' Sample values for feature(s) using external data
    #' @param feature (`character`) Feature name(s) to sample (can be single or multiple)
    #' @param newdata ([`data.table`][data.table::data.table]) External data to use for sampling.
    #' @param samples_per_row (`integer(1)`: `1L`) See `$sample()`. Output is `samples_per_row * nrow(newdata)`
    #'   rows in draw-major order.
    sample_newdata = function(feature, newdata, samples_per_row = 1L) {
      cli::cli_abort(c(
        "Abstract method",
        i = "Use a concrete implementation."
      ))
    },

    #' @description
    #' Print sampler
    #'
    #' @param ... Ignored.
    print = function(...) {
      cli::cli_h1(self$label)
      cli::cli_ul()
      cli::cli_li(
        "Task: {.val {self$task$id}} ({.strong {self$task$nrow}x{self$task$n_features}})"
      )
      cli::cli_li("Supported feature type{?s}: {.val {self$feature_types}}")
      cli::cli_end()
    }
  ),
  private = list(
    .get_task_data_by_row_id = function(row_ids) {
      if (!checkmate::test_subset(row_ids, self$task$row_ids)) {
        cli::cli_abort(c(
          x = "Requested {.val {length(setdiff(row_ids, self$task$row_ids))}} row_id{?s} not in stored {.cls Task}.",
          "!" = "For {.code $sample}, the row_ids must match those of the stored {.cls Task}."
        ))
      }
      self$task$data(rows = row_ids)
    },

    # Ensure feature types match task specification, fixing what can be fixed
    # and asserting correctness. Only modifies data when necessary.
    # mlr3 feature types correspond to R's class(), not typeof()
    .ensure_feature_types = function(data) {
      feature_types = self$task$feature_types

      for (feat in self$task$feature_names) {
        expected_type = feature_types[id == feat, type]
        actual_class = class(data[[feat]])[1]

        if (expected_type == actual_class) {
          next
        }

        # Attempt to fix known fixable cases
        if (expected_type == "integer" && actual_class == "numeric") {
          set(data, j = feat, value = as.integer(round(data[[feat]])))
          next
        }

        # Type mismatch that cannot be automatically fixed
        cli::cli_abort(c(
          x = "Feature {.val {feat}} has class {.val {actual_class}} but expected {.val {expected_type}}.",
          i = "This indicates a bug in the sampler implementation."
        ))
      }

      data
    }
  )
)

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import data.table
#' @importFrom checkmate %??%
#' @importFrom cli cli
#' @importFrom mlr3 resample PredictionClassif PredictionRegr
#' @importFrom mvtnorm rmvnorm
#' @importFrom paradox ps
#' @importFrom R6 R6Class
#' @importFrom stats var sd quantile qt binom.test runif rnorm
#' @importFrom utils packageVersion
## usethis namespace: end
NULL

#' @title ARF-based Conditional Sampler
#'
#' @description Implements conditional sampling using Adversarial Random Forests (ARF).
#' ARF can handle mixed data types (continuous and categorical) and provides
#' flexible conditional sampling by modeling the joint distribution.
#'
#' @details
#' The ConditionalARFSampler fits an [Adversarial Random Forest][arf::arf] model on the task data,
#' then uses it to generate samples from \eqn{P(X_j | X_{-j})} where \eqn{X_j} is the
#' feature of interest and \eqn{X_{-j}} are the conditioning features.
#'
#' @examplesIf requireNamespace("arf", quietly = TRUE)
#' library(mlr3)
#' task = tgen("2dnormals")$generate(n = 100)
#' # Create sampler with default parameters
#' sampler = ConditionalARFSampler$new(task, conditioning_set = "x2", verbose = FALSE)
#' # Sample using row_ids from stored task
#' sampled_data = sampler$sample("x1", row_ids = 1:10)
#' # Or use external data
#' data = task$data()
#' sampled_data_ext = sampler$sample_newdata("x1", newdata = data, conditioning_set = "x2")
#'
#' # Example with custom ARF parameters
#' sampler_custom = ConditionalARFSampler$new(
#'   task,
#'   min_node_size = 10L,
#'   finite_bounds = "local",
#'   verbose = FALSE
#' )
#' sampled_custom = sampler_custom$sample("x1", conditioning_set = "x2")
#' @references `r print_bib("watson_2023", "blesch_2025")`
#'
#' @export
ConditionalARFSampler = R6Class(
  "ConditionalARFSampler",
  inherit = ConditionalSampler,
  public = list(
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    #'   Will be checked against the provided [mlr3::Task] to ensure compatibility.
    feature_types = c(
      "numeric",
      "factor",
      "ordered",
      "integer",
      "logical",
      "character"
    ),
    #' @field arf_model Adversarial Random Forest model created by [arf::adversarial_rf].
    arf_model = NULL,
    #' @field psi Distribution parameters estimated from by [arf::forde].
    psi = NULL,

    #' @description
    #' Creates a new instance of the ConditionalARFSampler class.
    #' To fit the ARF in parallel, register a parallel backend first (see [arf::arf]) and set `parallel = TRUE`.
    #' @param task ([mlr3::Task]) Task to sample from.
    #' @param conditioning_set (`character` | `NULL`) Default conditioning set to use in `$sample()`. This parameter only affects the sampling behavior, not the ARF model fitting.
    #' @param num_trees (`integer(1)`: `10L`) Number of trees for ARF. Passed to [arf::adversarial_rf].
    #' @param min_node_size (`integer(1)`: `20L`) Minimum node size for ARF. Passed to [arf::adversarial_rf] and in turn to [ranger::ranger].
    #'   This is increased to 20 to mitigate overfitting.
    #' @param finite_bounds (`character(1)`: `"no"`) How to handle variable bounds. Passed to [arf::forde]. Default is `"no"` for compatibility. `"local"` may improve extrapolation but can cause issues with some data.
    #' @param epsilon (`numeric(1)`: `0`) Slack parameter for when `finite_bounds != "no"`. Passed to [arf::forde].
    #' @param round (`logical(1)`: `TRUE`) Whether to round continuous variables back to their original precision in sampling. Can be overridden in `$sample()` calls.
    #' @param stepsize (`numeric(1)`: `0`) Number of rows of evidence to process at a time when `parallel` is `TRUE`. Default (`0`) spreads evidence evenly over registered workers. Can be overridden in `$sample()` calls.
    #' @param verbose (`logical(1)`: `FALSE`) Whether to print progress messages. Default is `FALSE` (arf's default is `TRUE`). Can be overridden in `$sample()` calls.
    #' @param parallel (`logical(1)`: `FALSE`) Whether to use parallel processing via `foreach`. See examples in [arf::forge()]. Can be overridden in `$sample()` calls.
    #' @param ... Additional arguments passed to [arf::adversarial_rf].
    initialize = function(
      task,
      conditioning_set = NULL,
      num_trees = 10L,
      min_node_size = 20L,
      finite_bounds = "no",
      epsilon = 1e-15,
      round = TRUE,
      stepsize = 0,
      verbose = FALSE,
      parallel = FALSE,
      ...
    ) {
      require_package("arf")
      super$initialize(task, conditioning_set = conditioning_set)
      self$label = "Adversarial Random Forest sampler"

      # Extend param_set with ARF-specific parameters
      # Sampling parameters (can be overridden in $sample() calls)
      # Model fitting parameters (only used during initialization)
      self$param_set = c(
        self$param_set,
        paradox::ps(
          # Sampling parameters (stored for hierarchical resolution in $sample())
          round = paradox::p_lgl(default = TRUE),
          stepsize = paradox::p_dbl(lower = 0, default = 0),
          verbose = paradox::p_lgl(default = FALSE),
          parallel = paradox::p_lgl(default = FALSE),
          # Model fitting parameters (used only during initialization)
          num_trees = paradox::p_int(lower = 1L, default = 10L),
          min_node_size = paradox::p_int(lower = 1L, default = 20L),
          finite_bounds = paradox::p_fct(c("no", "local", "global"), default = "no"),
          epsilon = paradox::p_dbl(lower = 0, default = 1e-15)
        )
      )

      # Set parameter values
      self$param_set$set_values(
        round = round,
        stepsize = stepsize,
        verbose = verbose,
        parallel = parallel,
        num_trees = num_trees,
        min_node_size = min_node_size,
        finite_bounds = finite_bounds,
        epsilon = epsilon
      )

      # Register sequential backend in an attempt to silence foreach warning
      if (!foreach::getDoParRegistered() & !parallel) {
        foreach::registerDoSEQ()
      }

      # Fit ARF model on the task data, features only
      task_data = self$task$data(cols = self$task$feature_names)

      # Train ARF and estimate distribution parameters
      self$arf_model = arf::adversarial_rf(
        x = task_data,
        num_trees = num_trees,
        min_node_size = min_node_size,
        verbose = verbose,
        parallel = parallel,
        replace = FALSE,
        ...
      )
      self$psi = arf::forde(
        arf = self$arf_model,
        x = task_data,
        finite_bounds = finite_bounds,
        epsilon = epsilon,
        parallel = parallel
      )
    },

    #' @description
    #' Sample from stored task. Parameters use hierarchical resolution:
    #' function argument > stored `param_set` value > hard-coded default.
    #' @param feature (`character`) Feature(s) to sample.
    #' @param row_ids (`integer()` | `NULL`) Row IDs to use. If `NULL`, uses all rows.
    #' @param conditioning_set (`character` | `NULL`) Features to condition on.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param round (`logical(1)` | `NULL`) Round continuous variables.
    #' @param stepsize (`numeric(1)` | `NULL`) Batch size for parallel processing.
    #' @param verbose (`logical(1)` | `NULL`) Print progress messages.
    #' @param parallel (`logical(1)` | `NULL`) Use parallel processing.
    #' @return Modified copy with sampled feature(s).
    sample = function(
      feature,
      row_ids = NULL,
      conditioning_set = NULL,
      samples_per_row = 1L,
      round = NULL,
      stepsize = NULL,
      verbose = NULL,
      parallel = NULL
    ) {
      super$sample(
        feature,
        row_ids,
        conditioning_set,
        samples_per_row = samples_per_row,
        round = round,
        stepsize = stepsize,
        verbose = verbose,
        parallel = parallel
      )
    },

    #' @description
    #' Sample from external data. See `$sample()` for parameter details.
    #' @param feature (`character`) Feature(s) to sample.
    #' @param newdata ([`data.table`][data.table::data.table]) External data to use.
    #' @param conditioning_set (`character` | `NULL`) Features to condition on.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param round (`logical(1)` | `NULL`) Round continuous variables.
    #' @param stepsize (`numeric(1)` | `NULL`) Batch size for parallel processing.
    #' @param verbose (`logical(1)` | `NULL`) Print progress messages.
    #' @param parallel (`logical(1)` | `NULL`) Use parallel processing.
    #' @return Modified copy with sampled feature(s).
    sample_newdata = function(
      feature,
      newdata,
      conditioning_set = NULL,
      samples_per_row = 1L,
      round = NULL,
      stepsize = NULL,
      verbose = NULL,
      parallel = NULL
    ) {
      super$sample_newdata(
        feature,
        newdata,
        conditioning_set,
        samples_per_row = samples_per_row,
        round = round,
        stepsize = stepsize,
        verbose = verbose,
        parallel = parallel
      )
    }
  ),
  private = list(
    # Core sampling logic implementing ARF conditional sampling
    # Called by $sample() and $sample_newdata() which pass ARF-specific parameters
    .sample_conditional = function(
      data,
      feature,
      conditioning_set = NULL,
      samples_per_row = 1L,
      round = NULL,
      stepsize = NULL,
      verbose = NULL,
      parallel = NULL
    ) {
      round = resolve_param(round, self$param_set$values$round, TRUE)
      stepsize = resolve_param(stepsize, self$param_set$values$stepsize, 0)
      verbose = resolve_param(verbose, self$param_set$values$verbose, FALSE)
      parallel = resolve_param(parallel, self$param_set$values$parallel, FALSE)

      if (parallel && (!foreach::getDoParRegistered() || foreach::getDoParName() == "doSEQ")) {
        cli::cli_abort(c(
          x = "{.code parallel = TRUE} but no parallel backend is registered.",
          i = "This can happen when a sampler saved with {.fun saveRDS} is loaded in a new session.",
          i = "Register a backend first, e.g. {.code doParallel::registerDoParallel(cores = 4)}",
          i = "Or set {.code parallel = FALSE} in the {.fun $sample} call."
        ))
      }

      n = nrow(data)

      if (length(conditioning_set) == 0) {
        if (xplain_opt("debug")) {
          cli::cli_alert_info(
            "{.val conditioning_set} is length 0, passing {.code evidence = NULL} to {.fun arf::forge}"
          )
        }

        # Marginal case: arf::forge(evidence = NULL, n_synth = k) returns only k rows
        # (k unconditional draws), not n * k. We need n * samples_per_row independent
        # marginal draws so every (instance, draw) pair gets its own sample, so request
        # the full count up front; no reshape needed since marginal draws are
        # exchangeable across instances.
        synthetic = arf::forge(
          params = self$psi,
          n_synth = n * samples_per_row,
          evidence = NULL,
          evidence_row_mode = "separate",
          round = round,
          sample_NAs = FALSE,
          nomatch = "force",
          verbose = verbose,
          stepsize = stepsize,
          parallel = parallel
        )
      } else {
        evidence = data[, .SD, .SDcols = conditioning_set]

        if (xplain_opt("debug")) {
          cli::cli_inform(c(
            i = "Feature is {.val {feature}}",
            i = "Conditioning set is {.val {conditioning_set}}",
            i = "samples_per_row = {.val {samples_per_row}}"
          ))
        }

        synthetic = arf::forge(
          params = self$psi,
          n_synth = samples_per_row,
          evidence = evidence,
          evidence_row_mode = "separate",
          round = round,
          sample_NAs = FALSE,
          nomatch = "force",
          verbose = verbose,
          stepsize = stepsize,
          parallel = parallel
        )

        if (samples_per_row > 1L) {
          # arf returns evidence-major: e1d1, e1d2, ..., e1dk, e2d1, ..., endk
          # we want draw-major: e1d1, e2d1, ..., end1, e1d2, ..., endk
          draw_idx = ((seq_len(n * samples_per_row) - 1L) %% samples_per_row) + 1L
          synthetic = synthetic[order(draw_idx)]
        }
      }

      out = data[rep.int(seq_len(.N), times = samples_per_row)]
      out[, (feature) := synthetic[, .SD, .SDcols = feature]]

      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)

#' @title Shapley Additive Global Importance (SAGE) Base Class
#'
#' @description Base class for SAGE (Shapley Additive Global Importance)
#' feature importance based on Shapley values with marginalization.
#' This is an abstract class - use [MarginalSAGE] or [ConditionalSAGE].
#'
#' @details
#' SAGE uses Shapley values to fairly distribute the total prediction
#' performance among all features. Unlike perturbation-based methods,
#' SAGE marginalizes features by integrating over their distribution.
#' This is approximated by averaging predictions over a reference dataset.
#'
#' **Standard Error Calculation**: The standard errors (SE) reported in
#' `$convergence_history` reflect the uncertainty in Shapley value estimation
#' across different random permutations within a single resampling iteration.
#' These SEs quantify the Monte Carlo sampling error for a fixed trained model
#' and are only valid for inference about the importance of features for that
#' specific model. They do not capture broader uncertainty from model variability
#' across different train/test splits or resampling iterations.
#'
#' @references
#' `r print_bib("lundberg_2020")`
#'
#' @seealso [MarginalSAGE] [ConditionalSAGE]
#'
#' @export
SAGE = R6Class(
  "SAGE",
  inherit = FeatureImportanceMethod,
  public = list(
    #' @field n_permutations (`integer(1)`) Number of permutations to sample.
    n_permutations = NULL,
    #' @field convergence_history ([`data.table`][data.table::data.table]) History of SAGE values during computation.
    convergence_history = NULL,
    #' @field converged (`logical(1)`) Whether convergence was detected.
    converged = FALSE,
    #' @field n_permutations_used (`integer(1)`) Actual number of permutations used.
    n_permutations_used = NULL,

    #' @description
    #' Creates a new instance of the SAGE class.
    #' @param task,learner,measure,resampling,features Passed to FeatureImportanceMethod.
    #' @param n_permutations (`integer(1)`: `10L`) Number of permutations to sample for SAGE value estimation.
    #'   The total number of evaluated coalitions is `1 (empty) + n_permutations * n_features`.
    #' @param batch_size (`integer(1)`: `5000L`) Maximum number of observations to process in a single prediction call.
    #' @param n_samples (`integer(1)`: `100L`) Number of samples to use for marginalizing out-of-coalition features.
    #'   For [MarginalSAGE], this is the number of marginal data samples ("background data" in other implementations).
    #'   For [ConditionalSAGE], this is the number of conditional samples per test instance retrieved from `sampler`.
    #' @param early_stopping (`logical(1)`: `TRUE`) Whether to enable early stopping based on convergence detection.
    #' @param se_threshold (`numeric(1)`: `0.01`) Convergence threshold for relative standard error.
    #'   Convergence is detected when the maximum relative SE across all features falls below this threshold.
    #'   Relative SE is calculated as SE divided by the range of importance values (max - min),
    #'   making it scale-invariant across different loss metrics.
    #'   Default of `0.01` means convergence when relative SE is below 1% of the importance range.
    #' @param min_permutations (`integer(1)`: `10L`) Minimum permutations before checking for convergence. Convergence is judged based on the standard errors of the estimated SAGE values,
    #' which requires a sufficiently large number of samples (i.e., evaluated coalitions).
    #' @param check_interval (`integer(1)`: `1L`) Check convergence every N permutations.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      n_permutations = 10L,
      batch_size = 5000L,
      n_samples = 100L,
      early_stopping = TRUE,
      se_threshold = 0.01,
      min_permutations = 10L,
      check_interval = 1L
    ) {
      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        label = "Shapley Additive Global Importance"
      )

      self$n_permutations = checkmate::assert_int(n_permutations, lower = 1L)

      # For classification tasks, require predict_type = "prob"
      if (self$task$task_type == "classif") {
        if (learner$predict_type != "prob") {
          cli::cli_abort(c(
            "Classification learners require probability predictions for SAGE.",
            "i" = "Please set {.code learner$configure(predict_type = \"prob\")} before using SAGE."
          ))
        }
      }

      # Set parameters
      ps = ps(
        n_permutations = paradox::p_int(lower = 1L, default = 10L),
        batch_size = paradox::p_int(lower = 1L, default = 5000L),
        n_samples = paradox::p_int(lower = 1L, default = 100L),
        early_stopping = paradox::p_lgl(default = TRUE),
        se_threshold = paradox::p_dbl(lower = 0, upper = 1, default = 0.01),
        min_permutations = paradox::p_int(lower = 1L, default = 3L),
        check_interval = paradox::p_int(lower = 1L, default = 1L)
      )
      ps$values$n_permutations = n_permutations
      ps$values$batch_size = batch_size
      ps$values$n_samples = n_samples
      ps$values$early_stopping = early_stopping
      ps$values$se_threshold = se_threshold
      ps$values$min_permutations = min_permutations
      ps$values$check_interval = check_interval
      self$param_set = ps
    },

    #' @description
    #' Compute SAGE values.
    #' @param store_backends (`logical(1)`) Whether to store data backends.
    #' @param batch_size (`integer(1)`: `5000L`) Maximum number of observations to process in a single prediction call.
    #' @param early_stopping (`logical(1)`: `TRUE`) Whether to check for convergence and stop early.
    #' @param se_threshold (`numeric(1)`: `0.01`) Convergence threshold for relative standard error.
    #'   SE is normalized by the range of importance values (max - min) to make convergence
    #'   detection scale-invariant. Default `0.01` means convergence when relative SE < 1%.
    #' @param min_permutations (`integer(1)`: `10L`) Minimum permutations before checking convergence.
    #' @param check_interval (`integer(1)`: `1L`) Check convergence every N permutations.
    compute = function(
      store_backends = TRUE,
      batch_size = NULL,
      early_stopping = NULL,
      se_threshold = NULL,
      min_permutations = NULL,
      check_interval = NULL
    ) {
      # Reset convergence tracking
      self$convergence_history = NULL
      self$converged = FALSE
      self$n_permutations_used = NULL

      # Resolve parameters using hierarchical resolution
      batch_size = resolve_param(batch_size, self$param_set$values$batch_size, 5000L)
      early_stopping = resolve_param(
        early_stopping,
        self$param_set$values$early_stopping,
        TRUE
      )
      se_threshold = resolve_param(
        se_threshold,
        self$param_set$values$se_threshold,
        0.01
      )
      min_permutations = resolve_param(
        min_permutations,
        self$param_set$values$min_permutations,
        3L
      )
      check_interval = resolve_param(check_interval, self$param_set$values$check_interval, 1L)

      # Initial resampling to get trained learners
      rr = assemble_rr(
        task = self$task,
        learner = self$learner,
        resampling = self$resampling,
        store_models = TRUE,
        store_backends = store_backends
      )
      # Store results
      self$resample_result = rr

      # For convergence tracking, we'll use the first resampling iteration
      # (convergence is about permutation count, not resampling)
      iter_for_convergence = 1L

      # Compute SAGE values for convergence tracking (first iteration)
      first_result = private$.compute_sage_scores(
        learner = rr$learners[[iter_for_convergence]],
        test_dt = self$task$data(rows = rr$resampling$test_set(iter_for_convergence)),
        n_permutations = self$n_permutations,
        batch_size = batch_size,
        early_stopping = early_stopping,
        se_threshold = se_threshold,
        min_permutations = min_permutations,
        check_interval = check_interval
      )

      # Extract convergence data from first iteration
      # `convergence_data` exists even if early_stopping = FALSE
      self$convergence_history = first_result$convergence_data$convergence_history
      self$converged = first_result$convergence_data$converged
      self$n_permutations_used = first_result$convergence_data$n_permutations_used

      # If we have multiple resampling iterations, compute the rest without convergence tracking
      if (self$resampling$iters > 1) {
        remaining_results = lapply(seq_len(self$resampling$iters)[-iter_for_convergence], \(iter) {
          private$.compute_sage_scores(
            learner = rr$learners[[iter]],
            test_dt = self$task$data(rows = rr$resampling$test_set(iter)),
            # n_permutations_used is either same as n_permutations (if early_stopping = FALSE)
            # or its a smaller value if early_stopping = TRUE and it stopped early
            # The fallback to self$n_permutations should not be needed
            n_permutations = self$n_permutations_used %||% self$n_permutations,
            batch_size = batch_size,
            # Only track convergence etc. for first iteration
            early_stopping = FALSE
          )
        })

        # Extract scores from all results (always list format now)
        all_scores = c(list(first_result$scores), lapply(remaining_results, function(x) x$scores))
      } else {
        all_scores = list(first_result$scores)
      }

      # Combine results across resampling iterations
      scores = rbindlist(all_scores, idcol = "iter_rsmp")

      # iter_rsmp, feature, importance -- score_baseline or so don't apply here
      private$.scores = scores
    },

    #' @description
    #' Plot convergence history of SAGE values.
    #' @param features (`character` | `NULL`) Features to plot. If NULL, plots all features.
    #' @return A [ggplot2][ggplot2::ggplot] object
    plot_convergence = function(features = NULL) {
      require_package("ggplot2")

      if (is.null(self$convergence_history)) {
        cli::cli_abort("No convergence history available. Run $compute() first.")
      }

      # Create a copy to avoid modifying the original
      plot_data = copy(self$convergence_history)

      if (!is.null(features)) {
        plot_data = plot_data[feature %in% features]
      }

      p = ggplot2::ggplot(
        plot_data,
        ggplot2::aes(x = n_permutations, y = importance, fill = feature, color = feature)
      ) +
        ggplot2::geom_ribbon(
          ggplot2::aes(ymin = importance - se, ymax = importance + se),
          alpha = 1 / 3
        ) +
        ggplot2::geom_line(linewidth = 1) +
        ggplot2::geom_point(size = 2) +
        ggplot2::labs(
          title = "SAGE Value Convergence",
          subtitle = if (self$converged) {
            sprintf(
              "Converged after %d permutations (saved %d)",
              self$n_permutations_used,
              self$n_permutations - self$n_permutations_used
            )
          } else {
            sprintf("Completed all %d permutations", self$n_permutations)
          },
          x = "Number of Permutations",
          y = "SAGE Value",
          color = "Feature",
          fill = "Feature"
        ) +
        ggplot2::theme_minimal(base_size = 14)

      if (self$converged) {
        p = p +
          ggplot2::geom_vline(
            xintercept = self$n_permutations_used,
            linetype = "dashed",
            color = "red",
            alpha = 0.5
          )
      }

      p
    }
  ),

  private = list(
    # This function computes the SAGE values for a single resampling iteration.
    # It iterates through permutations of features, evaluates coalitions, and calculates marginal contributions.
    .compute_sage_scores = function(
      learner,
      test_dt,
      n_permutations,
      batch_size = NULL,
      early_stopping = FALSE,
      se_threshold = 0.01,
      min_permutations = 10L,
      check_interval = 1L
    ) {
      # Initialize numeric vectors to store marginal contributions and their squares for variance calculation.
      # We track both sum and sum of squares to calculate running variance and standard errors.
      sage_values = numeric(length(self$features)) # Sum of marginal contributions
      sage_values_sq = numeric(length(self$features)) # Sum of squared marginal contributions
      names(sage_values) = self$features
      names(sage_values_sq) = self$features

      # Pre-generate `n_permutations` permutations upfront
      # Relevant for reproducibility, especially when using early stopping or parallel processing.
      # Example: if self$features = c("x1", "x2", "x3") and n_permutations = 2,
      # all_permutations might be list(c("x2", "x1", "x3"), c("x3", "x1", "x2"))
      all_permutations = replicate(n_permutations, sample(self$features), simplify = FALSE)

      # Initialize variables for iterative checkpoint-based computation.
      # This allows for early stopping based on convergence and provides progress updates.
      convergence_history = list() # Stores SAGE values at each checkpoint for convergence tracking
      n_completed = 0 # Number of permutations processed so far
      converged = FALSE # Flag to indicate if convergence has been detected
      baseline_loss = NULL # Loss of the empty coalition (model with no features / all features marginalized)

      # Calculate total checkpoints for progress tracking.
      # A checkpoint is a group of 'check_interval' permutations.
      total_checkpoints = ceiling(n_permutations / check_interval)
      current_checkpoint = 0

      # Start checkpoint-based progress bar if progress display is enabled.
      if (xplain_opt("progress")) {
        cli::cli_progress_bar(
          "Computing SAGE values",
          total = total_checkpoints
        )
      }

      # Main loop: Process permutations in checkpoints until all permutations are done or convergence is reached.
      while (n_completed < n_permutations && !converged) {
        # Determine the size of the current checkpoint.
        # This ensures that the last checkpoint processes only the remaining permutations.
        checkpoint_size = min(check_interval, n_permutations - n_completed)
        # Define the indices of permutations to be processed in this checkpoint.
        checkpoint_perms = (n_completed + 1):(n_completed + checkpoint_size)

        # Get the actual permutation sequences for this checkpoint from the pre-generated list.
        checkpoint_permutations = all_permutations[checkpoint_perms]

        # Build this checkpoint's growing-prefix coalitions. The
        # empty coalition is prepended only in the first checkpoint;
        # its loss is the baseline anchor for marginal contributions.
        # Same single-batch call/order as before, so the RNG-bearing
        # marginal sampling inside .evaluate_coalitions_batch is
        # byte-identical to the pre-refactor scheme.
        checkpoint_coalitions = sage_growing_coalitions(checkpoint_permutations)
        offset = 0L
        if (n_completed == 0) {
          checkpoint_coalitions = c(list(character(0)), checkpoint_coalitions)
          offset = 1L
        }

        # Progress: one tick per checkpoint (unchanged cadence).
        current_checkpoint = current_checkpoint + 1

        # Evaluate all coalitions collected in this checkpoint in a single batch.
        # This is a performance optimization to minimize prediction calls to the learner.
        checkpoint_losses = private$.evaluate_coalitions_batch(
          learner,
          test_dt,
          checkpoint_coalitions,
          batch_size
        )

        # Update progress bar.
        if (xplain_opt("progress")) {
          cli::cli_progress_update(inc = 1)
        }

        # Store the baseline loss (loss of the empty coalition) from the first checkpoint.
        # This is the model's performance when no features are available.
        if (n_completed == 0) {
          baseline_loss = checkpoint_losses[1] # The first element is always the empty coalition's loss
        }

        # Closed-form accumulation over the growing-prefix losses.
        # Every permutation is a full feature permutation, so each
        # coalition's loss index is computed directly; `offset` skips
        # the leading empty-coalition slot present in the first
        # checkpoint. Replaces the former O(n^2) which(sapply())
        # coalition-map lookup.
        acc = sage_marginal_contributions(
          checkpoint_permutations,
          checkpoint_losses,
          baseline_loss,
          self$features,
          offset = offset
        )
        # Name-aligned add (defensive: positional add is only valid if
        # orders match; reindex by name to be safe).
        sage_values = sage_values + acc$sv[names(sage_values)]
        sage_values_sq = sage_values_sq + acc$sv_sq[names(sage_values_sq)]

        # Update the count of completed permutations.
        n_completed = n_completed + checkpoint_size

        # Calculate the current average SAGE values and standard errors based on completed permutations.
        current_avg = sage_values / n_completed

        # Calculate running variance and standard errors for each feature
        # Variance = E[X^2] - E[X]^2, SE = sqrt(Var / n)
        current_variance = (sage_values_sq / n_completed) - (current_avg^2)
        # Ensure variance is non-negative (numerical precision issues)
        current_variance[current_variance < 0] = 0
        current_se = sqrt(current_variance / n_completed)

        if (xplain_opt("debug")) {
          cli::cli_alert_info("SAGE values after {.val {n_completed}} permutations")
          cli::cli_ol(c(
            "SAGE values: {.val {round(current_avg, 4)}}",
            "current SE: {.val {round(current_se, 3)}}",
            "Completed: {.val {n_completed}}"
          ))
        }

        # Store the current average SAGE values and standard errors in the convergence history.
        # Used for plotting, early stopping, and uncertainty quantification.
        checkpoint_history = data.table(
          n_permutations = n_completed,
          feature = names(current_avg),
          importance = as.numeric(current_avg),
          se = as.numeric(current_se)
        )
        convergence_history[[length(convergence_history) + 1]] = checkpoint_history

        # Check for convergence if early stopping is enabled and enough permutations have been processed.
        if (early_stopping && n_completed >= min_permutations && length(convergence_history) > 1) {
          # Get SAGE values from the current checkpoint.
          curr_checkpoint = convergence_history[[length(convergence_history)]]

          # Ensure features are in the same order for comparison.
          curr_checkpoint_ordered = copy(curr_checkpoint)[order(feature)]
          curr_importance_values = curr_checkpoint_ordered$importance
          curr_se_values = curr_checkpoint_ordered$se

          # Calculate range of importance values across all features (matching fippy)
          importance_range = max(curr_importance_values, na.rm = TRUE) -
            min(curr_importance_values, na.rm = TRUE)

          # Normalize SE by range to get relative SE (matching fippy's formula)
          # fippy: ratio = SE / range, convergence if max(ratio) < threshold
          # https://github.com/gcskoenig/fippy/blob/a7a37aa5511f7074ead3289c89b1ae80036982cb/src/fippy/explainers/utils.py#L40-L42
          if (importance_range > 0 && is.finite(importance_range)) {
            relative_se_values = curr_se_values / importance_range
            max_relative_se = max(relative_se_values, na.rm = TRUE)
          } else {
            # If range is 0 or invalid, use absolute SE (fallback)
            max_relative_se = max(curr_se_values, na.rm = TRUE)
          }

          # Check convergence: max relative SE below threshold
          converged = is.finite(max_relative_se) && max_relative_se < se_threshold
          convergence_msg = c(
            "v" = "SAGE converged after {.val {n_completed}} permutations",
            "i" = "Maximum relative SE: {.val {round(max_relative_se, 4)}} (threshold: {.val {se_threshold}})",
            "i" = "Saved {.val {n_permutations - n_completed}} permutations"
          )

          if (xplain_opt("verbose") && converged) {
            cli::cli_inform(convergence_msg)
          }
        }
      }

      # Close the progress bar.
      if (xplain_opt("progress")) {
        cli::cli_progress_done()
      }

      # Calculate the final average SAGE values based on all completed permutations.
      final_sage_values = sage_values / n_completed

      # Return the computed scores and convergence data.
      list(
        scores = data.table(
          feature = names(final_sage_values),
          importance = as.numeric(final_sage_values)
        ),
        convergence_data = list(
          convergence_history = if (length(convergence_history) > 0) {
            rbindlist(convergence_history)
          } else {
            NULL
          },
          converged = converged,
          n_permutations_used = n_completed
        )
      )
    },

    # Template method: Defines the complete prediction and aggregation pipeline
    # Subclasses only need to implement .expand_coalitions_data()
    .evaluate_coalitions_batch = function(learner, test_dt, all_coalitions, batch_size = NULL) {
      n_test = nrow(test_dt)

      if (xplain_opt("debug")) {
        cli::cli_inform("Evaluating {.val {length(all_coalitions)}} coalitions")
      }

      # STEP 1: Subclass-specific data expansion (abstract method)
      # combined data has rows `n_samples * nrow(test_dt) * length(all_coalitions)`
      # Full coalition -> return is just test_dt
      combined_data = private$.expand_coalitions_data(test_dt, all_coalitions)
      # STEPS 2-5: Shared processing pipeline using general utilities
      predictions = sage_batch_predict(
        learner,
        combined_data,
        self$task,
        batch_size,
        self$task$task_type
      )
      if (anyNA(predictions)) {
        cli::cli_warn("Encountered missing values in model prediction")
      }
      avg_preds = sage_aggregate_predictions(
        combined_data,
        predictions,
        self$task$task_type,
        self$task$class_names
      )

      # Private method (needs self$task and self$measure)
      coalition_losses = private$.calculate_coalition_losses(avg_preds, n_test, test_dt)

      coalition_losses
    },

    # Abstract method - must be implemented by subclasses
    # Returns: data.table with all feature columns plus .coalition_id and .test_instance_id
    .expand_coalitions_data = function(test_dt, all_coalitions) {
      cli::cli_abort(c(
        "Abstract method not implemented",
        "i" = "Subclasses must implement {.fn .expand_coalitions_data}",
        "i" = "This method should return a data.table with:",
        "*" = "All feature columns (with marginalized features replaced/sampled)",
        "*" = "{.field .coalition_id}: integer identifying which coalition",
        "*" = "{.field .test_instance_id}: integer identifying original test instance"
      ))
    },

    # Private method - needs self$task and self$measure
    # Calculates losses from averaged predictions for each coalition.
    # Returns losses ordered by ascending `.coalition_id`, matching the
    # order coalitions were built (the closed-form accumulation indexes
    # this positionally).
    .calculate_coalition_losses = function(avg_preds, n_test, test_dt) {
      .coalition_id = .test_instance_id = NULL # data.table NSE NOTE tax

      coalition_ids = sort(unique(avg_preds$.coalition_id))
      truth = test_dt[[self$task$target_names]]

      # Key once and subset by key (binary search) per coalition,
      # rather than scanning `avg_preds[.coalition_id == i]` in a loop.
      # The key includes .test_instance_id so each block is
      # ordered to align with `truth` (test instance 1..n_test).
      setkey(avg_preds, .coalition_id, .test_instance_id)
      measure = self$measure
      is_classif = self$task$task_type == "classif"
      class_names = self$task$class_names

      # For pointwise (`obs_loss`) measures whose predict_type matches
      # the data we hold (regression response; classification prob), we
      # can call the measure's own mlr3measures function (`$fun`)
      # directly on (truth, response/prob), skipping the per-coalition
      # Prediction object. This reuses the mlr3 measure
      # implementation and does the
      # correct per-measure aggregation. Anything else (classification
      # response measures that need a prob->class step, measures needing
      # task/model context, non-decomposable measures like AUC) takes
      # the canonical Prediction$score() path, so correctness holds for
      # every measure.
      direct = !is.null(measure$fun) &&
        "obs_loss" %in% measure$properties &&
        ((!is_classif && measure$predict_type == "response") ||
          (is_classif && measure$predict_type == "prob"))

      score_block = if (direct && is_classif) {
        function(block) {
          measure$fun(truth = truth, prob = as.matrix(block[, .SD, .SDcols = class_names]))
        }
      } else if (direct) {
        function(block) measure$fun(truth = truth, response = block$avg_pred)
      } else if (is_classif) {
        function(block) {
          PredictionClassif$new(
            row_ids = seq_len(n_test),
            truth = truth,
            prob = as.matrix(block[, .SD, .SDcols = class_names])
          )$score(measure)
        }
      } else {
        function(block) {
          PredictionRegr$new(
            row_ids = seq_len(n_test),
            truth = truth,
            response = block$avg_pred
          )$score(measure)
        }
      }

      coalition_losses = numeric(length(coalition_ids))
      for (k in seq_along(coalition_ids)) {
        coalition_losses[k] = score_block(avg_preds[list(coalition_ids[k])])
      }

      coalition_losses
    }
  )
)

#' @title Create Feature Selection Design Matrix
#'
#' @description
#' Creates a logical design matrix for leave-in or leave-out feature evaluation.
#' Used internally with mlr3fselect to evaluate feature subsets.
#'
#' @param all_features (character()) All available feature names from the task.
#' @param feature_names (character() | list of character()) Features or feature
#'   groups to evaluate. Can be a vector for individual features or a named list
#'   for grouped features. Defaults to `all_features` if unspecified.
#' @param direction (character(1)) Either `"leave-in"` or `"leave-out"` (default).
#'   Controls which features are selected in the design matrix.
#'   `"leave-out"` sets features of interest to `FALSE`, and `"leave-in"` analogously
#'   sets them to `TRUE`.
#'
#' @return data.table with logical columns for each feature in `all_features` and
#'   `length(feature_names)` rows, one for each entry in `feature_names`
#'
#' @examples
#' task = mlr3::tsk("mtcars")
#'
#' # Individual features
#' feature_names = task$feature_names[1:3]
#' wvim_design_matrix(task$feature_names, feature_names, "leave-in")
#' wvim_design_matrix(task$feature_names, feature_names, "leave-out")
#'
#' # Feature groups
#' feature_groups = list(
#'   A = task$feature_names[1:2],
#'   B = task$feature_names[3:5]
#' )
#' wvim_design_matrix(task$feature_names, feature_groups, "leave-out")
#'
#' @export
wvim_design_matrix = function(
  all_features,
  feature_names = all_features,
  direction = c("leave-out", "leave-in")
) {
  if (is.list(feature_names)) {
    feature_names = check_groups(feature_names, all_features = all_features)
  } else {
    checkmate::assert_subset(feature_names, all_features)
  }

  direction = match.arg(direction)
  # Initialize all-FALSE matrix
  # one column for each feature in original task (must match)
  # one row for feature or group of features of interest (vector or list)
  design_mat = matrix(
    FALSE,
    nrow = length(feature_names),
    ncol = length(all_features),
    dimnames = list(NULL, all_features)
  )

  # feature_names can be vector or list, e.g.
  # feature_names = c("x1", "x2")
  # feature_names = list(A = c("x1", "x2"), B = c("x3"))
  for (feature_vec_idx in seq_along(feature_names)) {
    # Get matching index for original task feature names to ensure correct order
    # user might supply feature_names with arbitrary order
    fidx = match(feature_names[[feature_vec_idx]], all_features)
    # Flip value for corresponding feature index
    design_mat[feature_vec_idx, fidx] = TRUE
  }

  # When leaving features out we flip the logic
  # -> FALSE features are left out, TRUE are left in
  if (direction == "leave-out") {
    design_mat = !design_mat
  }

  design_dt = as.data.table(design_mat)

  stopifnot(nrow(design_dt) == length(feature_names))
  stopifnot(ncol(design_dt) == length(all_features))
  design_dt
}

#' @title Marginal Reference Sampler
#'
#' @description Samples complete observations from reference data to replace feature values.
#' This approach samples from the marginal distribution while preserving within-row
#' feature dependencies.
#'
#' @details
#' This sampler implements what is called "marginal imputation" in the SAGE literature
#' (Covert et al. 2020). For each observation, it samples a complete row from reference
#' data and takes the specified feature values from that row. This approach:
#'
#' - Samples from the marginal distribution \eqn{P(X_S)} where S is the set of features
#' - Preserves dependencies **within** the sampled reference row
#' - Breaks dependencies **between** test and reference data
#'
#' **Terminology note:** In SAGE literature, this is called "marginal imputation" because
#' features outside the coalition are "imputed" by sampling from their marginal distribution.
#' We use `MarginalReferenceSampler` to avoid confusion with missing data imputation and to
#' clarify that it samples from reference data.
#'
#' **Comparison with other samplers:**
#'
#' - `MarginalPermutationSampler`: Shuffles each feature independently, breaking all row structure
#' - `MarginalReferenceSampler`: Samples complete rows, preserving within-row dependencies
#' - `ConditionalSampler`: Samples from \eqn{P(X_S | X_{-S})}, conditioning on other features
#'
#' **Use in SAGE:**
#'
#' This is the default approach for `MarginalSAGE`. For a test observation x and features
#' to marginalize S, it samples a reference row x_ref and creates a "hybrid" observation
#' combining x's coalition features with x_ref's marginalized features.
#'
#' @examples
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#'
#' # Default: uses all task data as reference
#' sampler = MarginalReferenceSampler$new(task)
#' sampled = sampler$sample("important1", row_ids = 1:10)
#'
#' # Subsample reference data to 50 rows
#' sampler_subsampled = MarginalReferenceSampler$new(task, n_samples = 50L)
#' sampled2 = sampler_subsampled$sample("important1", row_ids = 1:10)
#'
#' @references `r print_bib("lundberg_2020")`

#'
#' @export
MarginalReferenceSampler = R6Class(
  "MarginalReferenceSampler",
  inherit = MarginalSampler,
  public = list(
    #' @field reference_data ([`data.table`][data.table::data.table])
    #'   Reference data to sample from for marginalization.
    reference_data = NULL,

    #' @description
    #' Creates a new instance of the MarginalReferenceSampler class.
    #' @param task ([mlr3::Task]) Task to sample from.
    #' @param n_samples (`integer(1)` | `NULL`) Number of reference samples to use.
    #'   If `NULL`, uses all task data as reference.
    initialize = function(task, n_samples = NULL) {
      super$initialize(task)

      if (is.null(n_samples)) {
        # Use all task data as reference
        self$reference_data = task$data(cols = task$feature_names)
      } else {
        # Subsample n_samples rows from task
        n_ref = min(n_samples, task$nrow)
        ref_indices = sample(task$nrow, n_ref)
        self$reference_data = task$data(rows = ref_indices, cols = task$feature_names)
      }

      self$label = "Marginal reference sampler"
    }
  ),

  private = list(
    # Implement marginal sampling via reference row sampling
    .sample_marginal = function(data, feature, samples_per_row = 1L) {
      n = nrow(data)
      total = n * samples_per_row

      sampled_indices = sample.int(
        nrow(self$reference_data),
        total,
        replace = TRUE
      )

      out = data[rep.int(seq_len(.N), times = samples_per_row)]
      out[, (feature) := self$reference_data[sampled_indices, .SD, .SDcols = feature]]

      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)

#' @title Conditional SAGE
#'
#' @description [SAGE] with conditional sampling (features are "marginalized" conditionally).
#' Uses [ConditionalARFSampler] as default [ConditionalSampler].
#'
#' @seealso [MarginalSAGE]
#'
#' @examplesIf requireNamespace("ranger", quietly = TRUE) && requireNamespace("mlr3learners", quietly = TRUE) && requireNamespace("arf", quietly = TRUE)
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#'
#' \donttest{
#' # Using default ConditionalARFSampler (also handles all mixed data)
#' sage = ConditionalSAGE$new(
#'   task = task,
#'   learner = lrn("regr.ranger", num.trees = 50),
#'   measure = msr("regr.mse"),
#'   n_permutations = 3L,
#'   n_samples = 20
#' )
#' sage$compute()
#' }
#' \donttest{
#' # For alternative conditional samplers:
#' custom_sampler = ConditionalGaussianSampler$new(
#'   task = task
#' )
#' sage_custom = ConditionalSAGE$new(
#'   task = task,
#'   learner = lrn("regr.ranger", num.trees = 50),
#'   measure = msr("regr.mse"),
#'   n_permutations = 5L,
#'   n_samples = 20,
#'   sampler = custom_sampler
#' )
#' sage_custom$compute()
#' }
#' @export
ConditionalSAGE = R6Class(
  "ConditionalSAGE",
  inherit = SAGE,
  public = list(
    #' @field sampler ([ConditionalSampler]) Sampler for conditional marginalization.
    sampler = NULL,

    #' @description
    #' Creates a new instance of the ConditionalSAGE class.
    #' @param task,learner,measure,resampling,features,n_permutations,batch_size,n_samples,early_stopping,se_threshold,min_permutations,check_interval Passed to [SAGE].
    #' @param sampler ([ConditionalSampler]) Optional custom sampler. Defaults to [ConditionalARFSampler].
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      n_permutations = 10L,
      sampler = NULL,
      batch_size = 5000L,
      n_samples = 100L,
      early_stopping = FALSE,
      se_threshold = 0.01,
      min_permutations = 10L,
      check_interval = 1L
    ) {
      # Use ConditionalARFSampler by default
      if (is.null(sampler)) {
        sampler = ConditionalARFSampler$new(task)
        if (xplain_opt("verbose")) {
          cli::cli_alert_info(
            "No {.cls ConditionalSampler} provided, using {.cls ConditionalARFSampler} with default settings."
          )
        }
      } else {
        checkmate::assert_class(sampler, "ConditionalSampler")
      }

      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        n_permutations = n_permutations,
        batch_size = batch_size,
        n_samples = n_samples,
        early_stopping = early_stopping,
        se_threshold = se_threshold,
        min_permutations = min_permutations,
        check_interval = check_interval
      )

      # Store sampler (specific to ConditionalSAGE)
      self$sampler = sampler

      self$label = "Conditional SAGE"
    }
  ),

  private = list(
    # Conditional-specific data expansion: Sample from conditional distribution
    # For each coalition, sample marginalized features conditioned on coalition features
    # Uses n_samples to create multiple samples per test instance for averaging
    .expand_coalitions_data = function(test_dt, all_coalitions) {
      n_test = nrow(test_dt)
      all_coalition_data = vector("list", length(all_coalitions))

      # For each coalition, sample from conditional distribution P(X_{-S} | X_S)
      for (i in seq_along(all_coalitions)) {
        coalition = all_coalitions[[i]]
        marginalize_features = setdiff(self$features, coalition)

        if (length(marginalize_features) > 0) {
          n_samples = self$param_set$values$n_samples

          # Sample conditionally using a sampler. Pass `samples_per_row = n_samples`
          # on the unique test rows so the sampler can use its native efficient
          # multi-draw path (e.g. ConditionalARFSampler invokes
          # `arf::forge(n_synth = n_samples)` on n_test evidence rows rather than
          # n_samples * n_test replicated rows).
          # Returns test_dt with "marginalized" features replaced by conditional samples
          # sampler: P(marginalize_features | coalition_features)
          marginalized_test = self$sampler$sample_newdata(
            feature = marginalize_features,
            newdata = test_dt,
            conditioning_set = coalition,
            samples_per_row = n_samples
          )

          # Tag rows with their originating test instance. Sampler output is
          # draw-major: rows 1..n_test = draw 1 across all test rows in order,
          # rows n_test+1..2*n_test = draw 2, etc. Hence `times`, not `each`.
          # Must be done AFTER sampling since sampler does not preserve extra columns.
          marginalized_test[,
            .test_instance_id := rep(seq_len(n_test), times = n_samples)
          ]
        } else {
          # If marginalize_features is empty then we evaluate the coalition of all features
          # in the task, so there's no room for any sampling and we just take all data
          # no need to duplicate+average anything since prediction is deterministic (usually (I hope (right?)))
          marginalized_test = copy(test_dt)
          marginalized_test[, .test_instance_id := seq_len(n_test)]
        }

        # Add coalition ID for tracking which coalition this data belongs to
        marginalized_test[, .coalition_id := i]

        all_coalition_data[[i]] = marginalized_test
      }

      # Combine all coalitions into single data.table for batch prediction
      rbindlist(all_coalition_data)
    }
  )
)

#' Perform batched predictions with automatic handling of fast vs regular predict methods
#'
#' @param learner mlr3 Learner object
#' @param data_list list of data.tables to predict on
#' @param task mlr3 Task object
#' @param test_row_ids integer vector of test row IDs for constructing predictions
#' @param batch_size integer or NULL. If NULL, predicts all data at once. Otherwise batches by this size.
#'
#' @return list of [Prediction][mlr3::Prediction] objects, one per element in `data_list`
#' @keywords internal
#' @noRd
predict_batched = function(learner, data_list, task, test_row_ids, batch_size = NULL) {
  n_data = length(data_list)
  test_size = length(test_row_ids)
  use_fast = is.function(learner$predict_newdata_fast)

  # Helper to construct Prediction object from raw prediction list
  construct_pred = function(raw_pred) {
    truth = task$truth(rows = test_row_ids)

    switch(
      task$task_type,
      classif = PredictionClassif$new(
        row_ids = test_row_ids,
        truth = truth,
        response = raw_pred$response,
        prob = raw_pred$prob
      ),
      regr = PredictionRegr$new(
        row_ids = test_row_ids,
        truth = truth,
        response = raw_pred$response
      )
    )
  }

  # If no batching or only one data element, predict directly
  if (is.null(batch_size) || n_data == 1) {
    combined_data = rbindlist(data_list)

    if (use_fast) {
      # Fast prediction on all data
      pred_raw = learner$predict_newdata_fast(newdata = combined_data, task = task)

      # Split into n_data predictions
      preds = lapply(seq_len(n_data), function(i) {
        start_idx = (i - 1) * test_size + 1
        end_idx = i * test_size

        # Extract slice for this repeat
        # Need to preserve matrix structure for prob predictions
        pred_slice = lapply(pred_raw, function(component) {
          if (!is.null(component) && length(component) > 0) {
            if (is.matrix(component)) {
              component[start_idx:end_idx, , drop = FALSE]
            } else {
              component[start_idx:end_idx]
            }
          } else {
            NULL
          }
        })

        construct_pred(pred_slice)
      })
    } else {
      # Regular prediction per data element
      preds = lapply(data_list, function(data_chunk) {
        learner$predict_newdata(newdata = data_chunk, task = task)
      })
    }

    return(preds)
  }

  # Batched prediction case
  # Calculate how many data elements fit in each batch
  data_per_batch = max(1, floor(batch_size / test_size))
  n_batches = ceiling(n_data / data_per_batch)

  preds = vector("list", n_data)

  for (batch_idx in seq_len(n_batches)) {
    # Determine which data elements are in this batch
    start_data_idx = (batch_idx - 1) * data_per_batch + 1
    end_data_idx = min(batch_idx * data_per_batch, n_data)
    data_indices = start_data_idx:end_data_idx

    # Combine data for this batch
    batch_data = rbindlist(data_list[data_indices])

    if (use_fast) {
      pred_raw = learner$predict_newdata_fast(newdata = batch_data, task = task)

      # Split back into individual predictions
      for (i in seq_along(data_indices)) {
        start_idx = (i - 1) * test_size + 1
        end_idx = i * test_size

        # Need to preserve matrix structure for prob predictions
        pred_slice = lapply(pred_raw, function(component) {
          if (!is.null(component) && length(component) > 0) {
            if (is.matrix(component)) {
              component[start_idx:end_idx, , drop = FALSE]
            } else {
              component[start_idx:end_idx]
            }
          } else {
            NULL
          }
        })

        preds[[data_indices[i]]] = construct_pred(pred_slice)
      }
    } else {
      # Regular prediction: need to predict each data element separately
      for (i in seq_along(data_indices)) {
        preds[[data_indices[i]]] = learner$predict_newdata(
          newdata = data_list[[data_indices[i]]],
          task = task
        )
      }
    }
  }

  preds
}

#' @title Knockoff Sampler
#'
#' @description Implements conditional sampling using Knockoffs.
#'
#' @details
#' The `KnockoffSampler` samples [Knockoffs][knockoff::knockoff] based on the task data.
#' This class allows arbitrary `knockoff_fun`, which also means that no input checking
#' against supported feature types can be done. Use [KnockoffGaussianSampler] for the
#' Gaussian knockoff sampler for numeric features.
#' Alternative knockoff samplers include `knockoff_seq()` from the `seqknockoff` package
#' available on GitHub: <https://github.com/kormama1/seqknockoff>.
#'
#' Knockoffs are related to the `ConditionalSampler` familty, with key differences:
#' They do not allow specifying a `conditioning_set`
#'
#' @examplesIf requireNamespace("knockoff", quietly = TRUE)
#' library(mlr3)
#' task = tgen("2dnormals")$generate(n = 100)
#' # Create sampler with default parameters
#' sampler = KnockoffSampler$new(task)
#' # Sample using row_ids from stored task
#' sampled_data = sampler$sample("x1")
#'
#' @references `r print_bib("watson_2021", "blesch_2023")`
#' @export
KnockoffSampler = R6Class(
  "KnockoffSampler",
  inherit = FeatureSampler,
  public = list(
    #' @field x_tilde Knockoff matrix with one (or `iters`) row(s) per original observation in `task`.
    x_tilde = NULL,
    #' @description
    #' Creates a new instance of the KnockoffSampler class.
    #' @param task ([mlr3::Task]) Task to sample from
    #' @param knockoff_fun (`function`) Function used to create knockoff matrix. Default are second-order Gaussian knockoffs (`knockoff::create.second_order()`)
    #' @param iters (`integer(1)`: `1`) Number of repetitions the `knockoff_fun` is applied to create multiple `x_tilde`
    #' instances per observation.
    initialize = function(
      task,
      knockoff_fun = function(x) knockoff::create.second_order(as.matrix(x)),
      iters = 1
    ) {
      super$initialize(task)
      self$label = "Knockoff sampler"

      require_package("knockoff")

      # Override param_set to include Knockoff-specific parameters
      self$param_set = paradox::ps(
        # conditioning_set = paradox::p_uty(default = NULL),
        knockoff_fun = paradox::p_uty(
          default = function(x) knockoff::create.second_order(as.matrix(x)),
          custom_check = function(x) {
            if (is.function(x)) TRUE else "knockoff_fun must be a function."
          }
        ),
        iters = paradox::p_int(lower = 1, default = 1)
      )

      # Set parameter values
      values_to_set = list()
      values_to_set$knockoff_fun = knockoff_fun
      values_to_set$iters = iters
      self$param_set$set_values(.values = values_to_set)

      # Create knockoff matrix, features only
      # No assertions here on feature types, the user has been warned in the doc
      if (iters == 1) {
        self$x_tilde = as.data.table(knockoff_fun(self$task$data(cols = self$task$feature_names)))
        self$x_tilde[, ..row_id := self$task$row_ids]
      } else {
        self$x_tilde = rbindlist(replicate(
          iters,
          {
            x_tilde = as.data.table(knockoff_fun(self$task$data(cols = self$task$feature_names)))
            x_tilde[, ..row_id := self$task$row_ids]
          },
          simplify = FALSE
        ))
      }

      checkmate::assert_subset(colnames(self$x_tilde), c(self$task$feature_names, "..row_id"))
      checkmate::assert_true(nrow(self$x_tilde) == (iters * self$task$nrow))
    },

    #' @description
    #' Sample from stored task using knockoff values. Replaces specified feature(s) with
    #' their knockoff counterparts from the pre-generated knockoff matrix.
    #' @param feature (`character`) Feature(s) to sample.
    #' @param row_ids (`integer()` | `NULL`) Row IDs to use. If `NULL`, uses all rows.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @return Modified copy with knockoff feature(s).
    sample = function(
      feature,
      row_ids = NULL,
      samples_per_row = 1L
    ) {
      checkmate::assert_count(samples_per_row, positive = TRUE)
      if (is.null(row_ids)) {
        row_ids = self$task$row_ids
      }
      iters = self$param_set$values$iters

      replace = samples_per_row > iters
      if (replace) {
        cli::cli_warn(c(
          "!" = "Some instances requested more often than they are present in generated knockoff matrix",
          i = "Will sample with replacement, so some knockoff values will be duplicated",
          i = "Create {.cls {class(self)[[1L]]}} with {.code iters = {samples_per_row}} or higher to prevent this"
        ))
      }

      data_copy = private$.get_task_data_by_row_id(row_ids)

      out = data_copy[rep.int(seq_len(.N), times = samples_per_row)]

      # Pick a knockoff iter index per draw. Without replacement when possible
      # (samples_per_row <= iters); with replacement otherwise (warned above).
      iter_idx_per_draw = if (replace) {
        sample.int(iters, samples_per_row, replace = TRUE)
      } else {
        sample.int(iters, samples_per_row, replace = FALSE)
      }

      # Collect per-draw feature blocks then assign in bulk so column type
      # from x_tilde (e.g. numeric) replaces any incompatible task type
      # (e.g. integer) without triggering data.table truncation warnings.
      feat_blocks = lapply(seq_len(samples_per_row), function(d) {
        iter_i = iter_idx_per_draw[d]
        # x_tilde is `iters` knockoff matrices stacked iter-major: rows
        # ((i-1)*task$nrow + 1) .. (i*task$nrow) belong to iter i.
        iter_block = self$x_tilde[((iter_i - 1L) * self$task$nrow + 1L):(iter_i * self$task$nrow)]
        row_idx_in_block = match(row_ids, iter_block[["..row_id"]])
        iter_block[row_idx_in_block, .SD, .SDcols = feature]
      })
      feat_values = data.table::rbindlist(feat_blocks)
      out[, (feature) := feat_values]

      out = private$.ensure_feature_types(out)
      setcolorder(out, self$task$feature_names)
      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)


#' @title Gaussian Knockoff Conditional Sampler
#'
#' @description
#' A [KnockoffSampler] defaulting to second-order Gaussian knockoffs
#' as created by [knockoff::create.second_order].
#'
#' @details
#' This is equivalent to [KnockoffSampler] using the default `knockoff_fun`.
#'
#' @examplesIf requireNamespace("knockoff", quietly = TRUE)
#' library(mlr3)
#' task = tgen("2dnormals")$generate(n = 100)
#' # Create sampler
#' sampler = KnockoffGaussianSampler$new(task)
#' # Sample using row_ids from stored task
#' sampled_data = sampler$sample("x1")
#' @references `r print_bib("watson_2021", "blesch_2023")`
#'
#' @export
KnockoffGaussianSampler = R6Class(
  "KnockoffGaussianSampler",
  inherit = KnockoffSampler,
  public = list(
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    #'   Will be checked against the provided [mlr3::Task] to ensure compatibility.
    feature_types = c("numeric", "integer"),
    #' @field x_tilde Knockoff matrix
    x_tilde = NULL,

    #' @description
    #' Creates a new instance using Gaussian knockoffs via [knockoff::create.second_order].
    #' @param task ([mlr3::Task]) Task to sample from.
    #' @param iters (`integer(1)`: `1`) Number of repetitions the `knockoff_fun` is applied to create multiple `x_tilde`
    #' instances per observation.
    initialize = function(
      task,
      iters = 1
    ) {
      require_package("knockoff")
      super$initialize(
        task = task,
        knockoff_fun = function(x) {
          knockoff::create.second_order(as.matrix(x))
        },
        iters = iters
      )
      self$label = "Gaussian Knockoff sampler"
    }
  )
)

#' Hierarchical parameter resolution helper
#'
#' Used to retrieve the highest-priority parameter value.
#'
#' Priority is `arg` > `stored` > `default`.
#'
#' @param arg,stored,default Scalar parameter values evlauted in that order
#'
#' @return First non-`NULL` parameter
#' @keywords internal
#' @noRd
#' @examples
#' resolve_param(arg = "foo", stored = "bar", default = "baz")
#' resolve_param(arg = NULL, stored = "bar", default = "baz")
#' resolve_param(arg = NULL, stored = NULL, default = "baz")
resolve_param = function(arg, stored, default) (arg %||% stored) %||% default

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

# Cap threads before any BLAS/OpenMP init: CRAN's Debian machine links a multithreaded (OpenMP) BLAS whose
# spin-pool inflates the CPU/elapsed ratio. Must be set here, before library() and any matrix op, to take effect.
Sys.setenv("OMP_THREAD_LIMIT" = 2)

library(testthat)
library(xplainfi)

test_check("xplainfi")

#' Print an Rd-formatted bib entry
#'
#' @keywords internal
#' @param ... (`character`) One or more quoted names of `bibentries` to print.
#' @importFrom tools toRd
#' @importFrom utils bibentry
print_bib = function(...) {
  str = sapply(list(...), function(entry) tools::toRd(bibentries[[entry]]))
  paste0(str, collapse = "\n\n")
}

#' @importFrom utils person
bibentries = c(
  ewald_2024 = bibentry(
    "inproceedings",
    title = "A Guide to Feature Importance Methods for Scientific Inference",
    author = c(
      person("Fiona Katharina", "Ewald"),
      person("Ludwig", "Bothmann"),
      person("Marvin N.", "Wright"),
      person("Bernd", "Bischl"),
      person("Giuseppe", "Casalicchio"),
      person("Gunnar", "K\u00f6nig")
    ),
    year = "2024",
    booktitle = "Explainable Artificial Intelligence",
    editor = c(
      person("Luca", "Longo"),
      person("Sebastian", "Lapuschkin"),
      person("Christin", "Seifert")
    ),
    pages = "440--464",
    publisher = "Springer Nature Switzerland",
    location = "Cham",
    doi = "10.1007/978-3-031-63797-1_22",
    isbn = "978-3-031-63797-1"
  ),

  konig_2021 = bibentry(
    "inproceedings",
    title = "Relative Feature Importance",
    author = c(
      person("Gunnar", "K\u00f6nig"),
      person("Christoph", "Molnar"),
      person("Bernd", "Bischl"),
      person("Moritz", "Grosse-Wentrup")
    ),
    year = "2021",
    booktitle = "2020 25th International Conference on Pattern Recognition (ICPR)",
    pages = "9318--9325",
    doi = "10.1109/ICPR48806.2021.9413090"
  ),

  blesch_2025 = bibentry(
    "article",
    title = "Conditional Feature Importance with Generative Modeling Using Adversarial Random Forests",
    author = c(
      person("Kristin", "Blesch"),
      person("Niklas", "Koenen"),
      person("Jan", "Kapar"),
      person("Pegah", "Golchian"),
      person("Lukas", "Burk"),
      person("Markus", "Loecher"),
      person("Marvin N.", "Wright")
    ),
    year = "2025",
    journal = "Proceedings of the AAAI Conference on Artificial Intelligence",
    volume = "39",
    number = "15",
    pages = "15596--15604",
    doi = "10.1609/aaai.v39i15.33712"
  ),

  watson_2023 = bibentry(
    "inproceedings",
    title = "Adversarial Random Forests for Density Estimation and Generative Modeling",
    author = c(
      person("David S.", "Watson"),
      person("Kristin", "Blesch"),
      person("Jan", "Kapar"),
      person("Marvin N.", "Wright")
    ),
    year = "2023",
    booktitle = "Proceedings of The 26th International Conference on Artificial Intelligence and Statistics",
    pages = "5357--5375",
    publisher = "PMLR",
    url = "https://proceedings.mlr.press/v206/watson23a.html"
  ),

  breiman_2001 = bibentry(
    "article",
    title = "Random Forests",
    author = person("Leo", "Breiman"),
    year = "2001",
    journal = "Machine Learning",
    volume = "45",
    number = "1",
    pages = "5--32",
    doi = "10.1023/A:1010933404324"
  ),

  fisher_2019 = bibentry(
    "article",
    title = "All Models Are Wrong, but Many Are Useful: Learning a Variable's Importance by Studying an Entire Class of Prediction Models Simultaneously",
    author = c(
      person("Aaron", "Fisher"),
      person("Cynthia", "Rudin"),
      person("Francesca", "Dominici")
    ),
    year = "2019",
    journal = "Journal of Machine Learning Research",
    volume = "20",
    pages = "177",
    url = "https://pmc.ncbi.nlm.nih.gov/articles/PMC8323609/"
  ),

  lundberg_2020 = bibentry(
    "inproceedings",
    title = "Understanding Global Feature Contributions With Additive Importance Measures",
    author = c(
      person("Ian", "Covert"),
      person("Scott M.", "Lundberg"),
      person("Su-In", "Lee")
    ),
    year = "2020",
    booktitle = "Advances in Neural Information Processing Systems",
    volume = "33",
    pages = "17212--17223",
    publisher = "Curran Associates, Inc.",
    url = "https://proceedings.neurips.cc/paper/2020/hash/c7bf0b7c1a86d5eb3be2c722cf2cf746-Abstract.html"
  ),

  watson_2021 = bibentry(
    "article",
    title = "Testing Conditional Independence in Supervised Learning Algorithms",
    author = c(
      person("David S.", "Watson"),
      person("Marvin N.", "Wright")
    ),
    year = "2021",
    journal = "Machine Learning",
    volume = "110",
    number = "8",
    pages = "2107--2129",
    doi = "10.1007/s10994-021-06030-6"
  ),

  blesch_2023 = bibentry(
    "article",
    title = "Conditional Feature Importance for Mixed Data",
    author = c(
      person("Kristin", "Blesch"),
      person("David S.", "Watson"),
      person("Marvin N.", "Wright")
    ),
    year = "2023",
    journal = "AStA Advances in Statistical Analysis",
    volume = "108",
    number = "2",
    pages = "259--278",
    doi = "10.1007/s10182-023-00477-9"
  ),

  lei_2018 = bibentry(
    "article",
    title = "Distribution-Free Predictive Inference for Regression",
    author = c(
      person("Jing", "Lei"),
      person("Max", "G'Sell"),
      person("Alessandro", "Rinaldo"),
      person("Ryan J.", "Tibshirani"),
      person("Larry", "Wasserman")
    ),
    year = "2018",
    journal = "Journal of the American Statistical Association",
    volume = "113",
    number = "523",
    pages = "1094--1111",
    doi = "10.1080/01621459.2017.1307116"
  ),

  strobl_2008 = bibentry(
    "article",
    title = "Conditional Variable Importance for Random Forests",
    author = c(
      person("Carolin", "Strobl"),
      person("Anne-Laure", "Boulesteix"),
      person("Thomas", "Kneib"),
      person("Thomas", "Augustin"),
      person("Achim", "Zeileis")
    ),
    year = "2008",
    journal = "BMC Bioinformatics",
    volume = "9",
    number = "1",
    pages = "307",
    doi = "10.1186/1471-2105-9-307"
  ),

  molnar_2023 = bibentry(
    "inproceedings",
    title = "Relating the Partial Dependence Plot and Permutation Feature Importance to the Data Generating Process",
    author = c(
      person("Christoph", "Molnar"),
      person("Timo", "Freiesleben"),
      person("Gunnar", "K\u00f6nig"),
      person("Julia", "Herbinger"),
      person("Tim", "Reisinger"),
      person("Giuseppe", "Casalicchio"),
      person("Marvin N.", "Wright"),
      person("Bernd", "Bischl")
    ),
    year = "2023",
    booktitle = "Explainable Artificial Intelligence",
    editor = person("Luca", "Longo"),
    pages = "456--479",
    publisher = "Springer Nature Switzerland",
    doi = "10.1007/978-3-031-44064-9_24",
    isbn = "978-3-031-44064-9"
  ),

  nadeau_2003 = bibentry(
    "article",
    title = "Inference for the Generalization Error",
    author = c(
      person("Claude", "Nadeau"),
      person("Yoshua", "Bengio")
    ),
    year = "2003",
    journal = "Machine Learning",
    volume = "52",
    number = "3",
    pages = "239--281",
    doi = "10.1023/A:1024068626366"
  ),

  little_2019 = bibentry(
    "book",
    title = "Statistical Analysis with Missing Data",
    author = c(
      person("Roderick J. A.", "Little"),
      person("Donald B.", "Rubin")
    ),
    year = "2019",
    edition = "3rd",
    publisher = "John Wiley & Sons",
    address = "Hoboken, NJ",
    isbn = "9780470526798"
  ),

  troyanskaya_2001 = bibentry(
    "article",
    title = "Missing Value Estimation Methods for DNA Microarrays",
    author = c(
      person("Olga", "Troyanskaya"),
      person("Michael", "Cantor"),
      person("Gavin", "Sherlock"),
      person("Pat", "Brown"),
      person("Trevor", "Hastie"),
      person("Robert", "Tibshirani"),
      person("David", "Botstein"),
      person("Russ B.", "Altman")
    ),
    year = "2001",
    journal = "Bioinformatics",
    volume = "17",
    number = "6",
    pages = "520--525",
    doi = "10.1093/bioinformatics/17.6.520"
  ),

  anderson_2003 = bibentry(
    "book",
    title = "An Introduction to Multivariate Statistical Analysis",
    author = person("Theodore W.", "Anderson"),
    year = "2003",
    edition = "3rd",
    publisher = "Wiley-Interscience",
    address = "Hoboken, NJ",
    isbn = "9780471360919"
  ),

  hothorn_2006 = bibentry(
    "article",
    title = "Unbiased Recursive Partitioning: A Conditional Inference Framework",
    author = c(
      person("Torsten", "Hothorn"),
      person("Kurt", "Hornik"),
      person("Achim", "Zeileis")
    ),
    year = "2006",
    journal = "Journal of Computational and Graphical Statistics",
    volume = "15",
    number = "3",
    pages = "651--674",
    doi = "10.1198/106186006X133933"
  ),

  aas_2021 = bibentry(
    "article",
    title = "Explaining Individual Predictions When Features Are Dependent: More Accurate Approximations to Shapley Values",
    author = c(
      person("Kjersti", "Aas"),
      person("Martin", "Jullum"),
      person("Anders", "L\u00f8land")
    ),
    year = "2021",
    journal = "Artificial Intelligence",
    volume = "298",
    pages = "103502",
    doi = "10.1016/j.artint.2021.103502"
  )
)

test_that("bibentries are valid", {
  # Don't run this on R-devel for now due to irrelevant formatting difference
  skip_if(as.numeric(R.version$minor) >= 6)

  expect_s3_class(bibentries, "bibentry")
  expect_gte(length(bibentries), 5)

  print_bib("ewald_2024", "konig_2021") |>
    expect_identical(
      "Ewald F, Bothmann L, Wright M, Bischl B, Casalicchio G, König G (2024).\n\\dQuote{A Guide to Feature Importance Methods for Scientific Inference.}\nIn Longo L, Lapuschkin S, Seifert C (eds.), \\emph{Explainable Artificial Intelligence}, 440--464.\nISBN 978-3-031-63797-1, \\doi{10.1007/978-3-031-63797-1_22}.\n\nKönig G, Molnar C, Bischl B, Grosse-Wentrup M (2021).\n\\dQuote{Relative Feature Importance.}\nIn \\emph{2020 25th International Conference on Pattern Recognition (ICPR)}, 9318--9325.\n\\doi{10.1109/ICPR48806.2021.9413090}."
    )
})

#' Package dependency helper
#'
#' Error if the requried package is not installed and show an informative
#' message to the user about how to install it, defaulting to installation
#' from CRAN via `install.packages()` or some other source, e.g. a GitHub URL.
#'
#' @param pkg (`character(1)`) Package name to check.
#' @param from (`character(1)`: `"cran"`) Package source. If not `"cran"`,
#'   should be the URL to the package remote, e.g. a GitHub URL.
#'
#' @return If package is available: `TRUE` (invisibly). Error otherwise.
#' @noRd
#' @keywords internal
#' @examples
#'
#' require_package("arf")
#' require_package("seqknockoff", from = "https://github.com/kormama1/seqknockoff")
#'
require_package = function(pkg, from = "cran") {
  has_pkg = length(find.package(pkg, quiet = TRUE)) > 0

  if (has_pkg) {
    return(invisible(TRUE))
  }

  msg = "Package {.pkg {pkg}} required but not found!"

  if (from == "cran") {
    instruct = "Install it from CRAN with {.code install.packages(\"{pkg}\")}"
  } else {
    instruct = "Install it from {.url {from}}"
  }

  cli::cli_abort(c(x = msg, i = instruct))
}

test_that("require_package works", {
  expect_true(require_package("stats"))

  expect_error(require_package("foobar"), regexp = "Package .* required.*Install it from CRAN with")

  # With custom remote
  expect_error(
    require_package("foobar", from = "https://github.com/foo/bar"),
    regexp = "Package .* required.*Install it from .*https:\\/\\/github\\.com\\/foo\\/bar"
  )
})

#' @title k-Nearest Neighbors Conditional Sampler
#'
#' @description Implements conditional sampling using k-nearest neighbors (kNN).
#' For each observation, finds the `k` most similar observations based on conditioning
#' features, then samples the target features from these neighbors.
#'
#' @details
#' This sampler approximates the conditional distribution \eqn{P(X_B | X_A = x_A)} by:
#' 1. Finding the k nearest neighbors of \eqn{x_A} in the training data
#' 2. Sampling uniformly from the target feature values \eqn{X_B} of these k neighbors
#'
#' This is a simple, non-parametric approach that:
#' - Requires no distributional assumptions
#' - Handles mixed feature types (numeric, integer, factor, ordered, logical)
#' - Is computationally efficient (no model fitting required)
#' - Adapts locally to the data structure
#'
#' The method is related to hot-deck imputation and kNN imputation techniques used in
#' missing data problems. As \eqn{k \to \infty} and \eqn{k/n \to 0}, the kNN conditional
#' distribution converges to the true conditional distribution under mild regularity
#' conditions (Lipschitz continuity).
#'
#' **Distance Metrics:**
#'
#' The sampler supports two distance metrics:
#' - **Euclidean**: For numeric/integer features only. Standardizes features before computing distances.
#' - **Gower**: For mixed feature types. Handles numeric, factor, ordered, and logical features.
#'   Numeric features are range-normalized, categorical features use exact matching (0/1).
#'
#' The `distance` parameter controls which metric to use:
#' - `"auto"` (default): Automatically selects Euclidean for all-numeric features, Gower otherwise
#' - `"euclidean"`: Forces Euclidean distance (errors if non-numeric features present)
#' - `"gower"`: Forces Gower distance (works with any feature types)
#'
#' **Advantages:**
#' - Very fast (no model training)
#' - Works with any feature types
#' - Automatic distance metric selection
#' - Naturally respects local data structure
#'
#' **Limitations:**
#' - Sensitive to choice of `k`
#' - The full task data is required for prediction
#' - Can produce duplicates if `k` is small
#' - May not extrapolate well to new regions
#'
#' @examples
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#' sampler = ConditionalKNNSampler$new(task, k = 5)
#'
#' # Sample features conditioned on others
#' test_data = task$data(rows = 1:5)
#' sampled = sampler$sample_newdata(
#'   feature = c("important2", "important3"),
#'   newdata = test_data,
#'   conditioning_set = "important1"
#' )
#'
#' @references `r print_bib("little_2019", "troyanskaya_2001")`
#'
#' @export
ConditionalKNNSampler = R6Class(
  "ConditionalKNNSampler",
  inherit = ConditionalSampler,
  public = list(
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    feature_types = c("numeric", "integer", "factor", "ordered", "logical"),

    #' @description
    #' Creates a new ConditionalKNNSampler.
    #' @param task ([mlr3::Task]) Task to sample from.
    #' @param conditioning_set (`character` | `NULL`) Default conditioning set to use in `$sample()`.
    #' @param k (`integer(1)`: `5L`) Number of nearest neighbors to sample from.
    initialize = function(task, conditioning_set = NULL, k = 5L) {
      super$initialize(task, conditioning_set = conditioning_set)

      # Extend param_set with k parameter
      self$param_set = c(
        self$param_set,
        paradox::ps(k = paradox::p_int(lower = 1L, default = 5L))
      )
      self$param_set$set_values(k = k)

      self$label = "k-Nearest Neighbors Conditional Sampler"
    },

    #' @description
    #' Sample features from their kNN-based conditional distribution.
    #'
    #' @param feature (`character()`) Feature name(s) to sample.
    #' @param row_ids (`integer()` | `NULL`) Row IDs from task to use as conditioning values.
    #' @param conditioning_set (`character()` | `NULL`) Features to condition on.
    #'   If `NULL`, samples from marginal distribution (random sampling from training data).
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param k (`integer(1)` | `NULL`) Number of neighbors. If `NULL`, uses stored parameter.
    #' @return Modified copy with sampled feature(s).
    sample = function(
      feature,
      row_ids = NULL,
      conditioning_set = NULL,
      samples_per_row = 1L,
      k = NULL
    ) {
      super$sample(feature, row_ids, conditioning_set, samples_per_row = samples_per_row, k = k)
    },

    #' @description
    #' Sample from external data conditionally.
    #'
    #' @param feature (`character()`) Feature(s) to sample.
    #' @param newdata ([`data.table`][data.table::data.table]) External data to use.
    #' @param conditioning_set (`character()` | `NULL`) Features to condition on.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @param k (`integer(1)` | `NULL`) Number of neighbors. If `NULL`, uses stored parameter.
    #' @return Modified copy with sampled feature(s).
    sample_newdata = function(
      feature,
      newdata,
      conditioning_set = NULL,
      samples_per_row = 1L,
      k = NULL
    ) {
      super$sample_newdata(
        feature,
        newdata,
        conditioning_set,
        samples_per_row = samples_per_row,
        k = k
      )
    }
  ),

  private = list(
    # Core kNN sampling logic implementing k-nearest neighbors conditional sampling
    .sample_conditional = function(
      data,
      feature,
      conditioning_set,
      samples_per_row = 1L,
      k = NULL,
      ...
    ) {
      k = resolve_param(k, self$param_set$values$k, 5L)
      training_data = self$task$data(cols = self$task$feature_names)

      # Marginal fallback (no conditioning): draw with replacement, draw-major
      if (length(conditioning_set) == 0) {
        n = nrow(data)
        out = data[rep.int(seq_len(.N), times = samples_per_row)]
        for (feat in feature) {
          out[, (feat) := sample(training_data[[feat]], n * samples_per_row, replace = TRUE)]
        }
        return(out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)])
      }

      cond_types = self$task$feature_types[id %in% conditioning_set, type]
      use_gower = !all(cond_types %in% c("numeric", "integer"))
      if (use_gower) {
        require_package("gower")
      }

      query_cond_dt = data[, .SD, .SDcols = conditioning_set]
      train_cond_dt = training_data[, .SD, .SDcols = conditioning_set]

      if (!use_gower) {
        query_cond = as.matrix(query_cond_dt)
        train_cond = as.matrix(train_cond_dt)
        numeric_cols = sapply(conditioning_set, function(col) is.numeric(training_data[[col]]))
        if (any(numeric_cols)) {
          means = colMeans(train_cond[, numeric_cols, drop = FALSE])
          sds = apply(train_cond[, numeric_cols, drop = FALSE], 2, stats::sd)
          sds[sds == 0] = 1
          query_cond[, numeric_cols] = scale(
            query_cond[, numeric_cols, drop = FALSE],
            center = means,
            scale = sds
          )
          train_cond[, numeric_cols] = scale(
            train_cond[, numeric_cols, drop = FALSE],
            center = means,
            scale = sds
          )
        }
      }

      n = nrow(data)
      # sampled_idx[d, i] = training row chosen for draw d, evidence row i
      sampled_idx = matrix(NA_integer_, nrow = samples_per_row, ncol = n)

      for (i in seq_len(n)) {
        if (use_gower) {
          dists = gower::gower_dist(query_cond_dt[i, ], train_cond_dt)
        } else {
          qp = query_cond[i, , drop = FALSE]
          dists = sqrt(rowSums((sweep(train_cond, 2, qp))^2))
        }
        k_actual = min(k, length(dists))
        neighbors = which(dists <= sort(dists, partial = k_actual)[k_actual])
        sampled_idx[, i] = sample(neighbors, samples_per_row, replace = TRUE)
      }

      # Draw-major flatten: rows of `t(sampled_idx)` are draws, so as.vector(t(...))
      # gives [draw 1 across all n rows, draw 2 across all n rows, ...].
      flat_idx = as.vector(t(sampled_idx))

      out = data[rep.int(seq_len(.N), times = samples_per_row)]
      out[, (feature) := training_data[flat_idx, .SD, .SDcols = feature]]
      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)

#' @title (experimental) Conditional Inference Tree Conditional Sampler
#'
#' @description Implements conditional sampling using conditional inference trees (ctree).
#' Builds a tree predicting target features from conditioning features, then samples
#' from the terminal node corresponding to each test observation.
#'
#' @details
#' This sampler approximates the conditional distribution \eqn{P(X_B | X_A = x_A)} by:
#' 1. Building a conditional inference tree with \eqn{X_B} as response and \eqn{X_A} as predictors
#' 2. For each test observation, finding its terminal (leaf) node in the tree
#' 3. Sampling uniformly from training observations in that same terminal node
#'
#' Conditional inference trees (ctree) use permutation tests to determine splits,
#' which helps avoid overfitting and handles mixed feature types naturally. The tree
#' partitions the feature space based on the conditioning variables, creating local
#' neighborhoods that respect the conditional distribution structure.
#'
#' **Key advantages over other samplers:**
#' - Handles mixed feature types (continuous and categorical)
#' - Non-parametric (no distributional assumptions)
#' - Automatic feature selection (splits only on informative features)
#' - Can capture non-linear conditional relationships
#' - Statistically principled splitting criteria
#'
#' **Hyperparameters** control tree complexity:
#' - `mincriterion`: Significance level for splits (higher = fewer splits)
#' - `minsplit`: Minimum observations required for a split
#' - `minbucket`: Minimum observations in terminal nodes
#'
#' This implementation is inspired by shapr's ctree approach but simplified for our
#' use case (we build trees on-demand rather than pre-computing all subsets).
#'
#' **Advantages:**
#' - Works with any feature types
#' - Robust to outliers
#' - Interpretable tree structure
#' - Handles high-dimensional conditioning
#'
#' **Limitations:**
#' - Requires model fitting (slower than kNN)
#' - Can produce duplicates if terminal nodes are small
#' - Tree building time increases with data size
#'
#' @examples
#' \donttest{
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#'
#' # Create sampler with default parameters
#' sampler = ConditionalCtreeSampler$new(task)
#'
#' # Sample features conditioned on others
#' test_data = task$data(rows = 1:5)
#' sampled = sampler$sample_newdata(
#'   feature = c("important2", "important3"),
#'   newdata = test_data,
#'   conditioning_set = "important1"
#' )
#' }
#'
#' @references `r print_bib("hothorn_2006", "aas_2021")`
#'
#' @export
ConditionalCtreeSampler = R6Class(
  "ConditionalCtreeSampler",
  inherit = ConditionalSampler,
  public = list(
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    #'   Will be checked against the provided [mlr3::Task] to ensure compatibility.
    feature_types = c("numeric", "integer", "factor", "ordered"),
    #' @field tree_cache (`environment`) Cache for fitted ctree models.
    tree_cache = NULL,
    #' @description
    #' Creates a new ConditionalCtreeSampler.
    #' @param task ([mlr3::Task]) Task to sample from.
    #' @param conditioning_set (`character` | `NULL`) Default conditioning set to use in `$sample()`.
    #' @param mincriterion (`numeric(1)`: `0.95`) Significance level threshold for splitting (1 - p-value).
    #'   Higher values result in fewer splits (simpler trees).
    #' @param minsplit (`integer(1)`: `20L`) Minimum number of observations required for a split.
    #' @param minbucket (`integer(1)`: `7L`) Minimum number of observations in terminal nodes.
    #' @param use_cache (`logical(1)`: `TRUE`) Whether to cache fitted trees.
    initialize = function(
      task,
      conditioning_set = NULL,
      mincriterion = 0.95,
      minsplit = 20L,
      minbucket = 7L,
      use_cache = TRUE
    ) {
      require_package("partykit")
      super$initialize(task, conditioning_set = conditioning_set)

      # Initialize tree cache
      self$tree_cache = new.env(parent = emptyenv())

      # Extend param_set with ctree-specific parameters
      self$param_set = c(
        self$param_set,
        paradox::ps(
          mincriterion = paradox::p_dbl(lower = 0, upper = 1, default = 0.95),
          minsplit = paradox::p_int(lower = 1L, default = 20L),
          minbucket = paradox::p_int(lower = 1L, default = 7L),
          use_cache = paradox::p_lgl(default = TRUE)
        )
      )

      self$param_set$set_values(
        mincriterion = mincriterion,
        minsplit = minsplit,
        minbucket = minbucket,
        use_cache = use_cache
      )

      self$label = "Conditional Inference Tree Conditional Sampler"
    }
  ),

  private = list(
    # Core ctree sampling logic implementing conditional inference tree sampling
    .sample_conditional = function(data, feature, conditioning_set, samples_per_row = 1L, ...) {
      training_data = self$task$data(cols = self$task$feature_names)

      # Marginal fallback (no conditioning): draw samples_per_row x nrow(data) values
      # per feature with replacement from the training column, organized draw-major.
      if (length(conditioning_set) == 0) {
        n = nrow(data)
        out = data[rep.int(seq_len(.N), times = samples_per_row)]
        for (feat in feature) {
          out[, (feat) := sample(training_data[[feat]], n * samples_per_row, replace = TRUE)]
        }
        return(out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)])
      }

      tree = private$.get_or_build_tree(feature, conditioning_set)

      n = nrow(data)
      data[, ..eval_id := seq_len(.N)]
      node_ids = predict(tree, newdata = data, type = "node")
      data[, ..node_id := node_ids]

      train_nodes = predict(tree, newdata = training_data, type = "node")
      train_data_with_nodes = data.table::copy(training_data)
      train_data_with_nodes[, ..node_id := train_nodes]

      # For each draw, sample one training observation per evidence row from its terminal node.
      # Stack the draws in draw-major order so positional alignment with row_ids is preserved.
      out = data[rep.int(seq_len(.N), times = samples_per_row)]

      for (i in seq_len(n)) {
        node_obs = train_data_with_nodes[..node_id == node_ids[i]]
        if (nrow(node_obs) == 0) {
          cli::cli_warn("No training observations in terminal node, using marginal sample")
          # Marginal fallback per draw: draw samples_per_row values for this row from training_data
          for (d in seq_len(samples_per_row)) {
            out_row_idx = (d - 1L) * n + i
            for (feat in feature) {
              sampled_value = sample(training_data[[feat]], 1L)
              data.table::set(out, i = out_row_idx, j = feat, value = sampled_value)
            }
          }
        } else {
          picks = node_obs[sample.int(.N, samples_per_row, replace = TRUE)]
          for (d in seq_len(samples_per_row)) {
            out_row_idx = (d - 1L) * n + i
            for (feat in feature) {
              data.table::set(out, i = out_row_idx, j = feat, value = picks[[feat]][d])
            }
          }
        }
      }

      data[, ..eval_id := NULL]
      data[, ..node_id := NULL]
      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    },

    # Get cached tree or build a new one
    .get_or_build_tree = function(feature, conditioning_set) {
      # Create cache key from feature names and conditioning set
      cache_key = paste(
        paste(sort(match(feature, self$task$feature_names)), collapse = "+"),
        "|",
        paste(sort(match(conditioning_set, self$task$feature_names)), collapse = "+")
      )

      use_cache = self$param_set$values$use_cache

      # Check cache
      if (use_cache && exists(cache_key, envir = self$tree_cache)) {
        return(get(cache_key, envir = self$tree_cache))
      }

      # Build new tree
      tree = private$.build_tree(feature, conditioning_set)

      # Cache if enabled
      if (use_cache) {
        assign(cache_key, tree, envir = self$tree_cache)
      }

      tree
    },

    # Build a ctree model
    .build_tree = function(feature, conditioning_set) {
      # Get training data from task
      training_data = self$task$data(cols = self$task$feature_names)

      # Create formula: features ~ conditioning_set
      # For multiple features, use cbind on LHS
      if (length(feature) == 1) {
        formula_str = sprintf(
          "`%s` ~ %s",
          feature,
          paste(sprintf("`%s`", conditioning_set), collapse = " + ")
        )
      } else {
        formula_str = sprintf(
          "cbind(%s) ~ %s",
          paste(sprintf("`%s`", feature), collapse = ", "),
          paste(sprintf("`%s`", conditioning_set), collapse = " + ")
        )
      }

      formula = as.formula(formula_str)

      # Extract control parameters
      mincriterion = self$param_set$values$mincriterion
      minsplit = self$param_set$values$minsplit
      minbucket = self$param_set$values$minbucket

      # Build tree with partykit::ctree
      tree = partykit::ctree(
        formula = formula,
        data = training_data,
        control = partykit::ctree_control(
          mincriterion = mincriterion,
          minsplit = minsplit,
          minbucket = minbucket
        )
      )

      tree
    }
  )
)

#' Batch Predict for SAGE
#'
#' Performs batched prediction on combined data to manage memory usage.
#' Supports both classification (probability predictions) and regression.
#'
#' @param learner ([`Learner`][mlr3::Learner]) Trained mlr3 learner.
#' @param combined_data (`data.table`) Data with feature columns to predict on.
#' @param task ([`Task`][mlr3::Task]) mlr3 task object.
#' @param batch_size (`integer(1)` or `NULL`) Batch size for predictions. If `NULL` or if
#'   `total_rows <= batch_size`, processes all data at once.
#' @param task_type (`character(1)`) Task type, either `"classif"` or `"regr"`.
#'
#' @return For classification: `matrix` of class probabilities (n_rows x n_classes).
#'   For regression: `numeric` vector of predictions (length n_rows).
#'
#' @keywords internal
sage_batch_predict = function(learner, combined_data, task, batch_size, task_type) {
  total_rows = nrow(combined_data)

  if (!is.null(batch_size) && total_rows > batch_size) {
    # Batched prediction
    n_batches = ceiling(total_rows / batch_size)
    all_predictions = vector("list", n_batches)

    for (batch_idx in seq_len(n_batches)) {
      start_row = (batch_idx - 1) * batch_size + 1
      end_row = min(batch_idx * batch_size, total_rows)
      batch_data = combined_data[start_row:end_row]

      if (xplain_opt("debug")) {
        cli::cli_inform(
          "Predicting on {.val {nrow(batch_data)}} instances in batch {.val {batch_idx}/{n_batches}}"
        )
      }

      pred_result = if (is.function(learner$predict_newdata_fast)) {
        learner$predict_newdata_fast(newdata = batch_data, task = task)
      } else {
        learner$predict_newdata(newdata = batch_data, task = task)
      }

      all_predictions[[batch_idx]] = if (task_type == "classif") {
        pred_result$prob
      } else {
        pred_result$response
      }
    }

    # Combine predictions from all batches
    if (task_type == "classif") {
      do.call(rbind, all_predictions)
    } else {
      do.call(c, all_predictions)
    }
  } else {
    # Single prediction without batching
    if (xplain_opt("debug")) {
      cli::cli_inform("Predicting on {.val {nrow(combined_data)}} instances at once")
    }

    pred_result = if (is.function(learner$predict_newdata_fast)) {
      learner$predict_newdata_fast(newdata = combined_data, task = task)
    } else {
      learner$predict_newdata(newdata = combined_data, task = task)
    }

    if (task_type == "classif") {
      pred_result$prob
    } else {
      pred_result$response
    }
  }
}

#' Aggregate Predictions by Coalition and Test Instance
#'
#' Averages predictions across multiple samples (reference data or conditional samples)
#' for each unique combination of coalition and test instance.
#'
#' @param combined_data (`data.table`) Data with columns `.coalition_id`, `.test_instance_id`,
#'   and feature columns.
#' @param predictions (`matrix` or `numeric`) For classification: matrix of class probabilities.
#'   For regression: numeric vector of predictions.
#' @param task_type (`character(1)`) Task type, either `"classif"` or `"regr"`.
#' @param class_names (`character()` or `NULL`: `NULL`) Character vector of class names. Required
#'   for classification, ignored for regression.
#'
#' @return `data.table` with columns:
#'   - `.coalition_id`: Coalition identifier (integer)
#'   - `.test_instance_id`: Test instance identifier (integer)
#'   - For classification: One column per class with averaged probabilities (numeric)
#'   - For regression: `avg_pred` column with averaged predictions (numeric)
#'
#' @keywords internal
sage_aggregate_predictions = function(combined_data, predictions, task_type, class_names = NULL) {
  if (task_type == "classif") {
    # Add prediction columns to combined_data
    n_classes = ncol(predictions)
    for (j in seq_len(n_classes)) {
      combined_data[, paste0(".pred_class_", j) := predictions[, j]]
    }

    # Aggregate: calculate mean probability for each class, grouped by coalition and test instance
    agg_cols = paste0(".pred_class_", seq_len(n_classes))
    avg_preds = combined_data[,
      lapply(.SD, function(x) mean(x, na.rm = TRUE)),
      .SDcols = agg_cols,
      by = c(".coalition_id", ".test_instance_id")
    ]

    # Rename aggregated columns to original class names
    setnames(avg_preds, agg_cols, class_names)
    avg_preds
  } else if (task_type == "regr") {
    # Regression: add predictions and aggregate
    .prediction = NULL # the data.table NSE NOTE tax
    combined_data[, .prediction := predictions]

    combined_data[,
      list(avg_pred = mean(.prediction, na.rm = TRUE)),
      by = c(".coalition_id", ".test_instance_id")
    ]
  }
}

#' Build row-major growing-prefix coalitions for a permutation list
#'
#' For each permutation, emit its growing prefixes
#' (`perm[1]`, `perm[1:2]`, ..., `perm`). Row-major over
#' `(permutation, step)`. Pure; no evaluation, no RNG.
#'
#' @param perm_sublist (`list`) Feature-name permutations.
#' @return `list` of character vectors (coalitions).
#' @keywords internal
#' @noRd
sage_growing_coalitions = function(perm_sublist) {
  # Pre-allocate to the exact coalition count (sum of permutation lengths)
  coalitions = vector("list", sum(lengths(perm_sublist)))
  k = 1L
  for (i in seq_along(perm_sublist)) {
    perm = perm_sublist[[i]]
    for (j in seq_along(perm)) {
      coalitions[[k]] = perm[seq_len(j)]
      k = k + 1L
    }
  }
  coalitions
}

#' Accumulate SAGE marginal contributions from a loss vector
#'
#' Given growing-prefix losses laid out row-major over
#' `(permutation, step)` (optionally preceded by `offset` leading
#' slots, e.g. an empty-coalition entry), accumulate per-feature SAGE
#' value sums and squared sums. Every permutation is a full feature
#' permutation, so the loss index is closed-form: no search/map.
#' Pure; order-independent across permutations.
#'
#' @param perm_sublist (`list`) Feature-name permutations.
#' @param losses (`numeric`) Losses for `offset` leading slots then the
#'   row-major growing-prefix coalitions of `perm_sublist`.
#' @param baseline (`numeric(1)`) Empty-coalition loss anchor.
#' @param feature_names (`character`) Names for the output vectors.
#' @param offset (`integer(1)`: `0L`) Leading loss slots to skip.
#' @return `list(sv, sv_sq)` named numeric vectors over `feature_names`.
#' @keywords internal
#' @noRd
sage_marginal_contributions = function(perm_sublist, losses, baseline, feature_names, offset = 0L) {
  sv = numeric(length(feature_names))
  sv_sq = numeric(length(feature_names))
  names(sv) = feature_names
  names(sv_sq) = feature_names

  for (i in seq_along(perm_sublist)) {
    perm = perm_sublist[[i]]
    p = length(perm)
    prev_loss = baseline
    for (j in seq_len(p)) {
      feature = perm[j]
      current_loss = losses[offset + (i - 1L) * p + j]
      contribution = prev_loss - current_loss
      sv[feature] = sv[feature] + contribution
      sv_sq[feature] = sv_sq[feature] + contribution^2
      prev_loss = current_loss
    }
  }

  list(sv = sv, sv_sq = sv_sq)
}

test_that("sage_batch_predict works for regression without batching", {
  skip_if_not_installed("rpart")

  set.seed(123)
  task = tgen("friedman1")$generate(n = 50)
  learner = lrn("regr.rpart")
  learner$train(task)

  test_data = task$data()
  predictions = sage_batch_predict(learner, test_data, task, batch_size = NULL, task_type = "regr")

  expect_type(predictions, "double")
  expect_length(predictions, 50)
  expect_true(all(is.finite(predictions)))
})

test_that("sage_batch_predict works for regression with batching", {
  skip_if_not_installed("rpart")

  set.seed(123)
  task = tgen("friedman1")$generate(n = 50)
  learner = lrn("regr.rpart")
  learner$train(task)

  test_data = task$data()

  # Predict without batching
  pred_no_batch = sage_batch_predict(
    learner,
    test_data,
    task,
    batch_size = NULL,
    task_type = "regr"
  )

  # Predict with batching
  pred_with_batch = sage_batch_predict(
    learner,
    test_data,
    task,
    batch_size = 10,
    task_type = "regr"
  )

  # Should be identical
  expect_equal(pred_no_batch, pred_with_batch)
})

test_that("sage_batch_predict works for classification without batching", {
  skip_if_not_installed("rpart")

  set.seed(123)
  task = tgen("2dnormals")$generate(n = 50)
  learner = lrn("classif.rpart", predict_type = "prob")
  learner$train(task)

  test_data = task$data()
  predictions = sage_batch_predict(
    learner,
    test_data,
    task,
    batch_size = NULL,
    task_type = "classif"
  )

  expect_true(is.matrix(predictions))
  expect_equal(nrow(predictions), 50)
  expect_equal(ncol(predictions), 2) # Binary classification
  expect_true(all(is.finite(predictions)))
  # Probabilities should sum to 1
  expect_true(all(abs(rowSums(predictions) - 1) < 1e-10))
})

test_that("sage_batch_predict works for classification with batching", {
  skip_if_not_installed("rpart")

  set.seed(123)
  task = tgen("2dnormals")$generate(n = 50)
  learner = lrn("classif.rpart", predict_type = "prob")
  learner$train(task)

  test_data = task$data()

  # Predict without batching
  pred_no_batch = sage_batch_predict(
    learner,
    test_data,
    task,
    batch_size = NULL,
    task_type = "classif"
  )

  # Predict with batching
  pred_with_batch = sage_batch_predict(
    learner,
    test_data,
    task,
    batch_size = 10,
    task_type = "classif"
  )

  # Should be identical
  expect_equal(pred_no_batch, pred_with_batch)
})

test_that("sage_aggregate_predictions works for regression", {
  # Create sample data with multiple samples per coalition/instance
  combined_data = data.table::data.table(
    .coalition_id = c(1, 1, 1, 2, 2, 2),
    .test_instance_id = c(1, 1, 1, 1, 1, 1),
    feature1 = c(1, 2, 3, 4, 5, 6)
  )

  predictions = c(1.0, 2.0, 3.0, 4.0, 5.0, 6.0)

  result = sage_aggregate_predictions(
    combined_data,
    predictions,
    task_type = "regr",
    class_names = NULL
  )

  expect_equal(nrow(result), 2) # 2 coalitions
  expect_true(all(c(".coalition_id", ".test_instance_id", "avg_pred") %in% names(result)))
  expect_equal(result[.coalition_id == 1]$avg_pred, 2.0) # mean(1,2,3)
  expect_equal(result[.coalition_id == 2]$avg_pred, 5.0) # mean(4,5,6)
})

test_that("sage_aggregate_predictions works for classification", {
  # Create sample data with multiple samples per coalition/instance
  combined_data = data.table::data.table(
    .coalition_id = c(1, 1, 1, 2, 2, 2),
    .test_instance_id = c(1, 1, 1, 1, 1, 1),
    feature1 = c(1, 2, 3, 4, 5, 6)
  )

  # Binary classification probabilities
  predictions = matrix(
    c(
      0.9,
      0.1,
      0.8,
      0.2,
      0.7,
      0.3,
      0.6,
      0.4,
      0.5,
      0.5,
      0.4,
      0.6
    ),
    ncol = 2,
    byrow = TRUE
  )

  class_names = c("A", "B")

  result = sage_aggregate_predictions(
    combined_data,
    predictions,
    task_type = "classif",
    class_names = class_names
  )

  expect_equal(nrow(result), 2) # 2 coalitions
  expect_true(all(c(".coalition_id", ".test_instance_id", "A", "B") %in% names(result)))

  # Check averaged probabilities for coalition 1
  expect_equal(result[.coalition_id == 1]$A, mean(c(0.9, 0.8, 0.7)))
  expect_equal(result[.coalition_id == 1]$B, mean(c(0.1, 0.2, 0.3)))

  # Check averaged probabilities for coalition 2
  expect_equal(result[.coalition_id == 2]$A, mean(c(0.6, 0.5, 0.4)))
  expect_equal(result[.coalition_id == 2]$B, mean(c(0.4, 0.5, 0.6)))
})

test_that("sage_aggregate_predictions handles multiple test instances", {
  # Create data with 2 coalitions and 2 test instances
  combined_data = data.table::data.table(
    .coalition_id = c(1, 1, 1, 1, 2, 2, 2, 2),
    .test_instance_id = c(1, 1, 2, 2, 1, 1, 2, 2),
    feature1 = 1:8
  )

  predictions = 1:8

  result = sage_aggregate_predictions(
    combined_data,
    predictions,
    task_type = "regr",
    class_names = NULL
  )

  expect_equal(nrow(result), 4) # 2 coalitions × 2 test instances
  expect_equal(result[.coalition_id == 1 & .test_instance_id == 1]$avg_pred, mean(c(1, 2)))
  expect_equal(result[.coalition_id == 1 & .test_instance_id == 2]$avg_pred, mean(c(3, 4)))
  expect_equal(result[.coalition_id == 2 & .test_instance_id == 1]$avg_pred, mean(c(5, 6)))
  expect_equal(result[.coalition_id == 2 & .test_instance_id == 2]$avg_pred, mean(c(7, 8)))
})

test_that("sage_growing_coalitions builds row-major growing prefixes", {
  perms = list(c("a", "b"), c("b", "a"))
  co = sage_growing_coalitions(perms)
  expect_length(co, 4L)
  expect_identical(co[[1]], "a")
  expect_identical(co[[2]], c("a", "b"))
  expect_identical(co[[3]], "b")
  expect_identical(co[[4]], c("b", "a"))
})

test_that("sage_marginal_contributions accumulates closed-form, offset-aware", {
  feats = c("a", "b")
  perms = list(c("a", "b"), c("b", "a"))
  baseline = 10
  # losses laid out row-major over (perm, step), no offset:
  # perm1: {a}=8, {a,b}=5 ; perm2: {b}=9, {b,a}=4
  losses = c(8, 5, 9, 4)
  r = sage_marginal_contributions(perms, losses, baseline, feats, offset = 0L)
  # perm1: a: 10-8=2 ; b: 8-5=3
  # perm2: b: 10-9=1 ; a: 9-4=5
  expect_equal(r$sv[["a"]], 2 + 5)
  expect_equal(r$sv[["b"]], 3 + 1)
  expect_equal(r$sv_sq[["a"]], 2^2 + 5^2)
  expect_equal(r$sv_sq[["b"]], 3^2 + 1^2)
  expect_setequal(names(r$sv), feats)

  # offset = 1L: an extra leading slot (e.g. empty coalition) shifts indices by 1
  losses_off = c(99, 8, 5, 9, 4)
  r2 = sage_marginal_contributions(perms, losses_off, baseline, feats, offset = 1L)
  expect_equal(r2$sv, r$sv)
  expect_equal(r2$sv_sq, r$sv_sq)
})

#' @title Marginal SAGE
#'
#' @description [SAGE] with marginal sampling (features are marginalized independently).
#' This is the standard SAGE implementation.
#'
#' @seealso [ConditionalSAGE]
#'
#' @examplesIf requireNamespace("ranger", quietly = TRUE) && requireNamespace("mlr3learners", quietly = TRUE)
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#' sage = MarginalSAGE$new(
#'   task = task,
#'   learner = lrn("regr.ranger", num.trees = 20),
#'   measure = msr("regr.mse"),
#'   n_permutations = 2L,
#'   n_samples = 10
#' )
#' sage$compute()
#' @export
MarginalSAGE = R6Class(
  "MarginalSAGE",
  inherit = SAGE,
  public = list(
    #' @description
    #' Creates a new instance of the MarginalSAGE class.
    #' @param task,learner,measure,resampling,features,n_permutations,batch_size,n_samples,early_stopping,se_threshold,min_permutations,check_interval Passed to [SAGE].
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      n_permutations = 10L,
      batch_size = 5000L,
      n_samples = 100L,
      early_stopping = FALSE,
      se_threshold = 0.01,
      min_permutations = 10L,
      check_interval = 1L
    ) {
      # No need to initialize sampler as marginal sampling is done differently here
      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        n_permutations = n_permutations,
        batch_size = batch_size,
        n_samples = n_samples,
        early_stopping = early_stopping,
        se_threshold = se_threshold,
        min_permutations = min_permutations,
        check_interval = check_interval
      )
      # Use training data as reference for later marginalization
      private$reference_data = self$task$data(cols = self$task$feature_names)

      # Subsample reference data if it's too large for efficiency
      if (nrow(private$reference_data) > n_samples) {
        sample_idx = sample(nrow(private$reference_data), size = n_samples)
        private$reference_data = private$reference_data[sample_idx, ]
      }

      self$label = "Marginal SAGE"
    }
  ),

  private = list(
    reference_data = NULL,

    # Marginal-specific data expansion: Cartesian product with reference data
    .expand_coalitions_data = function(test_dt, all_coalitions) {
      n_test = nrow(test_dt)
      n_reference = nrow(private$reference_data)
      all_expanded_data = vector("list", length(all_coalitions))

      for (i in seq_along(all_coalitions)) {
        coalition = all_coalitions[[i]]

        # Cartesian product: each test instance * all reference instances
        test_expanded = test_dt[rep(seq_len(n_test), each = n_reference)]
        reference_expanded = private$reference_data[rep(seq_len(n_reference), times = n_test)]

        # Add tracking columns BEFORE marginalization
        test_expanded[, .coalition_id := i]
        test_expanded[, .test_instance_id := rep(seq_len(n_test), each = n_reference)]

        # Replace marginalized features with reference values
        # Maintains correlation structure in out-of-coalition features (block-wise sampling)
        marginalize_features = setdiff(self$features, coalition)
        if (length(marginalize_features) > 0) {
          test_expanded[,
            (marginalize_features) := reference_expanded[, .SD, .SDcols = marginalize_features]
          ]
        }

        all_expanded_data[[i]] = test_expanded
      }

      rbindlist(all_expanded_data)
    }
  )
)

#' Parallel map function for xplainfi
#'
#' Unified parallelization interface supporting mirai, future, and sequential execution.
#' Follows the mlr3::future_map pattern for consistency.
#'
#' @param n Number of iterations (length of first varying argument)
#' @param .f Function to apply
#' @param ... Varying arguments to iterate over (passed to mapply)
#' @param .args Named list of fixed arguments passed to all iterations
#'
#' @return List of results from applying .f
#'
#' @details
#' Execution mode is determined automatically:
#' 1. If `getOption("xplainfi.sequential")` is TRUE, runs sequentially with mapply
#' 2. If mirai daemons are set, uses mirai_map for parallel execution
#' 3. If future plan is non-sequential, uses future.apply::future_mapply
#' 4. Otherwise, uses base mapply for sequential execution
#'
#' For mirai execution, required packages are loaded automatically based on context.
#'
#' @keywords internal
#' @noRd
xplainfi_map = function(n, .f, ..., .args = list()) {
  # Debug mode: sequential execution
  if (xplain_opt("sequential")) {
    return(mapply(.f, ..., MoreArgs = .args, SIMPLIFY = FALSE, USE.NAMES = TRUE))
  }

  # Mirai parallelization
  # Use same option as mlr3 for consistency (WVIM uses mlr3fselect which uses mlr3's parallelization)
  if (
    requireNamespace("mirai", quietly = TRUE) &&
      mirai::daemons_set(.compute = getOption("mlr3.mirai_parallelization", "mlr3_parallelization"))
  ) {
    # Capture names from first varying argument
    varying_args = list(...)
    result_names = if (length(varying_args) > 0) names(varying_args[[1]]) else NULL

    # Create data.table from varying arguments
    # Ensure first column name matches first parameter of .f for named argument passing
    varying_dt = data.table::data.table(...)
    first_param = names(formals(.f))[1]
    data.table::setnames(varying_dt, 1, first_param)

    result = mirai::mirai_map(
      varying_dt,
      .f,
      .args = c(.args, list(is_sequential = FALSE)),
      .compute = getOption("mlr3.mirai_parallelization", "mlr3_parallelization")
    )

    collected = mirai::collect_mirai(result)

    # Check for errors in parallel execution
    has_errors = vapply(collected, inherits, logical(1), "miraiError")
    if (any(has_errors)) {
      error_msgs = collected[has_errors]
      error_indices = which(has_errors)
      cli::cli_abort(c(
        "Errors occurred in mirai worker processes:",
        "x" = "{length(error_indices)} task{?s} failed: {.val {error_indices}}",
        "i" = "First error: {as.character(error_msgs[[1]])}"
      ))
    }

    # Restore names if they existed
    if (!is.null(result_names)) {
      names(collected) = result_names
    }

    return(collected)
  }

  # Future parallelization
  if (requireNamespace("future", quietly = TRUE)) {
    is_sequential = inherits(future::plan(), "sequential")

    if (!is_sequential) {
      # Get chunk size settings
      chunk_size = getOption("xplainfi.exec_chunk_size", 1)
      chunk_bins = getOption("xplainfi.exec_chunk_bins")
      if (!is.null(chunk_bins)) {
        chunk_size = ceiling(n / chunk_bins)
      }

      # Random scheduling for load balancing
      scheduling = if (isTRUE(getOption("xplainfi.exec_random", TRUE))) {
        structure(TRUE, ordering = "random")
      } else {
        TRUE
      }

      .args_future = c(.args, list(is_sequential = FALSE))

      return(future.apply::future_mapply(
        .f,
        ...,
        MoreArgs = .args_future,
        SIMPLIFY = FALSE,
        USE.NAMES = FALSE,
        future.globals = FALSE,
        future.packages = character(0), # Packages loaded explicitly in worker
        future.seed = TRUE,
        future.scheduling = scheduling,
        future.chunk.size = chunk_size,
        future.stdout = TRUE
      ))
    }
  }

  # Sequential fallback
  mapply(.f, ..., MoreArgs = .args, SIMPLIFY = FALSE, USE.NAMES = TRUE)
}

#' Feature Importance Method Class
#'
#' @export
FeatureImportanceMethod = R6Class(
  "FeatureImportanceMethod",
  public = list(
    #' @field label (`character(1)`) Method label.
    label = NA_character_,
    #' @field task ([mlr3::Task])
    task = NULL,
    #' @field learner ([mlr3::Learner])
    learner = NULL,
    #' @field measure ([mlr3::Measure])
    measure = NULL,
    #' @field resampling ([mlr3::Resampling]), instantiated upon construction.
    resampling = NULL,
    #' @field resample_result ([mlr3::ResampleResult]) of the original `learner` and `task`, used for baseline scores.
    resample_result = NULL,
    #' @field features (`character`: `NULL`) Features of interest. By default, importances will be computed for each feature
    #'   in `task`, but optionally this can be restricted to at least one feature. Ignored if `groups` is specified.
    features = NULL,
    #' @field groups (`list`: `NULL`) A (named) list of features (names or indices as in `task`).
    #'   If `groups` is specified, `features` is ignored.
    #'   Importances will be calculated for group of features at a time, e.g., in [PFI] not one but the group of features will be permuted at each step.
    #'   Analogously in [WVIM], each group of features will be left out (or in) for each model refit.
    #'   Not all methods support groups (e.g., [SAGE]).
    groups = NULL,
    #' @field param_set ([paradox::ps()])
    param_set = ps(),
    #' @field predictions ([data.table][data.table::data.table]) Feature-specific prediction objects provided for some methods ([PFI], [WVIM]). Contains columns for feature of interest, resampling iteration, refit or perturbation iteration, and [mlr3::Prediction] objects.
    predictions = NULL,

    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #' This is typically intended for use by derived classes.
    #' @param task,learner,measure,resampling,features,groups,param_set,label Used to set fields
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      param_set = paradox::ps(),
      label
    ) {
      self$task = mlr3::assert_task(task)
      self$learner = mlr3::assert_learner(learner, task = task, task_type = task$task_type)

      if (is.null(measure)) {
        self$measure = switch(
          task$task_type,
          "classif" = mlr3::msr("classif.ce"),
          "regr" = mlr3::msr("regr.mse")
        )
        if (xplain_opt("verbose")) {
          cli::cli_alert_info(
            "No {.cls Measure} provided, using {.code measure = msr(\"{self$measure$id}\")}"
          )
        }
      } else {
        self$measure = mlr3::assert_measure(measure, task = task, learner = learner)
      }
      self$param_set = paradox::assert_param_set(param_set)
      self$label = checkmate::assert_string(label, min.chars = 1)

      # Check features / groups
      # Default to using features, unless groups is specified
      if (is.null(groups)) {
        checkmate::assert_subset(features, self$task$feature_names, empty.ok = TRUE)
        self$features = features %||% self$task$feature_names
      } else {
        self$groups = check_groups(groups, all_features = self$task$feature_names)
        # check_groups ensures this produces a unique character vector
        self$features = unlist(groups, use.names = FALSE)
      }

      # resampling: default to holdout with default ratio if NULL
      if (is.null(resampling)) {
        resampling = mlr3::rsmp("holdout", ratio = 2 / 3)$instantiate(task)
        if (xplain_opt("verbose")) {
          cli::cli_inform(c(
            i = "No {.cls Resampling} provided, using {.code resampling = rsmp(\"holdout\", ratio = 2/3)} (test set size: {.val {length(resampling$test_set(1))}})"
          ))
        }
      } else {
        # Clone the resampling to avoid instantiating the resampling in the user's workspace
        resampling = mlr3::assert_resampling(resampling)$clone()

        # A pretrained learner requires a user-provided instantiated resampling
        # to define the test set explicitly. Auto-instantiation would pick an
        # arbitrary split unrelated to how the learner was trained.
        if (!is.null(learner$model) && !resampling$is_instantiated) {
          cli::cli_abort(c(
            "A pre-trained {.cls Learner} requires an instantiated {.cls Resampling}",
            i = "Instantiate the {.cls Resampling} before passing it, e.g. {.code rsmp(\"holdout\")$instantiate(task)}"
          ))
        }
      }
      if (!resampling$is_instantiated) {
        resampling$instantiate(task)
      }
      self$resampling = resampling

      # Check pretrained learner compatibility (multi-fold with pretrained learner)
      assert_pretrained(self$learner, self$task, self$resampling)
    },

    #' @description
    #' Compute feature importance scores
    #' @param store_backends (`logical(1): TRUE`) Whether to store backends.
    compute = function(store_backends = TRUE) {
      stop("Abstract method. Use a concrete implementation.")
    },

    #' @description
    #' Get aggregated importance scores.
    #' The stored [`measure`][mlr3::Measure] object's `aggregator` (default: `mean`) will be used to aggregated importance scores
    #' across resampling iterations and, depending on the method use, permutations ([PerturbationImportance] or refits [LOCO]).
    #' @param relation (character(1)) How to relate perturbed scores to originals ("difference" or "ratio").
    #'   If `NULL`, uses stored parameter value. This is only applicable for methods where importance is based on some
    #'   relation between baseline and post-modification loss, i.e. [PerturbationImportance] methods such as [PFI] or [WVIM] / [LOCO].
    #'   Not available for [SAGE] methods.
    #'   For `"ratio"`, features whose baseline score is `0` yield `NA` (with a warning) and therefore aggregate to `NA`.
    #' @param standardize (`logical(1)`: `FALSE`) If `TRUE`, importances are standardized by the highest score so all scores fall in `[-1, 1]`.
    #' @param ci_method (`character(1)`: `"none"`) Which confidence interval estimation method to use, defaulting to omitting
    #'   variance estimation (`"none"`).
    #'   If `"raw"`, uncorrected (too narrow) CIs are provided purely for informative purposes.
    #'   If `"nadeau_bengio"`, variance correction is performed according to Nadeau & Bengio (2003) as suggested by Molnar et al. (2023).
    #'   If `"quantile"`, empirical quantiles are used to construct confidence-like intervals.
    #'   These methods are model-agnostic and rely on suitable `resampling`s, e.g. subsampling with 15 repeats for `"nadeau_bengio"`.
    #'   See details.
    #' @param conf_level (`numeric(1)`: `0.95`) Confidence level to use for confidence interval construction when `ci_method != "none"`.
    #' @param alternative (`character(1)`: `"two.sided"`) Type of alternative hypothesis for statistical tests.
    #'   `"greater"` tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
    #'   `"two.sided"` tests H0: importance = 0 vs H1: importance != 0.
    #'   Only used when `ci_method != "none"`.
    #' @param ... Additional arguments passed to specialized methods, if any.
    #' @return ([data.table][data.table::data.table]) Aggregated importance scores with columns `"feature"`, `"importance"`,
    #' and depending on `ci_method` also `"se"`, `"statistic"`, `"p.value"`, `"conf_lower"`, `"conf_upper"`.
    #'
    #' @details
    #'
    #' ## Confidence Interval Methods
    #'
    #' The parametric methods (`"raw"`, `"nadeau_bengio"`) return standard error (`se`),
    #' test statistic (`statistic`), p-value (`p.value`), and confidence bounds
    #' (`conf_lower`, `conf_upper`). The `"quantile"` method returns only lower and upper bounds.
    #'
    #' **`"raw"`: Uncorrected (!) t-test**
    #' Uses a standard t-test assuming independence of resampling iterations.
    #' - SE = sd(resampling scores) / sqrt(n_iters)
    #' - Test statistic: t = importance / SE with df = n_iters - 1
    #' - P-value: From t-distribution (one-sided or two-sided depending on `alternative`)
    #' - CIs: importance +/- qt(1 - alpha, df) * SE
    #'
    #' **Warning**: These CIs are too narrow because resampling iterations share
    #' training data and are not independent.
    #' This method is included only for demonstration purposes.
    #'
    #' **`"nadeau_bengio"`: Corrected t-test**
    #' Applies the Nadeau & Bengio (2003) correction to account for correlation between
    #' resampling iterations due to overlapping training sets.
    #' - Correction factor: (1/n_iters + n_test/n_train)
    #' - SE = sqrt(correction_factor * var(resampling scores))
    #' - Test statistic and p-value: As in `"raw"`, but with corrected SE
    #'
    #' Recommended with bootstrap or subsampling (>= 10 iterations).
    #'

    #' **`"quantile"`: Non-parametric empirical method**
    #' Uses the resampling distribution directly without parametric assumptions.
    #' - CIs: Empirical quantiles of the resampling distribution
    #'
    #' This method does not provide `se`, `statistic`, or `p.value`.
    #'
    #' ## Method-Specific CI Methods
    #'
    #' Some importance methods provide additional CI methods tailored to their approach:
    #'
    #' - **[CFI]**: Adds `"cpi"` (Conditional Predictive Impact), which uses observation-wise
    #'   loss differences with holdout resampling. Supports t-test, Wilcoxon, Fisher permutation,
    #'   and binomial tests. See Watson & Wright (2021).
    #'
    #' ## Practical Recommendations
    #'
    #' Variance estimates for importance scores are biased due to the resampling procedure.
    #' Molnar et al. (2023) suggest using the Nadeau & Bengio correction with approximately
    #' 15 iterations of subsampling.
    #'
    #' Bootstrapping can cause information leakage with learners that bootstrap internally
    #' (e.g., Random Forests), as observations may appear in both train and test sets.
    #' Prefer subsampling in such cases:
    #'
    #' ```r
    #' PFI$new(
    #'   task = sim_dgp_interactions(n = 1000),
    #'   learner = lrn("regr.ranger", num.trees = 100),
    #'   measure = msr("regr.mse"),
    #'   resampling = rsmp("subsampling", repeats = 15),
    #'   n_repeats = 20
    #' )
    #' ```
    #'
    #' The `"nadeau_bengio"` correction was validated for PFI; its use with other methods
    #' like LOCO or SAGE is experimental.
    #'
    #' @param p_adjust (`character(1)`: `"none"`) Method for p-value adjustment for multiple comparisons.
    #'   Accepts any method supported by [stats::p.adjust.methods], e.g. `"holm"`, `"bonferroni"`, `"BH"`, `"none"`.
    #'   Applied to p-values from `"raw"` and `"nadeau_bengio"` methods.
    #'   When `"bonferroni"`, confidence intervals are also adjusted (alpha/k).
    #'   For other correction methods (e.g. `"holm"`, `"BH"`), only p-values are adjusted;
    #'   confidence intervals remain at the nominal `conf_level` because these sequential/adaptive
    #'   procedures do not have a clean per-comparison alpha for CI construction.
    #'
    #' @references
    #' `r print_bib("nadeau_2003")`
    #' `r print_bib("molnar_2023")`
    #'
    importance = function(
      relation = NULL,
      standardize = FALSE,
      ci_method = c("none", "raw", "nadeau_bengio", "quantile"),
      conf_level = 0.95,
      alternative = c("two.sided", "greater"),
      p_adjust = "none",
      ...
    ) {
      if (is.null(private$.scores)) {
        cli::cli_inform(c(
          x = "No importances computed yet!"
        ))
        return(invisible(NULL))
      }

      # Catch unknown arguments that were not consumed by subclass methods
      if (...length() > 0) {
        dots = list(...)
        cli::cli_abort(c(
          "Unknown argument{?s}: {.arg {names(dots)}}.",
          i = "These arguments are not used by {.fun $importance} with {.code ci_method = \"{ci_method}\"}."
        ))
      }

      # Validate ci_method
      if (length(ci_method) > 1) {
        ci_method = ci_method[1]
      }
      checkmate::assert_choice(ci_method, choices = private$.ci_methods)
      checkmate::assert_number(conf_level, lower = 0, upper = 1)
      checkmate::assert_choice(p_adjust, choices = stats::p.adjust.methods)
      alternative = match.arg(alternative)

      # Get aggregator and scores
      aggregator = self$measure$aggregator %||% mean
      scores = self$scores(relation = relation)

      # Standardize first so variance calculations use standardized values
      if (standardize) {
        scores[, importance := importance / max(abs(importance), na.rm = TRUE)]
      }

      # Dispatch to appropriate aggregation function
      agg_importance = switch(
        ci_method,
        none = importance_none(scores, aggregator, conf_level),
        raw = importance_raw(
          scores,
          aggregator,
          conf_level,
          alternative,
          self$resample_result$iters,
          p_adjust = p_adjust
        ),
        nadeau_bengio = importance_nadeau_bengio(
          scores,
          aggregator,
          conf_level,
          alternative,
          self$resampling,
          self$resample_result$iters,
          p_adjust = p_adjust
        ),
        quantile = importance_quantile(scores, aggregator, conf_level, alternative),
        cli::cli_abort(c(
          "Variance method {.val {ci_method}} not found.",
          i = "Available methods: {.val {private$.ci_methods}}"
        ))
      )

      setkeyv(agg_importance, "feature")
      agg_importance[]
    },

    #' @description
    #' Calculate observation-wise importance scores.
    #'
    #' Requires that `$compute()` was run and that `measure` is decomposable and
    #' has an observation-wise loss (`Measure$obs_loss()`) associated with it.
    #' This is not the case for measure like `classif.auc`, which is not decomposable.
    #'
    #' @param relation (character(1)) How to relate perturbed scores to originals ("difference" or "ratio"). If `NULL`, uses stored parameter value. This is only applicable for methods where importance is based on some
    #' relation between baseline and post-modification loss, i.e. [PerturbationImportance] methods such as [PFI] or [WVIM] / [LOCO]. Not available for [SAGE] methods.
    #'   For `"ratio"`, features whose baseline score is `0` yield `NA` (with a warning) and therefore aggregate to `NA`.
    #'
    #' @return ([data.table][data.table::data.table]) Observation-wise losses and importance scores with columns
    #'   `"feature"`, `"iter_rsmp"`, `"iter_repeat"` (if applicable), `"row_ids"`, `"loss_baseline"`, `"loss_post"`, and `"obs_importance"`.
    obs_loss = function(relation = NULL) {
      if (!has_obs_loss(self$measure)) {
        cli::cli_warn(c(
          x = "{.cls Measure} {.val {self$measure$id}} does not have an observation-wise loss:",
          i = "Is it decomposable?"
        ))
        return(invisible(NULL))
      }
      if (is.null(private$.obs_losses)) {
        cli::cli_warn(c(
          x = "No observation-wise losses stored!",
          i = "Did you run {.fun $compute}?",
          i = "Not all methods support observation-wise losses"
        ))
        return(invisible(NULL))
      }

      relation = resolve_param(relation, self$param_set$values$relation, "difference")

      # Prepare baseline losses
      obs_loss_baseline = self$resample_result$obs_loss(measures = self$measure)
      # obs_loss_baseline[, let(truth = NULL, response = NULL)]
      setnames(
        obs_loss_baseline,
        old = c("iteration", self$measure$id),
        new = c("iter_rsmp", "loss_baseline")
      )

      obs_loss_combined = obs_loss_baseline[
        private$.obs_losses,
        on = .(iter_rsmp, row_ids),
        allow.cartesian = TRUE
      ]

      obs_loss_combined[,
        obs_importance := private$.compute_score(
          loss_baseline,
          loss_post,
          relation = relation
        )
      ]

      # Select / reorder column names, some may be
      # specific to methods and may not be present
      names_to_keep = c(
        "feature",
        "iter_rsmp",
        "iter_repeat",
        "row_ids",
        "loss_baseline",
        "loss_post",
        "obs_importance"
      )
      names_to_keep = intersect(names_to_keep, colnames(obs_loss_combined))

      obs_loss_combined[, .SD, .SDcols = names_to_keep][]
    },

    #' @description
    #' Resets all stored fields populated by `$compute`: `$resample_result`, `$scores`, `$obs_losses`, and `$predictions`.
    reset = function() {
      self$resample_result = NULL
      private$.scores = NULL
      private$.obs_losses = NULL
      self$predictions = NULL
      # SAGE-specific fields (only reset if they exist)
      if ("n_permutations_used" %in% names(self)) {
        self$n_permutations_used = NULL
      }
    },

    #' @description
    #' Print importance scores
    #'
    #' @param ... Passed to `print()`
    print = function(...) {
      cli::cli_h2(self$label)
      cli::cli_ul()
      cli::cli_li("Learner: {.val {self$learner$id}}")
      cli::cli_li("Task: {.val {self$task$id}}")
      if (is.null(self$groups)) {
        cli::cli_li(
          "{.emph {length(self$features)}} feature{?s} of interest: {.val {self$features}}"
        )
      } else {
        cli::cli_li("{.emph {length(self$groups)}} feature group{?s} of interest:")
        ol = cli::cli_ol()
        for (i in seq_along(groups)) {
          cli::cli_li("{.strong {names(groups)[i]}}: {.val {groups[[i]]}}")
        }
        cli::cli_end(ol)
      }
      cli::cli_li("Resampling: {.val {self$resampling$id}} ({.val {self$resampling$iters}} iters)")

      cli::cli_li("Parameters:")

      pv = self$param_set$values
      pidx = seq_along(pv)
      sapply(pidx, \(i) {
        cli::cli_ul("{.code {names(pv)[i]}}: {.val {pv[i]}}")
      })
      cli::cli_end()
      self$importance()
    },

    #' @description
    #' Calculate importance scores for each resampling iteration and sub-iterations
    #' (`iter_rsmp` in [PFI] for example).
    #'
    #' Iteration-wise importance are computed on the fly depending on the chosen relation
    #' (`difference` or `ratio`) to avoid re-computation if only a different relation is needed.
    #'
    #' @param relation (character(1)) How to relate perturbed scores to originals ("difference" or "ratio"). If `NULL`, uses stored parameter value. This is only applicable for methods where importance is based on some
    #' relation between baseline and post-modification loss, i.e. [PerturbationImportance] methods such as [PFI] or [WVIM] / [LOCO]. Not available for [SAGE] methods.
    #'   For `"ratio"`, features whose baseline score is `0` yield `NA` (with a warning) and therefore aggregate to `NA`.
    #'
    #' @return ([data.table][data.table::data.table]) Iteration-wise importance scores with columns for
    #'   `"feature"`, iteration indices, baseline and post-modification scores, and `"importance"`.
    scores = function(relation = NULL) {
      if (is.null(private$.scores)) {
        cli::cli_warn(c(
          x = "No importances computed yet!",
          i = "Did you run {.fun $compute}?"
        ))
        return(invisible(NULL))
      }
      if ("importance" %in% colnames(private$.scores)) {
        # If there is already an importance variable in the stored scores like in SAGE,
        # we can't calculate pre/post scores like in PFI, LOCO etc,
        # individual "scores" would have different meaning there
        return(private$.scores)
      }

      relation = resolve_param(relation, self$param_set$values$relation, "difference")

      scores = data.table::copy(private$.scores)[,
        importance := private$.compute_score(
          score_baseline,
          score_post,
          relation = relation
        )
      ]

      setnames(
        scores,
        old = c("score_baseline", "score_post"),
        new = c(paste0(self$measure$id, c("_baseline", "_post")))
      )

      scores[]
    }
  ),
  private = list(
    # Registry of available variance methods
    .ci_methods = c("none", "raw", "nadeau_bengio", "quantile"),

    # Take the raw predictions as returned by $predict_newdata_fast and convert to Prediction object fitting the task type to simplify type-specific handling
    # @param raw_prediction `list` with elements `reponse` (vector) or `prob` (matrix) depending on task type.
    # @param test_row_ids `integer()` test set row ids, important to ensure predictions can be matched with original observations / baseline predictions
    # .construct_pred = function(raw_prediction, test_row_ids) {
    # 	truth = self$task$truth(rows = test_row_ids)

    # 	switch(
    # 		self$task$task_type,
    # 		classif = PredictionClassif$new(
    # 			row_ids = test_row_ids,
    # 			truth = truth,
    # 			response = raw_prediction$response, # vector of class names or NULL
    # 			prob = raw_prediction$prob # matrix for predict_type prob or NULL
    # 		),
    # 		regr = PredictionRegr$new(
    # 			row_ids = test_row_ids,
    # 			truth = truth,
    # 			response = raw_prediction$response # numeric
    # 		)
    # 	)
    # },

    # Utility to convert named list of groups of features into data.table to
    # make it a little easier to match group names and features in list columns etc
    # Used in WVIM where mlr3fselect stores "left in" features as list columns
    .groups_tbl = function() {
      group_tbl = data.table::data.table(
        group = names(self$groups),
        features_lst = unname(self$groups)
      )
      group_tbl[, features := vapply(features_lst, \(x) paste0(x, collapse = ";"), character(1))]
      group_tbl
    },

    # Scoring utility for computing importances
    #
    # Computes the `relation` of score before a change (e.g. PFI, LOCO, ...) and after.
    # If `minimize == TRUE`, then `scores_post - scores_pre` is computed for
    # `relation == "difference"`, otherwise `scores_pre - scores_post` is given.
    # If `minimize == FALSE`, then the order is flipped, ensuring that "higher value" means "more important".
    # @param scores_pre,scores_post (`numeric()`) Vector of scores or loss values at baseline / before (`_pre`) a modification, and after (`_post`) a modification (e.g., permutation or refit).
    # @param relation (`character(1)`: `"difference"`) Calculate the difference or `"ratio"` between pre and post modification value.
    .compute_score = function(
      scores_pre,
      scores_post,
      relation = c("difference", "ratio")
    ) {
      checkmate::assert_numeric(scores_pre, any.missing = FALSE)
      checkmate::assert_numeric(scores_post, any.missing = FALSE)
      checkmate::assert_true(length(scores_pre) == length(scores_post))
      relation = match.arg(relation)
      minimize = self$measure$minimize

      # General idea assuming a important feature:
      # For PFI and MSE (minimize == TRUE): post - baseline gives large value -> high importance
      # =="== and classif.acc (minimize == FALSE) -> post - baseline  = negative, so we flip
      # For WVIM when we "leave-in", the baseline scores are the empty model,
      # so we need to flip directions of the comparison as well to ensure "higher importance value" -> "more important"
      if (identical(self$direction, "leave-in")) {
        minimize = !minimize
      }

      # I know this could be more concise but for the time I prefer it to be very obvious in what happens when
      # General expectation -> higher score => more important
      if (minimize) {
        # Lower is better, e.g. ce, where scores_pre is expected to be smaller and scores_post larger
        switch(
          relation,
          difference = scores_post - scores_pre,
          ratio = safe_ratio(scores_post, scores_pre)
        )
      } else {
        # Higher is better, e.g. accuracy, where scores_pre is expected to be larger and scores_post smaller
        switch(
          relation,
          difference = scores_pre - scores_post,
          ratio = safe_ratio(scores_pre, scores_post)
        )
      }
    },

    # @field .scores ([data.table][data.table::data.table]) Iteration-wise importances scores. Essentially an aggregated form of .obs_losses (which may not be available), used as basis for the calculation in `$importance() `and `$scores()`.
    .scores = NULL,

    # @field .obs_losses ([data.table][data.table::data.table]) Observation-wise losses when available. Contains columns for row_ids, feature, iteration indices, individual loss values, and observation-wise losses for baseline and modified case.
    .obs_losses = NULL
  )
)

# Custom testthat expectations for importance result validation
#
# These follow testthat 3e conventions.
# See: https://testthat.r-lib.org/articles/custom-expectation.html

# -----------------------------------------------------------------------------
# expect_importance_dt
# -----------------------------------------------------------------------------

#' Expectation for aggregated importance score tables
#'
#' Validates columns:
#' - `feature` is a character value without missings
#' - `importance` is numeric vector without missings or infinite values
#' - Variance-related columns (se, estimate, conf_lower, conf_upper, statistic, p.value) may contain NA
#'
#' @param x (data.table()) Importance result table to validate.
#' @param features (character()) Feature names used to test names and order of importance scores.
expect_importance_dt = function(x, features) {
  checkmate::expect_data_table(
    x,
    types = c("character", "numeric"),
    nrows = length(features),
    min.cols = 2
  )

  # Core columns must not have missing values
  checkmate::expect_character(x$feature, any.missing = FALSE)
  checkmate::expect_numeric(x$importance, any.missing = FALSE)

  # Variance-related columns may contain NA (e.g., CPI test statistics can fail for some features)
  variance_cols = c("se", "sd", "estimate", "conf_lower", "conf_upper", "statistic", "p.value")
  for (col in intersect(variance_cols, colnames(x))) {
    checkmate::expect_numeric(x[[col]], any.missing = TRUE)
  }
}

# -----------------------------------------------------------------------------
# expect_scores_dt
# -----------------------------------------------------------------------------

#' Expectation for iteration-wise importance score tables
#'
#' Validates $scores() output. Works for all FeatureImportanceMethod subclasses
#' (PFI, CFI, RFI, WVIM/LOCO, SAGE) despite their different column structures.
#'
#' @param x (data.table()) Score result table from $scores().
#' @param features (character()) Feature names that should appear in the table.
expect_scores_dt = function(x, features) {
  checkmate::expect_data_table(x, min.rows = length(features), any.missing = FALSE)
  checkmate::expect_character(x$feature, any.missing = FALSE)
  checkmate::expect_numeric(x$importance, any.missing = FALSE)
  expect_true(all(features %in% x$feature))
}

# -----------------------------------------------------------------------------
# expect_obs_loss_dt
# -----------------------------------------------------------------------------

#' Expectation for observation-wise loss tables
#'
#' Validates $obs_loss() output. Only applicable for perturbation methods with
#' decomposable measures (e.g., regr.mse, classif.ce).
#'
#' @param x (data.table()) Observation-wise loss table from $obs_loss().
#' @param features (character()) Feature names that should appear in the table.
expect_obs_loss_dt = function(x, features) {
  checkmate::expect_data_table(x, min.rows = length(features), any.missing = FALSE)
  checkmate::expect_character(x$feature, any.missing = FALSE)
  expect_true(all(c("row_ids", "loss_baseline", "loss_post", "obs_importance") %in% names(x)))
  checkmate::expect_numeric(x$obs_importance, any.missing = FALSE)
  expect_true(all(features %in% x$feature))
}

# -----------------------------------------------------------------------------
# expect_method_output
# -----------------------------------------------------------------------------

#' Omnibus expectation for a computed FeatureImportanceMethod
#'
#' Validates all three main outputs of a computed method:
#' - $importance(): always checked
#' - $scores(): always checked
#' - $obs_loss(): checked if the method supports it (decomposable measure + perturbation method)
#'
#' @param method A computed FeatureImportanceMethod (must have had $compute() called)
expect_method_output = function(method) {
  features = method$features

  expect_importance_dt(method$importance(), features = features)
  expect_scores_dt(method$scores(), features = features)

  # obs_loss is only available for perturbation methods with decomposable measures
  if (has_obs_loss(method$measure) && !inherits(method, "SAGE")) {
    expect_obs_loss_dt(method$obs_loss(), features = features)
  }
}

# =============================================================================
# PFI Tests using higher-level test helpers
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality
# -----------------------------------------------------------------------------

test_that("PFI default behavior with minimal parameters", {
  test_default_behavior(PFI, task_type = "regr", n_repeats = 1L)
})

test_that("PFI basic workflow with classification", {
  task = tgen("2dnormals")$generate(n = 100)

  pfi = PFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    n_repeats = 1L
  )
  checkmate::expect_r6(pfi, c("FeatureImportanceMethod", "PFI"))

  pfi$compute()
  expect_method_output(pfi)
})

test_that("PFI featureless learner produces zero importance", {
  test_featureless_zero_importance(PFI, task_type = "classif", n_repeats = 1L)
})

# -----------------------------------------------------------------------------
# Repeats and scores
# -----------------------------------------------------------------------------

test_that("PFI multiple repeats and scores structure", {
  task = tgen("friedman1")$generate(n = 200)

  test_n_repeats_and_scores(
    PFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    n_repeats = 2L
  )
})

test_that("PFI single feature", {
  task = tgen("friedman1")$generate(n = 200)

  test_single_feature(
    PFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    feature = "important4",
    n_repeats = 2L
  )
})

# -----------------------------------------------------------------------------
# Relation parameter
# -----------------------------------------------------------------------------

test_that("PFI difference vs ratio relations", {
  task = tgen("2dnormals")$generate(n = 100)

  test_relation_parameter(
    PFI,
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("PFI friedman1 produces sensible ranking", {
  test_friedman1_sensible_ranking(PFI, n_repeats = 5L)
})

# -----------------------------------------------------------------------------
# Grouped importance
# -----------------------------------------------------------------------------

test_that("PFI with feature groups", {
  task = tgen("friedman1")$generate(n = 200)

  groups = list(
    important_group = c("important1", "important2", "important3"),
    unimportant_group = c("unimportant1", "unimportant2")
  )

  test_grouped_importance(
    PFI,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    groups = groups,
    expected_classes = c("FeatureImportanceMethod", "PFI"),
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# PFI-specific tests (not covered by generic helpers)
# -----------------------------------------------------------------------------

test_that("PFI scores and obs_losses agree", {
  task = tgen("friedman1")$generate(n = 200)

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("cv", folds = 3),
    n_repeats = 2
  )

  pfi$compute()

  importance_agg = pfi$importance()
  importance_scores = pfi$scores()[, .(iter_rsmp, iter_repeat, feature, importance)][
    order(iter_rsmp, iter_repeat, feature)
  ]
  importance_obs_loss = pfi$obs_loss()

  expect_equal(
    importance_agg,
    importance_scores[, list(importance = mean(importance)), by = "feature"],
    ignore_attr = TRUE
  )

  # Aggregate squared errors to get mse per iteration
  obs_agg = importance_obs_loss[,
    list(importance = mean(obs_importance)),
    by = c("iter_rsmp", "iter_repeat", "feature")
  ][order(iter_rsmp, iter_repeat, feature)]

  expect_equal(importance_scores, obs_agg, tolerance = sqrt(.Machine$double.eps))
})

test_that("PFI obs_loss uses specified measure, not default", {
  # Regression test for bug where pred$obs_loss() was called without
  # passing the measure, causing it to use the default measure (regr.mse)
  # instead of the user-specified one. With a non-default measure like
  # regr.mae, this caused obs_loss values to be NULL/wrong.
  task = mlr3::tsk("mtcars")

  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mae"),
    n_repeats = 1L
  )

  pfi$compute()

  obs_losses = pfi$obs_loss()
  expect_obs_loss_dt(obs_losses, features = task$feature_names)

  # obs_loss values should be absolute errors (MAE), not squared errors (MSE).
  # Verify by checking that aggregating obs losses matches the score-level
  # importance (which correctly uses the specified measure).
  scores = pfi$scores()[, .(iter_rsmp, feature, importance)][order(iter_rsmp, feature)]

  obs_agg = obs_losses[,
    list(importance = mean(obs_importance)),
    by = c("iter_rsmp", "feature")
  ][order(iter_rsmp, feature)]

  expect_equal(scores, obs_agg, tolerance = sqrt(.Machine$double.eps))
})

# =============================================================================
# MarginalSAGE Tests
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality
# -----------------------------------------------------------------------------

test_that("MarginalSAGE default behavior with minimal parameters", {
  # Use small params for test speed
  test_default_behavior(MarginalSAGE, task_type = "regr", n_permutations = 2L, n_samples = 20L)
})

test_that("MarginalSAGE works with classification tasks", {
  # Binary classification
  task_binary = tgen("2dnormals")$generate(n = 100)
  sage_binary = MarginalSAGE$new(
    task = task_binary,
    learner = lrn("classif.rpart", predict_type = "prob"),
    n_permutations = 2L
  )
  checkmate::expect_r6(sage_binary, c("FeatureImportanceMethod", "SAGE", "MarginalSAGE"))
  sage_binary$compute()
  expect_importance_dt(sage_binary$importance(), features = sage_binary$features)

  # Multiclass classification
  task_multi = tgen("cassini")$generate(n = 100)
  sage_multi = MarginalSAGE$new(
    task = task_multi,
    learner = lrn("classif.rpart", predict_type = "prob"),
    n_permutations = 2L
  )
  sage_multi$compute()
  expect_importance_dt(sage_multi$importance(), features = sage_multi$features)
  expect_length(task_multi$class_names, 3L)
})

test_that("MarginalSAGE featureless learner produces zero importance", {
  # Use small params for test speed
  test_featureless_zero_importance(
    MarginalSAGE,
    task_type = "regr",
    n_permutations = 2L,
    n_samples = 20L
  )
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("MarginalSAGE friedman1 produces sensible ranking", {
  # Use small params for test speed
  test_friedman1_sensible_ranking(MarginalSAGE, n = 200L, n_permutations = 2L, n_samples = 20L)
})

# -----------------------------------------------------------------------------
# Resampling
# -----------------------------------------------------------------------------

test_that("MarginalSAGE with cross-validation resampling", {
  task = tgen("friedman1")$generate(n = 200)

  sage = MarginalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    resampling = rsmp("cv", folds = 3),
    n_permutations = 2L
  )
  sage$compute()

  expect_importance_dt(sage$importance(), features = sage$features)
  checkmate::expect_data_table(
    sage$scores(),
    types = c("integer", "character", "numeric"),
    nrows = sage$resampling$iters * length(sage$features),
    ncols = 3,
    any.missing = FALSE
  )
})

# -----------------------------------------------------------------------------
# Single feature
# -----------------------------------------------------------------------------

test_that("MarginalSAGE with single feature", {
  task = tgen("friedman1")$generate(n = 100)

  sage = MarginalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    features = "important4",
    n_permutations = 2L
  )
  sage$compute()

  expect_importance_dt(sage$importance(), features = "important4")
  expect_equal(nrow(sage$importance()), 1L)
})

# -----------------------------------------------------------------------------
# n_samples parameter
# -----------------------------------------------------------------------------

test_that("MarginalSAGE with custom n_samples", {
  task = tgen("friedman1")$generate(n = 200)

  sage = MarginalSAGE$new(
    task = task,
    learner = lrn("regr.rpart"),
    n_samples = 30L,
    n_permutations = 2L
  )
  sage$compute()

  expect_importance_dt(sage$importance(), features = sage$features)
})

# -----------------------------------------------------------------------------
# Reproducibility
# -----------------------------------------------------------------------------

test_that("MarginalSAGE reproducibility with same seed", {
  task = tgen("2dnormals")$generate(n = 100)
  learner = lrn("classif.rpart", predict_type = "prob")
  measure = msr("classif.ce")

  set.seed(42)
  sage1 = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    n_permutations = 3L
  )
  sage1$compute()
  result1 = sage1$importance()

  set.seed(42)
  sage2 = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    n_permutations = 3L
  )
  sage2$compute()
  result2 = sage2$importance()

  # Results should be identical with same seed
  expect_equal(result1$importance, result2$importance, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# Parameter validation
# -----------------------------------------------------------------------------

test_that("MarginalSAGE parameter validation", {
  task = tgen("friedman1")$generate(n = 50)
  learner = lrn("regr.rpart")

  # n_permutations must be positive integer
  expect_error(MarginalSAGE$new(task = task, learner = learner, n_permutations = 0L))
  expect_error(MarginalSAGE$new(task = task, learner = learner, n_permutations = -1L))
})

test_that("MarginalSAGE requires predict_type='prob' for classification", {
  task = tgen("2dnormals")$generate(n = 50)

  # Should error for classification without predict_type = "prob"
  expect_error(
    MarginalSAGE$new(
      task = task,
      learner = lrn("classif.rpart", predict_type = "response")
    ),
    "Classification learners require probability predictions for SAGE."
  )
})

# -----------------------------------------------------------------------------
# Convergence tracking
# -----------------------------------------------------------------------------

test_that("MarginalSAGE SE tracking in convergence_history", {
  task = tgen("friedman1")$generate(n = 30)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  sage = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    n_permutations = 6L,
    n_samples = 20L
  )

  # Compute with early stopping to get convergence history
  sage$compute(early_stopping = TRUE, se_threshold = 0.05, check_interval = 2L)

  # Check that convergence_history exists and has SE column
  expect_false(is.null(sage$convergence_history))
  expect_contains(colnames(sage$convergence_history), "se")

  # Check structure of convergence_history
  expected_cols = c("n_permutations", "feature", "importance", "se")
  expect_setequal(colnames(sage$convergence_history), expected_cols)

  # SE values should be non-negative and finite
  se_values = sage$convergence_history$se
  checkmate::expect_numeric(se_values, lower = 0, finite = TRUE)

  # For each feature, SE should be in a reasonable range
  for (feat in unique(sage$convergence_history$feature)) {
    feat_data = sage$convergence_history[feature == feat]
    feat_data = feat_data[order(n_permutations)]

    if (nrow(feat_data) > 1) {
      # Just check that SE values are in a reasonable range and not exploding
      expect_lt(max(feat_data$se), 10)
      expect_lt(max(abs(diff(feat_data$se))), 5)
    }
  }

  # All features should be represented in convergence history
  expect_setequal(
    unique(sage$convergence_history$feature),
    sage$features
  )
})

test_that("MarginalSAGE SE-based convergence detection", {
  skip_on_cran() # ~1s - tests early stopping feature, not core SAGE

  task = tgen("friedman1")$generate(n = 100)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  sage = MarginalSAGE$new(
    task = task,
    learner = learner,
    measure = measure,
    n_permutations = 10L,
    n_samples = 20L
  )

  # Test with very loose SE threshold (should trigger convergence easily)
  sage$compute(
    early_stopping = TRUE,
    se_threshold = 100.0,
    min_permutations = 5L,
    check_interval = 1L
  )

  # Should converge early because SE will be well below 100.0
  expect_true(sage$converged)
  expect_lte(sage$n_permutations_used, 10L)

  # Reset for next test
  sage$reset()

  # Test with very strict SE threshold (should not converge)
  sage$compute(
    early_stopping = TRUE,
    se_threshold = 0.001,
    min_permutations = 5L,
    check_interval = 1L
  )

  # With very strict SE threshold, should not converge early
  expect_false(sage$converged)

  # Test with moderate SE threshold
  sage$reset()

  sage$compute(
    early_stopping = TRUE,
    se_threshold = 0.1,
    min_permutations = 5L,
    check_interval = 1L
  )

  # Should have convergence history with SE tracking regardless of convergence
  expect_false(is.null(sage$convergence_history))
  expect_contains(colnames(sage$convergence_history), "se")
})

test_that("coalition-loss direct measure$fun path matches per-coalition $score() (regression)", {
  # Guards the direct measure$fun path in .calculate_coalition_losses
  # (used for obs_loss regression measures) against the canonical
  # per-coalition Prediction + $score() path.
  task = sim_dgp_independent(40)
  # learner is irrelevant here: the private method under test only uses
  # task + measure, so use the dependency-free featureless learner.
  learner = lrn("regr.featureless")

  n_test = 12L
  n_coal = 5L
  set.seed(606)
  truth = rnorm(n_test)
  test_dt = data.table(y = truth)
  avg_preds = CJ(.coalition_id = seq_len(n_coal), .test_instance_id = seq_len(n_test))
  avg_preds[, avg_pred := rnorm(.N)]

  for (meas_id in c("regr.mse", "regr.mae")) {
    sage = MarginalSAGE$new(task, learner, measure = msr(meas_id), n_permutations = 2L)
    priv = sage$.__enclos_env__$private

    # copy(): the method keys avg_preds by reference
    fast = priv$.calculate_coalition_losses(copy(avg_preds), n_test, test_dt)

    slow = vapply(
      seq_len(n_coal),
      function(i) {
        cd = avg_preds[.coalition_id == i][order(.test_instance_id)]
        PredictionRegr$new(
          row_ids = seq_len(n_test),
          truth = truth,
          response = cd$avg_pred
        )$score(msr(meas_id))
      },
      numeric(1)
    )

    expect_equal(fast, slow, info = meas_id)
  }
})

test_that("coalition-loss direct measure$fun path matches per-coalition $score() (classif prob)", {
  # classif.logloss is a prob obs_loss measure -> direct measure$fun path.
  task = tgen("2dnormals")$generate(40)
  learner = lrn("classif.featureless", predict_type = "prob")
  classes = task$class_names

  n_test = 12L
  n_coal = 4L
  set.seed(707)
  truth = factor(sample(classes, n_test, replace = TRUE), levels = classes)
  test_dt = data.table(class = truth)
  setnames(test_dt, "class", task$target_names)

  # synthetic averaged class probabilities per (coalition, test instance)
  avg_preds = CJ(.coalition_id = seq_len(n_coal), .test_instance_id = seq_len(n_test))
  pm = matrix(runif(nrow(avg_preds) * length(classes)), ncol = length(classes))
  pm = pm / rowSums(pm)
  for (j in seq_along(classes)) {
    avg_preds[, (classes[j]) := pm[, j]]
  }

  sage = MarginalSAGE$new(task, learner, measure = msr("classif.logloss"), n_permutations = 2L)
  priv = sage$.__enclos_env__$private
  fast = priv$.calculate_coalition_losses(copy(avg_preds), n_test, test_dt)

  slow = vapply(
    seq_len(n_coal),
    function(i) {
      cd = avg_preds[.coalition_id == i][order(.test_instance_id)]
      PredictionClassif$new(
        row_ids = seq_len(n_test),
        truth = truth,
        prob = as.matrix(cd[, .SD, .SDcols = classes])
      )$score(msr("classif.logloss"))
    },
    numeric(1)
  )

  expect_equal(fast, slow)
})

# =============================================================================
# WVIM/LOCO Tests using higher-level test helpers
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality - WVIM
# -----------------------------------------------------------------------------

test_that("WVIM default behavior with minimal parameters", {
  test_default_behavior(WVIM, task_type = "regr", direction = "leave-out", n_repeats = 1L)
})

test_that("WVIM basic workflow with regression", {
  task = tgen("friedman1")$generate(n = 150)

  wvim = WVIM$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    direction = "leave-out",
    n_repeats = 1L
  )
  checkmate::expect_r6(wvim, c("FeatureImportanceMethod", "WVIM"))

  wvim$compute()
  expect_method_output(wvim)
})

test_that("WVIM direction parameter (leave-out vs leave-in)", {
  task = tgen("friedman1")$generate(n = 150)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")
  features = task$feature_names[1:3]

  # Test leave-out direction
  wvim_out = WVIM$new(
    task = task,
    learner = learner,
    measure = measure,
    features = features,
    direction = "leave-out",
    n_repeats = 1L
  )
  expect_equal(wvim_out$direction, "leave-out")
  wvim_out$compute()
  result_out = wvim_out$importance()
  expect_importance_dt(result_out, features = features)

  # Test leave-in direction
  wvim_in = WVIM$new(
    task = task,
    learner = learner,
    measure = measure,
    features = features,
    direction = "leave-in",
    n_repeats = 1L
  )
  expect_equal(wvim_in$direction, "leave-in")
  wvim_in$compute()
  result_in = wvim_in$importance()
  expect_importance_dt(result_in, features = features)

  # Results should differ
  expect_false(isTRUE(all.equal(result_out, result_in)))
})

test_that("WVIM with feature groups", {
  task = tgen("friedman1")$generate(n = 150)

  groups = list(
    important_set = c("important1", "important2", "important3"),
    unimportant_set = c("unimportant1", "unimportant2")
  )

  test_grouped_importance(
    WVIM,
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    groups = groups,
    expected_classes = c("FeatureImportanceMethod", "WVIM"),
    direction = "leave-out",
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Basic functionality - LOCO
# -----------------------------------------------------------------------------

test_that("LOCO default behavior with minimal parameters", {
  test_default_behavior(LOCO, task_type = "regr", n_repeats = 1L)
})

test_that("LOCO basic workflow with regression", {
  task = tgen("friedman1")$generate(n = 100)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    n_repeats = 1L
  )
  checkmate::expect_r6(loco, c("FeatureImportanceMethod", "WVIM", "LOCO"))

  loco$compute()
  expect_method_output(loco)

  # LOCO-specific checks
  expect_equal(loco$direction, "leave-out")
  expect_equal(loco$label, "Leave-One-Covariate-Out (LOCO)")
})

test_that("LOCO basic workflow with classification", {
  task = tgen("simplex", d = 5)$generate(n = 100)

  loco = LOCO$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.ce"),
    n_repeats = 1L
  )
  checkmate::expect_r6(loco, c("FeatureImportanceMethod", "WVIM", "LOCO"))

  loco$compute()
  expect_method_output(loco)
})

# -----------------------------------------------------------------------------
# Feature selection
# -----------------------------------------------------------------------------

test_that("LOCO with all features (features = NULL)", {
  task = tgen("friedman1")$generate(n = 100)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  loco = LOCO$new(task, learner, measure, n_repeats = 1L)
  expect_equal(loco$features, task$feature_names)

  loco$compute()
  expect_importance_dt(loco$importance(), features = task$feature_names)
})

test_that("LOCO with feature subset", {
  task = tgen("friedman1")$generate(n = 100)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  features_subset = task$feature_names[1:3]
  loco = LOCO$new(task, learner, measure, features = features_subset, n_repeats = 1L)
  expect_equal(loco$features, features_subset)

  loco$compute()
  expect_importance_dt(loco$importance(), features = features_subset)
})

# -----------------------------------------------------------------------------
# Repeats and resampling
# -----------------------------------------------------------------------------

test_that("LOCO with multiple refits", {
  task = tgen("friedman1")$generate(n = 150)
  learner = lrn("regr.rpart")
  measure = msr("regr.mse")

  loco = LOCO$new(
    task = task,
    learner = learner,
    measure = measure,
    features = task$feature_names[1:3],
    n_repeats = 3L
  )

  loco$compute()
  expect_importance_dt(loco$importance(), features = loco$features)

  # Scores should have multiple refits
  scores = loco$scores()
  expect_gte(nrow(scores), length(loco$features))
  expect_true(all(scores$iter_repeat %in% 1:3))
})

test_that("LOCO batch_size > 1 matches sequential (no cartesian join blow-up)", {
  # Regression: batch_size > 1 chunks multiple design points per internal
  # benchmark() call. The post-benchmark aggregation must key on a
  # per-design-point id, not the bbotk batch index, or the archive<->scores
  # join explodes (data.table cartesian guard) / yields wrong importances.
  make_loco = function(bs) {
    LOCO$new(
      task = sim_dgp_independent(n = 200),
      learner = lrn("regr.rpart"),
      measure = msr("regr.mse"),
      n_repeats = 3L,
      batch_size = bs
    )
  }

  withr::local_seed(6829)
  loco_seq = make_loco(NULL)
  loco_seq$compute()

  withr::local_seed(6829)
  loco_batched = make_loco(4L)
  # Must not raise the data.table cartesian-join error
  expect_no_error(loco_batched$compute())

  n_feat = length(loco_seq$features)
  # holdout -> 1 resampling iteration; expect exactly n_features * n_repeats rows,
  # NOT inflated by batch_size
  expect_equal(nrow(loco_batched$scores()), n_feat * 3L)
  expect_setequal(loco_batched$scores()$iter_repeat, 1:3)

  # Same seed + same per-design-point training => identical importances
  imp_seq = loco_seq$importance()
  imp_batched = loco_batched$importance()
  expect_equal(
    imp_batched[order(feature)]$importance,
    imp_seq[order(feature)]$importance
  )
})

test_that("LOCO batch_size > 1 obs_loss path (ci_method='lei') matches sequential", {
  # The obs_loss aggregation has a separate archive<->obs_loss join with the
  # same per-design-point keying requirement; exercised via Lei inference.
  # Re-seed and rebuild before each compute so both runs consume RNG from the
  # same state (rpart fit + holdout split are stochastic).
  make_loco = function(bs) {
    LOCO$new(
      task = sim_dgp_independent(n = 200),
      learner = lrn("regr.rpart"),
      measure = msr("regr.mae"),
      n_repeats = 1L,
      batch_size = bs
    )
  }

  withr::local_seed(415)
  loco_seq = make_loco(NULL)
  loco_seq$compute()
  imp_seq = loco_seq$importance(ci_method = "lei")

  withr::local_seed(415)
  loco_batched = make_loco(4L)
  expect_no_error(loco_batched$compute())
  imp_batched = loco_batched$importance(ci_method = "lei")

  expect_importance_dt(imp_batched, features = loco_batched$features)
  expect_equal(
    imp_batched[order(feature)]$importance,
    imp_seq[order(feature)]$importance
  )
})

test_that("LOCO with cross-validation", {
  task = tgen("friedman1")$generate(n = 150)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("cv", folds = 3),
    features = task$feature_names[1:2]
  )
  loco$compute()
  expect_importance_dt(loco$importance(), features = loco$features)
})

# -----------------------------------------------------------------------------
# Sensible results
# -----------------------------------------------------------------------------

test_that("LOCO friedman1 produces sensible ranking", {
  test_friedman1_sensible_ranking(
    LOCO,
    n_repeats = 1L
  )
})

# -----------------------------------------------------------------------------
# Lei et al. (2018) inference
# -----------------------------------------------------------------------------

test_that("LOCO ci_method='lei' with default wilcoxon test", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  imp = loco$importance(ci_method = "lei")

  expect_importance_dt(imp, features = loco$features)
  expected_cols = c(
    "feature",
    "importance",
    "se",
    "statistic",
    "p.value",
    "conf_lower",
    "conf_upper"
  )
  expect_true(all(expected_cols %in% names(imp)))
  expect_true(all(is.finite(imp$importance)))
  expect_true(all(is.finite(imp$p.value)))
  expect_true(all(imp$p.value >= 0 & imp$p.value <= 1))
})

test_that("LOCO ci_method='lei' with t-test", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  imp = loco$importance(ci_method = "lei", test = "t")

  expect_importance_dt(imp, features = loco$features)
  expect_true(all(c("statistic", "p.value", "conf_lower", "conf_upper") %in% names(imp)))
})

test_that("LOCO ci_method='lei' alternative='greater' produces one-sided CIs", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  imp_greater = suppressWarnings(loco$importance(ci_method = "lei", alternative = "greater"))
  imp_twosided = suppressWarnings(loco$importance(ci_method = "lei", alternative = "two.sided"))

  # One-sided: upper bound is Inf; lower bound is finite (or NA on
  # degenerate inputs that defeat the asymptotic variance estimator,
  # observed on R-devel 2026-05+).
  expect_true(all(is.infinite(imp_greater$conf_upper)))
  expect_true(all(is.finite(imp_greater$conf_lower) | is.na(imp_greater$conf_lower)))

  # Two-sided: both bounds finite (or NA on the same degenerate path).
  expect_true(all(is.finite(imp_twosided$conf_lower) | is.na(imp_twosided$conf_lower)))
  expect_true(all(is.finite(imp_twosided$conf_upper) | is.na(imp_twosided$conf_upper)))
})

test_that("LOCO ci_method='lei' custom aggregator", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  # Default aggregator is median
  imp_median = loco$importance(ci_method = "lei")
  # Override with mean
  imp_mean = loco$importance(ci_method = "lei", aggregator = mean)

  # Point estimates should generally differ (median != mean)
  expect_importance_dt(imp_median, features = loco$features)
  expect_importance_dt(imp_mean, features = loco$features)
})

test_that("LOCO ci_method='lei' warns about duplicated observations", {
  task = sim_dgp_independent(n = 200)

  # Subsampling with repeats causes observations to appear in multiple test sets
  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 3),
    n_repeats = 1L
  )
  loco$compute()

  expect_warning(
    loco$importance(ci_method = "lei"),
    regexp = "multiple test sets"
  )
})

test_that("LOCO ci_method='lei' works without warnings for cross-validation", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("cv", folds = 3),
    n_repeats = 1L
  )
  loco$compute()

  # CV has no duplicate observations, so no warning expected
  # (wilcox.test may warn about zeroes, which is expected/harmless)
  imp = suppressWarnings(loco$importance(ci_method = "lei"))
  expect_importance_dt(imp, features = loco$features)
  expect_true(all(c("statistic", "p.value", "conf_lower", "conf_upper") %in% names(imp)))
})

test_that("WVIM ci_method='lei' works for leave-out direction", {
  task = tgen("friedman1")$generate(n = 200)

  wvim = WVIM$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    features = task$feature_names[1:3],
    direction = "leave-out",
    n_repeats = 1L
  )
  wvim$compute()

  imp = wvim$importance(ci_method = "lei")
  expect_importance_dt(imp, features = wvim$features)
  expect_true(all(c("statistic", "p.value", "conf_lower", "conf_upper") %in% names(imp)))
})

test_that("LOCO ci_method='lei' other ci_methods still work", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("subsampling", repeats = 5),
    n_repeats = 1L
  )
  loco$compute()

  # Base methods should still work via super$importance()
  imp_none = loco$importance(ci_method = "none")
  imp_raw = loco$importance(ci_method = "raw")

  expect_importance_dt(imp_none, features = loco$features)
  expect_importance_dt(imp_raw, features = loco$features)
})

# -----------------------------------------------------------------------------
# Multiplicity correction
# -----------------------------------------------------------------------------

test_that("LOCO ci_method='lei' p_adjust corrects p-values", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  imp_none = suppressWarnings(loco$importance(ci_method = "lei", p_adjust = "none"))
  imp_holm = suppressWarnings(loco$importance(ci_method = "lei", p_adjust = "holm"))
  imp_bonf = suppressWarnings(loco$importance(ci_method = "lei", p_adjust = "bonferroni"))
  imp_bh = suppressWarnings(loco$importance(ci_method = "lei", p_adjust = "BH"))

  # All should produce valid output
  expect_importance_dt(imp_none, features = loco$features)
  expect_importance_dt(imp_holm, features = loco$features)
  expect_importance_dt(imp_bonf, features = loco$features)
  expect_importance_dt(imp_bh, features = loco$features)

  # Adjusted p-values should be >= unadjusted (or equal for single comparisons)
  expect_true(all(imp_holm$p.value >= imp_none$p.value - 1e-10))
  expect_true(all(imp_bonf$p.value >= imp_none$p.value - 1e-10))
})

test_that("LOCO ci_method='lei' default p_adjust is 'none'", {
  task = sim_dgp_independent(n = 200)

  loco = LOCO$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse"),
    resampling = rsmp("holdout"),
    n_repeats = 1L
  )
  loco$compute()

  imp_default = suppressWarnings(loco$importance(ci_method = "lei"))
  imp_none = suppressWarnings(loco$importance(ci_method = "lei", p_adjust = "none"))

  expect_equal(imp_default$p.value, imp_none$p.value)
})

# Higher-level test functions for FeatureImportanceMethod subclasses
#
# These functions encapsulate common test patterns to reduce duplication
# and make writing new tests easier. Works for PFI, CFI, RFI, WVIM, LOCO, etc.

# -----------------------------------------------------------------------------
# test_default_behavior
# -----------------------------------------------------------------------------

#' Test that a FeatureImportanceMethod works with minimal parameters
#'
#' Verifies that:
#' - Constructor fails without any arguments (basic validation)
#' - Method can be constructed with just task and learner
#' - Default measure is automatically selected based on task type
#' - Default resampling is holdout
#' - compute() and importance() work correctly
#'
#' @param method_class R6 class (PFI, CFI, RFI, WVIM, LOCO, etc.)
#' @param task_type "regr" or "classif"
#' @param ... Additional arguments passed to method constructor (e.g., conditioning_set for RFI)
test_default_behavior = function(method_class, task_type = "regr", ...) {
  # Constructor must fail without any arguments
  expect_error(method_class$new())

  if (task_type == "regr") {
    task = tgen("friedman1")$generate(n = 100)
    learner = lrn("regr.rpart")
    expected_measure_class = "MeasureRegr"
  } else {
    task = tgen("2dnormals")$generate(n = 100)
    learner = lrn("classif.rpart", predict_type = "prob")
    expected_measure_class = "MeasureClassif"
  }

  # Construct with minimal parameters
  method = method_class$new(
    task = task,
    learner = learner,
    ...
  )

  # Check default measure was selected
  checkmate::expect_r6(method$measure, expected_measure_class)

  # Check default resampling is holdout
  expect_equal(method$resampling$id, "holdout")

  # Compute and validate
  method$compute()
  expect_importance_dt(method$importance(), features = method$features)

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_featureless_zero_importance
# -----------------------------------------------------------------------------

#' Test that featureless learner produces zero importance for all features
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task_type "regr" or "classif"
#' @param ... Additional arguments passed to method constructor
test_featureless_zero_importance = function(method_class, task_type = "classif", ...) {
  if (task_type == "regr") {
    task = tgen("friedman1")$generate(n = 200)
    learner = lrn("regr.featureless")
    measure = msr("regr.mse")
  } else {
    task = tgen("xor")$generate(n = 200)
    learner = lrn("classif.featureless")
    measure = msr("classif.ce")
  }

  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    ...
  )

  method$compute()

  result = method$importance()

  # Verify structure
  expect_importance_dt(result, features = method$features)

  # All importance values should be essentially zero
  checkmate::expect_numeric(result$importance, lower = -1e-10, upper = 1e-10)

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_n_repeats_and_scores
# -----------------------------------------------------------------------------

#' Test multiple repeats and validate scores structure
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task mlr3 Task
#' @param learner mlr3 Learner
#' @param measure mlr3 Measure
#' @param n_repeats Number of repeats
#' @param resampling mlr3 Resampling (default: CV with 3 folds)
#' @param ... Additional arguments passed to method constructor
test_n_repeats_and_scores = function(
  method_class,
  task,
  learner,
  measure,
  n_repeats = 2L,
  resampling = rsmp("cv", folds = 3),
  ...
) {
  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_repeats = n_repeats,
    ...
  )

  method$compute()

  # Validate importance
  expect_importance_dt(method$importance(), features = method$features)

  # Validate scores structure
  expected_nrows = resampling$iters * n_repeats * length(method$features)

  checkmate::expect_data_table(
    method$scores(),
    types = c("character", "integer", "numeric"),
    nrows = expected_nrows,
    ncols = 6,
    any.missing = FALSE,
    min.cols = 6
  )

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_single_feature
# -----------------------------------------------------------------------------
#' Test computing importance for a single feature
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task mlr3 Task (should have feature named in `feature` param)
#' @param learner mlr3 Learner
#' @param measure mlr3 Measure
#' @param feature Single feature name to compute importance for
#' @param n_repeats Number of repeats
#' @param resampling mlr3 Resampling
#' @param ... Additional arguments passed to method constructor
test_single_feature = function(
  method_class,
  task,
  learner,
  measure,
  feature = "important4",
  n_repeats = 2L,
  resampling = rsmp("cv", folds = 3),
  ...
) {
  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    resampling = resampling,
    n_repeats = n_repeats,
    features = feature,
    ...
  )

  method$compute()

  # Validate importance for single feature
  expect_importance_dt(method$importance(), features = feature)

  # Validate scores structure (only 1 feature)
  expected_nrows = resampling$iters * n_repeats

  checkmate::expect_data_table(
    method$scores(),
    types = c("character", "integer", "numeric"),
    nrows = expected_nrows,
    ncols = 6,
    any.missing = FALSE,
    min.cols = 6
  )

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_friedman1_sensible_ranking
# -----------------------------------------------------------------------------

#' Test that friedman1 DGP produces sensible feature ranking
#'
#' Verifies that important features (important1-5) have higher mean importance
#' than unimportant features (unimportant1-5).
#'
#' @param method_class R6 class (PFI, CFI, RFI, MarginalSAGE, ConditionalSAGE, etc.)
#' @param learner mlr3 Learner for regression
#' @param measure mlr3 Measure for regression
#' @param n Sample size for friedman1 task (default 500 for stable ranking)
#' @param ... Additional arguments passed to method constructor (e.g., n_repeats for perturbation methods)
test_friedman1_sensible_ranking = function(
  method_class,
  learner = lrn("regr.rpart"),
  measure = msr("regr.mse"),
  n = 500L,
  ...
) {
  task = tgen("friedman1")$generate(n = n)

  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    ...
  )

  method$compute()
  result = method$importance()

  expect_importance_dt(result, features = method$features)

  # Extract scores by feature type
  important_features = grep("^important", result$feature, value = TRUE)
  unimportant_features = grep("^unimportant", result$feature, value = TRUE)

  important_scores = result[feature %in% important_features]$importance
  unimportant_scores = result[feature %in% unimportant_features]$importance

  # Important features should have higher mean importance
  expect_gt(mean(important_scores), mean(unimportant_scores))

  # Scores should be finite and not all zero
  checkmate::expect_numeric(result$importance, finite = TRUE)
  expect_gt(max(abs(result$importance)), 0)

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_relation_parameter
# -----------------------------------------------------------------------------

#' Test difference vs ratio relation parameter
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task mlr3 Task
#' @param learner mlr3 Learner
#' @param measure mlr3 Measure
#' @param ... Additional arguments passed to method constructor
test_relation_parameter = function(
  method_class,
  task,
  learner,
  measure,
  ...
) {
  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    ...
  )

  method$compute()

  res_diff = method$importance(relation = "difference")
  res_ratio = method$importance(relation = "ratio")
  res_default = method$importance()

  # All should be valid
  expect_importance_dt(res_diff, method$features)
  expect_importance_dt(res_ratio, method$features)

  # Default should be "difference"
  expect_identical(res_default, res_diff)

  # Different relations should give different results
  expect_false(isTRUE(all.equal(res_diff, res_ratio)))

  invisible(method)
}

# -----------------------------------------------------------------------------
# test_parameter_validation
# -----------------------------------------------------------------------------

#' Test that invalid parameters are rejected
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task mlr3 Task
#' @param learner mlr3 Learner
#' @param measure mlr3 Measure
#' @param ... Additional arguments passed to method constructor
test_parameter_validation = function(
  method_class,
  task,
  learner,
  measure,
  ...
) {
  # n_repeats = 0 should fail
  expect_error(
    method_class$new(
      task = task,
      learner = learner,
      measure = measure,
      n_repeats = 0L,
      ...
    )
  )

  # n_repeats = -1 should fail
  expect_error(
    method_class$new(
      task = task,
      learner = learner,
      measure = measure,
      n_repeats = -1L,
      ...
    )
  )
}

# -----------------------------------------------------------------------------
# test_grouped_importance
# -----------------------------------------------------------------------------

#' Test grouped feature importance
#'
#' @param method_class R6 class (PFI, CFI, or RFI)
#' @param task mlr3 Task (should have features matching group definitions)
#' @param learner mlr3 Learner
#' @param measure mlr3 Measure
#' @param groups Named list of feature groups
#' @param expected_classes Character vector of expected R6 class names
#' @param ... Additional arguments passed to method constructor
test_grouped_importance = function(
  method_class,
  task,
  learner,
  measure,
  groups,
  expected_classes = NULL,
  ...
) {
  method = method_class$new(
    task = task,
    learner = learner,
    measure = measure,
    groups = groups,
    ...
  )

  # Check class hierarchy if specified
  if (!is.null(expected_classes)) {
    checkmate::expect_r6(method, expected_classes)
  }

  # Groups should be stored
  expect_false(is.null(method$groups))
  expect_equal(names(method$groups), names(groups))

  method$compute()
  result = method$importance()

  # Should have one row per group
  expect_length(groups, nrow(result))
  expect_equal(result$feature, names(groups))
  expect_importance_dt(result, features = names(groups))

  invisible(method)
}

#' @title Marginal Permutation Sampler
#'
#' @description Implements marginal permutation-based sampling for Permutation Feature Importance (PFI).
#' Each specified feature is randomly shuffled (permuted) independently, breaking the
#' relationship between the feature and the target as well as between rows.
#'
#' @details
#' The permutation sampler randomly shuffles feature values across observations:
#' - Each feature is permuted **independently** within its column
#' - The association between feature values and target values is broken
#' - The association between feature values **across rows** is broken
#' - The marginal distribution of each feature is preserved
#'
#' **Important distinction from SAGE's "marginal" approach:**
#' - `MarginalPermutationSampler`: Shuffles features independently, breaking row structure
#' - `MarginalSAGE`: Uses reference data but keeps rows intact (features in coalition stay together)
#'
#' This is the classic approach used in Permutation Feature Importance (PFI) and
#' assumes features are independent.
#'
#' @export
#' @examples
#' library(mlr3)
#' task = tgen("2dnormals")$generate(n = 10)
#' task$data()
#' sampler = MarginalPermutationSampler$new(task)
#'
#' # Sample using row_ids from stored task
#' sampler$sample("x1")
#'
#' # Or use external data
#' data = task$data()
#' sampler$sample_newdata("x1", newdata = data)
MarginalPermutationSampler = R6Class(
  "MarginalPermutationSampler",
  inherit = MarginalSampler,
  public = list(
    #' @description
    #' Creates a new instance of the MarginalPermutationSampler class.
    #' @param task ([mlr3::Task]) Task to sample from.
    initialize = function(task) {
      super$initialize(task)
      self$label = "Permutation sampler"
    }
  ),

  private = list(
    # Implement marginal sampling via independent permutation.
    # CONTRACT: may mutate `data` by reference (the single-draw branch
    # permutes the feature column in place via `:=`). Callers MUST pass a
    # disposable table -- both entry points do: MarginalSampler$sample()
    # uses .get_task_data_by_row_id() -> task$data() (always a fresh
    # data.table) and $sample_newdata() takes data.table::copy(newdata). The in-place
    # branch is deliberate (avoids copying a potentially large table just
    # to shuffle one column); the draw-major branch must allocate anyway.
    .sample_marginal = function(data, feature, samples_per_row = 1L) {
      if (samples_per_row == 1L) {
        data[, (feature) := lapply(.SD, sample), .SDcols = feature]
        return(data[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)])
      }

      # Draw-major: stack `samples_per_row` independent permutations of the feature column(s).
      # Other columns (including the conditioning context / passthrough features) are
      # repeated positionally so that row alignment with `row_ids` is preserved.
      out = data[rep.int(seq_len(.N), times = samples_per_row)]

      # One `:=` assigns all target columns at once; the inner `lapply` builds one stacked
      # vector per feature (`samples_per_row` independent permutations concatenated).
      out[,
        (feature) := lapply(feature, function(feat) {
          unlist(
            lapply(seq_len(samples_per_row), function(d) sample(data[[feat]])),
            use.names = FALSE
          )
        })
      ]

      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)

# Importance Aggregation Methods ----
# These functions provide different methods for aggregating importance scores
# across resampling iterations and computing confidence intervals.

# Multiplicity Adjustment Helpers ----

#' Compute adjusted alpha for CI construction
#'
#' Only Bonferroni has a clean per-comparison alpha (alpha/k) for CI coverage.
#' For all other methods, CIs use the nominal alpha unchanged.
#'
#' @param alpha nominal significance level (1 - conf_level)
#' @param p_adjust p-value adjustment method name
#' @param k number of comparisons (features)
#' @return adjusted alpha for CI construction
#' @noRd
adjust_ci_alpha = function(alpha, p_adjust, k) {
  if (p_adjust == "bonferroni" && k > 1) {
    alpha / k
  } else {
    alpha
  }
}

#' Apply p-value adjustment to importance results
#'
#' @param dt data.table with a p.value column
#' @param p_adjust p-value adjustment method name
#' @return dt (modified in place) with adjusted p-values
#' @noRd
adjust_pvalues = function(dt, p_adjust) {
  p.value = NULL
  if (p_adjust != "none" && "p.value" %in% names(dt)) {
    dt[, p.value := stats::p.adjust(p.value, method = p_adjust)]
  }
  dt
}

#' Warn if resampling has more than one iteration
#'
#' Observation-wise inference methods (e.g. CPI) were validated with a single
#' train/test split. Multiple iterations introduce overlapping training data
#' (CV) or non-i.i.d. test observations (bootstrap/subsampling).
#'
#' @param resampling mlr3 Resampling object
#' @noRd
check_single_resampling_iter = function(resampling) {
  if (resampling$iters > 1) {
    cli::cli_warn(c(
      "Observation-wise inference was validated with a single test set.",
      "!" = "Current resampling has {.val {resampling$iters}} iterations.",
      "i" = "With cross-validation, models are fit on overlapping training data, which may affect coverage.",
      "i" = "With bootstrap or subsampling, test observations are not i.i.d.",
      "i" = "See {.code vignette(\"inference\", package = \"xplainfi\")} for details."
    ))
  }
}

#' Warn if any test observation appears in more than one resampling test set
#'
#' Lei et al. inference assumes unique test observations.
#' The iter_repeat dimension (refit/sampling repeats) is averaged over and
#' irrelevant here — what matters is whether a row_id appears in multiple
#' resampling iterations (iter_rsmp), e.g. with subsampling.
#'
#' @param obs_loss_data data.table with columns row_ids, iter_rsmp
#' @noRd
check_unique_test_obs = function(obs_loss_data) {
  N = row_ids = iter_rsmp = NULL
  obs_per_rsmp = unique(obs_loss_data[, list(row_ids, iter_rsmp)])
  dupes = obs_per_rsmp[, .N, by = "row_ids"][N > 1]

  if (nrow(dupes) >= 1) {
    cli::cli_warn(c(
      "Found {.val {nrow(dupes)}} observation{?s} appearing in multiple test sets.",
      x = "Inference assumes unique test observations; duplicates may invalidate inference.",
      i = "Use a resampling strategy where each observation appears at most once in the test set(s)."
    ))
  }
}

#' No variance estimation - just aggregated performance
#' @param scores ([data.table::data.table]) with feature and importance columns
#' @param aggregator (`function`) to aggregate importance scores
#' @param conf_level (`numeric(1)`) ignored for this method
#' @noRd
importance_none = function(scores, aggregator, conf_level) {
  # The data.table NSE tax
  importance = NULL

  agg_importance = scores[,
    list(importance = aggregator(importance)),
    by = "feature"
  ]
  agg_importance
}

#' Raw variance estimation without correction
#'
#' Standard t-test on resampling scores. CIs are too narrow due to non-independence.
#' SE = sd(scores) / sqrt(n), t-statistic with df = n - 1.
#'
#' @param scores data.table with feature, importance, iter_rsmp columns
#' @param aggregator function to aggregate importance scores
#' @param conf_level confidence level for intervals
#' @param alternative "greater" (one-sided) or "two.sided"
#' @param n_iters number of resampling iterations
#' @noRd
importance_raw = function(scores, aggregator, conf_level, alternative, n_iters, p_adjust = "none") {
  # The data.table NSE tax
  importance = se = statistic = p.value = NULL
  agg_importance = scores[,
    list(importance = aggregator(importance)),
    by = "feature"
  ]

  # Aggregate within resamplings first to get one row per resampling iter
  means_rsmp = scores[,
    list(importance = mean(importance)),
    by = c("iter_rsmp", "feature")
  ]

  sds = means_rsmp[,
    list(se = sqrt(stats::var(importance) / n_iters)),
    by = "feature"
  ]

  agg_importance = agg_importance[sds, on = "feature"]

  # t-test: H0: importance = 0
  df = n_iters - 1
  agg_importance[, statistic := importance / se]

  k = nrow(agg_importance)
  ci_alpha = adjust_ci_alpha(1 - conf_level, p_adjust, k)

  if (alternative == "greater") {
    agg_importance[, p.value := stats::pt(statistic, df = df, lower.tail = FALSE)]
    quant = stats::qt(1 - ci_alpha, df = df)
    agg_importance[, let(
      conf_lower = importance - quant * se,
      conf_upper = Inf
    )]
  } else {
    agg_importance[, p.value := 2 * stats::pt(abs(statistic), df = df, lower.tail = FALSE)]
    quant = stats::qt(1 - ci_alpha / 2, df = df)
    agg_importance[, let(
      conf_lower = importance - quant * se,
      conf_upper = importance + quant * se
    )]
  }

  adjust_pvalues(agg_importance, p_adjust)
}

#' Nadeau & Bengio (2003) corrected variance estimation
#'
#' Corrected t-test accounting for correlation between resampling iterations.
#' Correction factor: (1/n_iters + n_test/n_train). SE = sqrt(factor * var(scores)).
#'
#' @param scores data.table with feature, importance, iter_rsmp columns
#' @param aggregator function to aggregate importance scores
#' @param conf_level confidence level for intervals
#' @param alternative "greater" (one-sided) or "two.sided"
#' @param resampling mlr3 Resampling object (for test/train ratio)
#' @param n_iters number of resampling iterations
#' @noRd
importance_nadeau_bengio = function(
  scores,
  aggregator,
  conf_level,
  alternative,
  resampling,
  n_iters,
  p_adjust = "none"
) {
  # The data.table NSE tax
  importance = se = statistic = p.value = NULL

  # Validate resampling type
  if (!(resampling$id %in% c("bootstrap", "subsampling")) | resampling$iters < 10) {
    cli::cli_warn(c(
      "{.cls Resampling} is of type {.val {resampling$id}} with {.val {resampling$iters}} iterations.",
      i = "The Nadeau & Bengio corrected t-test is recommended for resampling types {.val {c('bootstrap', 'subsampling')}} with >= 10 iterations for valid inference"
    ))
  }

  if (resampling$id == "bootstrap") {
    test_train_ratio = 0.632
  } else {
    # Average test/train ration over iterations for simplicity as it works for any resampling strategy
    test_train_ratios = vapply(
      seq_len(resampling$iters),
      \(idx) {
        length(resampling$test_set(idx)) / length(resampling$train_set(idx))
      },
      FUN.VALUE = numeric(1)
    )
    test_train_ratio = mean(test_train_ratios)
  }

  agg_importance = scores[,
    list(importance = aggregator(importance)),
    by = "feature"
  ]

  # Aggregate within resamplings first
  means_rsmp = scores[,
    list(importance = aggregator(importance)),
    by = c("iter_rsmp", "feature")
  ]

  # Nadeau & Bengio adjustment factor ((1 / m) + c in Molnar et al.)
  adjustment_factor = 1 / n_iters + test_train_ratio

  sds = means_rsmp[,
    list(se = sqrt(adjustment_factor * stats::var(importance))),
    by = "feature"
  ]

  agg_importance = agg_importance[sds, on = "feature"]

  # t-test: H0: importance = 0, se is corrected by Nadeau & Bengio factor
  df = n_iters - 1
  agg_importance[, statistic := importance / se]

  # Bonferroni uses alpha / k, k = n_iters
  ci_alpha = adjust_ci_alpha(1 - conf_level, p_adjust, k = n_iters)

  if (alternative == "greater") {
    agg_importance[, p.value := stats::pt(statistic, df = df, lower.tail = FALSE)]
    quant = stats::qt(1 - ci_alpha, df = df)
    agg_importance[, let(
      conf_lower = importance - quant * se,
      conf_upper = Inf
    )]
  } else {
    agg_importance[, p.value := 2 * stats::pt(abs(statistic), df = df, lower.tail = FALSE)]
    quant = stats::qt(1 - ci_alpha / 2, df = df)
    agg_importance[, let(
      conf_lower = importance - quant * se,
      conf_upper = importance + quant * se
    )]
  }

  adjust_pvalues(agg_importance, p_adjust)
}

#' Empirical quantile-based confidence intervals
#'
#' Non-parametric inference from resampling distribution. CIs from empirical quantiles.
#' Does not provide p-values or test statistics as these would require a different
#' methodology than the quantile-based CIs.
#'
#' @param scores data.table with feature, importance, iter_rsmp columns
#' @param aggregator function to aggregate importance scores
#' @param conf_level confidence level for intervals
#' @param alternative "greater" (one-sided) or "two.sided"
#' @noRd
importance_quantile = function(scores, aggregator, conf_level, alternative) {
  # The data.table NSE tax
  importance = feature = NULL

  # Aggregate within resamplings first to get one value per resampling iteration
  means_rsmp = scores[,
    list(importance = aggregator(importance)),
    by = c("iter_rsmp", "feature")
  ]

  # For each feature, compute quantiles
  result_list = lapply(unique(means_rsmp$feature), function(feat) {
    feat_scores = means_rsmp[feature == feat, importance]

    # Point estimate using aggregator
    point_est = aggregator(feat_scores)

    # Compute empirical quantiles for CI
    alpha = 1 - conf_level
    if (alternative == "greater") {
      probs = alpha
      ci_lower = quantile(feat_scores, probs = probs, na.rm = TRUE)
      ci_upper = Inf
    } else {
      probs = c(alpha / 2, 1 - alpha / 2)
      ci_vals = quantile(feat_scores, probs = probs, na.rm = TRUE)
      ci_lower = ci_vals[1]
      ci_upper = ci_vals[2]
    }

    data.table(
      feature = feat,
      importance = point_est,
      conf_lower = ci_lower,
      conf_upper = ci_upper
    )
  })

  rbindlist(result_list)
}

#' Per-feature hypothesis tests on observation-wise importance scores
#'
#' Shared logic for CPI and Lei et al. inference. Runs a statistical test on the
#' observation-wise loss differences for each feature, computes p-values and CIs,
#' and applies multiplicity correction.
#'
#' SE is only reported for the t-test; for other tests it is `NA`.
#'
#' @param obs_loss_agg data.table with columns `feature`, `row_ids`, `obs_importance`
#'   (one row per observation per feature, already aggregated over repeats)
#' @param test "t", "wilcoxon", "fisher", or "binomial"
#' @param alternative "greater" or "two.sided"
#' @param conf_level confidence level for CI
#' @param p_adjust p-value adjustment method (any of stats::p.adjust.methods)
#' @param aggregator function to compute point estimate (default: mean)
#' @return data.table with feature, importance, se, statistic, p.value, conf_lower, conf_upper
#' @noRd
test_obs_importance = function(
  obs_loss_agg,
  test,
  alternative,
  conf_level,
  p_adjust,
  aggregator = mean
) {
  # The data.table NSE tax
  obs_importance = feature = NULL

  test_function = switch(
    test,
    t = stats::t.test,
    wilcoxon = stats::wilcox.test,
    fisher = fisher_test,
    binomial = binom_test
  )

  features_tested = unique(obs_loss_agg$feature)
  k = length(features_tested)
  ci_conf_level = 1 - adjust_ci_alpha(1 - conf_level, p_adjust, k)

  # SE is only meaningful for the t-test (sd / sqrt(n)).
  # For other tests (wilcoxon, fisher, binomial), SE is not well-defined.
  report_se = test == "t"

  result_list = lapply(features_tested, function(feat) {
    feat_obs = obs_loss_agg[feature == feat, obs_importance]

    if (mean(feat_obs) == 0) {
      if (alternative == "greater") {
        htest_result = list(
          statistic = 0,
          p.value = 1,
          conf.int = c(0, Inf)
        )
      } else {
        htest_result = list(
          statistic = 0,
          p.value = 1,
          conf.int = c(0, 0)
        )
      }
    } else {
      # R-devel (2026-05) rewrote stats:::wilcox.test internals; the
      # new asymptotic CI path errors with
      # "missing value where TRUE/FALSE needed" when the variance
      # estimator returns NA (many-ties / many-zeros inputs that the
      # previous implementation tolerated). Defensive fallback returns
      # a degenerate-shape htest so downstream code keeps going with
      # NA inference fields. Affects R-release runs only if R-devel
      # bug ships; safe no-op on stable R versions.
      htest_result = tryCatch(
        test_function(
          feat_obs,
          alternative = alternative,
          conf.level = ci_conf_level,
          conf.int = TRUE
        ),
        error = function(e) {
          cli::cli_warn(c(
            "Hypothesis test failed for feature {.val {feat}}; returning NA inference fields.",
            "i" = "{conditionMessage(e)}"
          ))
          list(
            statistic = NA_real_,
            p.value = NA_real_,
            conf.int = if (alternative == "greater") {
              c(NA_real_, Inf)
            } else {
              c(NA_real_, NA_real_)
            }
          )
        }
      )
    }

    data.table(
      feature = feat,
      importance = aggregator(feat_obs),
      se = if (report_se) sd(feat_obs) / sqrt(length(feat_obs)) else NA_real_,
      statistic = htest_result$statistic,
      p.value = htest_result$p.value,
      conf_lower = htest_result$conf.int[1],
      conf_upper = htest_result$conf.int[2]
    )
  })

  result = rbindlist(result_list, fill = TRUE)
  adjust_pvalues(result, p_adjust)
}

#' Conditional Predictive Impact (CPI) test
#'
#' Tests on observation-wise loss differences for CFI with knockoff samplers.
#' Available tests: t-test, Wilcoxon, Fisher permutation, binomial.
#' Fisher test uses Phipson & Smyth (2010) p-value correction.
#'
#' @param conf_level confidence level for CI
#' @param alternative "greater" (one-sided) or "two.sided"
#' @param test "t", "wilcoxon", "fisher", or "binomial"
#' @param p_adjust p-value adjustment method (any of stats::p.adjust.methods)
#' @param B number of replications for Fisher test
#' @param method_obj importance method object (needs $obs_loss())
#' @noRd
importance_cpi = function(
  conf_level,
  alternative = c("two.sided", "greater"),
  test = c("t", "wilcoxon", "fisher", "binomial"),
  p_adjust = "none",
  B = 1999,
  method_obj
) {
  # The data.table NSE tax
  N = obs_importance = feature = NULL

  alternative = match.arg(alternative)

  check_single_resampling_iter(method_obj$resampling)

  # CPI requires observation-wise losses
  if (is.null(method_obj$obs_loss())) {
    cli::cli_abort(c(
      "CPI requires observation-wise losses.",
      i = "Ensure {.code measure} has an {.fun $obs_loss} method."
    ))
  }
  if (class(method_obj)[[1]] != "CFI") {
    cli::cli_warn(c(
      "!" = "CPI is only known to yield valid inference for {.cls CFI}.",
      x = "Inference with {.cls PFI} is known to be invalid and other methods are not studied yet."
    ))
  }
  # Get observation-wise importances (already computed as differences)
  obs_loss_data = method_obj$obs_loss(relation = "difference")
  # Aggregate over n_repeats to get one value per observation per feature
  obs_loss_agg = obs_loss_data[,
    list(obs_importance = mean(obs_importance)),
    by = c("row_ids", "feature")
  ]

  test = match.arg(test)
  test_obs_importance(obs_loss_agg, test, alternative, conf_level, p_adjust)
}

#' Distribution-free inference for refit-based importance (Lei et al., 2018)
#'
#' Tests on observation-wise loss differences for WVIM/LOCO.
#' Available tests: Wilcoxon signed-rank (default), t-test, Fisher permutation, binomial.
#'
#' Lei et al. (2018) proposed this method specifically for LOCO with:
#' - L1 (absolute) loss as the measure
#' - Median as the aggregation function
#' - A single train/test split (holdout), since inference is conditional on
#'   the training data and requires i.i.d. test observations from that split
#'
#' While the implementation generalizes these choices (allowing other measures,
#' aggregators, and resampling strategies), deviating from the original proposal
#' may invalidate the theoretical guarantees.
#'
#' @param conf_level confidence level for CI
#' @param alternative "greater" (one-sided) or "two.sided"
#' @param test "wilcoxon", "t", "fisher", or "binomial"
#' @param aggregator function to compute point estimate from obs_importance (default: median)
#' @param p_adjust p-value adjustment method (any of stats::p.adjust.methods)
#' @param B number of replications for Fisher test
#' @param method_obj importance method object (needs $obs_loss())
#' @noRd
importance_loco = function(
  conf_level,
  alternative = c("two.sided", "greater"),
  test = c("wilcoxon", "t", "fisher", "binomial"),
  aggregator = stats::median,
  p_adjust = "none",
  B = 1999,
  method_obj
) {
  # The data.table NSE tax
  N = obs_importance = feature = NULL

  alternative = match.arg(alternative)

  # Lei et al. inference requires observation-wise losses
  if (is.null(method_obj$obs_loss())) {
    cli::cli_abort(c(
      "Lei et al. (2018) inference requires observation-wise losses.",
      i = "Ensure {.code measure} has an {.fun $obs_loss} method (i.e. is decomposable)."
    ))
  }

  # Get observation-wise importances (already computed as differences)
  obs_loss_data = method_obj$obs_loss(relation = "difference")

  check_unique_test_obs(obs_loss_data)

  # Aggregate over n_repeats (iter_repeat) to get one value per observation per feature
  obs_loss_agg = obs_loss_data[,
    list(obs_importance = mean(obs_importance)),
    by = c("row_ids", "feature")
  ]

  test = match.arg(test)
  test_obs_importance(obs_loss_agg, test, alternative, conf_level, p_adjust, aggregator)
}

# Score Relation Helpers ----

#' Ratio of two score vectors with a zero-denominator guard
#'
#' Used by `FeatureImportanceMethod$.compute_score()` for `relation = "ratio"`.
#' A zero denominator (zero baseline score) makes the ratio `Inf` (x/0) or
#' `NaN` (0/0). Those elements are set to `NA_real_` and a single warning is
#' emitted. Affected features then aggregate to `NA` under the default `mean`
#' aggregator -- intentional: an undefined ratio should surface, not propagate
#' as `Inf`/`NaN`. `NA` denominators already yield `NA` from the division and
#' do not trigger the zero-baseline warning.
#'
#' Kept as a standalone internal function (not an R6 private method) to keep
#' [FeatureImportanceMethod] objects lean for serialization.
#'
#' @param numerator,denominator (`numeric()`) Equal-length score vectors.
#' @return (`numeric()`) `numerator / denominator` with zero-denominator
#'   elements replaced by `NA_real_`.
#' @keywords internal
#' @noRd
safe_ratio = function(numerator, denominator) {
  out = numerator / denominator
  zero_denom = denominator == 0
  if (any(zero_denom, na.rm = TRUE)) {
    n_bad = sum(zero_denom, na.rm = TRUE)
    cli::cli_warn(c(
      "!" = "{.val ratio} relation undefined for {n_bad} score{?s} with a zero baseline.",
      "i" = "Setting {n_bad} result{?s} to {.val NA}; affected feature{?s} aggregate to {.val NA}."
    ))
    out[which(zero_denom)] = NA_real_
  }
  out
}

test_that(".compute_score guards zero baseline in ratio relation", {
  # regr.mse -> minimize = TRUE -> ratio = scores_post / scores_pre,
  # so scores_pre == 0 is the undefined (Inf / NaN) case.
  task = sim_dgp_independent(n = 60)
  pfi = PFI$new(
    task = task,
    learner = lrn("regr.rpart"),
    measure = msr("regr.mse")
  )
  cs = pfi$.__enclos_env__$private$.compute_score

  expect_warning(
    res <- cs(
      scores_pre = c(0, 1, 0),
      scores_post = c(1, 2, 0),
      relation = "ratio"
    ),
    regexp = "zero baseline"
  )
  # 1/0 -> NA, 2/1 -> 2, 0/0 -> NA
  checkmate::expect_numeric(res, len = 3L)
  expect_equal(res, c(NA_real_, 2, NA_real_))

  # difference relation must be unaffected: no NA, no warning
  expect_no_warning(
    diff_res <- cs(
      scores_pre = c(0, 1, 0),
      scores_post = c(1, 2, 0),
      relation = "difference"
    )
  )
  expect_equal(diff_res, c(1, 1, 0))
})

test_that(".compute_score guards zero baseline for minimize = FALSE (classif.acc)", {
  # classif.acc -> minimize = FALSE -> ratio = scores_pre / scores_post,
  # so scores_post == 0 is the undefined case.
  set.seed(531)
  task = tgen("2dnormals")$generate(n = 60)
  pfi = PFI$new(
    task = task,
    learner = lrn("classif.rpart", predict_type = "prob"),
    measure = msr("classif.acc")
  )
  cs = pfi$.__enclos_env__$private$.compute_score

  # minimize = FALSE: ratio = scores_pre / scores_post; scores_post == 0 is undefined
  # scores_pre = c(0, 1, 0), scores_post = c(1, 2, 0)
  # ratio = c(0/1, 1/2, 0/0) = c(0, 0.5, NaN) but scores_post[3] == 0 -> NA + warning
  expect_warning(
    res <- cs(
      scores_pre = c(0, 1, 0),
      scores_post = c(1, 2, 0),
      relation = "ratio"
    ),
    regexp = "zero baseline"
  )
  checkmate::expect_numeric(res, len = 3L)
  expect_equal(res, c(0, 0.5, NA_real_))
})

test_that("safe_ratio handles NA denominator without error", {
  # safe_ratio is a package-internal standalone function (not an R6 method),
  # kept off the object to keep FeatureImportanceMethod lean for serialization.

  # All-NA denominator: division produces NA, no zero-denom warning
  expect_no_error(r <- safe_ratio(1, NA_real_))
  expect_identical(r, NA_real_)

  # Mixed: one true zero, one NA denominator
  # zero denom -> warn + NA; NA denom -> NA via division (no extra warn)
  expect_warning(
    r2 <- safe_ratio(c(1, 2), c(0, NA_real_)),
    regexp = "zero baseline"
  )
  expect_equal(r2, c(NA_real_, NA_real_))
})

#' @title Marginal Sampler Base Class
#'
#' @description Abstract base class for marginal sampling strategies that do not
#' condition on other features. Marginal samplers sample from P(X_S), the marginal
#' distribution of features S, ignoring any dependencies with other features.
#'
#' @details
#' This class provides a common interface for different marginal sampling approaches:
#'
#' - **MarginalPermutationSampler**: Shuffles features independently within the dataset
#' - **MarginalReferenceSampler**: Samples complete rows from reference data
#'
#' Both approaches sample from the marginal distribution P(X_S), but differ in how
#' they preserve or break within-row dependencies:
#'
#' - Permutation breaks ALL dependencies (both with target and between features)
#' - Reference sampling preserves WITHIN-row dependencies but breaks dependencies with test data
#'
#' **Comparison with ConditionalSampler:**
#'
#' - `MarginalSampler`: Samples from \eqn{P(X_S)} - no conditioning
#' - `ConditionalSampler`: Samples from \eqn{P(X_S | X_{-S})}- conditions on other features
#'
#' This base class implements the public `$sample()` and `$sample_newdata()` methods,
#' delegating to private `.sample_marginal()` which subclasses must implement.
#'
#' @export
MarginalSampler = R6Class(
  "MarginalSampler",
  inherit = FeatureSampler,
  public = list(
    #' @description
    #' Sample features from their marginal distribution.
    #' @param feature (`character()`) Feature name(s) to sample.
    #' @param row_ids (`integer()` | `NULL`) Row IDs from task to use.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @return Modified copy with sampled feature(s).
    sample = function(feature, row_ids = NULL, samples_per_row = 1L) {
      checkmate::assert_count(samples_per_row, positive = TRUE)
      data_copy = private$.get_task_data_by_row_id(row_ids)
      private$.sample_marginal(data_copy, feature, samples_per_row = samples_per_row)
    },

    #' @description
    #' Sample from external data.
    #' @param feature (`character()`) Feature(s) to sample.
    #' @param newdata ([`data.table`][data.table::data.table]) External data to use.
    #' @param samples_per_row (`integer(1)`: `1L`) Number of independent samples per input row.
    #'   See [FeatureSampler]`$sample()` for output shape and ordering.
    #' @return Modified copy with sampled feature(s).
    sample_newdata = function(feature, newdata, samples_per_row = 1L) {
      checkmate::assert_count(samples_per_row, positive = TRUE)
      if (inherits(newdata, "data.table")) {
        data_copy = data.table::copy(newdata)
      } else {
        data_copy = as.data.table(newdata)
      }
      private$.sample_marginal(data_copy, feature, samples_per_row = samples_per_row)
    }
  ),

  private = list(
    # Abstract method for marginal sampling. Subclasses must implement it.
    # Implementations may mutate `data` by reference; callers MUST pass a
    # disposable table (this class's sample()/sample_newdata() guarantee it).
    .sample_marginal = function(data, feature, samples_per_row = 1L) {
      cli::cli_abort(c(
        "Abstract method",
        "i" = "Subclasses must implement the {.fn .sample_marginal} method"
      ))
    }
  )
)

#' @title Gaussian Conditional Sampler
#'
#' @description Implements conditional sampling assuming features follow a multivariate
#' Gaussian distribution. Computes conditional distributions analytically using standard
#' formulas for multivariate normal distributions.
#'
#' @details
#' For a joint Gaussian distribution \eqn{X \sim N(\mu, \Sigma)}, partitioned as
#' \eqn{X = (X_A, X_B)}, the conditional distribution is:
#'
#' \deqn{X_B | X_A = x_A \sim N(\mu_{B|A}, \Sigma_{B|A})}
#'
#' where:
#' \deqn{\mu_{B|A} = \mu_B + \Sigma_{BA} \Sigma_{AA}^{-1} (x_A - \mu_A)}
#' \deqn{\Sigma_{B|A} = \Sigma_{BB} - \Sigma_{BA} \Sigma_{AA}^{-1} \Sigma_{AB}}
#'
#' This is equivalent to the regression formulation used by fippy:
#' \deqn{\beta = \Sigma_{BA} \Sigma_{AA}^{-1}}
#' \deqn{\mu_{B|A} = \mu_B + \beta (x_A - \mu_A)}
#' \deqn{\Sigma_{B|A} = \Sigma_{BB} - \beta \Sigma_{AB}}
#'
#' **Assumptions:**
#' - Features are approximately multivariate normal
#' - Only continuous features are supported
#'
#' **Advantages:**
#' - Very fast (closed-form solution)
#' - Deterministic (given seed)
#' - No hyperparameters
#' - Memory efficient
#'
#' **Limitations:**
#' - Strong distributional assumption
#' - May produce out-of-range values for bounded features
#' - Cannot handle categorical features
#' - Integer features are treated as continuous and rounded back to integers
#'
#' @examples
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 100)
#' sampler = ConditionalGaussianSampler$new(task)
#'
#' # Sample x2, x3 conditioned on x1
#' test_data = task$data(rows = 1:5)
#' sampled = sampler$sample_newdata(
#'   feature = c("important2", "important3"),
#'   newdata = test_data,
#'   conditioning_set = "important1"
#' )
#'
#' @references `r print_bib("anderson_2003")`
#'
#' @export
ConditionalGaussianSampler = R6Class(
  "ConditionalGaussianSampler",
  inherit = ConditionalSampler,
  public = list(
    #' @field feature_types (`character()`) Feature types supported by the sampler.
    feature_types = c("numeric", "integer"),

    #' @field mu (`numeric()`) Mean vector estimated from training data.
    mu = NULL,

    #' @field sigma (`matrix()`) Covariance matrix estimated from training data.
    sigma = NULL,

    #' @description
    #' Creates a new ConditionalGaussianSampler.
    #' @param task ([mlr3::Task]) Task to sample from. Must have only numeric/integer features.
    #' @param conditioning_set (`character` | `NULL`) Default conditioning set to use in `$sample()`.
    initialize = function(task, conditioning_set = NULL) {
      super$initialize(task, conditioning_set = conditioning_set)

      # Extract feature data as matrix
      X = as.matrix(self$task$data(cols = self$task$feature_names))

      # Estimate mean and covariance
      self$mu = colMeans(X)
      self$sigma = stats::cov(X)

      self$label = "Gaussian Conditional Sampler"
    }
  ),

  private = list(
    # Core sampling logic implementing conditional Gaussian sampling
    .sample_conditional = function(data, feature, conditioning_set, samples_per_row = 1L, ...) {
      # Marginal case (no conditioning): identical mean for every row, just stack draws.
      if (length(conditioning_set) == 0) {
        mu_feature = self$mu[feature]
        sigma_feature = self$sigma[feature, feature, drop = FALSE]
        n = nrow(data)

        out = data[rep.int(seq_len(.N), times = samples_per_row)]
        samples = mvtnorm::rmvnorm(n * samples_per_row, mean = mu_feature, sigma = sigma_feature)
        out[, (feature) := as.data.table(samples)]
        out = private$.ensure_feature_types(out)
        return(out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)])
      }

      # Conditional case
      training_data = self$task$data(cols = self$task$feature_names)
      X_cond = as.matrix(training_data[, .SD, .SDcols = conditioning_set])
      X_cond = cbind(1, X_cond)

      n_features = length(feature)
      coef_list = vector("list", n_features)
      residuals_mat = matrix(0, nrow = nrow(training_data), ncol = n_features)
      for (j in seq_along(feature)) {
        y = training_data[[feature[j]]]
        fit = lm.fit(X_cond, y)
        coef_list[[j]] = fit$coefficients
        residuals_mat[, j] = fit$residuals
      }
      cond_cov = stats::cov(residuals_mat)
      if (n_features == 1) {
        cond_cov = matrix(cond_cov, 1, 1)
      }

      X_new = as.matrix(data[, .SD, .SDcols = conditioning_set])
      X_new = cbind(1, X_new)

      pred_mat = matrix(0, nrow = nrow(data), ncol = n_features)
      for (j in seq_along(feature)) {
        pred_mat[, j] = X_new %*% coef_list[[j]]
      }

      n = nrow(data)
      out = data[rep.int(seq_len(.N), times = samples_per_row)]

      # `samples_per_row` independent residual draws across all evidence rows, stacked in draw-major.
      # Means are repeated to align with the replicated evidence rows.
      noise = mvtnorm::rmvnorm(n * samples_per_row, mean = rep(0, n_features), sigma = cond_cov)
      means_replicated = pred_mat[rep.int(seq_len(n), times = samples_per_row), , drop = FALSE]
      samples = noise + means_replicated

      out[, (feature) := as.data.table(samples)]
      out = private$.ensure_feature_types(out)

      out[, .SD, .SDcols = c(self$task$target_names, self$task$feature_names)]
    }
  )
)

#' Fisher permutation test wrapper
#' Based on cpi pkg, extended to support two-sided tests
#' @param x numeric vector of observations
#' @param B number of permutations
#' @param alternative "greater" or "two.sided"
#' @param conf.level confidence level
#' @param ... ignored
#' @noRd
fisher_test = function(
  x,
  B = 1999,
  alternative = c("two.sided", "greater"),
  conf.level = 0.95,
  ...
) {
  alternative = match.arg(alternative)
  orig_mean = mean(x)

  # B permutations
  perm_means = replicate(B, {
    signs = sample(c(-1, 1), length(x), replace = TRUE)
    mean(signs * x)
  })

  if (alternative == "greater") {
    p.value = (sum(perm_means >= orig_mean) + 1) / (B + 1)
    conf.int = c(orig_mean - quantile(perm_means, conf.level), Inf)
  } else {
    # two.sided
    p.value = (sum(abs(perm_means) >= abs(orig_mean)) + 1) / (B + 1)
    alpha = 1 - conf.level
    conf.int = orig_mean - quantile(perm_means, c(1 - alpha / 2, alpha / 2))
  }

  list(
    statistic = orig_mean,
    p.value = p.value,
    conf.int = conf.int
  )
}

#' Binomial sign-test wrapper
#' Based on cpi pkg, extended to support two-sided tests
#' @param x numeric vector of observations
#' @param alternative "greater" or "two.sided"
#' @param conf.level confidence level
#' @param ... ignored
#' @noRd
binom_test = function(x, alternative = c("two.sided", "greater"), conf.level = 0.95, ...) {
  alternative = match.arg(alternative)
  stats::binom.test(sum(x > 0), length(x), alternative = alternative, conf.level = conf.level)
}

#' @title Perturbation Feature Importance Base Class
#'
#' @description Abstract base class for perturbation-based importance methods PFI, CFI, and RFI
#'
#' @export
PerturbationImportance = R6Class(
  "PerturbationImportance",
  inherit = FeatureImportanceMethod, # Inherit from existing base class
  public = list(
    #' @field sampler ([FeatureSampler]) Sampler object for feature perturbation
    sampler = NULL,

    #' @description
    #' Creates a new instance of the PerturbationImportance class
    #' @param task,learner,measure,resampling,features,groups Passed to [FeatureImportanceMethod].
    #' @param sampler ([FeatureSampler]) Sampler to use for feature perturbation.
    #' @param relation (`character(1)`: `"difference"`) How to relate perturbed and baseline scores. Can also be `"ratio"`.
    #' @param n_repeats (`integer(1)`: `30L`) Number of permutation/conditional sampling iterations. Can also be overridden in `$compute()`.
    #' @param batch_size (`integer(1)` | `NULL`: `NULL`) Maximum number of rows to predict at once. When `NULL`, predicts all `test_size * n_repeats` rows in one call. Use smaller values to reduce memory usage at the cost of more prediction calls. Can be overridden in `$compute()`.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      sampler = NULL,
      relation = "difference",
      n_repeats = 30L,
      batch_size = NULL
    ) {
      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        groups = groups,
        label = "Feature Importance (Abstract Class)"
      )

      # If no sampler is provided, create a default one (implementation dependent)
      self$sampler = sampler

      # Set up common parameters for all perturbation-based methods
      ps = paradox::ps(
        relation = paradox::p_fct(c("difference", "ratio"), default = "difference"),
        n_repeats = paradox::p_int(lower = 1, default = 1),
        batch_size = paradox::p_int(lower = 1, special_vals = list(NULL), default = NULL)
      )

      ps$values$relation = relation
      ps$values$n_repeats = n_repeats
      ps$values$batch_size = batch_size
      self$param_set = ps

      # Add CPI to variance methods registry
      private$.ci_methods = c(private$.ci_methods, "cpi")
    },

    #' @description
    #' Get aggregated importance scores.
    #' Extends the base `$importance()` method to support `ci_method = "cpi"`.
    #' For details, see [CFI], which is the only sub-method for which it is known to be valid.
    #' @param relation (`character(1)`) How to relate perturbed scores to originals ("difference" or "ratio"). If `NULL`, uses stored parameter value.
    #' @param standardize (`logical(1)`: `FALSE`) If `TRUE`, importances are standardized by the highest score so all scores fall in `[-1, 1]`.
    #' @param ci_method (`character(1)`: `"none"`) Variance estimation method. In addition to base methods (`"none"`, `"raw"`, `"nadeau_bengio"`, `"quantile"`),
    #'   perturbation methods support `"cpi"` (Conditional Predictive Impact).
    #'   CPI is specifically designed for [CFI] with knockoff samplers and uses one-sided hypothesis tests.
    #' @param conf_level (`numeric(1)`: `0.95`) Confidence level for confidence intervals when `ci_method != "none"`.
    #' @param alternative (`character(1)`: `"two.sided"`) Type of alternative hypothesis for statistical tests.
    #'   `"greater"` tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
    #'   `"two.sided"` tests H0: importance = 0 vs H1: importance != 0.
    #' @param test (`character(1)`: `"t"`) Test to use for CPI. One of `"t"`, `"wilcoxon"`, `"fisher"`, or `"binomial"`. Only used when `ci_method = "cpi"`.
    #' @param B (`integer(1)`: `1999`) Number of replications for Fisher test. Only used when `ci_method = "cpi"` and `test = "fisher"`.
    #' @param p_adjust (`character(1)`: `"none"`) Method for p-value adjustment for multiple comparisons.
    #'   Accepts any method supported by [stats::p.adjust.methods], e.g. `"holm"`, `"bonferroni"`, `"BH"`, `"none"`.
    #'   When `"bonferroni"`, confidence intervals are also adjusted (alpha/k).
    #'   For other correction methods (e.g. `"holm"`, `"BH"`), only p-values are adjusted;
    #'   confidence intervals remain at the nominal `conf_level` because these sequential/adaptive
    #'   procedures do not have a clean per-comparison alpha for CI construction.
    #' @param ... Additional arguments passed to the base method.
    #' @return ([data.table][data.table::data.table]) Aggregated importance scores.
    importance = function(
      relation = NULL,
      standardize = FALSE,
      ci_method = c("none", "raw", "nadeau_bengio", "quantile", "cpi"),
      conf_level = 0.95,
      alternative = c("two.sided", "greater"),
      test = c("t", "wilcoxon", "fisher", "binomial"),
      B = 1999,
      p_adjust = "none",
      ...
    ) {
      # Handle CPI separately, delegate rest to parent
      if (length(ci_method) > 1) {
        ci_method = ci_method[1]
      }
      alternative = match.arg(alternative)

      if (ci_method == "cpi") {
        # CPI requires special handling
        if (is.null(private$.scores)) {
          cli::cli_inform(c(
            x = "No importances computed yet!"
          ))
          return(invisible(NULL))
        }

        checkmate::assert_number(conf_level, lower = 0, upper = 1)

        # CPI does not support standardization - it uses obs-wise losses for inference
        if (standardize) {
          cli::cli_warn(c(
            "!" = "Standardization is not supported for CPI.",
            "i" = "CPI uses observation-wise losses for statistical inference.",
            "i" = "Ignoring {.code standardize = TRUE}."
          ))
        }

        # Call CPI function
        test = match.arg(test)
        agg_importance = importance_cpi(
          conf_level = conf_level,
          alternative = alternative,
          test = test,
          p_adjust = p_adjust,
          B = B,
          method_obj = self
        )

        setkeyv(agg_importance, "feature")
        return(agg_importance[])
      } else {
        # Delegate to parent for other methods
        super$importance(
          relation = relation,
          standardize = standardize,
          ci_method = ci_method,
          conf_level = conf_level,
          alternative = alternative,
          p_adjust = p_adjust,
          ...
        )
      }
    }
  ),

  private = list(
    .compute_baseline = function(store_backends = TRUE) {
      self$resample_result = assemble_rr(
        task = self$task,
        learner = self$learner,
        resampling = self$resampling,
        store_models = TRUE,
        store_backends = store_backends
      )
      # Prepare baseline scores
      scores_baseline = self$resample_result$score(self$measure)[,
        .SD,
        .SDcols = c("iteration", self$measure$id)
      ]
      setnames(scores_baseline, old = self$measure$id, "score_baseline")
      setnames(scores_baseline, old = "iteration", "iter_rsmp")
      scores_baseline[]
    },

    # Common computation method for all perturbation-based methods
    .compute_perturbation_importance = function(
      n_repeats = NULL,
      batch_size = NULL,
      store_models = TRUE,
      store_backends = TRUE,
      sampler = NULL
    ) {
      # Use provided sampler or default to self$sampler
      sampler = sampler %||% self$sampler

      n_repeats = resolve_param(n_repeats, self$param_set$values$n_repeats, 1L)
      batch_size = resolve_param(batch_size, self$param_set$values$batch_size, NULL)

      scores_baseline = private$.compute_baseline(store_backends = store_backends)

      # Get predictions for each resampling iter, permutation iter, feature
      # Create progress bar that tracks resampling_iter * feature/group combinations
      # if (xplain_opt("progress")) {
      # n_features_or_groups = length(self$groups %||% self$features)
      # total_iterations = self$resampling$iters * n_features_or_groups
      # progress_bar_id = cli::cli_progress_bar(
      # 	"Computing importances",
      # 	total = total_iterations
      # )
      # }

      all_preds = lapply(seq_len(self$resampling$iters), \(iter) {
        # Extract the learner here once because apparently reassembly is expensive
        this_learner = self$resample_result$learners[[iter]]
        test_row_ids = self$resampling$test_set(iter)
        test_size = length(test_row_ids)

        if (is.null(self$groups)) {
          iteration_proxy = self$features
          # name so lapply returns named list, used as idcol in rbindlist()
          names(iteration_proxy) = iteration_proxy
        } else {
          iteration_proxy = self$groups
        }

        # Use unified parallelization helper
        pred_per_feature = xplainfi_map(
          length(iteration_proxy),
          \(
            foi,
            task,
            learner,
            sampler,
            test_row_ids,
            n_repeats,
            batch_size,
            learner_packages,
            is_sequential = TRUE
          ) {
            # Load required packages in parallel workers
            if (!is_sequential) {
              library("data.table")
              library("mlr3")
              library("xplainfi")
              for (pkg in learner_packages) {
                library(pkg, character.only = TRUE)
              }
              # Force sequential forging inside the daemon. The outer mirai layer
              # already parallelizes across features, so letting the sampler's
              # stored `parallel` flag drive nested `arf::forge()` parallelism
              # here would only oversubscribe cores. That flag reflects fit-time
              # config (`adversarial_rf`/`forde`, done once in the caller) and
              # must not leak into sample-time forging. Set it on the daemon-local
              # sampler copy, samplers without a `parallel` param (all non-ARF
              # samplers) are unaffected.
              if ("parallel" %in% sampler$param_set$ids()) {
                sampler$param_set$set_values(parallel = FALSE)
              }
            }

            # Sampler produces `samples_per_row * test_size` rows in draw-major order;
            # see `FeatureSampler$sample()` for the contract.
            test_size = length(test_row_ids)
            perturbed_data = sampler$sample(
              foi,
              row_ids = test_row_ids,
              samples_per_row = n_repeats
            )

            # Split into n_repeats groups, each of test_size rows in test_row_ids order.
            perturbed_data_list = split(
              perturbed_data,
              rep(seq_len(n_repeats), each = test_size)
            )

            # Use batched prediction helper
            preds = predict_batched(
              learner = learner,
              data_list = perturbed_data_list,
              task = task,
              test_row_ids = test_row_ids,
              batch_size = batch_size
            )

            # Store predictions in data.table list column
            pred_per_perm = lapply(preds, \(pred) data.table::data.table(prediction = list(pred)))

            # Append iteration id for within-resampling permutations
            data.table::rbindlist(pred_per_perm, idcol = "iter_repeat")
          },
          iteration_proxy, # Varying argument
          .args = list(
            task = self$task,
            learner = this_learner,
            sampler = sampler,
            test_row_ids = test_row_ids,
            n_repeats = n_repeats,
            batch_size = batch_size,
            learner_packages = this_learner$packages
          )
        )

        # When groups are defined, "feature" is the group name
        # mild misnomer for convenience because if-else'ing the column name is annoying
        rbindlist(pred_per_feature, idcol = "feature")
      })
      # Append iteration id for resampling
      all_preds = rbindlist(all_preds, idcol = "iter_rsmp")
      # setkeyv(all_preds, cols = c("feature", "iter_rsmp"))

      # store predictions for future reference maybe?
      self$predictions = all_preds

      # Close progress bar
      # if (xplain_opt("progress")) {
      # 	cli::cli_progress_done(id = progress_bar_id)
      # }

      scores = data.table::copy(all_preds)[,
        score_post := vapply(
          prediction,
          \(p) p$score(measures = self$measure)[[self$measure$id]],
          FUN.VALUE = numeric(1)
        )
      ]
      vars_to_keep = c("feature", "iter_rsmp", "iter_repeat", "score_baseline", "score_post")
      scores = scores[scores_baseline, on = c("iter_rsmp")]
      private$.scores = scores[, .SD, .SDcols = vars_to_keep]

      # for obs_loss:
      # Not all losses are decomposable so this is optional and depends on the provided measure
      if (has_obs_loss(self$measure)) {
        grouping_vars = c("feature", "iter_rsmp", "iter_repeat")

        obs_loss_all = all_preds[,
          {
            pred = prediction[[1]]

            # Get only vector of obs losses, Prediction$obs_loss() returns full table
            obs_loss_vals = pred$obs_loss(measures = self$measure)[[self$measure$id]]

            list(
              row_ids = pred$row_ids,
              loss_post = obs_loss_vals
            )
          },
          by = grouping_vars
        ]

        private$.obs_losses = obs_loss_all
      }
    }
  )
)


#' @title Permutation Feature Importance
#'
#' @description
#' Implementation of Permutation Feature Importance (PFI) using modular sampling approach.
#' PFI measures the importance of a feature by calculating the increase in model error
#' when the feature's values are randomly permuted, breaking the relationship between
#' the feature and the target variable.
#'
#' @details
#' Permutation Feature Importance was originally introduced by Breiman (2001) as part of
#' the Random Forest algorithm. The method works by:
#' 1. Computing baseline model performance on the original dataset
#' 2. For each feature, randomly permuting its values while keeping other features unchanged
#' 3. Computing model performance on the permuted dataset
#' 4. Calculating importance as the difference (or ratio) between permuted and original performance
#'
#' @references
#' `r print_bib("breiman_2001")`
#' `r print_bib("fisher_2019")`
#' `r print_bib("strobl_2008")`
#'
#' @examples
#' library(mlr3)
#'
#' task <- sim_dgp_correlated(n = 500)
#'
#' pfi <- PFI$new(
#'   task = task,
#'   learner = lrn("regr.rpart"),
#'   measure = msr("regr.mse"),
#' n_repeats = 5
#' )
#' pfi$compute()
#' pfi$importance()
#' @export
PFI = R6Class(
  "PFI",
  inherit = PerturbationImportance,
  public = list(
    #' @description
    #' Creates a new instance of the PFI class
    #' @param task,learner,measure,resampling,features,groups,relation,n_repeats,batch_size Passed to [PerturbationImportance]
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      relation = "difference",
      n_repeats = 30L,
      batch_size = NULL
    ) {
      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        groups = groups,
        sampler = MarginalPermutationSampler$new(task),
        relation = relation,
        n_repeats = n_repeats,
        batch_size = batch_size
      )

      self$label = "Permutation Feature Importance"
    },

    #' @description
    #' Compute PFI scores
    #' @param n_repeats (`integer(1)`; `NULL`) Number of permutation iterations. If `NULL`, uses stored value.
    #' @param batch_size (`integer(1)` | `NULL`: `NULL`) Maximum number of rows to predict at once. If `NULL`, uses stored value.
    #' @param store_models,store_backends (`logical(1)`: `TRUE`) Whether to store fitted models / data backends, passed to [mlr3::resample] internally
    #' for the initial fit of the learner.
    #' This may be required for certain measures and is recommended to leave enabled unless really necessary.
    compute = function(
      n_repeats = NULL,
      batch_size = NULL,
      store_models = TRUE,
      store_backends = TRUE
    ) {
      # PFI uses the MarginalPermutationSampler directly
      private$.compute_perturbation_importance(
        n_repeats = n_repeats,
        batch_size = batch_size,
        store_models = store_models,
        store_backends = store_backends,
        sampler = self$sampler
      )
    }
  )
)

#' @title Conditional Feature Importance
#'
#' @description Implementation of CFI using modular sampling approach
#'
#' @details
#'
#' CFI replaces feature values with conditional samples from the distribution of
#' the feature given the other features. Any [ConditionalSampler] or [KnockoffSampler] can be used.
#'
#' ## Statistical Inference
#'
#' Two approaches for statistical inference are primarily supported via
#' `$importance(ci_method = "cpi")`:
#'
#' - **CPI** (Watson & Wright, 2021): The original Conditional Predictive Impact method,
#'   designed for use with knockoff samplers ([KnockoffGaussianSampler]).
#'
#' - **cARFi** (Blesch et al., 2025): CFI with ARF-based conditional sampling
#'   ([ConditionalARFSampler]), using the same CPI inference framework.
#'
#' Both require a decomposable measure (e.g., MSE) and out-of-sample evaluation.
#' CPI inference is guaranteed to be valid with holdout (a single train/test split).
#' With cross-validation, test observations are i.i.d. but models are fit on
#' overlapping training data, which may affect inference coverage. With bootstrap
#' or subsampling, both non-i.i.d. test observations and overlapping training data
#' can be an issue. See `vignette("inference", package = "xplainfi")` for details.
#'
#' Available tests: `"t"` (t-test), `"wilcoxon"` (signed-rank), `"fisher"` (permutation),
#' `"binomial"` (sign test). The Fisher test is recommended.
#'
#' Method-agnostic inference methods (`"raw"`, `"nadeau_bengio"`, `"quantile"`) are also
#' available; see [FeatureImportanceMethod] for details.
#'
#' For a comprehensive overview of inference methods including usage examples,
#' see `vignette("inference", package = "xplainfi")`.
#'
#' @references `r print_bib("watson_2021", "blesch_2025")`
#'
#' @examples
#' library(mlr3)
#'
#' task <- sim_dgp_correlated(n = 200)
#'
#' # Using default ConditionalARFSampler
#' cfi <- CFI$new(
#'   task = task,
#'   learner = lrn("regr.rpart"),
#'   measure = msr("regr.mse"),
#'   sampler = ConditionalGaussianSampler$new(task),
#'   n_repeats = 5
#' )
#' cfi$compute()
#' cfi$importance()
#' @export
CFI = R6Class(
  "CFI",
  inherit = PerturbationImportance,
  public = list(
    #' @description
    #' Creates a new instance of the CFI class
    #' @param task,learner,measure,resampling,features,groups,relation,n_repeats,batch_size Passed to [PerturbationImportance].
    #' @param sampler ([ConditionalSampler]) Optional custom sampler. Defaults to instantiating `ConditionalARFSampler` internally with default parameters.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      relation = "difference",
      n_repeats = 30L,
      batch_size = NULL,
      sampler = NULL
    ) {
      # Use ConditionalARFSampler by default for CFI
      if (is.null(sampler)) {
        sampler = ConditionalARFSampler$new(task)
        if (xplain_opt("verbose")) {
          cli::cli_alert_info(
            "No {.code sampler} provided, using {.cls ConditionalARFSampler} with default settings."
          )
        }
      }
      # checkmate::assert_class would expect sampler to inherit from all clases, but
      # the two are mutually exclusive (for now?)
      if (!inherits(sampler, c("ConditionalSampler", "KnockoffSampler"))) {
        cli::cli_abort(c(
          x = "Provided sampler is of class {.cls {class(sampler)[[1]]}}.",
          "!" = "Either a {.cls ConditionalSampler} or a {.cls KnockoffSampler} is needed for {.cls CFI}.",
          i = "Choose a supported {.cls FeatureSampler}, such as {.cls ConditionalARFSampler} or {.class KnockoffGaussianSampler}."
        ))
      }
      if (
        inherits(sampler, "ConditionalSampler") &&
          !is.null(sampler$param_set$values$conditioning_set)
      ) {
        cli::cli_warn(c(
          "!" = "Provided sampler has a pre-configured {.code conditioning_set}.",
          i = "To calculate {.cls CFI} correctly, {.code conditioning_set} will be reset such that sampling is performed conditionally on all remaining features."
        ))
        sampler$param_set$values$conditioning_set = NULL
      }

      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        groups = groups,
        sampler = sampler,
        n_repeats = n_repeats,
        batch_size = batch_size
      )

      self$label = "Conditional Feature Importance"
    },

    #' @description
    #' Compute CFI scores
    #' @param n_repeats (`integer(1)`) Number of permutation iterations. If `NULL`, uses stored value.
    #' @param batch_size (`integer(1)` | `NULL`: `NULL`) Maximum number of rows to predict at once. If `NULL`, uses stored value.
    #' @param store_models,store_backends (`logical(1)`: `TRUE`) Whether to store fitted models / data backends, passed to [mlr3::resample] internally
    #' for the initial fit of the learner.
    #' This may be required for certain measures and is recommended to leave enabled unless really necessary.
    compute = function(
      n_repeats = NULL,
      batch_size = NULL,
      store_models = TRUE,
      store_backends = TRUE
    ) {
      # CFI expects sampler configured to condition on all other features for each feature
      # Default for ConditionalARFSampler
      private$.compute_perturbation_importance(
        n_repeats = n_repeats,
        batch_size = batch_size,
        store_models = store_models,
        store_backends = store_backends,
        sampler = self$sampler
      )
    }
  )
)

#' @title Relative Feature Importance
#'
#' @description RFI generalizes CFI and PFI with arbitrary conditioning sets and samplers.
#'
#' @references `r print_bib("konig_2021")`
#'
#' @examples
#' library(mlr3)
#' task = tgen("friedman1")$generate(n = 200)
#' rfi = RFI$new(
#'   task = task,
#'   learner = lrn("regr.rpart"),
#'   measure = msr("regr.mse"),
#'   conditioning_set = c("important1"),
#'   sampler = ConditionalGaussianSampler$new(task),
#'   n_repeats = 5
#' )
#' rfi$compute()
#' rfi$importance()
#' @export
RFI = R6Class(
  "RFI",
  inherit = PerturbationImportance,
  public = list(
    #' @description
    #' Creates a new instance of the RFI class
    #' @param task,learner,measure,resampling,features,groups,relation,n_repeats,batch_size Passed to [PerturbationImportance].
    #' @param conditioning_set ([character()]) Set of features to condition on. Can be overridden in `$compute()`.
    #'   Default (`character(0)`) is equivalent to `PFI`. In `CFI`, this would be set to all features except that of interest.
    #' @param sampler ([ConditionalSampler]) Optional custom sampler. Defaults to `ConditionalARFSampler`.
    initialize = function(
      task,
      learner,
      measure = NULL,
      resampling = NULL,
      features = NULL,
      groups = NULL,
      conditioning_set = NULL,
      relation = "difference",
      n_repeats = 30L,
      batch_size = NULL,
      sampler = NULL
    ) {
      # Use ConditionalARFSampler by default for RFI
      if (is.null(sampler)) {
        sampler = ConditionalARFSampler$new(task)
        if (xplain_opt("verbose")) {
          cli::cli_alert_info(
            "No {.cls ConditionalSampler} provided, using {.cls ConditionalARFSampler} with default settings."
          )
        }
      } else {
        checkmate::assert_class(sampler, "ConditionalSampler")
      }

      super$initialize(
        task = task,
        learner = learner,
        measure = measure,
        resampling = resampling,
        features = features,
        groups = groups,
        sampler = sampler,
        relation = relation,
        n_repeats = n_repeats,
        batch_size = batch_size
      )

      # Validate and set up conditioning set after task is available
      if (!is.null(conditioning_set)) {
        conditioning_set = checkmate::assert_subset(conditioning_set, self$task$feature_names)
      } else {
        # Default to empty set (equivalent(ish) to PFI)
        cli::cli_warn(c(
          "Using empty conditioning set",
          i = "Set {.code conditioning_set} to condition on features."
        ))
        conditioning_set = character(0)
      }

      # Configure the sampler with the conditioning_set
      self$sampler$param_set$values$conditioning_set = conditioning_set

      # Create extended param_set for RFI with conditioning_set parameter
      rfi_ps = paradox::ps(
        conditioning_set = paradox::p_uty(default = character(0))
      )
      rfi_ps$values$conditioning_set = conditioning_set
      self$param_set = c(self$param_set, rfi_ps)

      self$label = "Relative Feature Importance"
    },

    #' @description
    #' Compute RFI scores
    #' @param conditioning_set (`character()`) Set of features to condition on. If `NULL`, uses the stored parameter value.
    #' @param n_repeats (`integer(1)`) Number of permutation iterations. If `NULL`, uses stored value.
    #' @param batch_size (`integer(1)` | `NULL`: `NULL`) Maximum number of rows to predict at once. If `NULL`, uses stored value.
    #' @param store_models,store_backends (`logical(1)`: `TRUE`) Whether to store fitted models / data backends, passed to [mlr3::resample] internally
    #' for the initial fit of the learner.
    #' This may be required for certain measures and is recommended to leave enabled unless really necessary.
    compute = function(
      conditioning_set = NULL,
      n_repeats = NULL,
      batch_size = NULL,
      store_models = TRUE,
      store_backends = TRUE
    ) {
      # Handle conditioning_set parameter override
      if (!is.null(conditioning_set)) {
        # Validate the provided conditioning_set
        conditioning_set = checkmate::assert_subset(conditioning_set, self$task$feature_names)

        # Clear cache and temporarily modify sampler's conditioning_set
        self$scores = NULL
        old_conditioning_set = self$sampler$param_set$values$conditioning_set
        self$sampler$param_set$values$conditioning_set = conditioning_set
        on.exit(self$sampler$param_set$values$conditioning_set <- old_conditioning_set)
      }

      # Use the (potentially modified) sampler
      private$.compute_perturbation_importance(
        n_repeats = n_repeats,
        batch_size = batch_size,
        store_models = store_models,
        store_backends = store_backends,
        sampler = self$sampler
      )
    }
  )
)

# Replicate a data.table for use by samplers that don't natively support
# `samples_per_row > 1`. Returns rows in draw-major order:
#   if `data` has rows r1, r2, r3 and `samples_per_row = 2`, returns
#   r1, r2, r3, r1, r2, r3.
# Intended as an opt-in convenience for user-implemented samplers.
#
# @param data (`data.table`) Data to replicate.
# @param samples_per_row (`integer(1)`) Number of draw blocks.
# @return data.table with `nrow(data) * samples_per_row` rows.
# @noRd
.replicate_evidence = function(data, samples_per_row) {
  checkmate::assert_data_table(data)
  checkmate::assert_count(samples_per_row, positive = TRUE)
  if (samples_per_row == 1L) {
    return(data)
  }
  data[rep.int(seq_len(.N), times = samples_per_row)]
}

test_that(".replicate_evidence preserves data unchanged when samples_per_row == 1", {
  dt = data.table::data.table(x = 1:3, y = letters[1:3])
  out = xplainfi:::.replicate_evidence(dt, samples_per_row = 1L)
  expect_identical(out, dt)
})

test_that(".replicate_evidence produces draw-major output for samples_per_row > 1", {
  dt = data.table::data.table(x = 1:3, y = letters[1:3])
  out = xplainfi:::.replicate_evidence(dt, samples_per_row = 4L)
  expect_equal(nrow(out), 12L)
  # Draw-major: rows 1-3 = draw1 (x = 1:3), rows 4-6 = draw2 (x = 1:3), etc.
  expect_identical(out$x, rep(1:3, times = 4L))
  expect_identical(out$y, rep(letters[1:3], times = 4L))
})

# https://testthat.r-lib.org/articles/special-files.html#setup-files

# Load mlr3 to avoid mlr3:: namespace clutter in tests
library(mlr3)

# CRAN policy: at most 2 cores in tests. data.table (OpenMP) and ranger (own thread pool via arf) need capping;
# ranger uses std::thread, so OMP_THREAD_LIMIT in testthat.R does not cover it. Restored in teardown.
old_dt_threads = data.table::getDTthreads()
old_ranger_threads = getOption("ranger.num.threads")
data.table::setDTthreads(2)
options(ranger.num.threads = 2)

# Get current log threshold
logger = lgr::get_logger("mlr3")
old_threshold = logger$threshold

# Reduce to warnings only for tests
logger$set_threshold("warn")

# Suppress xplainfi messages and progress during tests
old_opts = xplain_opt(verbose = FALSE, progress = FALSE)

# Restore after tests
withr::defer(
  {
    lgr::get_logger("mlr3")$set_threshold(old_threshold)
    xplain_opt(verbose = old_opts$verbose, progress = old_opts$progress)
    data.table::setDTthreads(old_dt_threads)
    options(ranger.num.threads = old_ranger_threads)
  },
  teardown_env()
)

#' xplainfi Package Options
#'
#' Get or set package-level options for xplainfi.
#'
#' @param ... Option names to retrieve (as character strings) or options to set (as named arguments).
#'   - To **get** an option: `xplain_opt("verbose")` returns the current value
#'   - To **set** an option: `xplain_opt(verbose = FALSE)` sets the value
#'   - To **get all** options: `xplain_opt()` returns a named list of all options
#'
#' @details
#' Options can be set in three ways (in order of precedence):
#' 1. Using `xplain_opt(option_name = value)` (recommended)
#' 2. Using `options("xplain.option_name" = value)`
#' 3. Using environment variables `XPLAIN_OPTION_NAME=value`
#'
#' ## Available Options
#'
#' | Option | Default | Description |
#' |--------|---------|-------------|
#' | `verbose` | `TRUE` | Show informational messages (e.g., when using default measure or resampling) |
#' | `progress` | `FALSE` | Show progress bars during computation |
#' | `sequential` | `FALSE` | Force sequential execution (disable parallelization) |
#' | `debug` | `FALSE` | Enable debug output for development and troubleshooting |
#'
#' @return
#' - When **getting** a single option: the option value (logical)
#' - When **getting** multiple options: a named list of option values
#' - When **setting** options: the previous values (invisibly)
#'
#' @examples
#' # Get current value of an option
#' xplain_opt("verbose")
#'
#' # Get all options
#' xplain_opt()
#'
#' # Set an option (returns previous value invisibly)
#' old <- xplain_opt(verbose = FALSE)
#' xplain_opt("verbose")  # Now FALSE
#'
#' # Restore previous value
#' xplain_opt(verbose = old$verbose)
#'
#' # Temporary option change with withr
#' if (requireNamespace("withr", quietly = TRUE)) {
#'   withr::with_options(
#'     list("xplain.verbose" = FALSE),
#'     {
#'       # Code here runs with verbose = FALSE
#'       xplain_opt("verbose")
#'     }
#'   )
#' }
#'
#' @export
xplain_opt = function(...) {
  # Define available options and their defaults
  option_defaults = list(
    verbose = TRUE,
    progress = FALSE,
    sequential = FALSE,
    debug = FALSE
  )

  args = list(...)

  # No arguments: return all options with current values
  if (length(args) == 0) {
    result = lapply(names(option_defaults), function(opt) {
      get_option_value(opt, option_defaults[[opt]])
    })
    names(result) = names(option_defaults)
    return(result)
  }

  # Check if we're getting or setting
  arg_names = names(args)

  if (is.null(arg_names) || all(arg_names == "")) {
    # All unnamed arguments: getting values
    # xplain_opt("verbose") or xplain_opt("verbose", "progress")
    opts_to_get = unlist(args)
    checkmate::assert_character(opts_to_get, min.len = 1)
    checkmate::assert_subset(opts_to_get, names(option_defaults))

    if (length(opts_to_get) == 1) {
      return(get_option_value(opts_to_get, option_defaults[[opts_to_get]]))
    } else {
      result = lapply(opts_to_get, function(opt) {
        get_option_value(opt, option_defaults[[opt]])
      })
      names(result) = opts_to_get
      return(result)
    }
  } else {
    # Named arguments: setting values
    # xplain_opt(verbose = FALSE, progress = TRUE)
    checkmate::assert_subset(arg_names, names(option_defaults))

    # Store old values to return
    old_values = lapply(arg_names, function(opt) {
      get_option_value(opt, option_defaults[[opt]])
    })
    names(old_values) = arg_names

    # Set new values
    for (opt in arg_names) {
      opt_list = list(args[[opt]])
      names(opt_list) = paste0("xplain.", tolower(opt))
      options(opt_list)
    }

    return(invisible(old_values))
  }
}

#' Get option value with precedence: R option > env var > default
#'
#' Coerces to the default's storage type, so logical options stay logical.
#' Anything that fails to coerce is treated as unset.
#'
#' @noRd
#' @keywords internal
get_option_value = function(name, default) {
  opt = getOption(paste0("xplain.", tolower(name)), default = NA)
  envvar = Sys.getenv(toupper(paste0("xplain_", name)), unset = NA)

  coerce = switch(
    typeof(default),
    logical = as.logical,
    integer = function(x) suppressWarnings(as.integer(x)),
    double = function(x) suppressWarnings(as.numeric(x)),
    character = as.character,
    identity
  )

  opt = coerce(opt)
  if (length(opt) == 0L || is.na(opt)) {
    opt = NULL
  }

  envvar = coerce(envvar)
  if (length(envvar) == 0L || is.na(envvar)) {
    envvar = NULL
  }

  # R option > env var > default
  opt %||% envvar %||% default
}

test_that("xplain_opt() returns all options when called without arguments", {
  result = xplain_opt()
  expect_type(result, "list")
  expect_true(all(c("verbose", "progress") %in% names(result)))
})

test_that("xplain_opt() gets single option value", {
  # verbose defaults to TRUE
  options("xplain.verbose" = NULL)
  Sys.unsetenv("XPLAIN_VERBOSE")
  expect_true(xplain_opt("verbose"))

  # progress defaults to FALSE
  options("xplain.progress" = NULL)
  Sys.unsetenv("XPLAIN_PROGRESS")
  expect_false(xplain_opt("progress"))
})

test_that("xplain_opt() gets multiple option values", {
  result = xplain_opt("verbose", "progress")
  expect_type(result, "list")
  expect_length(result, 2)
  expect_named(result, c("verbose", "progress"))
})

test_that("xplain_opt() sets options and returns old values", {
  # Store current values
  old_verbose = xplain_opt("verbose")

  # Set new value
  returned = xplain_opt(verbose = FALSE)
  expect_type(returned, "list")
  expect_length(returned, 1)
  expect_equal(returned$verbose, old_verbose)

  # Verify it was set
  expect_false(xplain_opt("verbose"))

  # Restore
  xplain_opt(verbose = old_verbose)
  expect_equal(xplain_opt("verbose"), old_verbose)
})

test_that("xplain_opt() sets multiple options", {
  # Store current values
  old = xplain_opt()

  # Set multiple
  xplain_opt(verbose = FALSE, progress = TRUE)

  expect_false(xplain_opt("verbose"))
  expect_true(xplain_opt("progress"))

  # Restore
  xplain_opt(verbose = old$verbose, progress = old$progress)
})

test_that("R option takes precedence over environment variable", {
  # Store current
  old_verbose = xplain_opt("verbose")

  # Set env to TRUE, option to FALSE
  Sys.setenv("XPLAIN_VERBOSE" = "TRUE")
  options("xplain.verbose" = FALSE)

  expect_false(xplain_opt("verbose"))

  # Set env to FALSE, option to TRUE
  Sys.setenv("XPLAIN_VERBOSE" = "FALSE")
  options("xplain.verbose" = TRUE)

  expect_true(xplain_opt("verbose"))

  # Cleanup
  Sys.unsetenv("XPLAIN_VERBOSE")
  options("xplain.verbose" = old_verbose)
})

test_that("environment variable is used when R option is not set", {
  # Store current
  old_verbose = xplain_opt("verbose")

  # Clear R option, set env
  options("xplain.verbose" = NULL)
  Sys.setenv("XPLAIN_VERBOSE" = "FALSE")

  expect_false(xplain_opt("verbose"))

  Sys.setenv("XPLAIN_VERBOSE" = "TRUE")
  expect_true(xplain_opt("verbose"))

  # Cleanup
  Sys.unsetenv("XPLAIN_VERBOSE")
  options("xplain.verbose" = old_verbose)
})

test_that("xplain_opt() errors on unknown options", {
  expect_error(xplain_opt("nonexistent"), "subset")
  expect_error(xplain_opt(nonexistent = TRUE), "subset")
})

#' Simulate data as in Ewald et al. (2024)
#'
#' @description
#' Reproduces the data generating process from Ewald et al. (2024) for benchmarking
#' feature importance methods. Includes correlated features and interaction effects.
#'
#' @details
#' **Mathematical Model:**
#' \deqn{X_1, X_3, X_5 \sim \text{Uniform}(0,1)}
#' \deqn{X_2 = X_1 + \varepsilon_2, \quad \varepsilon_2 \sim N(0, \sqrt{0.001})}
#' \deqn{X_4 = X_3 + \varepsilon_4, \quad \varepsilon_4 \sim N(0, \sqrt{0.1})}
#' \deqn{Y = X_4 + X_5 + X_4 \cdot X_5 + \varepsilon, \quad \varepsilon \sim N(0, \sqrt{0.1})}
#'
#' **Feature Properties:**
#' - X1, X3, X5: Independent uniform(0,1) distributions
#' - X2: Nearly perfect copy of X1 (correlation approximately 0.99)
#' - X4: Noisy copy of X3 (correlation approximately 0.94)
#' - Y depends on X4, X5, and their interaction
#'
#' @param n (`integer(1)`) Number of samples to create.
#'
#' @return A regression task ([mlr3::TaskRegr]) with [data.table][data.table::data.table] backend.
#' @export
#' @references `r print_bib("ewald_2024")`
#' @family simulation
#' @examples
#' sim_dgp_ewald(100)
#'
sim_dgp_ewald = function(n = 500) {
  x1 = runif(n)
  x3 = runif(n)
  x5 = runif(n)

  x2 = x1 + rnorm(n, 0, sd = 0.001)
  x4 = x3 + rnorm(n, 0, sd = 0.1)

  y = x4 + x5 + x4 * x5 + rnorm(n, 0, sd = 0.1)

  xdf = data.table::data.table(
    y,
    x1,
    x2,
    x3,
    x4,
    x5
  )

  mlr3::TaskRegr$new(backend = xdf, target = "y", id = paste0("Ewald_", n))
}

#' @title Simulation DGPs for Feature Importance Method Comparison
#' @name sim_dgp_scenarios
#' @description
#' These data generating processes (DGPs) are designed to illustrate specific
#' strengths and weaknesses of different feature importance methods like PFI, CFI, and RFI.
#' Each DGP focuses on one primary challenge to make the differences between methods clear.
#'
#' @references `r print_bib("ewald_2024")`
NULL

#' @describeIn sim_dgp_scenarios Correlated features demonstrating PFI's limitations
#'
#' @details
#' **Correlated Features DGP:**
#' This DGP creates highly correlated predictors where PFI will show artificially low
#' importance due to redundancy, while CFI will correctly identify each feature's
#' conditional contribution.
#'
#' **Mathematical Model:**
#' \deqn{(X_1, X_2)^T \sim \text{MVN}(0, \Sigma)}
#' where \eqn{\Sigma} is a \eqn{2 \times 2} covariance matrix with 1 on the diagonal and correlation \eqn{r} on the off-diagonal.
#' \deqn{X_3 \sim N(0,1), \quad X_4 \sim N(0,1)}
#' \deqn{Y = 2 \cdot X_1 + X_3 + \varepsilon}
#' where \eqn{\varepsilon \sim N(0, 0.2^2)}.
#'
#' **Feature Properties:**
#' - `x1`: Standard normal from MVN, direct causal effect on y (\eqn{\beta = 2.0})
#' - `x2`: Correlated with `x1` (correlation = `r`), NO causal effect on y (\eqn{\beta = 0})
#' - `x3`: Independent standard normal, direct causal effect on y (\eqn{\beta = 1.0})
#' - `x4`: Independent standard normal, no effect on y (\eqn{\beta = 0})
#'
#' **Expected Behavior:**
#' - Will depend on the used learner and the strength of correlation (`r`)
#' - **Marginal methods** (PFI, Marginal SAGE): Should falsely assign importance to x2 due to correlation with x1
#' - **CFI** Should correctly assign near-zero importance to x2
#' - x2 is a "spurious predictor" - correlated with causal feature but not causal itself
#'
#' @param n (`integer(1)`) Number of samples to generate.
#' @param r (`numeric(1)`: `0.9`) Correlation between x1 and x2. Must be between -1 and 1.
#' @return A regression task ([mlr3::TaskRegr]) with [data.table][data.table::data.table] backend.
#' @export
#' @family simulation
#' @examples
#' task = sim_dgp_correlated(200)
#' task$data()
#'
#' # With different correlation
#' task_high_cor = sim_dgp_correlated(200, r = 0.95)
#' cor(task_high_cor$data()$x1, task_high_cor$data()$x2)
sim_dgp_correlated = function(n = 500L, r = 0.9) {
  # Generate x1 (causal) and x2 (spurious) with fixed correlation
  x12 = mvtnorm::rmvnorm(n, sigma = stats::toeplitz(r^(0:1)))
  x1 = x12[, 1]
  # Spurious feature: correlated with x1 but no causal effect
  x2 = x12[, 2]

  # Independent features
  x3 = rnorm(n) # Has causal effect
  x4 = rnorm(n) # No causal effect (noise)

  # Outcome depends only on x1 and x3 (x2 has NO causal effect despite correlation)
  y = 2 * x1 + x3 + rnorm(n, 0, 0.2)

  data.table::data.table(
    y = y,
    x1 = x1,
    x2 = x2,
    x3 = x3,
    x4 = x4
  ) |>
    mlr3::TaskRegr$new(target = "y", id = paste0("correlated_", n))
}

#' @describeIn sim_dgp_scenarios Mediated effects showing direct vs total importance
#'
#' @details
#' **Mediated Effects DGP:**
#' This DGP demonstrates the difference between total and direct causal effects.
#' Some features affect the outcome only through mediators.
#'
#' **Mathematical Model:**
#' \deqn{\text{exposure} \sim N(0,1), \quad \text{direct} \sim N(0,1)}
#' \deqn{\text{mediator} = 0.8 \cdot \text{exposure} + 0.6 \cdot \text{direct} + \varepsilon_m}
#' \deqn{Y = 1.5 \cdot \text{mediator} + 0.5 \cdot \text{direct} + \varepsilon}
#' where \eqn{\varepsilon_m \sim N(0, 0.3^2)} and \eqn{\varepsilon \sim N(0, 0.2^2)}.
#'
#' **Feature Properties:**
#' - `exposure`: Has no direct effect on y, only through mediator (total effect = 1.2)
#' - `mediator`: Mediates the effect of exposure on y
#' - `direct`: Has both direct effect on y and effect on mediator
#' - `noise`: No causal relationship to y
#'
#' **Causal Structure:** exposure -> mediator -> y <- direct -> mediator
#'
#' @export
#' @family simulation
#' @examples
#' task = sim_dgp_mediated(200)
#' task$data()
sim_dgp_mediated = function(n = 500L) {
  # Initial exposure variable
  exposure = rnorm(n)

  # Direct predictor that affects both mediator and outcome
  direct = rnorm(n)

  # Mediator affected by both exposure and direct
  mediator = 0.8 * exposure + 0.6 * direct + rnorm(n, 0, 0.3)

  # Noise variable
  noise = rnorm(n)

  # Outcome depends on mediator and direct effect, but NOT directly on exposure
  y = 1.5 * mediator + 0.5 * direct + rnorm(n, 0, 0.2)

  data.table::data.table(
    y = y,
    exposure = exposure,
    mediator = mediator,
    direct = direct,
    noise = noise
  ) |>
    mlr3::TaskRegr$new(target = "y", id = paste0("mediated_", n))
}

#' @describeIn sim_dgp_scenarios Confounding scenario for conditional sampling
#'
#' @details
#' **Confounding DGP:**
#' This DGP includes a confounder that affects both a feature and the outcome.
#' Uses simple coefficients for easy interpretation.
#'
#' **Mathematical Model:**
#' \deqn{H \sim N(0,1)}
#' \deqn{X_1 = H + \varepsilon_1}
#' \deqn{\text{proxy} = H + \varepsilon_p, \quad \text{independent} \sim N(0,1)}
#' \deqn{Y = H + X_1 + \text{independent} + \varepsilon}
#' where all \eqn{\varepsilon \sim N(0, 0.5^2)} independently.
#'
#' **Model Structure:**
#' - Confounder H ~ N(0,1) (potentially unobserved)
#' - x1 = H + noise (affected by confounder)
#' - proxy = H + noise (noisy measurement of confounder)
#' - independent ~ N(0,1) (truly independent)
#' - y = H + x1 + independent + noise
#'
#' **Expected Behavior:**
#' - **PFI**: Will show inflated importance for x1 due to confounding
#' - **CFI**: Should partially account for confounding through conditional sampling and reduce its importance
#' - **RFI conditioning on proxy**: Should reduce confounding bias by conditioning on proxy
#'
#' @param n (`integer(1)`: `500L`) Number of observations to generate.
#' @param hidden (`logical(1)`: `TRUE`) Whether to hide the confounder from the returned task.
#'   If `FALSE`, the confounder is included as a feature, allowing direct adjustment.
#'   If `TRUE` (default), only the proxy is available, simulating unmeasured confounding.
#'
#' @export
#' @family simulation
#' @examples
#' # Hidden confounder scenario (traditional)
#' task_hidden = sim_dgp_confounded(200, hidden = TRUE)
#' task_hidden$feature_names  # proxy available but not confounder
#'
#' # Observable confounder scenario
#' task_observed = sim_dgp_confounded(200, hidden = FALSE)
#' task_observed$feature_names  # both confounder and proxy available
sim_dgp_confounded = function(n = 500L, hidden = TRUE) {
  # Confounder
  confounder = rnorm(n)

  # Feature affected by confounder
  x1 = confounder + rnorm(n, 0, 0.5)

  # Proxy measurement of confounder (observable but noisy)
  proxy = confounder + rnorm(n, 0, 0.5)

  # Independent feature unaffected by confounder
  independent = rnorm(n)

  # Outcome affected by confounder, x1, and independent feature
  y = confounder + x1 + independent + rnorm(n, 0, 0.5)

  # Create data.table conditionally including the confounder
  if (hidden) {
    # Traditional scenario: confounder is hidden, only proxy available
    dt = data.table::data.table(
      y = y,
      x1 = x1,
      proxy = proxy,
      independent = independent
    )
    task_id = paste0("confounded_hidden_", n)
  } else {
    # Observable confounder scenario: both confounder and proxy available
    dt = data.table::data.table(
      y = y,
      x1 = x1,
      confounder = confounder,
      proxy = proxy,
      independent = independent
    )
    task_id = paste0("confounded_observed_", n)
  }

  mlr3::TaskRegr$new(dt, target = "y", id = task_id)
}

#' @describeIn sim_dgp_scenarios Interaction effects between features
#'
#' @details
#' **Interaction Effects DGP:**
#' This DGP demonstrates a pure interaction effect where features have no main effects.
#'
#' **Mathematical Model:**
#' \deqn{Y = 2 \cdot X_1 \cdot X_2 + X_3 + \varepsilon}
#' where \eqn{X_j \sim N(0,1)} independently and \eqn{\varepsilon \sim N(0, 0.5^2)}.
#'
#' **Feature Properties:**
#' - `x1`, `x2`: Independent features with ONLY interaction effect (no main effects)
#' - `x3`: Independent feature with main effect only
#' - `noise1`, `noise2`: No causal effects
#'
#' **Expected Behavior:**
#' - Will depend on the used learner and its ability to model interactions
#' @export
#' @family simulation
#' @examples
#' task = sim_dgp_interactions(200)
#' task$data()
sim_dgp_interactions = function(n = 500L) {
  # Independent features for interaction
  x1 = rnorm(n)
  x2 = rnorm(n)

  # Independent feature with main effect
  x3 = rnorm(n)

  # Noise features
  noise1 = rnorm(n)
  noise2 = rnorm(n)

  # Outcome with ONLY interaction between x1 and x2 (no main effects), plus main effect of x3
  y = 2 * x1 * x2 + x3 + rnorm(n, 0, 0.5)

  data.table::data.table(
    y = y,
    x1 = x1,
    x2 = x2,
    x3 = x3,
    noise1 = noise1,
    noise2 = noise2
  ) |>
    mlr3::TaskRegr$new(target = "y", id = paste0("interactions_", n))
}

#' @describeIn sim_dgp_scenarios Independent features baseline scenario
#'
#' @details
#' **Independent Features DGP:**
#' This is a baseline scenario where all features are independent and their
#' effects are additive. All importance methods should give similar results.
#'
#' **Mathematical Model:**
#' \deqn{Y = 2.0 \cdot X_1 + 1.0 \cdot X_2 + 0.5 \cdot X_3 + \varepsilon}
#' where \eqn{X_j \sim N(0,1)} independently and \eqn{\varepsilon \sim N(0, 0.2^2)}.
#'
#' **Feature Properties:**
#' - `important1-3`: Independent features with different effect sizes
#' - `unimportant1-2`: Independent noise features with no effect
#'
#' **Expected Behavior:**
#' - **All methods**: Should rank features consistently by their true effect sizes
#' - **Ground truth**: important1 > important2 > important3 > unimportant1,2 (approximately 0)
#'
#' @export
#' @family simulation
#' @examples
#' task = sim_dgp_independent(200)
#' task$data()
sim_dgp_independent = function(n = 500L) {
  # Independent important features with different effect sizes
  important1 = rnorm(n)
  important2 = rnorm(n)
  important3 = rnorm(n)

  # Independent unimportant features
  unimportant1 = rnorm(n)
  unimportant2 = rnorm(n)

  # Additive linear outcome
  y = 2.0 * important1 + 1.0 * important2 + 0.5 * important3 + rnorm(n, 0, 0.2)

  data.table::data.table(
    y = y,
    important1 = important1,
    important2 = important2,
    important3 = important3,
    unimportant1 = unimportant1,
    unimportant2 = unimportant2
  ) |>
    mlr3::TaskRegr$new(target = "y", id = paste0("independent_", n))
}

test_that("sim_dgp_independent generates correct structure", {
  set.seed(42) # Fixed seed for reproducibility
  task = sim_dgp_independent(n = 200)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 200)
  expect_equal(task$target_names, "y")
  expect_setequal(
    task$feature_names,
    c("important1", "important2", "important3", "unimportant1", "unimportant2")
  )
  expect_true(grepl("^independent_", task$id))

  # Check data types
  expect_true(all(sapply(data, is.numeric)))

  # Check approximate independence (correlations should be low)
  cor_matrix = cor(data[, .(important1, important2, important3, unimportant1, unimportant2)])
  off_diagonal = cor_matrix[upper.tri(cor_matrix)]
  expect_lt(mean(abs(off_diagonal)), 0.2) # Low correlations

  # Check that y has reasonable correlation with predictors
  cor_y = cor(data$y, data[, .(important1, important2, important3)])
  # Check each correlation individually for better error messages
  expect_gt(abs(cor_y[1, 1]), 0.15) # important1 should have meaningful correlation
  expect_gt(abs(cor_y[1, 2]), 0.15) # important2 should have meaningful correlation
  expect_gt(abs(cor_y[1, 3]), 0.15) # important3 should have meaningful correlation

  # Noise features should have low correlation
  expect_lt(abs(cor(data$y, data$unimportant1)), 0.3)
  expect_lt(abs(cor(data$y, data$unimportant2)), 0.3)
})

test_that("sim_dgp_correlated generates correlated features", {
  task = sim_dgp_correlated(n = 200, r = 0.9)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 200)
  expect_setequal(task$feature_names, c("x1", "x2", "x3", "x4"))
  expect_true(grepl("^correlated_", task$id))

  # Check high correlation between x1 and x2
  cor_x1_x2 = cor(data$x1, data$x2)
  expect_gt(cor_x1_x2, 0.8) # Should be highly correlated

  # Check that x3 and x4 are less correlated with x1, x2
  expect_lt(abs(cor(data$x1, data$x3)), 0.3)
  expect_lt(abs(cor(data$x1, data$x4)), 0.3)

  # Check that x1 and x3 contribute to y (x2 is spurious, x4 is noise)
  lm_fit = lm(y ~ x1 + x2 + x3 + x4, data = data)
  expect_gt(summary(lm_fit)$r.squared, 0.7) # Should explain most variance

  # Also check a model with just the true causal features
  lm_causal = lm(y ~ x1 + x3, data = data)
  expect_gt(summary(lm_causal)$r.squared, 0.9) # Causal features alone should explain most variance
})

test_that("sim_dgp_correlated generates correlated features with different strength", {
  for (r in c(0.2, 0.5, 0.8)) {
    task = sim_dgp_correlated(n = 1000, r = r)
    cor_x1_x2 = cor(task$data(cols = c("x1", "x2")))[1, 2]
    expect_gt(cor_x1_x2, r - 0.1)
    expect_lt(cor_x1_x2, r + 0.1)
  }
})


test_that("sim_dgp_mediated generates mediation structure", {
  task = sim_dgp_mediated(n = 150)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 150)
  expect_setequal(task$feature_names, c("exposure", "mediator", "direct", "noise"))
  expect_true(grepl("^mediated_", task$id))

  # Check mediation relationships
  # Exposure should correlate with mediator
  expect_gt(cor(data$exposure, data$mediator), 0.3)

  # Direct should correlate with both mediator and y
  expect_gt(cor(data$direct, data$mediator), 0.3)
  expect_gt(cor(data$direct, data$y), 0.3)

  # Mediator should strongly correlate with y
  expect_gt(cor(data$mediator, data$y), 0.5)

  # Noise should have low correlation with everything
  expect_lt(abs(cor(data$noise, data$y)), 0.2)
})

test_that("sim_dgp_confounded with hidden=TRUE generates correct structure", {
  task = sim_dgp_confounded(n = 500, hidden = TRUE)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 500)
  expect_setequal(task$feature_names, c("x1", "proxy", "independent"))
  expect_true(grepl("^confounded_hidden_", task$id))
  expect_false("confounder" %in% task$feature_names) # Should be hidden

  # Check confounding structure through correlations
  # x1 and proxy should be correlated (due to hidden confounder)
  expect_gt(cor(data$x1, data$proxy), 0.4)

  # Independent should have low correlation with confounded variables
  expect_lt(abs(cor(data$independent, data$x1)), 0.3)
  expect_lt(abs(cor(data$independent, data$proxy)), 0.3)

  # All should correlate with y
  cor_y = cor(data$y, data[, .(x1, proxy, independent)])
  # Check all correlations are meaningful
  for (i in 1:ncol(cor_y)) {
    expect_gt(abs(cor_y[1, i]), 0.2)
  }
})

test_that("sim_dgp_confounded with hidden=FALSE includes confounder", {
  task = sim_dgp_confounded(n = 100, hidden = FALSE)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 100)
  expect_setequal(task$feature_names, c("x1", "confounder", "proxy", "independent"))
  expect_true(grepl("^confounded_observed_", task$id))
  expect_true("confounder" %in% task$feature_names) # Should be observable

  # Check that confounder is the source of correlations
  # Confounder should correlate highly with x1, proxy
  expect_gt(cor(data$confounder, data$x1), 0.6)
  expect_gt(cor(data$confounder, data$proxy), 0.6)
  expect_gt(cor(data$confounder, data$y), 0.6)

  # Independent should still be independent of confounder
  expect_lt(abs(cor(data$independent, data$confounder)), 0.3)
})

test_that("sim_dgp_interactions generates pure interaction effects", {
  task = sim_dgp_interactions(n = 300)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 300)
  expect_setequal(task$feature_names, c("x1", "x2", "x3", "noise1", "noise2"))
  expect_true(grepl("^interactions_", task$id))

  # Check that x1 and x2 have low marginal correlations with y
  expect_lt(abs(cor(data$x1, data$y)), 0.3) # Should have low marginal correlation
  expect_lt(abs(cor(data$x2, data$y)), 0.3) # Should have low marginal correlation

  # But x3 should have strong correlation (main effect)
  expect_gt(abs(cor(data$x3, data$y)), 0.3)

  # Check interaction effect by fitting model
  lm_fit = lm(y ~ x1 + x2 + x1:x2 + x3 + noise1 + noise2, data = data)
  coefs = coef(lm_fit)

  # Main effects of x1, x2 should be near zero
  expect_lt(abs(coefs["x1"]), 0.5)
  expect_lt(abs(coefs["x2"]), 0.5)

  # Interaction effect should be significant and close to 2
  expect_gt(abs(coefs["x1:x2"]), 1.5)
  expect_lt(abs(coefs["x1:x2"]), 2.5)

  # x3 effect should be close to 1
  expect_gt(abs(coefs["x3"]), 0.5)
  expect_lt(abs(coefs["x3"]), 1.5)

  # Noise effects should be small
  expect_lt(abs(coefs["noise1"]), 0.5)
  expect_lt(abs(coefs["noise2"]), 0.5)
})

test_that("all simulation functions handle different sample sizes", {
  sample_sizes = c(50, 500)

  for (n in sample_sizes) {
    # Test each function with different sample sizes
    task_ind = sim_dgp_independent(n = n)
    expect_equal(nrow(task_ind$data()), n)

    task_cor = sim_dgp_correlated(n = n)
    expect_equal(nrow(task_cor$data()), n)

    task_med = sim_dgp_mediated(n = n)
    expect_equal(nrow(task_med$data()), n)

    task_conf_h = sim_dgp_confounded(n = n, hidden = TRUE)
    expect_equal(nrow(task_conf_h$data()), n)

    task_conf_o = sim_dgp_confounded(n = n, hidden = FALSE)
    expect_equal(nrow(task_conf_o$data()), n)

    task_int = sim_dgp_interactions(n = n)
    expect_equal(nrow(task_int$data()), n)

    task_ewald = sim_dgp_ewald(n = n)
    expect_equal(nrow(task_ewald$data()), n)
  }
})

test_that("simulation functions produce finite values", {
  # Test that all functions produce finite (no NaN, Inf, NA) values
  functions_to_test = list(
    function() sim_dgp_independent(n = 50),
    function() sim_dgp_correlated(n = 50),
    function() sim_dgp_mediated(n = 50),
    function() sim_dgp_confounded(n = 50, hidden = TRUE),
    function() sim_dgp_confounded(n = 50, hidden = FALSE),
    function() sim_dgp_interactions(n = 50),
    function() sim_dgp_ewald(n = 50)
  )

  for (sim_fn in functions_to_test) {
    task = sim_fn()
    data = task$data()

    # Check for finite values
    expect_true(all(is.finite(as.matrix(data))))

    # Check for no missing values
    expect_false(anyNA(as.matrix(data)))
  }
})

test_that("simulation functions are reproducible with set.seed", {
  # Test reproducibility
  set.seed(123)
  task1 = sim_dgp_independent(n = 100)
  data1 = task1$data()

  set.seed(123)
  task2 = sim_dgp_independent(n = 100)
  data2 = task2$data()

  expect_equal(data1, data2)

  # Test with different function
  set.seed(456)
  task3 = sim_dgp_interactions(n = 50)
  data3 = task3$data()

  set.seed(456)
  task4 = sim_dgp_interactions(n = 50)
  data4 = task4$data()

  expect_equal(data3, data4)

  # Test with sim_dgp_ewald
  set.seed(789)
  task5 = sim_dgp_ewald(n = 75)
  data5 = task5$data()

  set.seed(789)
  task6 = sim_dgp_ewald(n = 75)
  data6 = task6$data()

  expect_equal(data5, data6)
})

test_that("sim_dgp_confounded default parameters work correctly", {
  # Test default parameters
  task_default = sim_dgp_confounded()
  expect_equal(nrow(task_default$data()), 500) # Default n
  expect_false("confounder" %in% task_default$feature_names) # Default hidden=TRUE

  # Test explicit parameters match defaults
  task_explicit = sim_dgp_confounded(n = 500L, hidden = TRUE)
  expect_equal(task_default$feature_names, task_explicit$feature_names)
  expect_equal(nrow(task_default$data()), nrow(task_explicit$data()))
})

test_that("sim_dgp_ewald generates correct structure", {
  task = sim_dgp_ewald(n = 500)
  data = task$data()

  # Basic structure tests
  expect_s3_class(task, "TaskRegr")
  expect_equal(nrow(data), 500)
  expect_equal(task$target_names, "y")
  expect_setequal(task$feature_names, c("x1", "x2", "x3", "x4", "x5"))
  expect_true(grepl("^Ewald_", task$id))

  # Check data types
  expect_true(all(sapply(data, is.numeric)))

  # Check relationships as specified in Ewald et al.
  # x2 should be highly correlated with x1 (noisy copy with sd=0.001)
  expect_gt(cor(data$x1, data$x2), 0.8)

  # x4 should be correlated with x3 (noisier copy with sd=0.1)
  expect_gt(cor(data$x3, data$x4), 0.8)

  # x1, x3, x5 should be relatively independent
  expect_lt(abs(cor(data$x1, data$x3)), 0.2)
  expect_lt(abs(cor(data$x1, data$x5)), 0.2)
  expect_lt(abs(cor(data$x3, data$x5)), 0.2)

  # Check that y depends on x4, x5, and their interaction
  # Simple check: both x4 and x5 should correlate with y
  expect_gt(abs(cor(data$x4, data$y)), 0.3)
  expect_gt(abs(cor(data$x5, data$y)), 0.3)

  # Fit a model to verify the interaction effect
  lm_fit = lm(y ~ x4 + x5 + x4:x5, data = data)
  expect_gt(summary(lm_fit)$r.squared, 0.75) # Should explain most variance

  # All coefficients should be significant and close to 1
  coefs = coef(lm_fit)
  expect_lt(abs(coefs["x4"] - 1), 0.3)
  expect_lt(abs(coefs["x5"] - 1), 0.3)
  expect_lt(abs(coefs["x4:x5"] - 1), 0.3)
})

#' Check if a measure has an obs_loss
#' @noRd
#' @keywords internal
#'
#' @return logical(1)
has_obs_loss = function(x) {
  res = FALSE
  if (inherits(x, "R6") && inherits(x, "Measure")) {
    if (utils::packageVersion("mlr3") >= package_version("1.3.0")) {
      res = "obs_loss" %in% x$properties
    } else {
      res = !is.null(x$obs_loss)
    }
  } else if (is.character(x)) {
    if (x %in% mlr3::mlr_measures$keys()) {
      res = has_obs_loss(mlr3::msr(x))
    }
  }

  isTRUE(res)
}

#' Check if a learner can be considered pretrained
#'
#' Returns `TRUE` if the learner has a model and the resampling is a compatible
#' single-iteration setup. Returns `FALSE` for untrained learners.
#' Errors if the learner is trained but the resampling is incompatible
#' (multi-fold or mismatched row IDs).
#'
#' Assumes `resampling` is already instantiated (enforced by
#' `FeatureImportanceMethod$initialize()` before this is called).
#'
#' @param learner,task,resampling ([mlr3::Learner], [mlr3::Task], [mlr3::Resampling])
#'
#' @return `logical(1)`: `TRUE` if pretrained and compatible, `FALSE` if untrained
#' @keywords internal
#' @noRd
assert_pretrained = function(learner, task, resampling) {
  # Untrained learner -> not pretrained, nothing to validate
  if (is.null(learner$model)) {
    return(FALSE)
  }

  # Learner is trained: resampling must have exactly 1 iteration
  if (resampling$iters != 1L) {
    cli::cli_abort(c(
      "Given {.code resampling} is not compatible with using a pre-trained {.cls Learner}",
      i = "If {.code learner} is pre-trained, {.code resampling} must have exactly 1 iteration (e.g. holdout)"
    ))
  }

  # Resampling test row IDs must be a subset of task row IDs
  if (length(setdiff(resampling$test_set(1), task$row_ids)) > 0) {
    cli::cli_abort(c(
      "Provided {.code task} has row_ids not compatible with provided {.code resampling}",
      i = "Make sure {.code resampling} was instantiated on the correct {.code task}"
    ))
  }

  TRUE
}

#' Create ResampleResult object
#'
#' @param task ([mlr3::Task])
#' @param learner ([mlr3::Learner]) Either untrained or trained learner. If trained, the `ResampleResult` will be constructed manually, otherwise [mlr3::resample()] is used.
#' @param resampling ([mlr3::Resampling]) If `learner` is trained, this must be an instantiated, single-iteration `Resampling`.
#' @param store_models,store_backends Passed to [mlr3::resample()] or [mlr3::as_resample_result()] respectively.
#'
#' @importFrom mlr3 as_resample_result
#' @noRd
#' @keywords internal
assemble_rr = function(
  task,
  learner,
  resampling,
  # Models are needed in any case and so are task backends I think
  store_models = TRUE,
  store_backends = TRUE
) {
  if (assert_pretrained(learner = learner, task = task, resampling = resampling)) {
    if (xplain_opt("debug")) {
      cli::cli_alert_info("Using pretrained learner")
    }
    pred = learner$predict(task, row_ids = resampling$test_set(1))

    # Clone learner: as_resample_result() clones internally but resets the model
    # on the object it receives, which would wipe the user's original via R6 reference
    # mlr3:::as_resample_result.list
    rdata = mlr3::as_result_data(
      task = task,
      learners = list(learner$clone()),
      resampling = resampling,
      iterations = seq_len(resampling$iters),
      store_backends = store_backends,
      predictions = list(list(test = pred))
    )
    mlr3::ResampleResult$new(rdata)
  } else {
    if (xplain_opt("debug")) {
      cli::cli_alert_info("Using {.fun resample}")
    }
    resample(
      task,
      learner,
      resampling,
      store_models = TRUE,
      store_backends = store_backends
    )
  }
}

#' Create a resampling with all data being test data
#'
#' Utility for use with a pretrained learner in importance methods which support it
#'
#' Note that the resulting Resampling will have an
#' empty train set, making it useless for any
#' other purpose than the use with a pretrained learner.
#' @param task ([mlr3::Task])
#' @return [mlr3::Resampling] with an empty `train_set` and a single `test_set` identical to all of the given `Task`.
#' @export
#' @importFrom mlr3 assert_task rsmp
#' @examples
#' library(mlr3)
#' # Create custom task from some data.frame
#' custom_task <- as_task_regr(mtcars, target = "mpg")
#' # Create matching Resampling with all-test data
#' resampling_custom <- rsmp_all_test(custom_task)
rsmp_all_test = function(task) {
  mlr3::assert_task(task)

  mlr3::rsmp("custom")$instantiate(
    task,
    train_sets = list(integer(0)),
    test_sets = list(task$row_ids)
  )
}

test_that("has_obs_loss returns TRUE for measures with obs_loss", {
  # Classification measures with obs_loss
  expect_true(has_obs_loss(msr("classif.ce")))
  expect_true(has_obs_loss(msr("classif.acc")))
  expect_true(has_obs_loss(msr("classif.logloss")))

  # Regression measures with obs_loss
  expect_true(has_obs_loss(msr("regr.mse")))
  expect_true(has_obs_loss(msr("regr.mae")))
})

test_that("has_obs_loss returns FALSE for measures without obs_loss", {
  # These measures don't have observation-wise losses
  expect_false(has_obs_loss(msr("classif.auc")))
  expect_false(has_obs_loss(msr("regr.rsq")))
  expect_false(has_obs_loss(msr("time_train")))
})
test_that("has_obs_loss works with character measure IDs", {
  # Should look up measure by ID and check
  expect_true(has_obs_loss("classif.ce"))
  expect_true(has_obs_loss("regr.mse"))
  expect_false(has_obs_loss("classif.auc"))
  expect_false(has_obs_loss("regr.rsq"))
})

test_that("has_obs_loss returns FALSE for invalid inputs", {
  # Non-existent measure ID
  expect_false(has_obs_loss("nonexistent_measure"))

  # Non-Measure objects
  expect_false(has_obs_loss(NULL))
  expect_false(has_obs_loss(42))
  expect_false(has_obs_loss(list()))
  expect_false(has_obs_loss(data.frame()))
})

test_that("has_obs_loss handles edge cases", {
  # Empty string
  expect_false(has_obs_loss(""))

  # NA values
  expect_false(has_obs_loss(NA))
  expect_false(has_obs_loss(NA_character_))
})

#' Default value for `NULL`
#'
#' A backport of `%||%` available in R versions from 4.4.0.
#'
#' @param x,y If `x` is NULL or length 0, will return `y`; otherwise returns `x`.
#' @rawNamespace if (getRversion() < "4.4.0") export(`%||%`)
#' @name op-null-default
#' @examples
#' 1 %||% 2
#' NULL %||% 2
`%||%` = function(x, y) {
  if (is.null(x)) y else x
}

---
title: "Simulation Settings for Feature Importance Methods"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Simulation Settings for Feature Importance Methods}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 7,
	fig.height = 5
)
```

```{r setup}
library(xplainfi)
library(DiagrammeR)
library(mlr3learners)
set.seed(123)
```

## Introduction

The `xplainfi` package provides several data generating processes (DGPs) designed to illustrate specific strengths and weaknesses of different feature importance methods. Each DGP focuses on one primary challenge to make the differences between methods clear.

This article provides a comprehensive overview of all simulation settings, including their mathematical formulations and causal structures visualized as directed acyclic graphs (DAGs).

## Overview of Simulation Settings

```{r overview-table, echo=FALSE}
dgp_overview <- data.frame(
	DGP = c(
		"sim_dgp_correlated",
		"sim_dgp_mediated",
		"sim_dgp_confounded",
		"sim_dgp_interactions",
		"sim_dgp_independent",
		"sim_dgp_ewald"
	),
	Challenge = c(
		"Spurious correlation",
		"Mediation effects",
		"Confounding",
		"Interaction effects",
		"Baseline (no challenges)",
		"Mixed effects"
	),
	`PFI Behavior` = c(
		"High for spurious x2",
		"Shows total effects",
		"Biased upward",
		"Low (no main effects)",
		"Accurate",
		"Mixed"
	),
	`CFI Behavior` = c(
		"Low for spurious x2",
		"Shows direct effects",
		"Less biased",
		"High (captures interactions)",
		"Accurate",
		"Mixed"
	),
	check.names = FALSE
)
knitr::kable(dgp_overview, caption = "Overview of simulation settings and expected method behavior")
```

## 1. Correlated Features DGP

This DGP creates a highly correlated spurious predictor to illustrate the fundamental difference between marginal and conditional importance methods.

### Mathematical Model

$$(X_1, X_2)^T \sim \text{MVN}(0, \Sigma)$$

where $\Sigma$ is a $2 \times 2$ covariance matrix with 1 on the diagonal and correlation $r$ (default 0.9) on the off-diagonal.

$$X_3 \sim N(0,1), \quad X_4 \sim N(0,1)$$
$$Y = 2 \cdot X_1 + X_3 + \varepsilon$$

where $\varepsilon \sim N(0, 0.2^2)$.

### Causal Structure

```{r dag-correlated, echo=FALSE, fig.cap="DAG for correlated features DGP", fig.width=10, fig.height=4}
#| fig.alt: "Directed acyclic graph with four feature nodes (X1-X4) and outcome Y. Arrows show causal paths from X1 and X3 to Y, with an edge between X1 and X2 showing r approx. 0.9"
grViz(
	"
  digraph Correlated {
    rankdir=LR;
    graph [ranksep=1.5];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightcoral', label='X₁\n(β=2.0)'];
    X2 [fillcolor='pink', label='X₂\n(β=0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=1.0)'];
    X4 [fillcolor='lightgray', label='X₄\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    X1 -> X2 [color=red, style=bold, label='r≈0.9'];
    X1 -> Y [label='2.0'];
    X2 -> Y [style=dashed, color=gray, label='0'];
    X3 -> Y [label='1.0'];
    X4 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X3; X4}
    {rank=same; X2}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r correlated-example}
set.seed(123)
task <- sim_dgp_correlated(n = 500)

# Check correlation between X1 and X2
cor(task$data()[, c("x1", "x2")])

# True coefficients: x1=2.0, x2=0, x3=1.0, x4=0
# Note: x2 is highly correlated with x1 but has NO causal effect!
```


## 2. Mediated Effects DGP

This DGP demonstrates the difference between total and direct causal effects. Some features affect the outcome only through mediators.

### Mathematical Model

$$\text{exposure} \sim N(0,1), \quad \text{direct} \sim N(0,1)$$
$$\text{mediator} = 0.8 \cdot \text{exposure} + 0.6 \cdot \text{direct} + \varepsilon_m$$
$$Y = 1.5 \cdot \text{mediator} + 0.5 \cdot \text{direct} + \varepsilon$$

where $\varepsilon_m \sim N(0, 0.3^2)$ and $\varepsilon \sim N(0, 0.2^2)$.

### Causal Structure

```{r dag-mediated, echo=FALSE, fig.cap="DAG for mediated effects DGP", fig.width=10, fig.height=4}
#| fig.alt: "Directed acyclic graph showing mediation structure. Exposure connects to mediator, which connects to Y. Direct connects to both mediator and Y. Noise node has dashed arrow to Y."
grViz(
	"
  digraph Mediated {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    E [fillcolor='orange', label='Exposure\n(β=0)'];
    D [fillcolor='lightblue', label='Direct\n(β=0.5)'];
    M [fillcolor='yellow', label='Mediator\n(β=1.5)'];
    N [fillcolor='lightgray', label='Noise\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    E -> M [label='0.8', color=purple, penwidth=2];
    D -> M [label='0.6', color=blue];
    D -> Y [label='0.5', color=blue];
    M -> Y [label='1.5', color=purple, penwidth=2];
    N -> Y [style=dashed, color=gray];
    
    {rank=source; E; D; N}
    {rank=same; M}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r mediated-example}
set.seed(123)
task <- sim_dgp_mediated(n = 500)

# Calculate total effect of exposure
# Total effect = 0.8 * 1.5 = 1.2 (through mediator)
# Direct effect = 0 (no direct path to Y)
```

## 3. Confounding DGP

This DGP includes a confounder that affects both a feature and the outcome.

### Mathematical Model

$$H \sim N(0,1) \quad \text{(confounder)}$$
$$X_1 = H + \varepsilon_1$$
$$\text{proxy} = H + \varepsilon_p, \quad \text{independent} \sim N(0,1)$$
$$Y = H + X_1 + \text{independent} + \varepsilon$$

where all $\varepsilon \sim N(0, 0.5^2)$ independently.

### Causal Structure

```{r dag-confounded, echo=FALSE, fig.cap="DAG for confounding DGP", fig.width=10, fig.height=5}
#| fig.alt: "Directed acyclic graph with hidden confounder H (dashed circle) connecting to X1, proxy, and Y. X1 and independent also have arrows to Y."
grViz(
	"
  digraph Confounded {
    rankdir=LR;
    graph [ranksep=1.2, nodesep=0.8];
    node [shape=circle, style=filled, fontsize=14, width=1.2];

    H [fillcolor='red', label='H\n(Confounder)', style='filled,dashed'];
    X1 [fillcolor='lightcoral', label='X₁\n(β=1.0)'];
    P [fillcolor='pink', label='Proxy\n(β=0)'];
    I [fillcolor='lightblue', label='Independent\n(β=1.0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];

    H -> X1 [color=red, label='1.0'];
    H -> P [color=red, style=dashed, label='1.0'];
    H -> Y [color=red, label='1.0', penwidth=2];
    X1 -> Y [label='1.0'];
    I -> Y [label='1.0'];

    {rank=source; H}
    {rank=same; X1; P; I}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r confounded-example}
set.seed(123)
# Hidden confounder scenario (default)
task_hidden <- sim_dgp_confounded(n = 500, hidden = TRUE)
task_hidden$feature_names # proxy available but not confounder

# Observable confounder scenario
task_observed <- sim_dgp_confounded(n = 500, hidden = FALSE)
task_observed$feature_names # both confounder and proxy available
```

## 4. Interaction Effects DGP

This DGP demonstrates a pure interaction effect where features have no main effects.

### Mathematical Model

$$Y = 2 \cdot X_1 \cdot X_2 + X_3 + \varepsilon$$

where $X_j \sim N(0,1)$ independently and $\varepsilon \sim N(0, 0.5^2)$.

### Causal Structure

```{r dag-interactions, echo=FALSE, fig.cap="DAG for interaction effects DGP", fig.width=10, fig.height=4}
#| fig.alt: "Directed acyclic graph with X1 and X2 connecting to a diamond-shaped interaction node, which connects to Y. X3 connects directly to Y. N1 and N2 have dashed arrows to Y."
grViz(
	"
  digraph Interaction {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='orange', label='X₁\n(β=0)'];
    X2 [fillcolor='orange', label='X₂\n(β=0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=1.0)'];
    N1 [fillcolor='lightgray', label='N₁\n(β=0)'];
    N2 [fillcolor='lightgray', label='N₂\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    INT [fillcolor='red', shape=diamond, label='X₁×X₂\n(β=2.0)', width=1.5];
    
    X1 -> INT [color=red, penwidth=2];
    X2 -> INT [color=red, penwidth=2];
    INT -> Y [color=red, label='2.0', penwidth=2];
    X3 -> Y [label='1.0'];
    N1 -> Y [style=dashed, color=gray];
    N2 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X2; X3; N1; N2}
    {rank=same; INT}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r interactions-example}
set.seed(123)
task <- sim_dgp_interactions(n = 500)

# Note: X1 and X2 have NO main effects
# Their importance comes ONLY through their interaction
```

## 5. Independent Features DGP (Baseline)

This is a baseline scenario where all features are independent and their effects are additive. All importance methods should give similar results.

### Mathematical Model

$$Y = 2.0 \cdot X_1 + 1.0 \cdot X_2 + 0.5 \cdot X_3 + \varepsilon$$

where $X_j \sim N(0,1)$ independently and $\varepsilon \sim N(0, 0.2^2)$.

### Causal Structure

```{r dag-independent, echo=FALSE, fig.cap="DAG for independent features DGP", fig.width=10, fig.height=4}
#| fig.alt: "Directed acyclic graph with five feature nodes in a row, each with arrows pointing to outcome Y. X1-X3 have solid arrows with effect sizes labeled, N1-N2 have dashed arrows."
grViz(
	"
  digraph Independent {
    rankdir=LR;
    graph [ranksep=1.5];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightblue', label='X₁\n(β=2.0)'];
    X2 [fillcolor='lightblue', label='X₂\n(β=1.0)'];
    X3 [fillcolor='lightblue', label='X₃\n(β=0.5)'];
    N1 [fillcolor='lightgray', label='N₁\n(β=0)'];
    N2 [fillcolor='lightgray', label='N₂\n(β=0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    
    X1 -> Y [label='2.0', penwidth=3];
    X2 -> Y [label='1.0', penwidth=2];
    X3 -> Y [label='0.5'];
    N1 -> Y [style=dashed, color=gray];
    N2 -> Y [style=dashed, color=gray];
    
    {rank=source; X1; X2; X3; N1; N2}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r independent-example}
set.seed(123)
task <- sim_dgp_independent(n = 500)

# All methods should rank features consistently:
# important1 > important2 > important3 > unimportant1,2 (approx. 0)
```

### Expected Behavior

- **All methods**: Should rank features consistently by their true effect sizes
- **Ground truth**: important1 (2.0) > important2 (1.0) > important3 (0.5) > unimportant1,2 (0)

## 6. Ewald et al. (2024) DGP

Reproduces the data generating process from Ewald et al. (2024) for benchmarking feature importance methods. Includes correlated features and interaction effects.

### Mathematical Model

$$X_1, X_3, X_5 \sim \text{Uniform}(0,1)$$
$$X_2 = X_1 + \varepsilon_2, \quad \varepsilon_2 \sim N(0, 0.001)$$
$$X_4 = X_3 + \varepsilon_4, \quad \varepsilon_4 \sim N(0, 0.1)$$
$$Y = X_4 + X_5 + X_4 \cdot X_5 + \varepsilon, \quad \varepsilon \sim N(0, 0.1)$$

### Causal Structure

```{r dag-ewald, echo=FALSE, fig.cap="DAG for Ewald et al. (2024) DGP", fig.width=10, fig.height=4}
#| fig.alt: "Directed acyclic graph with X1-X2 correlation, X3-X4 correlation. X4 and X5 connect to a diamond interaction node and directly to Y."
grViz(
	"
  digraph Ewald {
    rankdir=LR;
    graph [ranksep=1.2];
    node [shape=circle, style=filled, fontsize=14, width=1.2];
    
    X1 [fillcolor='lightgray', label='X₁\n(β=0)'];
    X2 [fillcolor='lightgray', label='X₂\n(β=0)'];
    X3 [fillcolor='lightgray', label='X₃\n(β=0)'];
    X4 [fillcolor='lightblue', label='X₄\n(β=1.0)'];
    X5 [fillcolor='lightblue', label='X₅\n(β=1.0)'];
    Y [fillcolor='greenyellow', label='Y', width=1.5];
    INT [fillcolor='red', shape=diamond, label='X₄×X₅\n(β=1.0)', width=1.5];
    
    X1 -> X2 [color=gray, label='≈1.0'];
    X3 -> X4 [color=gray, label='≈1.0'];
    X4 -> Y [label='1.0'];
    X5 -> Y [label='1.0'];
    X4 -> INT [color=red];
    X5 -> INT [color=red];
    INT -> Y [color=red, label='1.0'];
    
    {rank=source; X1; X3; X5}
    {rank=same; X2; X4}
    {rank=same; INT}
    {rank=sink; Y}
  }"
)
```

### Usage Example

```{r ewald-example}
sim_dgp_ewald(n = 500)
```

---
title: "Feature Samplers"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Feature Samplers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 7,
	fig.height = 5
)
```

Feature samplers are a core component of perturbation-based feature importance methods (PFI, CFI, RFI) and any other methods based on some form of marginalization (SAGE).
They determine how features are replaced or perturbed when evaluating feature importance. 
This vignette introduces the different types of feature samplers available in `xplainfi` and demonstrates their use.

## Setup

We create two tasks: one with mixed features (the penguins data), and one with all-numeric features.

```{r load-libraries}
library(xplainfi)
library(mlr3)
library(mlr3learners)
library(data.table)

# Create a task for demonstration
task_mixed = tsk("penguins")
task_numeric = sim_dgp_correlated(n = 200)
```

## Base Class: FeatureSampler

All feature samplers inherit from the `FeatureSampler` base class, which provides a common interface for sampling features.

### Key Properties

**Feature Type Support**: Each sampler declares which feature types it supports:

```{r feature-types}
# Check supported feature types for different samplers
task_mixed$feature_types
permutation = MarginalPermutationSampler$new(task_mixed)
permutation$feature_types
```

**Two Sampling Methods**:

1. `$sample(feature, row_ids)` - Sample from the stored task
2. `$sample_newdata(feature, newdata)` - Sample using external data

Let's demonstrate both with the permutation sampler:

```{r sample-methods}
# Sample from stored task (using row_ids)
sampled_task = permutation$sample(
	feature = "bill_length",
	row_ids = 40:45
)
sampled_task

# Sample from "external" data
test_data = task_mixed$data(rows = 40:45)
sampled_external = permutation$sample_newdata(
	feature = "bill_length",
	newdata = test_data
)
sampled_external
```

Notice that:

- The sampled feature values change (they are permuted)
- Other features and the target remain unchanged
- The data structure is preserved

## Permutation Sampler

The `MarginalPermutationSampler` performs simple random permutation of features, breaking their relationship with the target and other features. This is the classic approach used in Permutation Feature Importance (PFI).

**How it works:**

- Each feature is randomly shuffled (permuted) independently
- The association between feature values and target values is broken
- The association between feature values **across rows** is broken
- The marginal distribution of each feature is preserved


```{r permutation-example}
# Create permutation sampler
permutation = MarginalPermutationSampler$new(task_mixed)

# Sample a continuous feature
original = task_mixed$data(rows = 1:10)
sampled = permutation$sample("bill_length", row_ids = 1:10)

# Compare original and sampled values
data.table(
	original_bill = original$bill_length,
	sampled_bill = sampled$bill_length,
	sex = original$sex # Unchanged
)
```

Note that the permutation is only performed _within_ the requested `row_ids`.

**Use in PFI**: The permutation sampler is used by default in Permutation Feature Importance

## Marginal Reference Sampler

The `MarginalReferenceSampler` is another type of marginal sampler that samples complete rows from reference data.
Unlike `MarginalPermutationSampler` which shuffles feature values independently, this sampler preserves within-row dependencies by sampling intact observations.

**Key differences from MarginalPermutationSampler:**

- **MarginalPermutationSampler**: Shuffles each feature independently -- breaks dependencies between sampled features
- **MarginalReferenceSampler**: Samples complete reference rows -- preserves dependencies of the sampled features

This is the approach used in [SAGE (Shapley Additive Global importancE)](https://github.com/iancovert/sage) for marginal feature importance. In this SAGE implementation, the [`MarginalImputer`](https://github.com/iancovert/sage/blob/master/sage/imputers.py#L44-L78), uses this approach.
The "imputer" name comes from the "imputation" of model predictions using out-of-coalition features by sampling from their marginal distribution.
In `xplainfi`, we separate the "feature sampling" and "model prediction" steps (for now), which is why we keep the sampling infrastructure independent.

**How it works:**

```{r marginal-ref-example}
# Create marginal reference sampler with n_samples reference pool
marginal_ref = MarginalReferenceSampler$new(task_mixed, n_samples = 30L)

# Sample a feature - each row gets values from a randomly sampled reference row
original = task_mixed$data(rows = 1:5)
sampled = marginal_ref$sample("bill_length", row_ids = 1:5)

# Compare
data.table(
	original_bill = original$bill_length,
	sampled_bill = sampled$bill_length,
	sex = original$sex # Unchanged
)
```

**Parameters:**

- `n_samples`: Controls the size of the reference data pool
  - If `NULL`: uses all task data
  - If specified: subsamples that many rows from task

**Preserving within-row correlations:**

To demonstrate the difference, consider features that are correlated as in `task_numeric`.
Here, `x1` and `x2` are correlated, and if we sample them jointly, only `MarginalReferenceSampler` retains their original correlation (approximately).

```{r correlation-preservation}
# Sample with MarginalPermutationSampler (breaks correlations)
perm = MarginalPermutationSampler$new(task_numeric)
sampled_perm = perm$sample(c("x1", "x2"), row_ids = 1:10)

# Sample with MarginalReferenceSampler (preserves within-row correlations)
ref = MarginalReferenceSampler$new(task_numeric, n_samples = 50L)
sampled_ref = ref$sample(c("x1", "x2"), row_ids = 1:10)

# Check correlations
cor_original = cor(task_numeric$data()$x1, task_numeric$data()$x2)
cor_perm = cor(sampled_perm$x1, sampled_perm$x2)
cor_ref = cor(sampled_ref$x1, sampled_ref$x2)

data.table(
	method = c("Original", "Permutation", "Reference"),
	correlation = c(cor_original, cor_perm, cor_ref)
)
```

The reference sampler better preserves the correlation structure because it samples complete rows, while permutation completely breaks the dependency.

## Conditional Samplers

Conditional samplers account for dependencies between features by sampling from $P(X_j | X_{-j})$ rather than the marginal $P(X_j)$.
This is relevant when features are correlated.

All conditional samplers inherit from `ConditionalSampler` and support:

- Specifying which features to condition on via `conditioning_set`
- Both `$sample()` and `$sample_newdata()` methods
- Mixed feature types (depending on the specific sampler)

### Gaussian Conditional Sampler

The `ConditionalGaussianSampler` assumes features follow a multivariate Gaussian distribution and uses closed-form conditional distributions.

**Advantages**:

- Very fast (no model fitting during sampling)
- Deterministic given a seed
- No hyperparameters

**Limitations**:

- Assumes multivariate normality
- Only supports continuous features
- May produce out-of-range values

```{r gaussian-sampler}
# Create Gaussian conditional sampler
gaussian = ConditionalGaussianSampler$new(task_numeric)

# Sample x1 conditioned on other features
sampled = gaussian$sample(
	feature = "x1",
	row_ids = 1:10,
	conditioning_set = c("x2", "x3", "x4")
)

# Compare original and conditionally sampled values
original = task_numeric$data(rows = 1:10)
data.table(
	original = original$x1,
	sampled = sampled$x1,
	x2 = original$x2 # Conditioning feature (unchanged)
)
```

Notice that the sampled values respect the conditional distribution - they're different from the original but plausible given the conditioning features.

### ARF Sampler

The `ConditionalARFSampler` uses Adversarial Random Forests to model complex conditional distributions.
It's the most flexible conditional sampler.

**Advantages**:

- Handles mixed feature types (continuous, categorical, ordered)
- No distributional assumptions
- Captures non-linear relationships

**Limitations**:

- Requires fitting ARF model (slower initialization and sampling for large data)
- More hyperparameters to tune
- Stochastic sampling

```{r arf-sampler}
# Create ARF sampler (works with full task including categorical features)
arf = ConditionalARFSampler$new(task_mixed, num_trees = 20, verbose = FALSE)

# Sample island conditioned on body measurements
sampled = arf$sample(
	feature = "island",
	row_ids = 1:10,
	conditioning_set = c("bill_length", "body_mass")
)

# Compare original and sampled island
original = task_mixed$data(rows = 1:10)
data.table(
	original_island = original$island,
	sampled_island = sampled$island,
	bill_length = original$bill_length, # Conditioning feature
	body_mass = original$body_mass # Conditioning feature
)
```

**Use in CFI**: ConditionalARFSampler is the default for Conditional Feature Importance since it can be used with any task, unlike other samplers.

### Ctree Conditional Sampler

The `ConditionalCtreeSampler` uses conditional inference trees to partition the feature space and sample from local neighborhoods.

**Advantages**:

- Handles mixed feature types
- Interpretable tree structure
- Automatic feature selection (only splits on informative features)

**Limitations**:

- Requires tree building (slower than kNN)
- May produce duplicates if terminal nodes are small

```{r ctree-sampler}
# Create ctree sampler
ctree = ConditionalCtreeSampler$new(task_mixed)

# Sample with default parameters
sampled = ctree$sample(
	feature = "bill_length",
	row_ids = 1:10,
	conditioning_set = "island"
)

original = task_mixed$data(rows = 1:10)
data.table(
	island = original$island, # Conditioning feature
	original = original$bill_length,
	sampled = sampled$bill_length
)
```

The ctree sampler partitions observations based on the conditioning features and samples from within the same partition (terminal node).

### kNN Conditional Sampler

The `ConditionalKNNSampler` finds k nearest neighbors based on conditioning features and samples from them.

**Advantages**:

- Very simple and intuitive
- Fast (no model fitting)
- Single hyperparameter (k)
- Automatic distance metric selection
- Supports mixed feature types (numeric, categorical, ordered, logical)

**Limitations**:

- Sensitive to choice of k
- May produce duplicates if k is small

**Distance metric**:

The sampler automatically selects the appropriate distance metric based on conditioning features:

- **Euclidean distance** when all conditioning features are numeric/integer (standardized)
- **Gower distance** when any conditioning features are categorical (factor, ordered, logical)

#### Example 1: All-numeric conditioning (Euclidean distance)

```{r knn-sampler-numeric}
# Create kNN sampler with k=5 neighbors
knn_numeric = ConditionalKNNSampler$new(task_numeric, k = 5)

# Sample x1 based on nearest neighbors in (x2, x3) space
sampled_numeric = knn_numeric$sample(
	feature = "x1",
	row_ids = 1:5,
	conditioning_set = c("x2", "x3")
)

original_numeric = task_numeric$data(rows = 1:5)
data.table(
	x2 = original_numeric$x2,
	x3 = original_numeric$x3,
	original_x1 = original_numeric$x1,
	sampled_x1 = sampled_numeric$x1
)
```

#### Example 2: Mixed-type conditioning (Gower distance)

```{r knn-sampler-mixed}
# Use task with categorical features
knn_mixed = ConditionalKNNSampler$new(task_mixed, k = 5)

# Sample bill_length conditioning on island (categorical) and body_mass (numeric)
sampled_mixed = knn_mixed$sample(
	feature = "bill_length",
	row_ids = 1:5,
	conditioning_set = c("island", "body_mass")
)

original_mixed = task_mixed$data(rows = 1:5)
data.table(
	island = original_mixed$island,
	body_mass = original_mixed$body_mass,
	original_bill = original_mixed$bill_length,
	sampled_bill = sampled_mixed$bill_length
)
```

The kNN sampler finds the k most similar observations (based on conditioning features) and samples from their feature values. The distance metric is chosen automatically based on feature types.

## Knockoff Samplers

Now that we've seen conditional samplers, we can understand an important limitation of knockoff samplers: unlike the conditional samplers above, knockoffs **don't support arbitrary conditioning sets**.

Knockoff samplers create synthetic features (knockoffs) that satisfy specific statistical properties. They must fulfill the **knockoff swap property**: swapping a feature with its knockoff should not change the joint distribution.

Knockoffs are a separate category because:

1. They require special construction to satisfy theoretical guarantees, and implementations are limited
2. They don't support conditional sampling with arbitrary conditioning sets
3. They're used primarily for feature selection with FDR control, not general importance

### Gaussian Knockoffs

For multivariate Gaussian data, we can construct exact knockoffs:

```{r knockoff-sampler}
# Create Gaussian knockoff sampler (using task_numeric from earlier)
knockoff = KnockoffGaussianSampler$new(task_numeric)

# Generate knockoffs
original = task_numeric$data(rows = 1:5)
knockoffs = knockoff$sample(
	feature = task_numeric$feature_names,
	row_ids = 1:5
)

# Original vs knockoff values
data.table(
	x1_original = original$x1,
	x1_knockoff = knockoffs$x1,
	x2_original = original$x2,
	x2_knockoff = knockoffs$x2
)
```

Key properties of knockoffs:

- Different values but similar statistical properties
- Pairwise exchangeability with originals
- Preserve correlation structure
- Cannot specify which features to condition on (determined by the knockoff construction)

**Conditional Independence Testing**: Knockoffs are particularly relevant for conditional independence testing as implemented in the [cpi package](https://cran.r-project.org/package=cpi). You can combine knockoff samplers with CFI and perform inference:

```{r cfi-knockoff, eval = FALSE}
# CFI with knockoff sampler for conditional independence testing
cfi_knockoff = CFI$new(
	task = task_numeric,
	learner = lrn("regr.ranger"),
	measure = msr("regr.mse"),
	sampler = knockoff
)

# Compute importance with CPI-based inference
cfi_knockoff$compute()
cfi_knockoff$importance(ci_method = "cpi")
```

See `vignette("inference")` for more details on statistical inference with feature importance.


**Key takeaways**:

- **Permutation sampling** produces any value from the marginal distribution
- **Conditional samplers** produce values consistent with conditioning features
- Choice of sampler affects feature importance estimates, especially when features are correlated

## Summary

| Sampler | Feature Types | Assumptions | Speed | Use Case |
|---------|--------------|-------------|-------|----------|
| `MarginalPermutationSampler` | All | None | Very fast | PFI, uncorrelated features |
| `KnockoffGaussianSampler` | Continuous | Multivariate normal | Fast | Model-X knockoffs |
| `ConditionalGaussianSampler` | Continuous | Multivariate normal | Very fast | CFI with continuous features |
| `ConditionalARFSampler` | All | None | Moderate | CFI, complex dependencies |
| `ConditionalCtreeSampler` | All | None | Moderate | CFI, interpretable sampling |
| `ConditionalKNNSampler` | All | None (auto-selects distance) | Fast | CFI, simple local structure |

**General guidelines**:

1. Use **permutation sampling** when features are independent or for baseline PFI
2. Use **Gaussian conditional** for fast conditional sampling with continuous features
3. Use **ARF** for the most flexible conditional sampling with mixed types
4. Use **kNN** for simple, fast conditional sampling
5. Use **ctree** when you want interpretable conditional sampling
6. Use **knockoffs** when you need theoretical guarantees for feature selection and inference

---
title: "Getting Started with xplainfi"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with xplainfi}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
	collapse = TRUE,
	comment = "#>",
	fig.width = 8,
	fig.height = 6
)
set.seed(123)
# Quiet down
lgr::get_logger("mlr3")$set_threshold("warn")
options("xplain.progress" = interactive())
```

```{r setup}
library(xplainfi)
library(mlr3)
library(mlr3learners)
library(data.table)
library(ggplot2)
```

`xplainfi` provides feature importance methods for machine learning models. It implements several approaches for measuring how much each feature contributes to model performance, with a focus on model-agnostic methods that work with any learner.

## Core Concepts

Feature importance methods in `xplainfi` address different but related questions:

- **How much does each feature contribute to model performance?** (Permutation Feature Importance)
- **What happens when we remove features and retrain?** (Leave-One-Covariate-Out)  
- **How do features depend on each other?** (Conditional and Relative methods)

All methods share a common interface built on [mlr3](https://mlr3.mlr-org.com/), making them easy to use with any task, learner, measure, and resampling strategy. 

The general pattern is to call `$compute()` to calculate importance (which _always re-computes_), then `$importance()` to retrieve the aggregated results, with intermediate results available in `$scores()` and, if the chosen measure supports it, `$obs_loss()`.

## Basic Example

Let's use the Friedman1 task to demonstrate feature importance methods with known ground truth:

```{r setup-problem}
task <- tgen("friedman1")$generate(n = 300)
learner <- lrn("regr.ranger", num.trees = 100)
measure <- msr("regr.mse")
resampling <- rsmp("cv", folds = 3)
```

The task has `r task$nrow` observations with `r length(task$feature_names)` features. Features `important1` through `important5` truly affect the target, while `unimportant1` through `unimportant5` are pure noise. We'll use a random forest learner with cross-validation for more stable estimates.

The target function is: $y = 10 \cdot \operatorname{sin}(\pi  x_1  x_2) + 20  (x_3 - 0.5)^2 + 10  x_4 + 5  x_5 + \epsilon$


## Permutation Feature Importance (PFI)

PFI is the most straightforward method: for each feature, we permute (shuffle) its values and measure how much model performance deteriorates. More important features cause larger performance drops when shuffled.

```{r pfi-basic}
pfi <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 10
)

pfi$compute()
pfi$importance()
```

The `importance` column shows the performance difference when each feature is permuted. Higher values indicate more important features.

For more stable estimates, we can use multiple permutation iterations per resampling fold with `n_repeats`, which is set to `30` by default for all methods.
Note that in this case "more is more", and while there is no clear "good enough" value, setting `n_repeats` to a small value like 1 will most definitely yield unreliable results.


```{r pfi-parameters}
pfi_stable <- PFI$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 50
)

pfi_stable$compute()
pfi_stable$importance()
```

To illustrate why this is important, we can take a look at the variability of PFI scores for feature `important2` within each resampling iteration using individual importance scores via `$score()` (see below):

```{r pif-nrepeats}
pfi_stable$scores()[feature == "important2", ] |>
	ggplot(aes(y = importance, x = factor(iter_rsmp))) +
	geom_boxplot() +
	labs(
		title = "PFI variability within resampling iterations",
		subtitle = "Setting n_repeats higher improves PFI estimates",
		y = "PFI score (important2)",
		x = "Resampling iteration (3-fold CV)"
	) +
	theme_minimal()
```

```{r pfi-scores-tmp, echo=FALSE}
pfi_important2_range = round(range(pfi_stable$scores()[feature == "important2", importance]), 2)
```

The aggregated importance score for this feature is approximately `r round(pfi_stable$importance()[feature == "important2", importance], 1)`, but across all resamplings the estimated PFI scores range from `r pfi_important2_range[1]` to `r pfi_important2_range[2]`, and with insufficient resampling or low `n_repeats`, we might have over- or underestimated the features PFI by some margin.


We can also use the ratio of performance scores instead of their difference for the importance calculation, meaning that an unimportant feature is now expected to get an importance score of 1 rather than 0:

```{r pfi-ratio}
pfi_stable$importance(relation = "ratio")
```

## Leave-One-Covariate-Out (LOCO)

LOCO measures importance by retraining the model without each feature and comparing performance to the full model.
This shows the contribution of each feature when all other features are present.

```{r loco-basic}
loco <- LOCO$new(
	task = task,
	learner = learner,
	measure = measure,
	resampling = resampling,
	n_repeats = 10
)

loco$compute()
loco$importance()
```

LOCO is computationally expensive as it requires retraining for each feature, but provides clear interpretation: higher values mean larger performance drop when the feature is removed.
However, it cannot distinguish between direct effects and indirect effects through correlated features.

## Feature Samplers

For advanced methods that account for feature dependencies, `xplainfi` provides different sampling strategies. 
While PFI uses simple permutation (marginal sampling), conditional samplers can preserve feature relationships.

Let's demonstrate conditional sampling using adversarial random forests (ARF), which preserves relationships between features when sampling:

```{r samplers-demo}
arf_sampler <- ConditionalARFSampler$new(task)

sample_data <- task$data(rows = 1:5)
sample_data[, .(important1, important2)]
```

Now we'll conditionally sample the `important1` feature given the values of `important2` and `important3`:

```{r conditional-sampling}
sampled_conditional <- arf_sampler$sample_newdata(
	feature = "important1",
	newdata = sample_data,
	conditioning_set = c("important2", "important3")
)

sample_data[, .(important1, important2, important3)]
sampled_conditional[, .(important1, important2, important3)]
```

This conditional sampling is essential for methods like CFI and RFI that need to preserve feature dependencies.
See the [perturbation-importance article](https://mlr-org.github.io/xplainfi/articles/perturbation-importance.html) for detailed comparisons and `vignette("feature-samplers")` for more details on implemented samplers.

## Detailed Scoring Information

All methods store detailed scoring information from each resampling iteration for further analysis. Let's examine the structure of PFI's detailed scores:

```{r detailed-scores}
pfi$scores() |>
	head(10) |>
	knitr::kable(digits = 4, caption = "Detailed PFI scores (first 10 rows)")
```

We can also summarize the scoring structure:

```{r scoring-summary}
pfi$scores()[, .(
	features = uniqueN(feature),
	resampling_folds = uniqueN(iter_rsmp),
	permutation_iters = uniqueN(iter_repeat),
	total_scores = .N
)]
```

So `$importance()` always gives us the aggregated importances across multiple resampling- and permutation-/refitting iterations, whereas `$scores()` gives you the individual scores as calculated by the supplied `measure` and the corresponding importance calculated from the difference of these scores by default.

Analogous to `$importance()`, you can also use `relation = "ratio"` here:

```{r detailed-scores-ratio}
pfi$scores(relation = "ratio") |>
	head(10) |>
	knitr::kable(digits = 4, caption = "PFI scores using the ratio (first 10 rows)")
```

## Observation-wise losses and importances

For methods where importances are calculated based on observation-level comparisons and with decomposable measures, we can also retrieve observation-level information with `$obs_loss()`, which works analogously to `$scores()` and `$importance()` but at an even more detailed level:

```{r pfi-obs-scores}
pfi$obs_loss()
```

Since we computed PFI using the mean squared error (`msr("regr.mse")`), we can use the associated `Measure$obs_loss()`, the squared error.  
In the resulting table we see 

- `loss_baseline`: The loss (squared error) for the baseline model before permutation
- `loss_post`: The loss for this observation after permutation (or in the case of `LOCO`, after refit)
- `obs_importance`: The difference (or ratio if `relation = "ratio"`) of the two losses

Note that not all measures have a `Measure$obs_loss()`:
Some measures like `msr("classif.auc")` are not decomposable, so observation-wise loss values are not available.  
In other cases, the corresponding `obs_loss()` is just not yet implemented in [`mlr3measures`](https://CRAN.R-project.org/package=mlr3measures), but will likely be in the future.

## Statistical Inference

All importance methods support confidence intervals and p-values via the `ci_method` argument in `$importance()`.
Available approaches range from empirical quantiles and corrected t-tests (Nadeau & Bengio) for resampling-based variability, to observation-wise inference methods like CPI/cARFi (for `CFI`) and Lei et al. (2018) (for `LOCO`).
Multiplicity correction via `p_adjust` is supported for all methods that produce p-values.

For a comprehensive guide covering all inference methods, see the [Inference for Feature Importance](https://mlr-org.github.io/xplainfi/articles/inference.html) article.

## Using Pre-trained Learners

By default, `xplainfi` trains the learner internally via `mlr3::resample()`.
However, if you have already trained a learner (for example because training is expensive or you want to explain a specific model) you can pass it directly to perturbation-based methods (`PFI`, `CFI`, `RFI`) and `SAGE` methods. Refit-based methods (`LOCO` / `WVIM`) require retraining by design and will warn if given a pretrained learner.
The only requirement is that the `resampling` must be instantiated and have exactly one iteration (i.e., a single test set).
This is necessary because a pre-trained learner corresponds to a single fitted model, and there is no meaningful way to associate it with multiple resampling folds.

A holdout resampling is the natural choice here.
We first train the learner on the train set and `PFI` will calculate importance using the trained learner and the corresponding test set defined by the `resampling`:

```{r pretrained-pfi}
resampling_holdout <- rsmp("holdout")$instantiate(task)
learner_trained <- lrn("regr.ranger", num.trees = 100)
learner_trained$train(task, row_ids = resampling_holdout$train_set(1))

pfi_pretrained <- PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_holdout,
	n_repeats = 10
)

pfi_pretrained$compute()
pfi_pretrained$importance()
```

A common real-world scenario is that the learner was trained on some dataset and you want to explain the model on entirely new, unseen data.
In that case, create a task from the new data (via `as_task_regr()` for example) and use `rsmp("custom")` to designate all rows as the test set.
The resampling here is purely a technicality used for internal consistency, and the train set is irrelevant since the learner is already trained.
A utility function `rsmp_all_test()` can be used as a shortcut to achieve the same goal.

```{r pretrained-custom}
# Simulate: learner was trained elsewhere, we have new data to use
new_data <- tgen("friedman1")$generate(n = 100)

# Same as rsmp_all_test(task)
resampling_custom <- rsmp("custom")$instantiate(
	new_data,
	train_sets = list(integer(0)),
	test_sets = list(new_data$row_ids)
)

pfi_newdata <- PFI$new(
	task = new_data,
	learner = learner_trained,
	measure = measure,
	resampling = resampling_custom,
	n_repeats = 10
)

pfi_newdata$compute()
pfi_newdata$importance()
```

If you pass a trained learner with a multi-fold or non-instantiated resampling, you will get an informative error at construction time:

```{r pretrained-error, error = TRUE}
PFI$new(
	task = task,
	learner = learner_trained,
	measure = measure,
	resampling = rsmp("cv", folds = 3)
)
```

## Parallelization

Both PFI/CFI/RFI and LOCO/WVIM support parallel execution to speed up computation when working with multiple features or expensive learners. The parallelization follows mlr3's approach, allowing users to choose between `mirai` and `future` backends.

### Example with future

The `future` package provides a simple interface for parallel and distributed computing:

```{r parallel-future, eval = FALSE}
library(future)
plan("multisession", workers = 2)

# PFI with parallelization across features
pfi_parallel = PFI$new(
	task,
	learner = lrn("regr.ranger"),
	measure = msr("regr.mse"),
	n_repeats = 10
)
pfi_parallel$compute()
pfi_parallel$importance()

# LOCO with parallelization (uses mlr3fselect internally)
loco_parallel = LOCO$new(
	task,
	learner = lrn("regr.ranger"),
	measure = msr("regr.mse")
)
loco_parallel$compute()
loco_parallel$importance()
```

### Example with mirai

The `mirai` package offers a modern alternative for parallel computing:

```{r parallel-mirai, eval = FALSE}
library(mirai)
daemons(n = 2)

# Same PFI/LOCO code works with mirai backend
pfi_parallel = PFI$new(
	task,
	learner = lrn("regr.ranger"),
	measure = msr("regr.mse"),
	n_repeats = 10
)
pfi_parallel$compute()
pfi_parallel$importance()

# Clean up daemons when done
daemons(0)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PerturbationImportance.R
\name{PerturbationImportance}
\alias{PerturbationImportance}
\title{Perturbation Feature Importance Base Class}
\description{
Abstract base class for perturbation-based importance methods PFI, CFI, and RFI
}
\section{Super class}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{PerturbationImportance}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{sampler}}{(\link{FeatureSampler}) Sampler object for feature perturbation}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PerturbationImportance-initialize}{\code{PerturbationImportance$new()}}
    \item \href{#method-PerturbationImportance-importance}{\code{PerturbationImportance$importance()}}
    \item \href{#method-PerturbationImportance-clone}{\code{PerturbationImportance$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="compute"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-compute'><code>FeatureImportanceMethod$compute()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PerturbationImportance-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PerturbationImportance-initialize}{}}}
\subsection{\code{PerturbationImportance$new()}}{
  Creates a new instance of the PerturbationImportance class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PerturbationImportance$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  sampler = NULL,
  relation = "difference",
  n_repeats = 30L,
  batch_size = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups}}{Passed to \link{FeatureImportanceMethod}.}
      \item{\code{sampler}}{(\link{FeatureSampler}) Sampler to use for feature perturbation.}
      \item{\code{relation}}{(\code{character(1)}: \code{"difference"}) How to relate perturbed and baseline scores. Can also be \code{"ratio"}.}
      \item{\code{n_repeats}}{(\code{integer(1)}: \code{30L}) Number of permutation/conditional sampling iterations. Can also be overridden in \verb{$compute()}.}
      \item{\code{batch_size}}{(\code{integer(1)} | \code{NULL}: \code{NULL}) Maximum number of rows to predict at once. When \code{NULL}, predicts all \code{test_size * n_repeats} rows in one call. Use smaller values to reduce memory usage at the cost of more prediction calls. Can be overridden in \verb{$compute()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PerturbationImportance-importance"></a>}}
\if{latex}{\out{\hypertarget{method-PerturbationImportance-importance}{}}}
\subsection{\code{PerturbationImportance$importance()}}{
  Get aggregated importance scores.
Extends the base \verb{$importance()} method to support \code{ci_method = "cpi"}.
For details, see \link{CFI}, which is the only sub-method for which it is known to be valid.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PerturbationImportance$importance(
  relation = NULL,
  standardize = FALSE,
  ci_method = c("none", "raw", "nadeau_bengio", "quantile", "cpi"),
  conf_level = 0.95,
  alternative = c("two.sided", "greater"),
  test = c("t", "wilcoxon", "fisher", "binomial"),
  B = 1999,
  p_adjust = "none",
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{relation}}{(\code{character(1)}) How to relate perturbed scores to originals ("difference" or "ratio"). If \code{NULL}, uses stored parameter value.}
      \item{\code{standardize}}{(\code{logical(1)}: \code{FALSE}) If \code{TRUE}, importances are standardized by the highest score so all scores fall in \verb{[-1, 1]}.}
      \item{\code{ci_method}}{(\code{character(1)}: \code{"none"}) Variance estimation method. In addition to base methods (\code{"none"}, \code{"raw"}, \code{"nadeau_bengio"}, \code{"quantile"}),
perturbation methods support \code{"cpi"} (Conditional Predictive Impact).
CPI is specifically designed for \link{CFI} with knockoff samplers and uses one-sided hypothesis tests.}
      \item{\code{conf_level}}{(\code{numeric(1)}: \code{0.95}) Confidence level for confidence intervals when \code{ci_method != "none"}.}
      \item{\code{alternative}}{(\code{character(1)}: \code{"two.sided"}) Type of alternative hypothesis for statistical tests.
\code{"greater"} tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
\code{"two.sided"} tests H0: importance = 0 vs H1: importance != 0.}
      \item{\code{test}}{(\code{character(1)}: \code{"t"}) Test to use for CPI. One of \code{"t"}, \code{"wilcoxon"}, \code{"fisher"}, or \code{"binomial"}. Only used when \code{ci_method = "cpi"}.}
      \item{\code{B}}{(\code{integer(1)}: \code{1999}) Number of replications for Fisher test. Only used when \code{ci_method = "cpi"} and \code{test = "fisher"}.}
      \item{\code{p_adjust}}{(\code{character(1)}: \code{"none"}) Method for p-value adjustment for multiple comparisons.
Accepts any method supported by \link[stats:p.adjust.methods]{stats::p.adjust.methods}, e.g. \code{"holm"}, \code{"bonferroni"}, \code{"BH"}, \code{"none"}.
When \code{"bonferroni"}, confidence intervals are also adjusted (alpha/k).
For other correction methods (e.g. \code{"holm"}, \code{"BH"}), only p-values are adjusted;
confidence intervals remain at the nominal \code{conf_level} because these sequential/adaptive
procedures do not have a clean per-comparison alpha for CI construction.}
      \item{\code{...}}{Additional arguments passed to the base method.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\link[data.table:data.table]{data.table}) Aggregated importance scores.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PerturbationImportance-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PerturbationImportance-clone}{}}}
\subsection{\code{PerturbationImportance$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PerturbationImportance$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-mlr3.R
\name{rsmp_all_test}
\alias{rsmp_all_test}
\title{Create a resampling with all data being test data}
\usage{
rsmp_all_test(task)
}
\arguments{
\item{task}{(\link[mlr3:Task]{mlr3::Task})}
}
\value{
\link[mlr3:Resampling]{mlr3::Resampling} with an empty \code{train_set} and a single \code{test_set} identical to all of the given \code{Task}.
}
\description{
Utility for use with a pretrained learner in importance methods which support it
}
\details{
Note that the resulting Resampling will have an
empty train set, making it useless for any
other purpose than the use with a pretrained learner.
}
\examples{
library(mlr3)
# Create custom task from some data.frame
custom_task <- as_task_regr(mtcars, target = "mpg")
# Create matching Resampling with all-test data
resampling_custom <- rsmp_all_test(custom_task)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WVIM.R
\name{WVIM}
\alias{WVIM}
\title{Williamson's Variable Importance Measure (WVIM)}
\description{
Base class generalizing refit-based variable importance measures.
Default corresponds to leaving out each feature \code{n_repeats = 1} times, which
corresponds to LOCO (Leave One Covariate Out).
}
\examples{
\dontshow{if (requireNamespace("ranger", quietly = TRUE) && requireNamespace("mlr3learners", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(mlr3)
library(mlr3learners)

task <- sim_dgp_correlated(n = 500)

# Group correlated features together, independent features separately
groups <- list(
  correlated = c("x1", "x2"),
  independent = c("x3", "x4")
)

wvim <- WVIM$new(
  task = task,
  learner = lrn("regr.ranger", num.trees = 10),
  groups = groups
)
wvim$compute()
wvim$importance()
\dontshow{\}) # examplesIf}
}
\references{
Lei J, G'Sell M, Rinaldo A, Tibshirani R, Wasserman L (2018).
\dQuote{Distribution-Free Predictive Inference for Regression.}
\emph{Journal of the American Statistical Association}, \bold{113}(523), 1094--1111.
\doi{10.1080/01621459.2017.1307116}.
}
\keyword{internal}
\section{Super class}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{WVIM}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{direction}}{(\code{character(1)}) Either "leave-out" or "leave-in".}

    \item{\code{design}}{(\code{logical()}) Feature selection design matrix where \code{TRUE} equals "left in" and \code{FALSE} "left out".
Columns correspond to \code{task$feature_names} and the number of rows corresponds to \code{length(features) * n_repeats}.
The base matrix is created by \link{wvim_design_matrix} and then replicated \code{n_repeats} times before.}

    \item{\code{instance}}{(\code{FSelectInstanceBatchSingleCrit}) The \code{mlr3fselect} feature selection instance containing
also the archive of all evaluations, possible useful for future use. Only stored if \code{store_instance} is \code{TRUE}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-WVIM-initialize}{\code{WVIM$new()}}
    \item \href{#method-WVIM-importance}{\code{WVIM$importance()}}
    \item \href{#method-WVIM-compute}{\code{WVIM$compute()}}
    \item \href{#method-WVIM-clone}{\code{WVIM$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WVIM-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-WVIM-initialize}{}}}
\subsection{\code{WVIM$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{WVIM$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  direction = c("leave-out", "leave-in"),
  label = "Williamson's Variable Importance Measure (WVIM)",
  n_repeats = 1L,
  batch_size = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups}}{Passed to \code{FeatureImportanceMethod} for construction.}
      \item{\code{direction}}{(\code{character(1)}) Either "leave-out" or "leave-in".}
      \item{\code{label}}{(\code{character(1)}) Method label.}
      \item{\code{n_repeats}}{(\code{integer(1)}: \code{1L}) Number of refit iterations per resampling iteration.
This may be useful for some stochastic learners, but in general it is advisable to increase resampling iterations instead of
repeatedly refitting on the same data.}
      \item{\code{batch_size}}{(\code{integer(1)}: \code{NULL}) Number of design points (refits) dispatched per internal
\code{mlr3::benchmark()} call by the \code{mlr3fselect::fs("design_points")} fselector. \code{NULL} (default) keeps
the bbotk default of one design point per call, i.e. sequential single-refit evaluation. Set to a
positive integer to batch refits so mlr3's \code{future}/\code{mirai} parallelization can spread them across
workers; a sensible value is the number of available workers/daemons.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WVIM-importance"></a>}}
\if{latex}{\out{\hypertarget{method-WVIM-importance}{}}}
\subsection{\code{WVIM$importance()}}{
  Get aggregated importance scores.
Extends the base \verb{$importance()} method to support \code{ci_method = "lei"}.

This implements distribution-free inference based on Lei et al. (2018),
testing observation-wise loss differences using the Wilcoxon signed-rank test by default.

Lei et al. (2018) proposed this method specifically for LOCO with L1 (absolute) loss,
median aggregation, and a single train/test split (holdout). The inference is
conditional on the training data, requiring i.i.d. test observations from a single
split. While the aggregation function, statistical test, and resampling strategy are
parameterizable, deviating from these defaults may invalidate the theoretical guarantees.

For a comprehensive overview of inference methods, see
\code{vignette("inference", package = "xplainfi")}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{WVIM$importance(
  relation = NULL,
  standardize = FALSE,
  ci_method = c("none", "raw", "nadeau_bengio", "quantile", "lei"),
  conf_level = 0.95,
  alternative = c("two.sided", "greater"),
  test = c("wilcoxon", "t", "fisher", "binomial"),
  B = 1999,
  aggregator = NULL,
  p_adjust = "none",
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{relation}}{(\code{character(1)}) How to relate perturbed scores to originals ("difference" or "ratio"). If \code{NULL}, uses stored parameter value.}
      \item{\code{standardize}}{(\code{logical(1)}: \code{FALSE}) If \code{TRUE}, importances are standardized by the highest score so all scores fall in \verb{[-1, 1]}.}
      \item{\code{ci_method}}{(\code{character(1)}: \code{"none"}) Variance estimation method. In addition to base methods (\code{"none"}, \code{"raw"}, \code{"nadeau_bengio"}, \code{"quantile"}),
WVIM methods support \code{"lei"} for distribution-free inference (Lei et al., 2018).}
      \item{\code{conf_level}}{(\code{numeric(1)}: \code{0.95}) Confidence level to use for confidence interval construction when \code{ci_method != "none"}.}
      \item{\code{alternative}}{(\code{character(1)}: \code{"two.sided"}) Type of alternative hypothesis for statistical tests.
\code{"greater"} tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
\code{"two.sided"} tests H0: importance = 0 vs H1: importance != 0.}
      \item{\code{test}}{(\code{character(1)}: \code{"wilcoxon"}) Test to use for Lei et al. inference. One of \code{"wilcoxon"}, \code{"t"}, \code{"fisher"}, or \code{"binomial"}.
Only used when \code{ci_method = "lei"}.}
      \item{\code{B}}{(\code{integer(1)}: \code{1999}) Number of replications for Fisher test. Only used when \code{ci_method = "lei"} and \code{test = "fisher"}.}
      \item{\code{aggregator}}{(\code{function}: \code{stats::median}) Aggregation function for computing the point estimate from observation-wise
importance values. Defaults to \code{stats::median} as proposed by Lei et al. (2018). Only used when \code{ci_method = "lei"}.}
      \item{\code{p_adjust}}{(\code{character(1)}: \code{"none"}) Method for p-value adjustment for multiple comparisons.
Accepts any method supported by \link[stats:p.adjust.methods]{stats::p.adjust.methods}, e.g. \code{"holm"}, \code{"bonferroni"}, \code{"BH"}, \code{"none"}.
When \code{"bonferroni"}, confidence intervals are also adjusted (alpha/k).
For other correction methods (e.g. \code{"holm"}, \code{"BH"}), only p-values are adjusted;
confidence intervals remain at the nominal \code{conf_level} because these sequential/adaptive
procedures do not have a clean per-comparison alpha for CI construction.}
      \item{\code{...}}{Additional arguments passed to the base method.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\link[data.table:data.table]{data.table}) Aggregated importance scores.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WVIM-compute"></a>}}
\if{latex}{\out{\hypertarget{method-WVIM-compute}{}}}
\subsection{\code{WVIM$compute()}}{
  Computes leave-out or leave-in feature importance.
\code{wvim_design_matrix(task$feature_names, "leave-out")} corresponds to LOCO.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{WVIM$compute(
  store_models = TRUE,
  store_backends = TRUE,
  store_instance = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{store_models, store_backends}}{(\code{logical(1)}: \code{TRUE}) Whether to store fitted models / data backends, passed to \link[mlr3:resample]{mlr3::resample} internally
backends in resample result.
Required for some measures, but may increase memory footprint.}
      \item{\code{store_instance}}{(\code{logical(1)}: \code{FALSE}) Whether to store the \link[mlr3fselect:mlr3fselect]{mlr3fselect} instance in \verb{$instance}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-WVIM-clone"></a>}}
\if{latex}{\out{\hypertarget{method-WVIM-clone}{}}}
\subsection{\code{WVIM$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{WVIM$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PerturbationImportance.R
\name{PFI}
\alias{PFI}
\title{Permutation Feature Importance}
\description{
Implementation of Permutation Feature Importance (PFI) using modular sampling approach.
PFI measures the importance of a feature by calculating the increase in model error
when the feature's values are randomly permuted, breaking the relationship between
the feature and the target variable.
}
\details{
Permutation Feature Importance was originally introduced by Breiman (2001) as part of
the Random Forest algorithm. The method works by:
\enumerate{
\item Computing baseline model performance on the original dataset
\item For each feature, randomly permuting its values while keeping other features unchanged
\item Computing model performance on the permuted dataset
\item Calculating importance as the difference (or ratio) between permuted and original performance
}
}
\examples{
library(mlr3)

task <- sim_dgp_correlated(n = 500)

pfi <- PFI$new(
  task = task,
  learner = lrn("regr.rpart"),
  measure = msr("regr.mse"),
n_repeats = 5
)
pfi$compute()
pfi$importance()
}
\references{
Breiman L (2001).
\dQuote{Random Forests.}
\emph{Machine Learning}, \bold{45}(1), 5--32.
\doi{10.1023/A:1010933404324}.
Fisher A, Rudin C, Dominici F (2019).
\dQuote{All Models Are Wrong, but Many Are Useful: Learning a Variable's Importance by Studying an Entire Class of Prediction Models Simultaneously.}
\emph{Journal of Machine Learning Research}, \bold{20}, 177.
\url{https://pmc.ncbi.nlm.nih.gov/articles/PMC8323609/}.
Strobl C, Boulesteix A, Kneib T, Augustin T, Zeileis A (2008).
\dQuote{Conditional Variable Importance for Random Forests.}
\emph{BMC Bioinformatics}, \bold{9}(1), 307.
\doi{10.1186/1471-2105-9-307}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{\link[xplainfi:PerturbationImportance]{PerturbationImportance}} -> \code{PFI}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-PFI-initialize}{\code{PFI$new()}}
    \item \href{#method-PFI-compute}{\code{PFI$compute()}}
    \item \href{#method-PFI-clone}{\code{PFI$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="PerturbationImportance" data-id="importance"><a href='../../xplainfi/html/PerturbationImportance.html#method-PerturbationImportance-importance'><code>PerturbationImportance$importance()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PFI-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-PFI-initialize}{}}}
\subsection{\code{PFI$new()}}{
  Creates a new instance of the PFI class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PFI$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  relation = "difference",
  n_repeats = 30L,
  batch_size = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups, relation, n_repeats, batch_size}}{Passed to \link{PerturbationImportance}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PFI-compute"></a>}}
\if{latex}{\out{\hypertarget{method-PFI-compute}{}}}
\subsection{\code{PFI$compute()}}{
  Compute PFI scores
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PFI$compute(
  n_repeats = NULL,
  batch_size = NULL,
  store_models = TRUE,
  store_backends = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n_repeats}}{(\code{integer(1)}; \code{NULL}) Number of permutation iterations. If \code{NULL}, uses stored value.}
      \item{\code{batch_size}}{(\code{integer(1)} | \code{NULL}: \code{NULL}) Maximum number of rows to predict at once. If \code{NULL}, uses stored value.}
      \item{\code{store_models, store_backends}}{(\code{logical(1)}: \code{TRUE}) Whether to store fitted models / data backends, passed to \link[mlr3:resample]{mlr3::resample} internally
for the initial fit of the learner.
This may be required for certain measures and is recommended to leave enabled unless really necessary.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PFI-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PFI-clone}{}}}
\subsection{\code{PFI$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{PFI$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KnockoffSampler.R
\name{KnockoffSampler}
\alias{KnockoffSampler}
\title{Knockoff Sampler}
\description{
Implements conditional sampling using Knockoffs.
}
\details{
The \code{KnockoffSampler} samples \link[knockoff:knockoff]{Knockoffs} based on the task data.
This class allows arbitrary \code{knockoff_fun}, which also means that no input checking
against supported feature types can be done. Use \link{KnockoffGaussianSampler} for the
Gaussian knockoff sampler for numeric features.
Alternative knockoff samplers include \code{knockoff_seq()} from the \code{seqknockoff} package
available on GitHub: \url{https://github.com/kormama1/seqknockoff}.

Knockoffs are related to the \code{ConditionalSampler} familty, with key differences:
They do not allow specifying a \code{conditioning_set}
}
\examples{
\dontshow{if (requireNamespace("knockoff", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(mlr3)
task = tgen("2dnormals")$generate(n = 100)
# Create sampler with default parameters
sampler = KnockoffSampler$new(task)
# Sample using row_ids from stored task
sampled_data = sampler$sample("x1")
\dontshow{\}) # examplesIf}
}
\references{
Watson D, Wright M (2021).
\dQuote{Testing Conditional Independence in Supervised Learning Algorithms.}
\emph{Machine Learning}, \bold{110}(8), 2107--2129.
\doi{10.1007/s10994-021-06030-6}.

Blesch K, Watson D, Wright M (2023).
\dQuote{Conditional Feature Importance for Mixed Data.}
\emph{AStA Advances in Statistical Analysis}, \bold{108}(2), 259--278.
\doi{10.1007/s10182-023-00477-9}.
}
\section{Super class}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{KnockoffSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{x_tilde}}{Knockoff matrix with one (or \code{iters}) row(s) per original observation in \code{task}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-KnockoffSampler-initialize}{\code{KnockoffSampler$new()}}
    \item \href{#method-KnockoffSampler-sample}{\code{KnockoffSampler$sample()}}
    \item \href{#method-KnockoffSampler-clone}{\code{KnockoffSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="sample_newdata"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-sample_newdata'><code>FeatureSampler$sample_newdata()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-KnockoffSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-KnockoffSampler-initialize}{}}}
\subsection{\code{KnockoffSampler$new()}}{
  Creates a new instance of the KnockoffSampler class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{KnockoffSampler$new(
  task,
  knockoff_fun = function(x) knockoff::create.second_order(as.matrix(x)),
  iters = 1
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from}
      \item{\code{knockoff_fun}}{(\code{function}) Function used to create knockoff matrix. Default are second-order Gaussian knockoffs (\code{knockoff::create.second_order()})}
      \item{\code{iters}}{(\code{integer(1)}: \code{1}) Number of repetitions the \code{knockoff_fun} is applied to create multiple \code{x_tilde}
instances per observation.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-KnockoffSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-KnockoffSampler-sample}{}}}
\subsection{\code{KnockoffSampler$sample()}}{
  Sample from stored task using knockoff values. Replaces specified feature(s) with
their knockoff counterparts from the pre-generated knockoff matrix.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{KnockoffSampler$sample(feature, row_ids = NULL, samples_per_row = 1L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature(s) to sample.}
      \item{\code{row_ids}}{(\code{integer()} | \code{NULL}) Row IDs to use. If \code{NULL}, uses all rows.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with knockoff feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-KnockoffSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-KnockoffSampler-clone}{}}}
\subsection{\code{KnockoffSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{KnockoffSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KnockoffSampler.R
\name{KnockoffGaussianSampler}
\alias{KnockoffGaussianSampler}
\title{Gaussian Knockoff Conditional Sampler}
\description{
A \link{KnockoffSampler} defaulting to second-order Gaussian knockoffs
as created by \link[knockoff:create.second_order]{knockoff::create.second_order}.
}
\details{
This is equivalent to \link{KnockoffSampler} using the default \code{knockoff_fun}.
}
\examples{
\dontshow{if (requireNamespace("knockoff", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(mlr3)
task = tgen("2dnormals")$generate(n = 100)
# Create sampler
sampler = KnockoffGaussianSampler$new(task)
# Sample using row_ids from stored task
sampled_data = sampler$sample("x1")
\dontshow{\}) # examplesIf}
}
\references{
Watson D, Wright M (2021).
\dQuote{Testing Conditional Independence in Supervised Learning Algorithms.}
\emph{Machine Learning}, \bold{110}(8), 2107--2129.
\doi{10.1007/s10994-021-06030-6}.

Blesch K, Watson D, Wright M (2023).
\dQuote{Conditional Feature Importance for Mixed Data.}
\emph{AStA Advances in Statistical Analysis}, \bold{108}(2), 259--278.
\doi{10.1007/s10182-023-00477-9}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:KnockoffSampler]{KnockoffSampler}} -> \code{KnockoffGaussianSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.
Will be checked against the provided \link[mlr3:Task]{mlr3::Task} to ensure compatibility.}

    \item{\code{x_tilde}}{Knockoff matrix}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-KnockoffGaussianSampler-initialize}{\code{KnockoffGaussianSampler$new()}}
    \item \href{#method-KnockoffGaussianSampler-clone}{\code{KnockoffGaussianSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="sample_newdata"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-sample_newdata'><code>FeatureSampler$sample_newdata()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="KnockoffSampler" data-id="sample"><a href='../../xplainfi/html/KnockoffSampler.html#method-KnockoffSampler-sample'><code>KnockoffSampler$sample()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-KnockoffGaussianSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-KnockoffGaussianSampler-initialize}{}}}
\subsection{\code{KnockoffGaussianSampler$new()}}{
  Creates a new instance using Gaussian knockoffs via \link[knockoff:create.second_order]{knockoff::create.second_order}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{KnockoffGaussianSampler$new(task, iters = 1)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
      \item{\code{iters}}{(\code{integer(1)}: \code{1}) Number of repetitions the \code{knockoff_fun} is applied to create multiple \code{x_tilde}
instances per observation.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-KnockoffGaussianSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-KnockoffGaussianSampler-clone}{}}}
\subsection{\code{KnockoffGaussianSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{KnockoffGaussianSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PerturbationImportance.R
\name{CFI}
\alias{CFI}
\title{Conditional Feature Importance}
\description{
Implementation of CFI using modular sampling approach
}
\details{
CFI replaces feature values with conditional samples from the distribution of
the feature given the other features. Any \link{ConditionalSampler} or \link{KnockoffSampler} can be used.
\subsection{Statistical Inference}{

Two approaches for statistical inference are primarily supported via
\verb{$importance(ci_method = "cpi")}:
\itemize{
\item \strong{CPI} (Watson & Wright, 2021): The original Conditional Predictive Impact method,
designed for use with knockoff samplers (\link{KnockoffGaussianSampler}).
\item \strong{cARFi} (Blesch et al., 2025): CFI with ARF-based conditional sampling
(\link{ConditionalARFSampler}), using the same CPI inference framework.
}

Both require a decomposable measure (e.g., MSE) and out-of-sample evaluation.
CPI inference is guaranteed to be valid with holdout (a single train/test split).
With cross-validation, test observations are i.i.d. but models are fit on
overlapping training data, which may affect inference coverage. With bootstrap
or subsampling, both non-i.i.d. test observations and overlapping training data
can be an issue. See \code{vignette("inference", package = "xplainfi")} for details.

Available tests: \code{"t"} (t-test), \code{"wilcoxon"} (signed-rank), \code{"fisher"} (permutation),
\code{"binomial"} (sign test). The Fisher test is recommended.

Method-agnostic inference methods (\code{"raw"}, \code{"nadeau_bengio"}, \code{"quantile"}) are also
available; see \link{FeatureImportanceMethod} for details.

For a comprehensive overview of inference methods including usage examples,
see \code{vignette("inference", package = "xplainfi")}.
}
}
\examples{
library(mlr3)

task <- sim_dgp_correlated(n = 200)

# Using default ConditionalARFSampler
cfi <- CFI$new(
  task = task,
  learner = lrn("regr.rpart"),
  measure = msr("regr.mse"),
  sampler = ConditionalGaussianSampler$new(task),
  n_repeats = 5
)
cfi$compute()
cfi$importance()
}
\references{
Watson D, Wright M (2021).
\dQuote{Testing Conditional Independence in Supervised Learning Algorithms.}
\emph{Machine Learning}, \bold{110}(8), 2107--2129.
\doi{10.1007/s10994-021-06030-6}.

Blesch K, Koenen N, Kapar J, Golchian P, Burk L, Loecher M, Wright M (2025).
\dQuote{Conditional Feature Importance with Generative Modeling Using Adversarial Random Forests.}
\emph{Proceedings of the AAAI Conference on Artificial Intelligence}, \bold{39}(15), 15596--15604.
\doi{10.1609/aaai.v39i15.33712}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{\link[xplainfi:PerturbationImportance]{PerturbationImportance}} -> \code{CFI}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-CFI-initialize}{\code{CFI$new()}}
    \item \href{#method-CFI-compute}{\code{CFI$compute()}}
    \item \href{#method-CFI-clone}{\code{CFI$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="PerturbationImportance" data-id="importance"><a href='../../xplainfi/html/PerturbationImportance.html#method-PerturbationImportance-importance'><code>PerturbationImportance$importance()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFI-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-CFI-initialize}{}}}
\subsection{\code{CFI$new()}}{
  Creates a new instance of the CFI class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{CFI$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  relation = "difference",
  n_repeats = 30L,
  batch_size = NULL,
  sampler = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups, relation, n_repeats, batch_size}}{Passed to \link{PerturbationImportance}.}
      \item{\code{sampler}}{(\link{ConditionalSampler}) Optional custom sampler. Defaults to instantiating \code{ConditionalARFSampler} internally with default parameters.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFI-compute"></a>}}
\if{latex}{\out{\hypertarget{method-CFI-compute}{}}}
\subsection{\code{CFI$compute()}}{
  Compute CFI scores
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{CFI$compute(
  n_repeats = NULL,
  batch_size = NULL,
  store_models = TRUE,
  store_backends = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n_repeats}}{(\code{integer(1)}) Number of permutation iterations. If \code{NULL}, uses stored value.}
      \item{\code{batch_size}}{(\code{integer(1)} | \code{NULL}: \code{NULL}) Maximum number of rows to predict at once. If \code{NULL}, uses stored value.}
      \item{\code{store_models, store_backends}}{(\code{logical(1)}: \code{TRUE}) Whether to store fitted models / data backends, passed to \link[mlr3:resample]{mlr3::resample} internally
for the initial fit of the learner.
This may be required for certain measures and is recommended to leave enabled unless really necessary.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-CFI-clone"></a>}}
\if{latex}{\out{\hypertarget{method-CFI-clone}{}}}
\subsection{\code{CFI$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{CFI$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bibentries.R
\name{print_bib}
\alias{print_bib}
\title{Print an Rd-formatted bib entry}
\usage{
print_bib(...)
}
\arguments{
\item{...}{(\code{character}) One or more quoted names of \code{bibentries} to print.}
}
\description{
Print an Rd-formatted bib entry
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xplainfi-package.R
\docType{package}
\name{xplainfi-package}
\alias{xplainfi}
\alias{xplainfi-package}
\title{xplainfi: Feature Importance Methods for Global Explanations}
\description{
Provides a consistent interface for common feature importance methods as described in Ewald et al. (2024) \doi{10.1007/978-3-031-63797-1_22}, including permutation feature importance (PFI), conditional and relative feature importance (CFI, RFI), leave one covariate out (LOCO), and Shapley additive global importance (SAGE), as well as feature sampling mechanisms to support conditional importance methods.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mlr-org.github.io/xplainfi/}
  \item \url{https://github.com/mlr-org/xplainfi}
  \item Report bugs at \url{https://github.com/mlr-org/xplainfi/issues}
}

}
\author{
\strong{Maintainer}: Lukas Burk \email{cran@lukasburk.de} (\href{https://orcid.org/0000-0001-7528-3795}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Lukas Burk \email{cran@lukasburk.de} (\href{https://orcid.org/0000-0001-7528-3795}{ORCID}) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PerturbationImportance.R
\name{RFI}
\alias{RFI}
\title{Relative Feature Importance}
\description{
RFI generalizes CFI and PFI with arbitrary conditioning sets and samplers.
}
\examples{
library(mlr3)
task = tgen("friedman1")$generate(n = 200)
rfi = RFI$new(
  task = task,
  learner = lrn("regr.rpart"),
  measure = msr("regr.mse"),
  conditioning_set = c("important1"),
  sampler = ConditionalGaussianSampler$new(task),
  n_repeats = 5
)
rfi$compute()
rfi$importance()
}
\references{
König G, Molnar C, Bischl B, Grosse-Wentrup M (2021).
\dQuote{Relative Feature Importance.}
In \emph{2020 25th International Conference on Pattern Recognition (ICPR)}, 9318--9325.
\doi{10.1109/ICPR48806.2021.9413090}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{\link[xplainfi:PerturbationImportance]{PerturbationImportance}} -> \code{RFI}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-RFI-initialize}{\code{RFI$new()}}
    \item \href{#method-RFI-compute}{\code{RFI$compute()}}
    \item \href{#method-RFI-clone}{\code{RFI$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="PerturbationImportance" data-id="importance"><a href='../../xplainfi/html/PerturbationImportance.html#method-PerturbationImportance-importance'><code>PerturbationImportance$importance()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RFI-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-RFI-initialize}{}}}
\subsection{\code{RFI$new()}}{
  Creates a new instance of the RFI class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{RFI$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  conditioning_set = NULL,
  relation = "difference",
  n_repeats = 30L,
  batch_size = NULL,
  sampler = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups, relation, n_repeats, batch_size}}{Passed to \link{PerturbationImportance}.}
      \item{\code{conditioning_set}}{(\code{\link[=character]{character()}}) Set of features to condition on. Can be overridden in \verb{$compute()}.
Default (\code{character(0)}) is equivalent to \code{PFI}. In \code{CFI}, this would be set to all features except that of interest.}
      \item{\code{sampler}}{(\link{ConditionalSampler}) Optional custom sampler. Defaults to \code{ConditionalARFSampler}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RFI-compute"></a>}}
\if{latex}{\out{\hypertarget{method-RFI-compute}{}}}
\subsection{\code{RFI$compute()}}{
  Compute RFI scores
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{RFI$compute(
  conditioning_set = NULL,
  n_repeats = NULL,
  batch_size = NULL,
  store_models = TRUE,
  store_backends = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{conditioning_set}}{(\code{character()}) Set of features to condition on. If \code{NULL}, uses the stored parameter value.}
      \item{\code{n_repeats}}{(\code{integer(1)}) Number of permutation iterations. If \code{NULL}, uses stored value.}
      \item{\code{batch_size}}{(\code{integer(1)} | \code{NULL}: \code{NULL}) Maximum number of rows to predict at once. If \code{NULL}, uses stored value.}
      \item{\code{store_models, store_backends}}{(\code{logical(1)}: \code{TRUE}) Whether to store fitted models / data backends, passed to \link[mlr3:resample]{mlr3::resample} internally
for the initial fit of the learner.
This may be required for certain measures and is recommended to leave enabled unless really necessary.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RFI-clone"></a>}}
\if{latex}{\out{\hypertarget{method-RFI-clone}{}}}
\subsection{\code{RFI$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{RFI$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SAGE.R
\name{SAGE}
\alias{SAGE}
\title{Shapley Additive Global Importance (SAGE) Base Class}
\description{
Base class for SAGE (Shapley Additive Global Importance)
feature importance based on Shapley values with marginalization.
This is an abstract class - use \link{MarginalSAGE} or \link{ConditionalSAGE}.
}
\details{
SAGE uses Shapley values to fairly distribute the total prediction
performance among all features. Unlike perturbation-based methods,
SAGE marginalizes features by integrating over their distribution.
This is approximated by averaging predictions over a reference dataset.

\strong{Standard Error Calculation}: The standard errors (SE) reported in
\verb{$convergence_history} reflect the uncertainty in Shapley value estimation
across different random permutations within a single resampling iteration.
These SEs quantify the Monte Carlo sampling error for a fixed trained model
and are only valid for inference about the importance of features for that
specific model. They do not capture broader uncertainty from model variability
across different train/test splits or resampling iterations.
}
\references{
Covert I, Lundberg S, Lee S (2020).
\dQuote{Understanding Global Feature Contributions With Additive Importance Measures.}
In \emph{Advances in Neural Information Processing Systems}, volume 33, 17212--17223.
\url{https://proceedings.neurips.cc/paper/2020/hash/c7bf0b7c1a86d5eb3be2c722cf2cf746-Abstract.html}.
}
\seealso{
\link{MarginalSAGE} \link{ConditionalSAGE}
}
\section{Super class}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{SAGE}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{n_permutations}}{(\code{integer(1)}) Number of permutations to sample.}

    \item{\code{convergence_history}}{(\code{\link[data.table:data.table]{data.table}}) History of SAGE values during computation.}

    \item{\code{converged}}{(\code{logical(1)}) Whether convergence was detected.}

    \item{\code{n_permutations_used}}{(\code{integer(1)}) Actual number of permutations used.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-SAGE-initialize}{\code{SAGE$new()}}
    \item \href{#method-SAGE-compute}{\code{SAGE$compute()}}
    \item \href{#method-SAGE-plot_convergence}{\code{SAGE$plot_convergence()}}
    \item \href{#method-SAGE-clone}{\code{SAGE$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="importance"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-importance'><code>FeatureImportanceMethod$importance()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SAGE-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-SAGE-initialize}{}}}
\subsection{\code{SAGE$new()}}{
  Creates a new instance of the SAGE class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{SAGE$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  n_permutations = 10L,
  batch_size = 5000L,
  n_samples = 100L,
  early_stopping = TRUE,
  se_threshold = 0.01,
  min_permutations = 10L,
  check_interval = 1L
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features}}{Passed to FeatureImportanceMethod.}
      \item{\code{n_permutations}}{(\code{integer(1)}: \code{10L}) Number of permutations to sample for SAGE value estimation.
The total number of evaluated coalitions is \code{1 (empty) + n_permutations * n_features}.}
      \item{\code{batch_size}}{(\code{integer(1)}: \code{5000L}) Maximum number of observations to process in a single prediction call.}
      \item{\code{n_samples}}{(\code{integer(1)}: \code{100L}) Number of samples to use for marginalizing out-of-coalition features.
For \link{MarginalSAGE}, this is the number of marginal data samples ("background data" in other implementations).
For \link{ConditionalSAGE}, this is the number of conditional samples per test instance retrieved from \code{sampler}.}
      \item{\code{early_stopping}}{(\code{logical(1)}: \code{TRUE}) Whether to enable early stopping based on convergence detection.}
      \item{\code{se_threshold}}{(\code{numeric(1)}: \code{0.01}) Convergence threshold for relative standard error.
Convergence is detected when the maximum relative SE across all features falls below this threshold.
Relative SE is calculated as SE divided by the range of importance values (max - min),
making it scale-invariant across different loss metrics.
Default of \code{0.01} means convergence when relative SE is below 1\% of the importance range.}
      \item{\code{min_permutations}}{(\code{integer(1)}: \code{10L}) Minimum permutations before checking for convergence. Convergence is judged based on the standard errors of the estimated SAGE values,
which requires a sufficiently large number of samples (i.e., evaluated coalitions).}
      \item{\code{check_interval}}{(\code{integer(1)}: \code{1L}) Check convergence every N permutations.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SAGE-compute"></a>}}
\if{latex}{\out{\hypertarget{method-SAGE-compute}{}}}
\subsection{\code{SAGE$compute()}}{
  Compute SAGE values.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{SAGE$compute(
  store_backends = TRUE,
  batch_size = NULL,
  early_stopping = NULL,
  se_threshold = NULL,
  min_permutations = NULL,
  check_interval = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{store_backends}}{(\code{logical(1)}) Whether to store data backends.}
      \item{\code{batch_size}}{(\code{integer(1)}: \code{5000L}) Maximum number of observations to process in a single prediction call.}
      \item{\code{early_stopping}}{(\code{logical(1)}: \code{TRUE}) Whether to check for convergence and stop early.}
      \item{\code{se_threshold}}{(\code{numeric(1)}: \code{0.01}) Convergence threshold for relative standard error.
SE is normalized by the range of importance values (max - min) to make convergence
detection scale-invariant. Default \code{0.01} means convergence when relative SE < 1\%.}
      \item{\code{min_permutations}}{(\code{integer(1)}: \code{10L}) Minimum permutations before checking convergence.}
      \item{\code{check_interval}}{(\code{integer(1)}: \code{1L}) Check convergence every N permutations.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SAGE-plot_convergence"></a>}}
\if{latex}{\out{\hypertarget{method-SAGE-plot_convergence}{}}}
\subsection{\code{SAGE$plot_convergence()}}{
  Plot convergence history of SAGE values.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{SAGE$plot_convergence(features = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{features}}{(\code{character} | \code{NULL}) Features to plot. If NULL, plots all features.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    A \link[ggplot2:ggplot]{ggplot2} object
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SAGE-clone"></a>}}
\if{latex}{\out{\hypertarget{method-SAGE-clone}{}}}
\subsection{\code{SAGE$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{SAGE$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SAGE-conditional.R
\name{ConditionalSAGE}
\alias{ConditionalSAGE}
\title{Conditional SAGE}
\description{
\link{SAGE} with conditional sampling (features are "marginalized" conditionally).
Uses \link{ConditionalARFSampler} as default \link{ConditionalSampler}.
}
\examples{
\dontshow{if (requireNamespace("ranger", quietly = TRUE) && requireNamespace("mlr3learners", quietly = TRUE) && requireNamespace("arf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(mlr3)
task = tgen("friedman1")$generate(n = 100)

\donttest{
# Using default ConditionalARFSampler (also handles all mixed data)
sage = ConditionalSAGE$new(
  task = task,
  learner = lrn("regr.ranger", num.trees = 50),
  measure = msr("regr.mse"),
  n_permutations = 3L,
  n_samples = 20
)
sage$compute()
}
\donttest{
# For alternative conditional samplers:
custom_sampler = ConditionalGaussianSampler$new(
  task = task
)
sage_custom = ConditionalSAGE$new(
  task = task,
  learner = lrn("regr.ranger", num.trees = 50),
  measure = msr("regr.mse"),
  n_permutations = 5L,
  n_samples = 20,
  sampler = custom_sampler
)
sage_custom$compute()
}
\dontshow{\}) # examplesIf}
}
\seealso{
\link{MarginalSAGE}
}
\section{Super classes}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{\link[xplainfi:SAGE]{SAGE}} -> \code{ConditionalSAGE}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{sampler}}{(\link{ConditionalSampler}) Sampler for conditional marginalization.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalSAGE-initialize}{\code{ConditionalSAGE$new()}}
    \item \href{#method-ConditionalSAGE-clone}{\code{ConditionalSAGE$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="importance"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-importance'><code>FeatureImportanceMethod$importance()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="SAGE" data-id="compute"><a href='../../xplainfi/html/SAGE.html#method-SAGE-compute'><code>SAGE$compute()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="SAGE" data-id="plot_convergence"><a href='../../xplainfi/html/SAGE.html#method-SAGE-plot_convergence'><code>SAGE$plot_convergence()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSAGE-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSAGE-initialize}{}}}
\subsection{\code{ConditionalSAGE$new()}}{
  Creates a new instance of the ConditionalSAGE class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSAGE$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  n_permutations = 10L,
  sampler = NULL,
  batch_size = 5000L,
  n_samples = 100L,
  early_stopping = FALSE,
  se_threshold = 0.01,
  min_permutations = 10L,
  check_interval = 1L
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, n_permutations, batch_size, n_samples, early_stopping, se_threshold, min_permutations, check_interval}}{Passed to \link{SAGE}.}
      \item{\code{sampler}}{(\link{ConditionalSampler}) Optional custom sampler. Defaults to \link{ConditionalARFSampler}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSAGE-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSAGE-clone}{}}}
\subsection{\code{ConditionalSAGE$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSAGE$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-sim.R
\name{sim_dgp_scenarios}
\alias{sim_dgp_scenarios}
\alias{sim_dgp_correlated}
\alias{sim_dgp_mediated}
\alias{sim_dgp_confounded}
\alias{sim_dgp_interactions}
\alias{sim_dgp_independent}
\title{Simulation DGPs for Feature Importance Method Comparison}
\usage{
sim_dgp_correlated(n = 500L, r = 0.9)

sim_dgp_mediated(n = 500L)

sim_dgp_confounded(n = 500L, hidden = TRUE)

sim_dgp_interactions(n = 500L)

sim_dgp_independent(n = 500L)
}
\arguments{
\item{n}{(\code{integer(1)}: \code{500L}) Number of observations to generate.}

\item{r}{(\code{numeric(1)}: \code{0.9}) Correlation between x1 and x2. Must be between -1 and 1.}

\item{hidden}{(\code{logical(1)}: \code{TRUE}) Whether to hide the confounder from the returned task.
If \code{FALSE}, the confounder is included as a feature, allowing direct adjustment.
If \code{TRUE} (default), only the proxy is available, simulating unmeasured confounding.}
}
\value{
A regression task (\link[mlr3:TaskRegr]{mlr3::TaskRegr}) with \link[data.table:data.table]{data.table} backend.
}
\description{
These data generating processes (DGPs) are designed to illustrate specific
strengths and weaknesses of different feature importance methods like PFI, CFI, and RFI.
Each DGP focuses on one primary challenge to make the differences between methods clear.
}
\details{
\strong{Correlated Features DGP:}
This DGP creates highly correlated predictors where PFI will show artificially low
importance due to redundancy, while CFI will correctly identify each feature's
conditional contribution.

\strong{Mathematical Model:}
\deqn{(X_1, X_2)^T \sim \text{MVN}(0, \Sigma)}
where \eqn{\Sigma} is a \eqn{2 \times 2} covariance matrix with 1 on the diagonal and correlation \eqn{r} on the off-diagonal.
\deqn{X_3 \sim N(0,1), \quad X_4 \sim N(0,1)}
\deqn{Y = 2 \cdot X_1 + X_3 + \varepsilon}
where \eqn{\varepsilon \sim N(0, 0.2^2)}.

\strong{Feature Properties:}
\itemize{
\item \code{x1}: Standard normal from MVN, direct causal effect on y (\eqn{\beta = 2.0})
\item \code{x2}: Correlated with \code{x1} (correlation = \code{r}), NO causal effect on y (\eqn{\beta = 0})
\item \code{x3}: Independent standard normal, direct causal effect on y (\eqn{\beta = 1.0})
\item \code{x4}: Independent standard normal, no effect on y (\eqn{\beta = 0})
}

\strong{Expected Behavior:}
\itemize{
\item Will depend on the used learner and the strength of correlation (\code{r})
\item \strong{Marginal methods} (PFI, Marginal SAGE): Should falsely assign importance to x2 due to correlation with x1
\item \strong{CFI} Should correctly assign near-zero importance to x2
\item x2 is a "spurious predictor" - correlated with causal feature but not causal itself
}

\strong{Mediated Effects DGP:}
This DGP demonstrates the difference between total and direct causal effects.
Some features affect the outcome only through mediators.

\strong{Mathematical Model:}
\deqn{\text{exposure} \sim N(0,1), \quad \text{direct} \sim N(0,1)}
\deqn{\text{mediator} = 0.8 \cdot \text{exposure} + 0.6 \cdot \text{direct} + \varepsilon_m}
\deqn{Y = 1.5 \cdot \text{mediator} + 0.5 \cdot \text{direct} + \varepsilon}
where \eqn{\varepsilon_m \sim N(0, 0.3^2)} and \eqn{\varepsilon \sim N(0, 0.2^2)}.

\strong{Feature Properties:}
\itemize{
\item \code{exposure}: Has no direct effect on y, only through mediator (total effect = 1.2)
\item \code{mediator}: Mediates the effect of exposure on y
\item \code{direct}: Has both direct effect on y and effect on mediator
\item \code{noise}: No causal relationship to y
}

\strong{Causal Structure:} exposure -> mediator -> y <- direct -> mediator

\strong{Confounding DGP:}
This DGP includes a confounder that affects both a feature and the outcome.
Uses simple coefficients for easy interpretation.

\strong{Mathematical Model:}
\deqn{H \sim N(0,1)}
\deqn{X_1 = H + \varepsilon_1}
\deqn{\text{proxy} = H + \varepsilon_p, \quad \text{independent} \sim N(0,1)}
\deqn{Y = H + X_1 + \text{independent} + \varepsilon}
where all \eqn{\varepsilon \sim N(0, 0.5^2)} independently.

\strong{Model Structure:}
\itemize{
\item Confounder H ~ N(0,1) (potentially unobserved)
\item x1 = H + noise (affected by confounder)
\item proxy = H + noise (noisy measurement of confounder)
\item independent ~ N(0,1) (truly independent)
\item y = H + x1 + independent + noise
}

\strong{Expected Behavior:}
\itemize{
\item \strong{PFI}: Will show inflated importance for x1 due to confounding
\item \strong{CFI}: Should partially account for confounding through conditional sampling and reduce its importance
\item \strong{RFI conditioning on proxy}: Should reduce confounding bias by conditioning on proxy
}

\strong{Interaction Effects DGP:}
This DGP demonstrates a pure interaction effect where features have no main effects.

\strong{Mathematical Model:}
\deqn{Y = 2 \cdot X_1 \cdot X_2 + X_3 + \varepsilon}
where \eqn{X_j \sim N(0,1)} independently and \eqn{\varepsilon \sim N(0, 0.5^2)}.

\strong{Feature Properties:}
\itemize{
\item \code{x1}, \code{x2}: Independent features with ONLY interaction effect (no main effects)
\item \code{x3}: Independent feature with main effect only
\item \code{noise1}, \code{noise2}: No causal effects
}

\strong{Expected Behavior:}
\itemize{
\item Will depend on the used learner and its ability to model interactions
}

\strong{Independent Features DGP:}
This is a baseline scenario where all features are independent and their
effects are additive. All importance methods should give similar results.

\strong{Mathematical Model:}
\deqn{Y = 2.0 \cdot X_1 + 1.0 \cdot X_2 + 0.5 \cdot X_3 + \varepsilon}
where \eqn{X_j \sim N(0,1)} independently and \eqn{\varepsilon \sim N(0, 0.2^2)}.

\strong{Feature Properties:}
\itemize{
\item \code{important1-3}: Independent features with different effect sizes
\item \code{unimportant1-2}: Independent noise features with no effect
}

\strong{Expected Behavior:}
\itemize{
\item \strong{All methods}: Should rank features consistently by their true effect sizes
\item \strong{Ground truth}: important1 > important2 > important3 > unimportant1,2 (approximately 0)
}
}
\section{Functions}{
\itemize{
\item \code{sim_dgp_correlated()}: Correlated features demonstrating PFI's limitations

\item \code{sim_dgp_mediated()}: Mediated effects showing direct vs total importance

\item \code{sim_dgp_confounded()}: Confounding scenario for conditional sampling

\item \code{sim_dgp_interactions()}: Interaction effects between features

\item \code{sim_dgp_independent()}: Independent features baseline scenario

}}
\examples{
task = sim_dgp_correlated(200)
task$data()

# With different correlation
task_high_cor = sim_dgp_correlated(200, r = 0.95)
cor(task_high_cor$data()$x1, task_high_cor$data()$x2)
task = sim_dgp_mediated(200)
task$data()
# Hidden confounder scenario (traditional)
task_hidden = sim_dgp_confounded(200, hidden = TRUE)
task_hidden$feature_names  # proxy available but not confounder

# Observable confounder scenario
task_observed = sim_dgp_confounded(200, hidden = FALSE)
task_observed$feature_names  # both confounder and proxy available
task = sim_dgp_interactions(200)
task$data()
task = sim_dgp_independent(200)
task$data()
}
\references{
Ewald F, Bothmann L, Wright M, Bischl B, Casalicchio G, König G (2024).
\dQuote{A Guide to Feature Importance Methods for Scientific Inference.}
In Longo L, Lapuschkin S, Seifert C (eds.), \emph{Explainable Artificial Intelligence}, 440--464.
ISBN 978-3-031-63797-1.
\doi{10.1007/978-3-031-63797-1_22}.
}
\seealso{
Other simulation:
\code{\link[=sim_dgp_ewald]{sim_dgp_ewald()}}
}
\concept{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConditionalARFSampler.R
\name{ConditionalARFSampler}
\alias{ConditionalARFSampler}
\title{ARF-based Conditional Sampler}
\description{
Implements conditional sampling using Adversarial Random Forests (ARF).
ARF can handle mixed data types (continuous and categorical) and provides
flexible conditional sampling by modeling the joint distribution.
}
\details{
The ConditionalARFSampler fits an \link[arf:arf]{Adversarial Random Forest} model on the task data,
then uses it to generate samples from \eqn{P(X_j | X_{-j})} where \eqn{X_j} is the
feature of interest and \eqn{X_{-j}} are the conditioning features.
}
\examples{
\dontshow{if (requireNamespace("arf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(mlr3)
task = tgen("2dnormals")$generate(n = 100)
# Create sampler with default parameters
sampler = ConditionalARFSampler$new(task, conditioning_set = "x2", verbose = FALSE)
# Sample using row_ids from stored task
sampled_data = sampler$sample("x1", row_ids = 1:10)
# Or use external data
data = task$data()
sampled_data_ext = sampler$sample_newdata("x1", newdata = data, conditioning_set = "x2")

# Example with custom ARF parameters
sampler_custom = ConditionalARFSampler$new(
  task,
  min_node_size = 10L,
  finite_bounds = "local",
  verbose = FALSE
)
sampled_custom = sampler_custom$sample("x1", conditioning_set = "x2")
\dontshow{\}) # examplesIf}
}
\references{
Watson D, Blesch K, Kapar J, Wright M (2023).
\dQuote{Adversarial Random Forests for Density Estimation and Generative Modeling.}
In \emph{Proceedings of The 26th International Conference on Artificial Intelligence and Statistics}, 5357--5375.
\url{https://proceedings.mlr.press/v206/watson23a.html}.

Blesch K, Koenen N, Kapar J, Golchian P, Burk L, Loecher M, Wright M (2025).
\dQuote{Conditional Feature Importance with Generative Modeling Using Adversarial Random Forests.}
\emph{Proceedings of the AAAI Conference on Artificial Intelligence}, \bold{39}(15), 15596--15604.
\doi{10.1609/aaai.v39i15.33712}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:ConditionalSampler]{ConditionalSampler}} -> \code{ConditionalARFSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.
Will be checked against the provided \link[mlr3:Task]{mlr3::Task} to ensure compatibility.}

    \item{\code{arf_model}}{Adversarial Random Forest model created by \link[arf:adversarial_rf]{arf::adversarial_rf}.}

    \item{\code{psi}}{Distribution parameters estimated from by \link[arf:forde]{arf::forde}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalARFSampler-initialize}{\code{ConditionalARFSampler$new()}}
    \item \href{#method-ConditionalARFSampler-sample}{\code{ConditionalARFSampler$sample()}}
    \item \href{#method-ConditionalARFSampler-sample_newdata}{\code{ConditionalARFSampler$sample_newdata()}}
    \item \href{#method-ConditionalARFSampler-clone}{\code{ConditionalARFSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalARFSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalARFSampler-initialize}{}}}
\subsection{\code{ConditionalARFSampler$new()}}{
  Creates a new instance of the ConditionalARFSampler class.
To fit the ARF in parallel, register a parallel backend first (see \link[arf:arf]{arf::arf}) and set \code{parallel = TRUE}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalARFSampler$new(
  task,
  conditioning_set = NULL,
  num_trees = 10L,
  min_node_size = 20L,
  finite_bounds = "no",
  epsilon = 1e-15,
  round = TRUE,
  stepsize = 0,
  verbose = FALSE,
  parallel = FALSE,
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Default conditioning set to use in \verb{$sample()}. This parameter only affects the sampling behavior, not the ARF model fitting.}
      \item{\code{num_trees}}{(\code{integer(1)}: \code{10L}) Number of trees for ARF. Passed to \link[arf:adversarial_rf]{arf::adversarial_rf}.}
      \item{\code{min_node_size}}{(\code{integer(1)}: \code{20L}) Minimum node size for ARF. Passed to \link[arf:adversarial_rf]{arf::adversarial_rf} and in turn to \link[ranger:ranger]{ranger::ranger}.
This is increased to 20 to mitigate overfitting.}
      \item{\code{finite_bounds}}{(\code{character(1)}: \code{"no"}) How to handle variable bounds. Passed to \link[arf:forde]{arf::forde}. Default is \code{"no"} for compatibility. \code{"local"} may improve extrapolation but can cause issues with some data.}
      \item{\code{epsilon}}{(\code{numeric(1)}: \code{0}) Slack parameter for when \code{finite_bounds != "no"}. Passed to \link[arf:forde]{arf::forde}.}
      \item{\code{round}}{(\code{logical(1)}: \code{TRUE}) Whether to round continuous variables back to their original precision in sampling. Can be overridden in \verb{$sample()} calls.}
      \item{\code{stepsize}}{(\code{numeric(1)}: \code{0}) Number of rows of evidence to process at a time when \code{parallel} is \code{TRUE}. Default (\code{0}) spreads evidence evenly over registered workers. Can be overridden in \verb{$sample()} calls.}
      \item{\code{verbose}}{(\code{logical(1)}: \code{FALSE}) Whether to print progress messages. Default is \code{FALSE} (arf's default is \code{TRUE}). Can be overridden in \verb{$sample()} calls.}
      \item{\code{parallel}}{(\code{logical(1)}: \code{FALSE}) Whether to use parallel processing via \code{foreach}. See examples in \code{\link[arf:forge]{arf::forge()}}. Can be overridden in \verb{$sample()} calls.}
      \item{\code{...}}{Additional arguments passed to \link[arf:adversarial_rf]{arf::adversarial_rf}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalARFSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalARFSampler-sample}{}}}
\subsection{\code{ConditionalARFSampler$sample()}}{
  Sample from stored task. Parameters use hierarchical resolution:
function argument > stored \code{param_set} value > hard-coded default.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalARFSampler$sample(
  feature,
  row_ids = NULL,
  conditioning_set = NULL,
  samples_per_row = 1L,
  round = NULL,
  stepsize = NULL,
  verbose = NULL,
  parallel = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature(s) to sample.}
      \item{\code{row_ids}}{(\code{integer()} | \code{NULL}) Row IDs to use. If \code{NULL}, uses all rows.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Features to condition on.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{round}}{(\code{logical(1)} | \code{NULL}) Round continuous variables.}
      \item{\code{stepsize}}{(\code{numeric(1)} | \code{NULL}) Batch size for parallel processing.}
      \item{\code{verbose}}{(\code{logical(1)} | \code{NULL}) Print progress messages.}
      \item{\code{parallel}}{(\code{logical(1)} | \code{NULL}) Use parallel processing.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalARFSampler-sample_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalARFSampler-sample_newdata}{}}}
\subsection{\code{ConditionalARFSampler$sample_newdata()}}{
  Sample from external data. See \verb{$sample()} for parameter details.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalARFSampler$sample_newdata(
  feature,
  newdata,
  conditioning_set = NULL,
  samples_per_row = 1L,
  round = NULL,
  stepsize = NULL,
  verbose = NULL,
  parallel = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature(s) to sample.}
      \item{\code{newdata}}{(\code{\link[data.table:data.table]{data.table}}) External data to use.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Features to condition on.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{round}}{(\code{logical(1)} | \code{NULL}) Round continuous variables.}
      \item{\code{stepsize}}{(\code{numeric(1)} | \code{NULL}) Batch size for parallel processing.}
      \item{\code{verbose}}{(\code{logical(1)} | \code{NULL}) Print progress messages.}
      \item{\code{parallel}}{(\code{logical(1)} | \code{NULL}) Use parallel processing.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalARFSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalARFSampler-clone}{}}}
\subsection{\code{ConditionalARFSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalARFSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/op-null-default.R
\name{op-null-default}
\alias{op-null-default}
\alias{\%||\%}
\title{Default value for \code{NULL}}
\usage{
x \%||\% y
}
\arguments{
\item{x, y}{If \code{x} is NULL or length 0, will return \code{y}; otherwise returns \code{x}.}
}
\description{
A backport of \verb{\%||\%} available in R versions from 4.4.0.
}
\examples{
1 \%||\% 2
NULL \%||\% 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-sim.R
\name{sim_dgp_ewald}
\alias{sim_dgp_ewald}
\title{Simulate data as in Ewald et al. (2024)}
\usage{
sim_dgp_ewald(n = 500)
}
\arguments{
\item{n}{(\code{integer(1)}) Number of samples to create.}
}
\value{
A regression task (\link[mlr3:TaskRegr]{mlr3::TaskRegr}) with \link[data.table:data.table]{data.table} backend.
}
\description{
Reproduces the data generating process from Ewald et al. (2024) for benchmarking
feature importance methods. Includes correlated features and interaction effects.
}
\details{
\strong{Mathematical Model:}
\deqn{X_1, X_3, X_5 \sim \text{Uniform}(0,1)}
\deqn{X_2 = X_1 + \varepsilon_2, \quad \varepsilon_2 \sim N(0, \sqrt{0.001})}
\deqn{X_4 = X_3 + \varepsilon_4, \quad \varepsilon_4 \sim N(0, \sqrt{0.1})}
\deqn{Y = X_4 + X_5 + X_4 \cdot X_5 + \varepsilon, \quad \varepsilon \sim N(0, \sqrt{0.1})}

\strong{Feature Properties:}
\itemize{
\item X1, X3, X5: Independent uniform(0,1) distributions
\item X2: Nearly perfect copy of X1 (correlation approximately 0.99)
\item X4: Noisy copy of X3 (correlation approximately 0.94)
\item Y depends on X4, X5, and their interaction
}
}
\examples{
sim_dgp_ewald(100)

}
\references{
Ewald F, Bothmann L, Wright M, Bischl B, Casalicchio G, König G (2024).
\dQuote{A Guide to Feature Importance Methods for Scientific Inference.}
In Longo L, Lapuschkin S, Seifert C (eds.), \emph{Explainable Artificial Intelligence}, 440--464.
ISBN 978-3-031-63797-1.
\doi{10.1007/978-3-031-63797-1_22}.
}
\seealso{
Other simulation:
\code{\link{sim_dgp_scenarios}}
}
\concept{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wvim-design-matrix.R
\name{wvim_design_matrix}
\alias{wvim_design_matrix}
\title{Create Feature Selection Design Matrix}
\usage{
wvim_design_matrix(
  all_features,
  feature_names = all_features,
  direction = c("leave-out", "leave-in")
)
}
\arguments{
\item{all_features}{(character()) All available feature names from the task.}

\item{feature_names}{(character() | list of character()) Features or feature
groups to evaluate. Can be a vector for individual features or a named list
for grouped features. Defaults to \code{all_features} if unspecified.}

\item{direction}{(character(1)) Either \code{"leave-in"} or \code{"leave-out"} (default).
Controls which features are selected in the design matrix.
\code{"leave-out"} sets features of interest to \code{FALSE}, and \code{"leave-in"} analogously
sets them to \code{TRUE}.}
}
\value{
data.table with logical columns for each feature in \code{all_features} and
\code{length(feature_names)} rows, one for each entry in \code{feature_names}
}
\description{
Creates a logical design matrix for leave-in or leave-out feature evaluation.
Used internally with mlr3fselect to evaluate feature subsets.
}
\examples{
task = mlr3::tsk("mtcars")

# Individual features
feature_names = task$feature_names[1:3]
wvim_design_matrix(task$feature_names, feature_names, "leave-in")
wvim_design_matrix(task$feature_names, feature_names, "leave-out")

# Feature groups
feature_groups = list(
  A = task$feature_names[1:2],
  B = task$feature_names[3:5]
)
wvim_design_matrix(task$feature_names, feature_groups, "leave-out")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-opt.R
\name{xplain_opt}
\alias{xplain_opt}
\title{xplainfi Package Options}
\usage{
xplain_opt(...)
}
\arguments{
\item{...}{Option names to retrieve (as character strings) or options to set (as named arguments).
\itemize{
\item To \strong{get} an option: \code{xplain_opt("verbose")} returns the current value
\item To \strong{set} an option: \code{xplain_opt(verbose = FALSE)} sets the value
\item To \strong{get all} options: \code{xplain_opt()} returns a named list of all options
}}
}
\value{
\itemize{
\item When \strong{getting} a single option: the option value (logical)
\item When \strong{getting} multiple options: a named list of option values
\item When \strong{setting} options: the previous values (invisibly)
}
}
\description{
Get or set package-level options for xplainfi.
}
\details{
Options can be set in three ways (in order of precedence):
\enumerate{
\item Using \code{xplain_opt(option_name = value)} (recommended)
\item Using \code{options("xplain.option_name" = value)}
\item Using environment variables \code{XPLAIN_OPTION_NAME=value}
}
\subsection{Available Options}{\tabular{lll}{
   Option \tab Default \tab Description \cr
   \code{verbose} \tab \code{TRUE} \tab Show informational messages (e.g., when using default measure or resampling) \cr
   \code{progress} \tab \code{FALSE} \tab Show progress bars during computation \cr
   \code{sequential} \tab \code{FALSE} \tab Force sequential execution (disable parallelization) \cr
   \code{debug} \tab \code{FALSE} \tab Enable debug output for development and troubleshooting \cr
}

}
}
\examples{
# Get current value of an option
xplain_opt("verbose")

# Get all options
xplain_opt()

# Set an option (returns previous value invisibly)
old <- xplain_opt(verbose = FALSE)
xplain_opt("verbose")  # Now FALSE

# Restore previous value
xplain_opt(verbose = old$verbose)

# Temporary option change with withr
if (requireNamespace("withr", quietly = TRUE)) {
  withr::with_options(
    list("xplain.verbose" = FALSE),
    {
      # Code here runs with verbose = FALSE
      xplain_opt("verbose")
    }
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConditionalGaussianSampler.R
\name{ConditionalGaussianSampler}
\alias{ConditionalGaussianSampler}
\title{Gaussian Conditional Sampler}
\description{
Implements conditional sampling assuming features follow a multivariate
Gaussian distribution. Computes conditional distributions analytically using standard
formulas for multivariate normal distributions.
}
\details{
For a joint Gaussian distribution \eqn{X \sim N(\mu, \Sigma)}, partitioned as
\eqn{X = (X_A, X_B)}, the conditional distribution is:

\deqn{X_B | X_A = x_A \sim N(\mu_{B|A}, \Sigma_{B|A})}

where:
\deqn{\mu_{B|A} = \mu_B + \Sigma_{BA} \Sigma_{AA}^{-1} (x_A - \mu_A)}
\deqn{\Sigma_{B|A} = \Sigma_{BB} - \Sigma_{BA} \Sigma_{AA}^{-1} \Sigma_{AB}}

This is equivalent to the regression formulation used by fippy:
\deqn{\beta = \Sigma_{BA} \Sigma_{AA}^{-1}}
\deqn{\mu_{B|A} = \mu_B + \beta (x_A - \mu_A)}
\deqn{\Sigma_{B|A} = \Sigma_{BB} - \beta \Sigma_{AB}}

\strong{Assumptions:}
\itemize{
\item Features are approximately multivariate normal
\item Only continuous features are supported
}

\strong{Advantages:}
\itemize{
\item Very fast (closed-form solution)
\item Deterministic (given seed)
\item No hyperparameters
\item Memory efficient
}

\strong{Limitations:}
\itemize{
\item Strong distributional assumption
\item May produce out-of-range values for bounded features
\item Cannot handle categorical features
\item Integer features are treated as continuous and rounded back to integers
}
}
\examples{
library(mlr3)
task = tgen("friedman1")$generate(n = 100)
sampler = ConditionalGaussianSampler$new(task)

# Sample x2, x3 conditioned on x1
test_data = task$data(rows = 1:5)
sampled = sampler$sample_newdata(
  feature = c("important2", "important3"),
  newdata = test_data,
  conditioning_set = "important1"
)

}
\references{
Anderson T (2003).
\emph{An Introduction to Multivariate Statistical Analysis}, 3rd edition.
Wiley-Interscience, Hoboken, NJ.
ISBN 9780471360919.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:ConditionalSampler]{ConditionalSampler}} -> \code{ConditionalGaussianSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.}

    \item{\code{mu}}{(\code{numeric()}) Mean vector estimated from training data.}

    \item{\code{sigma}}{(\code{matrix()}) Covariance matrix estimated from training data.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalGaussianSampler-initialize}{\code{ConditionalGaussianSampler$new()}}
    \item \href{#method-ConditionalGaussianSampler-clone}{\code{ConditionalGaussianSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="ConditionalSampler" data-id="sample"><a href='../../xplainfi/html/ConditionalSampler.html#method-ConditionalSampler-sample'><code>ConditionalSampler$sample()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="ConditionalSampler" data-id="sample_newdata"><a href='../../xplainfi/html/ConditionalSampler.html#method-ConditionalSampler-sample_newdata'><code>ConditionalSampler$sample_newdata()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalGaussianSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalGaussianSampler-initialize}{}}}
\subsection{\code{ConditionalGaussianSampler$new()}}{
  Creates a new ConditionalGaussianSampler.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalGaussianSampler$new(task, conditioning_set = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from. Must have only numeric/integer features.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Default conditioning set to use in \verb{$sample()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalGaussianSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalGaussianSampler-clone}{}}}
\subsection{\code{ConditionalGaussianSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalGaussianSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FeatureImportanceMethod.R
\name{FeatureImportanceMethod}
\alias{FeatureImportanceMethod}
\title{Feature Importance Method Class}
\description{
Feature Importance Method Class
}
\references{
Nadeau C, Bengio Y (2003).
\dQuote{Inference for the Generalization Error.}
\emph{Machine Learning}, \bold{52}(3), 239--281.
\doi{10.1023/A:1024068626366}.
Molnar C, Freiesleben T, König G, Herbinger J, Reisinger T, Casalicchio G, Wright M, Bischl B (2023).
\dQuote{Relating the Partial Dependence Plot and Permutation Feature Importance to the Data Generating Process.}
In Longo L (ed.), \emph{Explainable Artificial Intelligence}, 456--479.
ISBN 978-3-031-44064-9.
\doi{10.1007/978-3-031-44064-9_24}.
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{label}}{(\code{character(1)}) Method label.}

    \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task})}

    \item{\code{learner}}{(\link[mlr3:Learner]{mlr3::Learner})}

    \item{\code{measure}}{(\link[mlr3:Measure]{mlr3::Measure})}

    \item{\code{resampling}}{(\link[mlr3:Resampling]{mlr3::Resampling}), instantiated upon construction.}

    \item{\code{resample_result}}{(\link[mlr3:ResampleResult]{mlr3::ResampleResult}) of the original \code{learner} and \code{task}, used for baseline scores.}

    \item{\code{features}}{(\code{character}: \code{NULL}) Features of interest. By default, importances will be computed for each feature
in \code{task}, but optionally this can be restricted to at least one feature. Ignored if \code{groups} is specified.}

    \item{\code{groups}}{(\code{list}: \code{NULL}) A (named) list of features (names or indices as in \code{task}).
If \code{groups} is specified, \code{features} is ignored.
Importances will be calculated for group of features at a time, e.g., in \link{PFI} not one but the group of features will be permuted at each step.
Analogously in \link{WVIM}, each group of features will be left out (or in) for each model refit.
Not all methods support groups (e.g., \link{SAGE}).}

    \item{\code{param_set}}{(\code{\link[paradox:ps]{paradox::ps()}})}

    \item{\code{predictions}}{(\link[data.table:data.table]{data.table}) Feature-specific prediction objects provided for some methods (\link{PFI}, \link{WVIM}). Contains columns for feature of interest, resampling iteration, refit or perturbation iteration, and \link[mlr3:Prediction]{mlr3::Prediction} objects.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-FeatureImportanceMethod-initialize}{\code{FeatureImportanceMethod$new()}}
    \item \href{#method-FeatureImportanceMethod-compute}{\code{FeatureImportanceMethod$compute()}}
    \item \href{#method-FeatureImportanceMethod-importance}{\code{FeatureImportanceMethod$importance()}}
    \item \href{#method-FeatureImportanceMethod-obs_loss}{\code{FeatureImportanceMethod$obs_loss()}}
    \item \href{#method-FeatureImportanceMethod-reset}{\code{FeatureImportanceMethod$reset()}}
    \item \href{#method-FeatureImportanceMethod-print}{\code{FeatureImportanceMethod$print()}}
    \item \href{#method-FeatureImportanceMethod-scores}{\code{FeatureImportanceMethod$scores()}}
    \item \href{#method-FeatureImportanceMethod-clone}{\code{FeatureImportanceMethod$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-initialize}{}}}
\subsection{\code{FeatureImportanceMethod$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
This is typically intended for use by derived classes.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  groups = NULL,
  param_set = paradox::ps(),
  label
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task, learner, measure, resampling, features, groups, param_set, label}}{Used to set fields}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-compute"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-compute}{}}}
\subsection{\code{FeatureImportanceMethod$compute()}}{
  Compute feature importance scores
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$compute(store_backends = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{store_backends}}{(\code{logical(1): TRUE}) Whether to store backends.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-importance"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-importance}{}}}
\subsection{\code{FeatureImportanceMethod$importance()}}{
  Get aggregated importance scores.
The stored \code{\link[mlr3:Measure]{measure}} object's \code{aggregator} (default: \code{mean}) will be used to aggregated importance scores
across resampling iterations and, depending on the method use, permutations (\link{PerturbationImportance} or refits \link{LOCO}).
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$importance(
  relation = NULL,
  standardize = FALSE,
  ci_method = c("none", "raw", "nadeau_bengio", "quantile"),
  conf_level = 0.95,
  alternative = c("two.sided", "greater"),
  p_adjust = "none",
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{relation}}{(character(1)) How to relate perturbed scores to originals ("difference" or "ratio").
If \code{NULL}, uses stored parameter value. This is only applicable for methods where importance is based on some
relation between baseline and post-modification loss, i.e. \link{PerturbationImportance} methods such as \link{PFI} or \link{WVIM} / \link{LOCO}.
Not available for \link{SAGE} methods.
For \code{"ratio"}, features whose baseline score is \code{0} yield \code{NA} (with a warning) and therefore aggregate to \code{NA}.}
      \item{\code{standardize}}{(\code{logical(1)}: \code{FALSE}) If \code{TRUE}, importances are standardized by the highest score so all scores fall in \verb{[-1, 1]}.}
      \item{\code{ci_method}}{(\code{character(1)}: \code{"none"}) Which confidence interval estimation method to use, defaulting to omitting
variance estimation (\code{"none"}).
If \code{"raw"}, uncorrected (too narrow) CIs are provided purely for informative purposes.
If \code{"nadeau_bengio"}, variance correction is performed according to Nadeau & Bengio (2003) as suggested by Molnar et al. (2023).
If \code{"quantile"}, empirical quantiles are used to construct confidence-like intervals.
These methods are model-agnostic and rely on suitable \code{resampling}s, e.g. subsampling with 15 repeats for \code{"nadeau_bengio"}.
See details.}
      \item{\code{conf_level}}{(\code{numeric(1)}: \code{0.95}) Confidence level to use for confidence interval construction when \code{ci_method != "none"}.}
      \item{\code{alternative}}{(\code{character(1)}: \code{"two.sided"}) Type of alternative hypothesis for statistical tests.
\code{"greater"} tests H0: importance <= 0 vs H1: importance > 0 (one-sided).
\code{"two.sided"} tests H0: importance = 0 vs H1: importance != 0.
Only used when \code{ci_method != "none"}.}
      \item{\code{p_adjust}}{(\code{character(1)}: \code{"none"}) Method for p-value adjustment for multiple comparisons.
Accepts any method supported by \link[stats:p.adjust.methods]{stats::p.adjust.methods}, e.g. \code{"holm"}, \code{"bonferroni"}, \code{"BH"}, \code{"none"}.
Applied to p-values from \code{"raw"} and \code{"nadeau_bengio"} methods.
When \code{"bonferroni"}, confidence intervals are also adjusted (alpha/k).
For other correction methods (e.g. \code{"holm"}, \code{"BH"}), only p-values are adjusted;
confidence intervals remain at the nominal \code{conf_level} because these sequential/adaptive
procedures do not have a clean per-comparison alpha for CI construction.}
      \item{\code{...}}{Additional arguments passed to specialized methods, if any.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Details}{
    \subsection{Confidence Interval Methods}{

The parametric methods (\code{"raw"}, \code{"nadeau_bengio"}) return standard error (\code{se}),
test statistic (\code{statistic}), p-value (\code{p.value}), and confidence bounds
(\code{conf_lower}, \code{conf_upper}). The \code{"quantile"} method returns only lower and upper bounds.

\strong{\code{"raw"}: Uncorrected (!) t-test}
Uses a standard t-test assuming independence of resampling iterations.
\itemize{
\item SE = sd(resampling scores) / sqrt(n_iters)
\item Test statistic: t = importance / SE with df = n_iters - 1
\item P-value: From t-distribution (one-sided or two-sided depending on \code{alternative})
\item CIs: importance +/- qt(1 - alpha, df) * SE
}

\strong{Warning}: These CIs are too narrow because resampling iterations share
training data and are not independent.
This method is included only for demonstration purposes.

\strong{\code{"nadeau_bengio"}: Corrected t-test}
Applies the Nadeau & Bengio (2003) correction to account for correlation between
resampling iterations due to overlapping training sets.
\itemize{
\item Correction factor: (1/n_iters + n_test/n_train)
\item SE = sqrt(correction_factor * var(resampling scores))
\item Test statistic and p-value: As in \code{"raw"}, but with corrected SE
}

Recommended with bootstrap or subsampling (>= 10 iterations).

\strong{\code{"quantile"}: Non-parametric empirical method}
Uses the resampling distribution directly without parametric assumptions.
\itemize{
\item CIs: Empirical quantiles of the resampling distribution
}

This method does not provide \code{se}, \code{statistic}, or \code{p.value}.
}

\subsection{Method-Specific CI Methods}{

Some importance methods provide additional CI methods tailored to their approach:
\itemize{
\item \strong{\link{CFI}}: Adds \code{"cpi"} (Conditional Predictive Impact), which uses observation-wise
loss differences with holdout resampling. Supports t-test, Wilcoxon, Fisher permutation,
and binomial tests. See Watson & Wright (2021).
}
}

\subsection{Practical Recommendations}{

Variance estimates for importance scores are biased due to the resampling procedure.
Molnar et al. (2023) suggest using the Nadeau & Bengio correction with approximately
15 iterations of subsampling.

Bootstrapping can cause information leakage with learners that bootstrap internally
(e.g., Random Forests), as observations may appear in both train and test sets.
Prefer subsampling in such cases:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{PFI$new(
  task = sim_dgp_interactions(n = 1000),
  learner = lrn("regr.ranger", num.trees = 100),
  measure = msr("regr.mse"),
  resampling = rsmp("subsampling", repeats = 15),
  n_repeats = 20
)
}\if{html}{\out{</div>}}

The \code{"nadeau_bengio"} correction was validated for PFI; its use with other methods
like LOCO or SAGE is experimental.
}
  }
  \subsection{Returns}{
    (\link[data.table:data.table]{data.table}) Aggregated importance scores with columns \code{"feature"}, \code{"importance"},
and depending on \code{ci_method} also \code{"se"}, \code{"statistic"}, \code{"p.value"}, \code{"conf_lower"}, \code{"conf_upper"}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-obs_loss"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-obs_loss}{}}}
\subsection{\code{FeatureImportanceMethod$obs_loss()}}{
  Calculate observation-wise importance scores.

Requires that \verb{$compute()} was run and that \code{measure} is decomposable and
has an observation-wise loss (\code{Measure$obs_loss()}) associated with it.
This is not the case for measure like \code{classif.auc}, which is not decomposable.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$obs_loss(relation = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{relation}}{(character(1)) How to relate perturbed scores to originals ("difference" or "ratio"). If \code{NULL}, uses stored parameter value. This is only applicable for methods where importance is based on some
relation between baseline and post-modification loss, i.e. \link{PerturbationImportance} methods such as \link{PFI} or \link{WVIM} / \link{LOCO}. Not available for \link{SAGE} methods.
For \code{"ratio"}, features whose baseline score is \code{0} yield \code{NA} (with a warning) and therefore aggregate to \code{NA}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\link[data.table:data.table]{data.table}) Observation-wise losses and importance scores with columns
\code{"feature"}, \code{"iter_rsmp"}, \code{"iter_repeat"} (if applicable), \code{"row_ids"}, \code{"loss_baseline"}, \code{"loss_post"}, and \code{"obs_importance"}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-reset"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-reset}{}}}
\subsection{\code{FeatureImportanceMethod$reset()}}{
  Resets all stored fields populated by \verb{$compute}: \verb{$resample_result}, \verb{$scores}, \verb{$obs_losses}, and \verb{$predictions}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$reset()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-print"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-print}{}}}
\subsection{\code{FeatureImportanceMethod$print()}}{
  Print importance scores
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{Passed to \code{print()}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-scores"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-scores}{}}}
\subsection{\code{FeatureImportanceMethod$scores()}}{
  Calculate importance scores for each resampling iteration and sub-iterations
(\code{iter_rsmp} in \link{PFI} for example).

Iteration-wise importance are computed on the fly depending on the chosen relation
(\code{difference} or \code{ratio}) to avoid re-computation if only a different relation is needed.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$scores(relation = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{relation}}{(character(1)) How to relate perturbed scores to originals ("difference" or "ratio"). If \code{NULL}, uses stored parameter value. This is only applicable for methods where importance is based on some
relation between baseline and post-modification loss, i.e. \link{PerturbationImportance} methods such as \link{PFI} or \link{WVIM} / \link{LOCO}. Not available for \link{SAGE} methods.
For \code{"ratio"}, features whose baseline score is \code{0} yield \code{NA} (with a warning) and therefore aggregate to \code{NA}.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    (\link[data.table:data.table]{data.table}) Iteration-wise importance scores with columns for
\code{"feature"}, iteration indices, baseline and post-modification scores, and \code{"importance"}.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureImportanceMethod-clone"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureImportanceMethod-clone}{}}}
\subsection{\code{FeatureImportanceMethod$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureImportanceMethod$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConditionalCtreeSampler.R
\name{ConditionalCtreeSampler}
\alias{ConditionalCtreeSampler}
\title{(experimental) Conditional Inference Tree Conditional Sampler}
\description{
Implements conditional sampling using conditional inference trees (ctree).
Builds a tree predicting target features from conditioning features, then samples
from the terminal node corresponding to each test observation.
}
\details{
This sampler approximates the conditional distribution \eqn{P(X_B | X_A = x_A)} by:
\enumerate{
\item Building a conditional inference tree with \eqn{X_B} as response and \eqn{X_A} as predictors
\item For each test observation, finding its terminal (leaf) node in the tree
\item Sampling uniformly from training observations in that same terminal node
}

Conditional inference trees (ctree) use permutation tests to determine splits,
which helps avoid overfitting and handles mixed feature types naturally. The tree
partitions the feature space based on the conditioning variables, creating local
neighborhoods that respect the conditional distribution structure.

\strong{Key advantages over other samplers:}
\itemize{
\item Handles mixed feature types (continuous and categorical)
\item Non-parametric (no distributional assumptions)
\item Automatic feature selection (splits only on informative features)
\item Can capture non-linear conditional relationships
\item Statistically principled splitting criteria
}

\strong{Hyperparameters} control tree complexity:
\itemize{
\item \code{mincriterion}: Significance level for splits (higher = fewer splits)
\item \code{minsplit}: Minimum observations required for a split
\item \code{minbucket}: Minimum observations in terminal nodes
}

This implementation is inspired by shapr's ctree approach but simplified for our
use case (we build trees on-demand rather than pre-computing all subsets).

\strong{Advantages:}
\itemize{
\item Works with any feature types
\item Robust to outliers
\item Interpretable tree structure
\item Handles high-dimensional conditioning
}

\strong{Limitations:}
\itemize{
\item Requires model fitting (slower than kNN)
\item Can produce duplicates if terminal nodes are small
\item Tree building time increases with data size
}
}
\examples{
\donttest{
library(mlr3)
task = tgen("friedman1")$generate(n = 100)

# Create sampler with default parameters
sampler = ConditionalCtreeSampler$new(task)

# Sample features conditioned on others
test_data = task$data(rows = 1:5)
sampled = sampler$sample_newdata(
  feature = c("important2", "important3"),
  newdata = test_data,
  conditioning_set = "important1"
)
}

}
\references{
Hothorn T, Hornik K, Zeileis A (2006).
\dQuote{Unbiased Recursive Partitioning: A Conditional Inference Framework.}
\emph{Journal of Computational and Graphical Statistics}, \bold{15}(3), 651--674.
\doi{10.1198/106186006X133933}.

Aas K, Jullum M, Løland A (2021).
\dQuote{Explaining Individual Predictions When Features Are Dependent: More Accurate Approximations to Shapley Values.}
\emph{Artificial Intelligence}, \bold{298}, 103502.
\doi{10.1016/j.artint.2021.103502}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:ConditionalSampler]{ConditionalSampler}} -> \code{ConditionalCtreeSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.
Will be checked against the provided \link[mlr3:Task]{mlr3::Task} to ensure compatibility.}

    \item{\code{tree_cache}}{(\code{environment}) Cache for fitted ctree models.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalCtreeSampler-initialize}{\code{ConditionalCtreeSampler$new()}}
    \item \href{#method-ConditionalCtreeSampler-clone}{\code{ConditionalCtreeSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="ConditionalSampler" data-id="sample"><a href='../../xplainfi/html/ConditionalSampler.html#method-ConditionalSampler-sample'><code>ConditionalSampler$sample()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="ConditionalSampler" data-id="sample_newdata"><a href='../../xplainfi/html/ConditionalSampler.html#method-ConditionalSampler-sample_newdata'><code>ConditionalSampler$sample_newdata()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalCtreeSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalCtreeSampler-initialize}{}}}
\subsection{\code{ConditionalCtreeSampler$new()}}{
  Creates a new ConditionalCtreeSampler.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalCtreeSampler$new(
  task,
  conditioning_set = NULL,
  mincriterion = 0.95,
  minsplit = 20L,
  minbucket = 7L,
  use_cache = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Default conditioning set to use in \verb{$sample()}.}
      \item{\code{mincriterion}}{(\code{numeric(1)}: \code{0.95}) Significance level threshold for splitting (1 - p-value).
Higher values result in fewer splits (simpler trees).}
      \item{\code{minsplit}}{(\code{integer(1)}: \code{20L}) Minimum number of observations required for a split.}
      \item{\code{minbucket}}{(\code{integer(1)}: \code{7L}) Minimum number of observations in terminal nodes.}
      \item{\code{use_cache}}{(\code{logical(1)}: \code{TRUE}) Whether to cache fitted trees.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalCtreeSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalCtreeSampler-clone}{}}}
\subsection{\code{ConditionalCtreeSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalCtreeSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MarginalPermutationSampler.R
\name{MarginalPermutationSampler}
\alias{MarginalPermutationSampler}
\title{Marginal Permutation Sampler}
\description{
Implements marginal permutation-based sampling for Permutation Feature Importance (PFI).
Each specified feature is randomly shuffled (permuted) independently, breaking the
relationship between the feature and the target as well as between rows.
}
\details{
The permutation sampler randomly shuffles feature values across observations:
\itemize{
\item Each feature is permuted \strong{independently} within its column
\item The association between feature values and target values is broken
\item The association between feature values \strong{across rows} is broken
\item The marginal distribution of each feature is preserved
}

\strong{Important distinction from SAGE's "marginal" approach:}
\itemize{
\item \code{MarginalPermutationSampler}: Shuffles features independently, breaking row structure
\item \code{MarginalSAGE}: Uses reference data but keeps rows intact (features in coalition stay together)
}

This is the classic approach used in Permutation Feature Importance (PFI) and
assumes features are independent.
}
\examples{
library(mlr3)
task = tgen("2dnormals")$generate(n = 10)
task$data()
sampler = MarginalPermutationSampler$new(task)

# Sample using row_ids from stored task
sampler$sample("x1")

# Or use external data
data = task$data()
sampler$sample_newdata("x1", newdata = data)
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:MarginalSampler]{MarginalSampler}} -> \code{MarginalPermutationSampler}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MarginalPermutationSampler-initialize}{\code{MarginalPermutationSampler$new()}}
    \item \href{#method-MarginalPermutationSampler-clone}{\code{MarginalPermutationSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="MarginalSampler" data-id="sample"><a href='../../xplainfi/html/MarginalSampler.html#method-MarginalSampler-sample'><code>MarginalSampler$sample()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="MarginalSampler" data-id="sample_newdata"><a href='../../xplainfi/html/MarginalSampler.html#method-MarginalSampler-sample_newdata'><code>MarginalSampler$sample_newdata()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalPermutationSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalPermutationSampler-initialize}{}}}
\subsection{\code{MarginalPermutationSampler$new()}}{
  Creates a new instance of the MarginalPermutationSampler class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalPermutationSampler$new(task)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalPermutationSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalPermutationSampler-clone}{}}}
\subsection{\code{MarginalPermutationSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalPermutationSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sage-internals.R
\name{sage_batch_predict}
\alias{sage_batch_predict}
\title{Batch Predict for SAGE}
\usage{
sage_batch_predict(learner, combined_data, task, batch_size, task_type)
}
\arguments{
\item{learner}{(\code{\link[mlr3:Learner]{Learner}}) Trained mlr3 learner.}

\item{combined_data}{(\code{data.table}) Data with feature columns to predict on.}

\item{task}{(\code{\link[mlr3:Task]{Task}}) mlr3 task object.}

\item{batch_size}{(\code{integer(1)} or \code{NULL}) Batch size for predictions. If \code{NULL} or if
\code{total_rows <= batch_size}, processes all data at once.}

\item{task_type}{(\code{character(1)}) Task type, either \code{"classif"} or \code{"regr"}.}
}
\value{
For classification: \code{matrix} of class probabilities (n_rows x n_classes).
For regression: \code{numeric} vector of predictions (length n_rows).
}
\description{
Performs batched prediction on combined data to manage memory usage.
Supports both classification (probability predictions) and regression.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_groups.R
\name{check_groups}
\alias{check_groups}
\title{Check group specification}
\usage{
check_groups(groups, all_features)
}
\arguments{
\item{groups}{(\code{list}) A (named) list of groups}

\item{all_features}{(character()) All available feature names from the task.}
}
\value{
The input list \code{group}, with each element now named.
}
\description{
Check group specification
}
\examples{
task <- sim_dgp_interactions(n = 100)
task$feature_names

# Intended use
groups1 = list(effects = c("x1", "x2", "x3"), noise = c("noise1", "noise2"))
check_groups(groups1, task$feature_names)

# Names are auto-generated where needed
check_groups(list(a = "x1",  c("x2", "x1")), task$feature_names)

# Examples for cases that throw errors:

# Unexpected features
groups2 = list(effects = c("x1", "foo", "bar", "x1"))
try(check_groups(groups2, task$feature_names))
# Too deeply nested
groups3 = list(effects = c("x1", "x2", "x3"), noise = c("noise1", list(c("noise2"))))
try(check_groups(groups2, task$feature_names))

}
\concept{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MarginalReferenceSampler.R
\name{MarginalReferenceSampler}
\alias{MarginalReferenceSampler}
\title{Marginal Reference Sampler}
\description{
Samples complete observations from reference data to replace feature values.
This approach samples from the marginal distribution while preserving within-row
feature dependencies.
}
\details{
This sampler implements what is called "marginal imputation" in the SAGE literature
(Covert et al. 2020). For each observation, it samples a complete row from reference
data and takes the specified feature values from that row. This approach:
\itemize{
\item Samples from the marginal distribution \eqn{P(X_S)} where S is the set of features
\item Preserves dependencies \strong{within} the sampled reference row
\item Breaks dependencies \strong{between} test and reference data
}

\strong{Terminology note:} In SAGE literature, this is called "marginal imputation" because
features outside the coalition are "imputed" by sampling from their marginal distribution.
We use \code{MarginalReferenceSampler} to avoid confusion with missing data imputation and to
clarify that it samples from reference data.

\strong{Comparison with other samplers:}
\itemize{
\item \code{MarginalPermutationSampler}: Shuffles each feature independently, breaking all row structure
\item \code{MarginalReferenceSampler}: Samples complete rows, preserving within-row dependencies
\item \code{ConditionalSampler}: Samples from \eqn{P(X_S | X_{-S})}, conditioning on other features
}

\strong{Use in SAGE:}

This is the default approach for \code{MarginalSAGE}. For a test observation x and features
to marginalize S, it samples a reference row x_ref and creates a "hybrid" observation
combining x's coalition features with x_ref's marginalized features.
}
\examples{
library(mlr3)
task = tgen("friedman1")$generate(n = 100)

# Default: uses all task data as reference
sampler = MarginalReferenceSampler$new(task)
sampled = sampler$sample("important1", row_ids = 1:10)

# Subsample reference data to 50 rows
sampler_subsampled = MarginalReferenceSampler$new(task, n_samples = 50L)
sampled2 = sampler_subsampled$sample("important1", row_ids = 1:10)

}
\references{
Covert I, Lundberg S, Lee S (2020).
\dQuote{Understanding Global Feature Contributions With Additive Importance Measures.}
In \emph{Advances in Neural Information Processing Systems}, volume 33, 17212--17223.
\url{https://proceedings.neurips.cc/paper/2020/hash/c7bf0b7c1a86d5eb3be2c722cf2cf746-Abstract.html}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:MarginalSampler]{MarginalSampler}} -> \code{MarginalReferenceSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{reference_data}}{(\code{\link[data.table:data.table]{data.table}})
Reference data to sample from for marginalization.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MarginalReferenceSampler-initialize}{\code{MarginalReferenceSampler$new()}}
    \item \href{#method-MarginalReferenceSampler-clone}{\code{MarginalReferenceSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="MarginalSampler" data-id="sample"><a href='../../xplainfi/html/MarginalSampler.html#method-MarginalSampler-sample'><code>MarginalSampler$sample()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="MarginalSampler" data-id="sample_newdata"><a href='../../xplainfi/html/MarginalSampler.html#method-MarginalSampler-sample_newdata'><code>MarginalSampler$sample_newdata()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalReferenceSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalReferenceSampler-initialize}{}}}
\subsection{\code{MarginalReferenceSampler$new()}}{
  Creates a new instance of the MarginalReferenceSampler class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalReferenceSampler$new(task, n_samples = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
      \item{\code{n_samples}}{(\code{integer(1)} | \code{NULL}) Number of reference samples to use.
If \code{NULL}, uses all task data as reference.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalReferenceSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalReferenceSampler-clone}{}}}
\subsection{\code{MarginalReferenceSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalReferenceSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sage-internals.R
\name{sage_aggregate_predictions}
\alias{sage_aggregate_predictions}
\title{Aggregate Predictions by Coalition and Test Instance}
\usage{
sage_aggregate_predictions(
  combined_data,
  predictions,
  task_type,
  class_names = NULL
)
}
\arguments{
\item{combined_data}{(\code{data.table}) Data with columns \code{.coalition_id}, \code{.test_instance_id},
and feature columns.}

\item{predictions}{(\code{matrix} or \code{numeric}) For classification: matrix of class probabilities.
For regression: numeric vector of predictions.}

\item{task_type}{(\code{character(1)}) Task type, either \code{"classif"} or \code{"regr"}.}

\item{class_names}{(\code{character()} or \code{NULL}: \code{NULL}) Character vector of class names. Required
for classification, ignored for regression.}
}
\value{
\code{data.table} with columns:
\itemize{
\item \code{.coalition_id}: Coalition identifier (integer)
\item \code{.test_instance_id}: Test instance identifier (integer)
\item For classification: One column per class with averaged probabilities (numeric)
\item For regression: \code{avg_pred} column with averaged predictions (numeric)
}
}
\description{
Averages predictions across multiple samples (reference data or conditional samples)
for each unique combination of coalition and test instance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConditionalKNNSampler.R
\name{ConditionalKNNSampler}
\alias{ConditionalKNNSampler}
\title{k-Nearest Neighbors Conditional Sampler}
\description{
Implements conditional sampling using k-nearest neighbors (kNN).
For each observation, finds the \code{k} most similar observations based on conditioning
features, then samples the target features from these neighbors.
}
\details{
This sampler approximates the conditional distribution \eqn{P(X_B | X_A = x_A)} by:
\enumerate{
\item Finding the k nearest neighbors of \eqn{x_A} in the training data
\item Sampling uniformly from the target feature values \eqn{X_B} of these k neighbors
}

This is a simple, non-parametric approach that:
\itemize{
\item Requires no distributional assumptions
\item Handles mixed feature types (numeric, integer, factor, ordered, logical)
\item Is computationally efficient (no model fitting required)
\item Adapts locally to the data structure
}

The method is related to hot-deck imputation and kNN imputation techniques used in
missing data problems. As \eqn{k \to \infty} and \eqn{k/n \to 0}, the kNN conditional
distribution converges to the true conditional distribution under mild regularity
conditions (Lipschitz continuity).

\strong{Distance Metrics:}

The sampler supports two distance metrics:
\itemize{
\item \strong{Euclidean}: For numeric/integer features only. Standardizes features before computing distances.
\item \strong{Gower}: For mixed feature types. Handles numeric, factor, ordered, and logical features.
Numeric features are range-normalized, categorical features use exact matching (0/1).
}

The \code{distance} parameter controls which metric to use:
\itemize{
\item \code{"auto"} (default): Automatically selects Euclidean for all-numeric features, Gower otherwise
\item \code{"euclidean"}: Forces Euclidean distance (errors if non-numeric features present)
\item \code{"gower"}: Forces Gower distance (works with any feature types)
}

\strong{Advantages:}
\itemize{
\item Very fast (no model training)
\item Works with any feature types
\item Automatic distance metric selection
\item Naturally respects local data structure
}

\strong{Limitations:}
\itemize{
\item Sensitive to choice of \code{k}
\item The full task data is required for prediction
\item Can produce duplicates if \code{k} is small
\item May not extrapolate well to new regions
}
}
\examples{
library(mlr3)
task = tgen("friedman1")$generate(n = 100)
sampler = ConditionalKNNSampler$new(task, k = 5)

# Sample features conditioned on others
test_data = task$data(rows = 1:5)
sampled = sampler$sample_newdata(
  feature = c("important2", "important3"),
  newdata = test_data,
  conditioning_set = "important1"
)

}
\references{
Little R, Rubin D (2019).
\emph{Statistical Analysis with Missing Data}, 3rd edition.
John Wiley & Sons, Hoboken, NJ.
ISBN 9780470526798.

Troyanskaya O, Cantor M, Sherlock G, Brown P, Hastie T, Tibshirani R, Botstein D, Altman R (2001).
\dQuote{Missing Value Estimation Methods for DNA Microarrays.}
\emph{Bioinformatics}, \bold{17}(6), 520--525.
\doi{10.1093/bioinformatics/17.6.520}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{\link[xplainfi:ConditionalSampler]{ConditionalSampler}} -> \code{ConditionalKNNSampler}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalKNNSampler-initialize}{\code{ConditionalKNNSampler$new()}}
    \item \href{#method-ConditionalKNNSampler-sample}{\code{ConditionalKNNSampler$sample()}}
    \item \href{#method-ConditionalKNNSampler-sample_newdata}{\code{ConditionalKNNSampler$sample_newdata()}}
    \item \href{#method-ConditionalKNNSampler-clone}{\code{ConditionalKNNSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalKNNSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalKNNSampler-initialize}{}}}
\subsection{\code{ConditionalKNNSampler$new()}}{
  Creates a new ConditionalKNNSampler.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalKNNSampler$new(task, conditioning_set = NULL, k = 5L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Default conditioning set to use in \verb{$sample()}.}
      \item{\code{k}}{(\code{integer(1)}: \code{5L}) Number of nearest neighbors to sample from.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalKNNSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalKNNSampler-sample}{}}}
\subsection{\code{ConditionalKNNSampler$sample()}}{
  Sample features from their kNN-based conditional distribution.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalKNNSampler$sample(
  feature,
  row_ids = NULL,
  conditioning_set = NULL,
  samples_per_row = 1L,
  k = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character()}) Feature name(s) to sample.}
      \item{\code{row_ids}}{(\code{integer()} | \code{NULL}) Row IDs from task to use as conditioning values.}
      \item{\code{conditioning_set}}{(\code{character()} | \code{NULL}) Features to condition on.
If \code{NULL}, samples from marginal distribution (random sampling from training data).}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{k}}{(\code{integer(1)} | \code{NULL}) Number of neighbors. If \code{NULL}, uses stored parameter.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalKNNSampler-sample_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalKNNSampler-sample_newdata}{}}}
\subsection{\code{ConditionalKNNSampler$sample_newdata()}}{
  Sample from external data conditionally.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalKNNSampler$sample_newdata(
  feature,
  newdata,
  conditioning_set = NULL,
  samples_per_row = 1L,
  k = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character()}) Feature(s) to sample.}
      \item{\code{newdata}}{(\code{\link[data.table:data.table]{data.table}}) External data to use.}
      \item{\code{conditioning_set}}{(\code{character()} | \code{NULL}) Features to condition on.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{k}}{(\code{integer(1)} | \code{NULL}) Number of neighbors. If \code{NULL}, uses stored parameter.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalKNNSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalKNNSampler-clone}{}}}
\subsection{\code{ConditionalKNNSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalKNNSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/WVIM.R
\name{LOCO}
\alias{LOCO}
\title{Leave-One-Covariate-Out (LOCO)}
\description{
Calculates Leave-One-Covariate-Out (LOCO) scores.
}
\details{
LOCO measures feature importance by comparing model performance with and without
each feature. For each feature, the model is retrained without that feature and
the performance difference (reduced_model_loss - full_model_loss) indicates the
feature's importance. Higher values indicate more important features.
}
\examples{
library(mlr3)
library(mlr3learners)

task <- sim_dgp_correlated(n = 500)

loco <- LOCO$new(
  task = task,
  learner = lrn("regr.rpart"),
  measure = msr("regr.mse")
)
loco$compute()
loco$importance()
}
\references{
Lei J, G'Sell M, Rinaldo A, Tibshirani R, Wasserman L (2018).
\dQuote{Distribution-Free Predictive Inference for Regression.}
\emph{Journal of the American Statistical Association}, \bold{113}(523), 1094--1111.
\doi{10.1080/01621459.2017.1307116}.
}
\section{Super classes}{
\code{\link[xplainfi:FeatureImportanceMethod]{FeatureImportanceMethod}} -> \code{\link[xplainfi:WVIM]{WVIM}} -> \code{LOCO}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-LOCO-initialize}{\code{LOCO$new()}}
    \item \href{#method-LOCO-compute}{\code{LOCO$compute()}}
    \item \href{#method-LOCO-clone}{\code{LOCO$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="obs_loss"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-obs_loss'><code>FeatureImportanceMethod$obs_loss()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="print"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-print'><code>FeatureImportanceMethod$print()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="reset"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-reset'><code>FeatureImportanceMethod$reset()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureImportanceMethod" data-id="scores"><a href='../../xplainfi/html/FeatureImportanceMethod.html#method-FeatureImportanceMethod-scores'><code>FeatureImportanceMethod$scores()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="WVIM" data-id="importance"><a href='../../xplainfi/html/WVIM.html#method-WVIM-importance'><code>WVIM$importance()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LOCO-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-LOCO-initialize}{}}}
\subsection{\code{LOCO$new()}}{
  Creates a new instance of this \link[R6:R6Class]{R6} class.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LOCO$new(
  task,
  learner,
  measure = NULL,
  resampling = NULL,
  features = NULL,
  n_repeats = 1L,
  batch_size = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to compute importance for.}
      \item{\code{learner}}{(\link[mlr3:Learner]{mlr3::Learner}) Learner to use for prediction.}
      \item{\code{measure}}{(\link[mlr3:Measure]{mlr3::Measure}: \code{NULL}) Measure to use for scoring. Defaults to
\code{classif.ce} for classification and \code{regr.mse} for regression.}
      \item{\code{resampling}}{(\link[mlr3:Resampling]{mlr3::Resampling}) Resampling strategy. Defaults to holdout.}
      \item{\code{features}}{(\code{character()}) Features to compute importance for. Defaults to all features.}
      \item{\code{n_repeats}}{(\code{integer(1)}: \code{1L}) Number of refit iterations per resampling iteration.
This may be useful for some stochastic learners, but in general it is advisable to increase resampling iterations instead of
repeatedly refitting on the same data.}
      \item{\code{batch_size}}{(\code{integer(1)}: \code{NULL}) Number of refits dispatched per internal
\code{mlr3::benchmark()} call. \code{NULL} (default) keeps sequential single-refit evaluation; set to a
positive integer (e.g. the number of \code{future}/\code{mirai} workers) to enable parallel refits.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LOCO-compute"></a>}}
\if{latex}{\out{\hypertarget{method-LOCO-compute}{}}}
\subsection{\code{LOCO$compute()}}{
  Compute LOCO importances.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LOCO$compute(store_models = TRUE, store_backends = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{store_models, store_backends}}{(\code{logical(1)}: \code{TRUE}) Whether to store fitted models / data backends, passed to \link[mlr3:resample]{mlr3::resample} internally}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-LOCO-clone"></a>}}
\if{latex}{\out{\hypertarget{method-LOCO-clone}{}}}
\subsection{\code{LOCO$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{LOCO$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MarginalSampler.R
\name{MarginalSampler}
\alias{MarginalSampler}
\title{Marginal Sampler Base Class}
\description{
Abstract base class for marginal sampling strategies that do not
condition on other features. Marginal samplers sample from P(X_S), the marginal
distribution of features S, ignoring any dependencies with other features.
}
\details{
This class provides a common interface for different marginal sampling approaches:
\itemize{
\item \strong{MarginalPermutationSampler}: Shuffles features independently within the dataset
\item \strong{MarginalReferenceSampler}: Samples complete rows from reference data
}

Both approaches sample from the marginal distribution P(X_S), but differ in how
they preserve or break within-row dependencies:
\itemize{
\item Permutation breaks ALL dependencies (both with target and between features)
\item Reference sampling preserves WITHIN-row dependencies but breaks dependencies with test data
}

\strong{Comparison with ConditionalSampler:}
\itemize{
\item \code{MarginalSampler}: Samples from \eqn{P(X_S)} - no conditioning
\item \code{ConditionalSampler}: Samples from \eqn{P(X_S | X_{-S})}- conditions on other features
}

This base class implements the public \verb{$sample()} and \verb{$sample_newdata()} methods,
delegating to private \code{.sample_marginal()} which subclasses must implement.
}
\section{Super class}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{MarginalSampler}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-MarginalSampler-sample}{\code{MarginalSampler$sample()}}
    \item \href{#method-MarginalSampler-sample_newdata}{\code{MarginalSampler$sample_newdata()}}
    \item \href{#method-MarginalSampler-clone}{\code{MarginalSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="initialize"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-initialize'><code>FeatureSampler$initialize()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalSampler-sample}{}}}
\subsection{\code{MarginalSampler$sample()}}{
  Sample features from their marginal distribution.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalSampler$sample(feature, row_ids = NULL, samples_per_row = 1L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character()}) Feature name(s) to sample.}
      \item{\code{row_ids}}{(\code{integer()} | \code{NULL}) Row IDs from task to use.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalSampler-sample_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalSampler-sample_newdata}{}}}
\subsection{\code{MarginalSampler$sample_newdata()}}{
  Sample from external data.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalSampler$sample_newdata(feature, newdata, samples_per_row = 1L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character()}) Feature(s) to sample.}
      \item{\code{newdata}}{(\code{\link[data.table:data.table]{data.table}}) External data to use.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MarginalSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MarginalSampler-clone}{}}}
\subsection{\code{MarginalSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{MarginalSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConditionalSampler.R
\name{ConditionalSampler}
\alias{ConditionalSampler}
\title{Conditional Feature Sampler}
\description{
Base class for conditional sampling methods where features
are sampled conditionally on other features. This is an abstract class
that should be extended by concrete implementations.
}
\section{Super class}{
\code{\link[xplainfi:FeatureSampler]{FeatureSampler}} -> \code{ConditionalSampler}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-ConditionalSampler-initialize}{\code{ConditionalSampler$new()}}
    \item \href{#method-ConditionalSampler-sample}{\code{ConditionalSampler$sample()}}
    \item \href{#method-ConditionalSampler-sample_newdata}{\code{ConditionalSampler$sample_newdata()}}
    \item \href{#method-ConditionalSampler-clone}{\code{ConditionalSampler$clone()}}
  }
}
\if{html}{\out{<details open><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="xplainfi" data-topic="FeatureSampler" data-id="print"><a href='../../xplainfi/html/FeatureSampler.html#method-FeatureSampler-print'><code>FeatureSampler$print()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSampler-initialize}{}}}
\subsection{\code{ConditionalSampler$new()}}{
  Creates a new instance of the ConditionalSampler class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSampler$new(task, conditioning_set = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Default conditioning set to use in \verb{$sample()}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSampler-sample}{}}}
\subsection{\code{ConditionalSampler$sample()}}{
  Sample from stored task conditionally on other features.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSampler$sample(
  feature,
  row_ids = NULL,
  conditioning_set = NULL,
  samples_per_row = 1L,
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature(s) to sample.}
      \item{\code{row_ids}}{(\code{integer()} | \code{NULL}) Row IDs to use. If \code{NULL}, uses all rows.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Features to condition on.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{...}}{Additional arguments passed to the sampler implementation.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSampler-sample_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSampler-sample_newdata}{}}}
\subsection{\code{ConditionalSampler$sample_newdata()}}{
  Sample from external data conditionally.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSampler$sample_newdata(
  feature,
  newdata,
  conditioning_set = NULL,
  samples_per_row = 1L,
  ...
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature(s) to sample.}
      \item{\code{newdata}}{(\code{\link[data.table:data.table]{data.table}}) External data to use.}
      \item{\code{conditioning_set}}{(\code{character} | \code{NULL}) Features to condition on.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples per input row.
See \link{FeatureSampler}\verb{$sample()} for output shape and ordering.}
      \item{\code{...}}{Additional arguments passed to the sampler implementation.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy with sampled feature(s).
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ConditionalSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ConditionalSampler-clone}{}}}
\subsection{\code{ConditionalSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{ConditionalSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FeatureSampler.R
\name{FeatureSampler}
\alias{FeatureSampler}
\title{Feature Sampler Class}
\description{
Base class for implementing different sampling strategies
for feature importance methods like PFI and CFI
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Original task.}

    \item{\code{label}}{(\code{character(1)}) Name of the sampler.}

    \item{\code{feature_types}}{(\code{character()}) Feature types supported by the sampler.
Will be checked against the provided \link[mlr3:Task]{mlr3::Task} to ensure compatibility.}

    \item{\code{param_set}}{(\link[paradox:ParamSet]{paradox::ParamSet}) Parameter set for the sampler.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-FeatureSampler-initialize}{\code{FeatureSampler$new()}}
    \item \href{#method-FeatureSampler-sample}{\code{FeatureSampler$sample()}}
    \item \href{#method-FeatureSampler-sample_newdata}{\code{FeatureSampler$sample_newdata()}}
    \item \href{#method-FeatureSampler-print}{\code{FeatureSampler$print()}}
    \item \href{#method-FeatureSampler-clone}{\code{FeatureSampler$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureSampler-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureSampler-initialize}{}}}
\subsection{\code{FeatureSampler$new()}}{
  Creates a new instance of the FeatureSampler class
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureSampler$new(task)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{task}}{(\link[mlr3:Task]{mlr3::Task}) Task to sample from}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureSampler-sample"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureSampler-sample}{}}}
\subsection{\code{FeatureSampler$sample()}}{
  Sample values for feature(s) from stored task
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureSampler$sample(feature, row_ids = NULL, samples_per_row = 1L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature name(s) to sample (can be single or multiple). Must match those in the stored \link[mlr3:Task]{Task}.}
      \item{\code{row_ids}}{(\code{integer()}: \code{NULL}) Row IDs of the stored \link[mlr3:Task]{Task} to use as basis for sampling.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) Number of independent samples to draw per input row.
When \verb{> 1}, the returned data.table has \code{samples_per_row * length(row_ids)} rows in \strong{draw-major}
order: rows \verb{1..n} are draw 1 across all input rows in order, rows \verb{n+1..2n} are draw 2, and so on.}
    }
    \if{html}{\out{</div>}}
  }
  \subsection{Returns}{
    Modified copy of the input features with the feature(s) sampled. A
\link[data.table:data.table]{data.table} with \code{samples_per_row * length(row_ids)} rows in draw-major order.
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureSampler-sample_newdata"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureSampler-sample_newdata}{}}}
\subsection{\code{FeatureSampler$sample_newdata()}}{
  Sample values for feature(s) using external data
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureSampler$sample_newdata(feature, newdata, samples_per_row = 1L)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{feature}}{(\code{character}) Feature name(s) to sample (can be single or multiple)}
      \item{\code{newdata}}{(\code{\link[data.table:data.table]{data.table}}) External data to use for sampling.}
      \item{\code{samples_per_row}}{(\code{integer(1)}: \code{1L}) See \verb{$sample()}. Output is \code{samples_per_row * nrow(newdata)}
rows in draw-major order.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureSampler-print"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureSampler-print}{}}}
\subsection{\code{FeatureSampler$print()}}{
  Print sampler
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureSampler$print(...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{...}}{Ignored.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-FeatureSampler-clone"></a>}}
\if{latex}{\out{\hypertarget{method-FeatureSampler-clone}{}}}
\subsection{\code{FeatureSampler$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{FeatureSampler$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}
